# FILE: Earth_flavour_mapmod.py 
# VERSION: 2.1
# AUTHOR: Nicolas "lumpthing" Holzapfel (but 'see CREDIT' below)
# PURPOSE: Have civs start in regions that match their flavour.
# DISCUSSION: http://forums.civfanatics.com/showthread.php?t=251580
# CREDIT:  90% of the credit for this mod goes to David "dreiche2" Reichert.
# He did all the proper scripting stuff to create his Fall from Heaven flavor
# mod. All I did was take his mod and change the settings (this time for
# standard Earth civs, rather than fantasy civs) until I got something
# that worked reasonably well.

# DESCRIPTION: 
# This module can be used in conjunction with other map scripts.
# It has the following effects:
# 1) Civs tend to start in appropriate climates: i.e. Euopean civs start in
# colder regions,
# African civs start in warmer climates
# 2) East Asian civs tend to start close together
# 3) 'New World' civs tend to start close together
# 4) The Portuguese, Dutch, Carthagians, Vikings and Americans are more likely
# to start near the sea (The former four because of unique units or buildings,
# the latter for historical flavour)
# 5) The Khmer are more likely to start close to ivory
# (because of its unique unit)
# Note that this module does not change the map itself, and only uses the
# starting locations
# assigned by the base map script. There will only be a tendency (albeit a
# clear one), no guarantee 
# that all the relevant civs have 'flavoury' starting positions. An advantage
# is that this module minimally interfers with what the map scripts does.


# USAGE:
# It is very simple to use this module in a map script of your choice
# (e.g. the default continents  script). Just copy this file into the
# <civ4>/PublicMaps folder, and then open the map script file 
# (e.g. Continents.py) with a text editor and copy/paste the small code block
# below this paragraph into the script (preferably at the top), and uncomment
# the 'import' line by removing the '#'. Of course you're advised to backup
# the map script and rename your modified version to something like 
# "Earth_Flavour_whatever.py" so that you don't confuse the various versions.

## copy below, remove '#' in 'from...import' line.

##############################################################################
# This map script was modified to use the Earth Flavour Map Mod              #
# See Earth_flavour_mapmod.py for details                                    #
##############################################################################

#from Earth_flavour_mapmod import normalizeStartingPlotLocations

##############################################################################

## copy above


# COMPATIBILITY
# The flavour module should work with most map scripts. It uses the
# normalizeStartingPlotLocations() function, hence civ does not use its
# default implementation. That means that members of a team are no longer
# placed close together as by default.

# A conflict occurs if the map script itselfs implements
# normalizeStartingPlotLocations(), though only few do. If that's the case
# you can simply remove/comment out the relevant code in the map script,
# although of course you remove any functionality that was implemented there.

# CUSTOMIZATION
# It should be easy to modify the settings to your liking, and add new flavor
# civs. Just look at the code below. Everything relevant happens in
# normalizeStartingPlotLocations().

# VERSION HISTORY
# Version 2.1: Added flood plain preferences for several Middle Eastern civs.
# Increased the Vikings' preference for tundra relative to their preference
# for ocean. Made all Northern European civs prefer tundra.
# Version 2.0: Corrected a bug identified by Seven05. Many major changes:
# Replaced most of the bonus preferences with terrain and feature preferences.
# Version 1.02: Corrected a bug identified by SevenSpirits
# Version 1.02: Corrected a major typo in the usage instructions
# Version 1.01: Added sheep to Russian and Viking bonus preferences, in case
# no tundra starting positions are available




# MAP SCRIPT FOLLOWS BELOW

from CvPythonExtensions import *
import CvUtil
import CvMapGeneratorUtil


###################### some parameters ########################
#
# these are default settings. They can be changed for single civs
# in normalizeStartingPlotLocations(), e.g. InfoMalakim.regionSize = 4


def regionSize():
	return 6 #region in which to look is a square reaching regionSize tiles in any direction
			 #from the center tile

def closeRegionSize():
	return 4 #closer squares have larger impact
	
def closeFactor():
	return 1.5 #weight for closer squares

###############################################################

FFH_DEBUG=False


#helper function
#unfortunately there seems to be no inverse procedure in general
#could have stored strings in the dictionary instead of the type ints :/
def IT(string):
	cgc = CyGlobalContext()
	return cgc.getInfoTypeForString(string)

	

	
############################################################################
# in this function the locatons are redistributed. Make your changes here...

def normalizeStartingPlotLocations():

	#creating flavor civs

	#possible parameters to change:
	#.terrains[]
	#.boni[]
	#.plotTypes[]
	#.features[]
	#.setPriority()
	#.coastalStartValue
	#.regionSize   (etc.)
	#
	#examples see below, and see the civFlavorInfo class 
	#after normalizeStartingPlotLocations()
	
			
	civInfos = {} #hurray for dictionaries!
	#info = CivFlavorInfo(IT("CIVILIZATION_LUMPLAND"))
	#infoLumpland.terrains[IT("TERRAIN_TUNDRA")] = 1
	#infoLumpland.setPriority(3)
			# default priority is 2, possible 1-3
			# priority 1 civs choose their starting position first,
			# followed priority 2, with priority 3 civs choosing last
	#infoLumpland.closeFactor = 1  # don't value close tiles stronger as by default
	#civInfos[IT("CIVILIZATION_LUMPLAND")] = infoLumpland
	################


#EURO-COLD
	infoRussia = CivFlavorInfo(IT("CIVILIZATION_RUSSIA"))
	infoRussia.setPriority(1)
	infoRussia.terrains[IT("TERRAIN_TUNDRA")] = 1
	infoRussia.features[IT("FEATURE_JUNGLE")] = -1
	infoRussia.terrains[IT("TERRAIN_DESERT")] = -1
	civInfos[IT("CIVILIZATION_RUSSIA")] = infoRussia
	#################
	infoViking = CivFlavorInfo(IT("CIVILIZATION_VIKING"))
	infoViking.setPriority(1)
	infoViking.terrains[IT("TERRAIN_TUNDRA")] = 1
	infoViking.features[IT("FEATURE_JUNGLE")] = -1
	infoViking.terrains[IT("TERRAIN_DESERT")] = -1
	infoViking.terrains[IT("TERRAIN_OCEAN")] = 0.5
	civInfos[IT("CIVILIZATION_VIKING")] = infoViking
	#################

#EURO-TEMPERATE
	infoCelt = CivFlavorInfo(IT("CIVILIZATION_CELT"))
	infoCelt.setPriority(2)
	infoCelt.terrains[IT("TERRAIN_TUNDRA")] = 1
	infoCelt.features[IT("FEATURE_JUNGLE")] = -1
	infoCelt.terrains[IT("TERRAIN_DESERT")] = -1
	civInfos[IT("CIVILIZATION_CELT")] = infoCelt
	################
	infoEngland = CivFlavorInfo(IT("CIVILIZATION_ENGLAND"))
	infoEngland.setPriority(2)
	infoEngland.terrains[IT("TERRAIN_TUNDRA")] = 1
	infoEngland.features[IT("FEATURE_JUNGLE")] = -1
	infoEngland.terrains[IT("TERRAIN_DESERT")] = -1
	civInfos[IT("CIVILIZATION_ENGLAND")] = infoEngland
	#################
	infoFrance = CivFlavorInfo(IT("CIVILIZATION_FRANCE"))
	infoFrance.setPriority(2)
	infoFrance.terrains[IT("TERRAIN_TUNDRA")] = 1
	infoFrance.features[IT("FEATURE_JUNGLE")] = -1
	infoFrance.terrains[IT("TERRAIN_DESERT")] = -1
	civInfos[IT("CIVILIZATION_FRANCE")] = infoFrance
	#################
	infoGermany = CivFlavorInfo(IT("CIVILIZATION_GERMANY"))
	infoGermany.setPriority(2)
	infoGermany.terrains[IT("TERRAIN_TUNDRA")] = 1
	infoGermany.features[IT("FEATURE_JUNGLE")] = -1
	infoGermany.terrains[IT("TERRAIN_DESERT")] = -1
	civInfos[IT("CIVILIZATION_GERMANY")] = infoGermany
	#################
	infoHolyRoman = CivFlavorInfo(IT("CIVILIZATION_HOLY_ROMAN"))
	infoHolyRoman.setPriority(2)
	infoHolyRoman.terrains[IT("TERRAIN_TUNDRA")] = 1
	infoHolyRoman.features[IT("FEATURE_JUNGLE")] = -1
	infoHolyRoman.terrains[IT("TERRAIN_DESERT")] = -1
	civInfos[IT("CIVILIZATION_HOLY_ROMAN")] = infoHolyRoman
	#################
	infoNetherlands = CivFlavorInfo(IT("CIVILIZATION_NETHERLANDS"))
	infoNetherlands.setPriority(2)
	infoNetherlands.terrains[IT("TERRAIN_TUNDRA")] = 1
	infoNetherlands.features[IT("FEATURE_JUNGLE")] = -1
	infoNetherlands.terrains[IT("TERRAIN_DESERT")] = -1
	infoNetherlands.terrains[IT("TERRAIN_OCEAN")] = 1
	civInfos[IT("CIVILIZATION_NETHERLANDS")] = infoNetherlands
	#################

#EURO-MEDITERRANEAN
	infoByzantium = CivFlavorInfo(IT("CIVILIZATION_BYZANTIUM"))
	infoByzantium.setPriority(3)
	infoByzantium.features[IT("FEATURE_JUNGLE")] = -1
	civInfos[IT("CIVILIZATION_BYZANTIUM")] = infoByzantium
	#################
	infoGreece = CivFlavorInfo(IT("CIVILIZATION_GREECE"))
	infoGreece.setPriority(3)
	infoGreece.features[IT("FEATURE_JUNGLE")] = -1
	civInfos[IT("CIVILIZATION_GREECE")] = infoGreece
	#################
	infoPortugal = CivFlavorInfo(IT("CIVILIZATION_PORTUGAL"))
	infoPortugal.setPriority(3)
	infoPortugal.features[IT("FEATURE_JUNGLE")] = -1
	infoPortugal.terrains[IT("TERRAIN_OCEAN")] = 1
	civInfos[IT("CIVILIZATION_PORTUGAL")] = infoPortugal
	#################
	infoRome = CivFlavorInfo(IT("CIVILIZATION_ROME"))
	infoRome.setPriority(3)
	infoRome.features[IT("FEATURE_JUNGLE")] = -1
	civInfos[IT("CIVILIZATION_ROME")] = infoRome
	#################
	infoSpain = CivFlavorInfo(IT("CIVILIZATION_SPAIN"))
	infoSpain.setPriority(3)
	infoSpain.features[IT("FEATURE_JUNGLE")] = -1
	civInfos[IT("CIVILIZATION_SPAIN")] = infoSpain
	#################

#ARID MIDDLE EAST/NORTH AFRICA
	infoArabia = CivFlavorInfo(IT("CIVILIZATION_ARABIA"))
	infoArabia.setPriority(3)
	infoArabia.terrains[IT("TERRAIN_DESERT")] = 1
	infoArabia.terrains[IT("TERRAIN_TUNDRA")] = -1
	civInfos[IT("CIVILIZATION_ARABIA")] = infoArabia
	#################
	infoBabylon = CivFlavorInfo(IT("CIVILIZATION_BABYLON"))
	infoBabylon.setPriority(3)
	infoBabylon.terrains[IT("TERRAIN_DESERT")] = 1
	infoBabylon.features[IT("FEATURE_FLOOD_PLAINS")] = 1
	infoBabylon.terrains[IT("TERRAIN_TUNDRA")] = -1
	civInfos[IT("CIVILIZATION_BABYLON")] = infoBabylon
	#################	
	infoCarthage = CivFlavorInfo(IT("CIVILIZATION_CARTHAGE"))
	infoCarthage.setPriority(3)
	infoCarthage.terrains[IT("TERRAIN_DESERT")] = 1
	infoCarthage.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoCarthage.terrains[IT("TERRAIN_OCEAN")] = 1
	civInfos[IT("CIVILIZATION_CARTHAGE")] = infoCarthage
	#################
	infoEgypt = CivFlavorInfo(IT("CIVILIZATION_EGYPT"))
	infoEgypt.setPriority(3)
	infoEgypt.terrains[IT("TERRAIN_DESERT")] = 1
	infoEgypt.features[IT("FEATURE_FLOOD_PLAINS")] = 1
	infoEgypt.terrains[IT("TERRAIN_TUNDRA")] = -1
	civInfos[IT("CIVILIZATION_EGYPT")] = infoEgypt
	#################
	infoOttoman = CivFlavorInfo(IT("CIVILIZATION_OTTOMAN"))
	infoOttoman.setPriority(3)
	infoOttoman.terrains[IT("TERRAIN_DESERT")] = 1
	infoOttoman.terrains[IT("TERRAIN_TUNDRA")] = -1
	civInfos[IT("CIVILIZATION_OTTOMAN")] = infoOttoman
	#################
	infoPersia = CivFlavorInfo(IT("CIVILIZATION_PERSIA"))
	infoPersia.setPriority(3)
	infoPersia.terrains[IT("TERRAIN_DESERT")] = 1
	infoPersia.features[IT("FEATURE_FLOOD_PLAINS")] = 1
	infoPersia.terrains[IT("TERRAIN_TUNDRA")] = -1
	civInfos[IT("CIVILIZATION_PERSIA")] = infoPersia
	###############
	infoSumeria = CivFlavorInfo(IT("CIVILIZATION_SUMERIA"))
	infoSumeria.setPriority(3)
	infoSumeria.terrains[IT("TERRAIN_DESERT")] = 1
	infoSumeria.features[IT("FEATURE_FLOOD_PLAINS")] = 1
	infoSumeria.terrains[IT("TERRAIN_TUNDRA")] = -1
	civInfos[IT("CIVILIZATION_SUMERIA")] = infoSumeria
	#################

#TROPICAL
	infoEthiopia = CivFlavorInfo(IT("CIVILIZATION_ETHIOPIA"))
	infoEthiopia.setPriority(1)
	infoEthiopia.features[IT("FEATURE_JUNGLE")] = 1
	infoEthiopia.terrains[IT("TERRAIN_TUNDRA")] = -1
	civInfos[IT("CIVILIZATION_ETHIOPIA")] = infoEthiopia
	#################
	infoMali = CivFlavorInfo(IT("CIVILIZATION_MALI"))
	infoMali.setPriority(1)
	infoMali.features[IT("FEATURE_JUNGLE")] = 1
	infoMali.terrains[IT("TERRAIN_TUNDRA")] = -1
	civInfos[IT("CIVILIZATION_MALI")] = infoMali
	################
	infoZulu = CivFlavorInfo(IT("CIVILIZATION_ZULU"))
	infoZulu.setPriority(1)
	infoZulu.features[IT("FEATURE_JUNGLE")] = 1
	infoZulu.terrains[IT("TERRAIN_TUNDRA")] = -1
	civInfos[IT("CIVILIZATION_ZULU")] = infoZulu
	#################
	infoIndia = CivFlavorInfo(IT("CIVILIZATION_INDIA"))
	infoIndia.setPriority(1)
	infoIndia.features[IT("FEATURE_JUNGLE")] = 1
	infoIndia.terrains[IT("TERRAIN_TUNDRA")] = -1
	civInfos[IT("CIVILIZATION_INDIA")] = infoIndia
	#################

#EAST ASIA
	#################
	infoKhmer = CivFlavorInfo(IT("CIVILIZATION_KHMER"))
	infoKhmer.setPriority(1)
	infoKhmer.boni[IT("BONUS_SILK")] = 1
	infoKhmer.boni[IT("BONUS_IVORY")] = 0.5
	infoKhmer.closeFactor = 1
	infoKhmer.regionSize = 18
	civInfos[IT("CIVILIZATION_KHMER")] = infoKhmer
	#################
	infoChina = CivFlavorInfo(IT("CIVILIZATION_CHINA"))
	infoChina.setPriority(1)
	infoChina.boni[IT("BONUS_SILK")] = 1
	infoChina.closeFactor = 1.5
	infoChina.regionSize = 12
	civInfos[IT("CIVILIZATION_CHINA")] = infoChina
	#################
	infoJapan = CivFlavorInfo(IT("CIVILIZATION_JAPAN"))
	infoJapan.setPriority(1)
	infoJapan.boni[IT("BONUS_SILK")] = 1
	infoJapan.closeFactor = 1.5
	infoJapan.regionSize = 12
	civInfos[IT("CIVILIZATION_JAPAN")] = infoJapan
	#################
	infoKorea = CivFlavorInfo(IT("CIVILIZATION_KOREA"))
	infoKorea.setPriority(1)
	infoKorea.boni[IT("BONUS_SILK")] = 1
	infoKorea.closeFactor = 1.5
	infoKorea.regionSize = 12
	civInfos[IT("CIVILIZATION_KOREA")] = infoKorea
	#################
	infoMongol = CivFlavorInfo(IT("CIVILIZATION_MONGOL"))
	infoMongol.setPriority(1)
	infoMongol.boni[IT("BONUS_SILK")] = 1
	infoMongol.closeFactor = 1
	infoMongol.regionSize = 18
	civInfos[IT("CIVILIZATION_MONGOL")] = infoMongol
	#################

#AMERICAS
	infoAmerica = CivFlavorInfo(IT("CIVILIZATION_AMERICA"))
	infoAmerica.setPriority(2)
	infoAmerica.boni[IT("BONUS_DYE")] = 1
	infoAmerica.boni[IT("BONUS_CRAB")] = 0.5
	infoAmerica.closeFactor = 1
	infoAmerica.regionSize = 20
	civInfos[IT("CIVILIZATION_AMERICA")] = infoAmerica
	#################
	infoNativeAmerica = CivFlavorInfo(IT("CIVILIZATION_NATIVE_AMERICA"))
	infoNativeAmerica.setPriority(2)
	infoNativeAmerica.boni[IT("BONUS_DYE")] = 1
	infoNativeAmerica.closeFactor = 1
	infoNativeAmerica.regionSize = 12
	civInfos[IT("CIVILIZATION_NATIVE_AMERICA")] = infoNativeAmerica
	#################
	infoAztec = CivFlavorInfo(IT("CIVILIZATION_AZTEC"))
	infoAztec.setPriority(2)
	infoAztec.boni[IT("BONUS_DYE")] = 1
	infoAztec.regionSize = 12
	civInfos[IT("CIVILIZATION_AZTEC")] = infoAztec
	#################
	infoInca = CivFlavorInfo(IT("CIVILIZATION_INCA"))
	infoInca.setPriority(2)
	infoInca.boni[IT("BONUS_DYE")] = 1
	infoInca.regionSize = 12
	civInfos[IT("CIVILIZATION_INCA")] = infoInca
	#################
	infoMaya = CivFlavorInfo(IT("CIVILIZATION_MAYA"))
	infoMaya.setPriority(2)
	infoMaya.boni[IT("BONUS_DYE")] = 1
	infoMaya.regionSize = 12
	civInfos[IT("CIVILIZATION_MAYA")] = infoMaya
	#################

	
	debugOut("flavor civs: ")
	debugOut(civInfos.keys())
	
	#reallocation
	NiTextOut("Python: Redistributing starting positions (FfH_script_version) ...")	
	cgc = CyGlobalContext()
	dice = cgc.getGame().getMapRand()
	cymap = CyMap() 
	width = cymap.getGridWidth()
	height = cymap.getGridHeight()
	iPlayers = cgc.getGame().countCivPlayersEverAlive()

	startingPlots = []
	playersLeft = range(iPlayers)
	playersToRemove = []

		
	#creating list of starting plots
	for playerIndex in range(iPlayers):
		pPlayer = cgc.getPlayer(playerIndex)
		pStartPlot = pPlayer.getStartingPlot()	  
		startingPlots.append(pStartPlot)
		

	#reallocation...
	for priorityCount in [1,2,3]:
		debugOut("priority count: " + repr(priorityCount))
		for pRem in playersToRemove:
			playersLeft.remove(pRem)
		playersToRemove = []
		debugOut("players left:")
		debugOut(playersLeft)
		for playerIndex in playersLeft:
			player=cgc.getPlayer(playerIndex)
			civName=player.getCivilizationType()
			debugOut("checking the following civ: " + repr(civName))
			if civName in civInfos.keys() and  civInfos[civName]._priority == priorityCount:
				debugOut("ok found a candidate")
				bestIndex=findBestPlot(civInfos[civName], startingPlots)
				player.setStartingPlot(startingPlots[bestIndex],True)	
				del startingPlots[bestIndex]	
				playersToRemove.append(playerIndex)	#ok some code could be more elegant						
	for pRem in playersToRemove:
		playersLeft.remove(pRem)	

	#remaining players	
	for Index in range(len(playersLeft)):
		player=cgc.getPlayer(playersLeft[Index])
		player.setStartingPlot(startingPlots[Index],True)	


	return None

	
	

################# for reallocation: helper functions & classes below #######################
		



class CivFlavorInfo: # use setter function to change member values with underscore

	def __init__(self, name):
		self._name = name
		self._priority = 2 
		self.terrains = {}
		self.features = {}
		self.boni = {}
		self.plotTypes = {}
		self.regionSize = regionSize()
		self.closeRegionSize = closeRegionSize()
		self.closeFactor = closeFactor()
		self.coastalStartValue = 0
	
	def setPriority(self,priority): # the lower the earlier
		if priority < 1:
			priority = 1
		if priority > 3:
			priority = 3
		self._priority = priority


# returns index of best plot for the civ
def findBestPlot(civInfo, plotList):
	debugOut("finding now best spot for: " + repr(civInfo._name))
	debugOut("civ info:")
	debugOut(civInfo.terrains)
	debugOut(civInfo.plotTypes)
	debugOut(civInfo.features)
	debugOut(civInfo.boni)
	debugOut("coastalStartValue:" + repr(civInfo.coastalStartValue))
	cymap = CyMap() 
	bestIndex=0
	foundMax=0

	for plotIndex in range(len(plotList)):
		startPlot=plotList[plotIndex]
		x0=startPlot.getX()
		y0=startPlot.getY()	
		# print "start tile at ", x0," ", y0
		debugOut("checking start at " +repr(x0) +" " + repr(y0))

		score=0
		if isCoast(startPlot):
			score+=civInfo.coastalStartValue


		for x in range(x0-regionSize(),x0+regionSize()+1):
			for y in range(y0-regionSize(),y0+regionSize()+1):
				plotValue=0
				plot=cymap.plot(x,y)

				terrain = plot.getTerrainType()
				plotType = plot.getPlotType()				
				feature = plot.getFeatureType()				
				bonus = plot.getBonusType(-1)				

				if terrain in civInfo.terrains.keys():
					plotValue+=civInfo.terrains[terrain]
				if plotType in civInfo.plotTypes.keys():
					plotValue+=civInfo.plotTypes[plotType]
				if feature in civInfo.features.keys():
					plotValue+=civInfo.features[feature]
				if bonus in civInfo.boni.keys():
					plotValue+=civInfo.boni[bonus]

				if x >= x0-closeRegionSize() and x <= x0+closeRegionSize() \
and y >= y0-closeRegionSize() and y <= y0+closeRegionSize():
					plotValue*=closeFactor()

				score+=plotValue
						
		debugOut("found here: " + repr(score))
		if foundMax<score:
			foundMax=score
			bestIndex=plotIndex				
	debugOut("found max: " +repr(foundMax))
	
	return bestIndex	
	



# coastal plot?
def isCoast(plot):
	WaterArea = plot.waterArea()
	if not WaterArea.isNone():
		if not WaterArea.isLake():
			return	True
	return False
		

def debugOut(output):
	if FFH_DEBUG:
		print(output)
	return	

