#
#	FILE:	 SmartMap.py
#	VERSION: 2.0
#	AUTHOR:  Doug McCreary (surt@uga.edu)
#	PURPOSE: Allow a finely detailed user control of map generation
#-----------------------------------------------------------------------------
#	Copyright (c) 2005-2006 Doug's Software All rights reserved.
#-----------------------------------------------------------------------------
#

#todo great lakes

# Version History
# 8.7
#     Added 3 great lakes options (few, normal, many).  This will generate a map with a number
#       of large lakes (large enough to typically have ocean tiles).
#     Edge sea and corner sea were reversed in what they generated, fixed.
#     Corrected the usage of cardinal directions, thanks to Sto for noticing.  This turned out to
#       be harmless, but would show up in the xml error log.
#     Fixed bug with generating inland sea maps, thanks DG for noticing.
#     Corrected minor math error in hill placement that was generating more hills than specified.
#     Corrected disagreement between ocean percentage and size of central sea in such maps.
#     Fixed problem with counting number of available sea resource spots that would place too
#       many sea resources on certain styles of maps.
#     Removed all used of CyPlot.isCoastalLand() because this method doesn't do what you'd expect.
#     Forced central sea maps to keep the central sea away from the poles to prevent splitting
#       the map in two.
#     Speed up altitude calculation by simplifying the search for nearby ocean.
#     Speed up wetness calculation by skipping water tiles.
#     Speed up hash lookups by increasing step size.
#     These speed improvements should knock about 10-20% off the total generation time.



##########################################################################
#the imports required for SmartMap or the borrowed fractal generators
import operator
from CvPythonExtensions import *
import CvUtil
import CvMapGeneratorUtil
from CvMapGeneratorUtil import FractalWorld
from CvMapGeneratorUtil import TerrainGenerator
from CvMapGeneratorUtil import FeatureGenerator
import pickle

##########################################################################
#stuff the user may want to configure

#these are the resources you find only in water tiles
resourcesWater = ('BONUS_CLAM', 'BONUS_FISH', 'BONUS_WHALE', 'BONUS_CRAB')
#these resources should be given more often than others
resourcesCommon = ('BONUS_CORN', 'BONUS_COW', 'BONUS_DEER', 'BONUS_PIG', 'BONUS_RICE', 
'BONUS_SHEEP', 'BONUS_WHEAT')
#these resources are important early on, and will place extras unless using one of the 'stricter' options
resourcesImportantEarly = ('BONUS_STONE', 'BONUS_MARBLE', 'BONUS_COPPER', 'BONUS_IRON', 'BONUS_HORSE')
#these are the strategic resources restricted by the 'restricted' settings
resourcesStrategic = ('BONUS_IRON', 'BONUS_OIL', 'BONUS_URANIUM') 

##########################################################################
#global variables	
wrapX = False #is map wrapped on x
wrapY = False #is map wrapped on y
centralPole = False #a central pole means the point at the center of the map is a pole
width = 0 #width of the map in plots
height = 0 #height of the map in plots
mapDimensionSize = 0
#the eight tiles that immediately surround a given tile
cardinalDirs = [(-1,0),(0,-1),(0,1),(1,0)]	
#the eight tiles that immediately surround a given tile
nearDirs = [(-1,-1),(-1,0),(-1,1),(0,-1),(0,1),(1,-1),(1,0),(1,1)]	
#the set of tiles that can be used by a given tile if a city is built there
playerStartCross = [(-1,-1),(-1,0),(-1,1),(0,-1),(0,1),(1,-1),(1,0),(1,1),
                    (-2,-1),(-2,0),(-2,1),(2,-1),(2,0),(2,1),(-1,2),(0,2),(1,2),(-1,-2),(0,-2),(1,-2)]

#a list of lists: of the continents, and all their land tiles
postContinentLists = []	
#remember how many total land tiles we have
totalLandCount = 0
#remember how many coastal tiles there are
totalCoastCount = 0
#pre-computed altitude of all plots
preComputedAltitude = []
#which map size was selected
sizeSelected = 0
#pre-computed wetness of all plots
preComputedWetness = []

#spiral search support vars
spiralCenterX = 0
spiralCenterY = 0
spiralStep = 0
spiralSign = 0
spiralX = 0
spiralY = 0
spiralQueue = []

#cache the random menu choices so we don't give inconsistent answers
cachedMenuChoices = []

#type used to fill empty tiles, allowing for ocean filled or land filled maps
fillType = PlotTypes.PLOT_OCEAN
landPlotTypes = [PlotTypes.PLOT_LAND,PlotTypes.PLOT_HILLS,PlotTypes.PLOT_PEAK]

##########################################################################
#Step 1: the map grid size is determined. your map will contain 4 tiles
# for every grid size, so an 8x4 grid size is 32x16 tiles
def getGridSize(argsList):
	if (argsList[0] == -1): # (-1,) is passed to function on loads
		return (-1,-1)
	beforeStepOne()
	OutputMessage("Python: SmartMap: Step 1 Setting Map Size")
	global wrapX
	global wrapY
	gc = CyGlobalContext()
	dice = gc.getGame().getMapRand()
	
	
	#scale the world size to match the area supplied by typical maps
	#this data is (map area, minDimension)
	#         Areas     Duel Tiny Small Standard Large Huge
	#         SmartMap: 80   140  220   360      560   860
	#         Standard: 60   104  160   273      416   640
	#note that a 60x60 which is bigger than my 1Gig machine can handle 
	#would be 3600, or 3x the area of this huge
	#the min dimension i've chosen is rounddown(sqrt((area / pi)))
	grid_infos = {
		WorldSizeTypes.WORLDSIZE_DUEL:		(80,5),
		WorldSizeTypes.WORLDSIZE_TINY:		(140,6),
		WorldSizeTypes.WORLDSIZE_SMALL:		(220,8),
		WorldSizeTypes.WORLDSIZE_STANDARD:	(360,10),
		WorldSizeTypes.WORLDSIZE_LARGE:		(560,13),
		WorldSizeTypes.WORLDSIZE_HUGE:		(860,16)
	}
	
	
	#extract the right world size
	[eWorldSize] = argsList
	global sizeSelected
	sizeSelected = eWorldSize
	print "    [eWorldSize]",eWorldSize	

	mapArea, minDimension = grid_infos[eWorldSize]
	maxDimension = int(mapArea / minDimension)
	print "    mapArea",mapArea	
	print "    minDimension",minDimension	
	print "    maxDimension",maxDimension	

	#calculate a random width that will fit the world size, base width is minDimension, with up to extra minDimension
	calcwidth = minDimension + dice.get(maxDimension - minDimension, "Python: SmartMap: Dice: Random Continent X")
	#bias the width or height toward the wrap direction if the map wraps in only one direction
	if wrapX and not wrapY:
		calcwidth += int(calcwidth/5)
	if not wrapX and wrapY:
		calcwidth -= int(calcwidth/5)
	if calcwidth > maxDimension:
		calcwidth = maxDimension
	if calcwidth < minDimension:
		calcwidth = minDimension
	
	calcheight = int(mapArea / calcwidth)
	
	if calcheight > maxDimension:
		calcheight = maxDimension
	if calcheight < minDimension:
		calcheight = minDimension
		
	#make sure the map isn't too small		
	calculatedArea = calcwidth * calcheight
	while (calculatedArea < int((mapArea*86)/100)):
		if calcwidth < maxDimension:
			calcwidth += 1
		if calcheight < maxDimension:
			calcheight += 1
		calculatedArea = calcwidth * calcheight

	#make sure the map isn't too big	
	calculatedArea = calcwidth * calcheight
	while (calculatedArea > int((mapArea*114)/100)):
		if calcwidth > minDimension:
			calcwidth -= 1
		if calcheight > minDimension:
			calcheight -= 1
		calculatedArea = calcwidth * calcheight
	
	#finally, recheck to make sure our width and height are legal
	if calcwidth > maxDimension:
		calcwidth = maxDimension
	if calcwidth < minDimension:
		calcwidth = minDimension
	if calcheight > maxDimension:
		calcheight = maxDimension
	if calcheight < minDimension:
		calcheight = minDimension
		
	#allow for overrides of the width and height
	overrideWidth = getSelectedMapValue("Override Width:")
	overrideHeight = getSelectedMapValue("Override Height:")
	
	if overrideWidth > 0:
		calcwidth = overrideWidth
		print "    override width",overrideWidth
	if overrideHeight > 0:
		calcheight = overrideHeight
		print "    override height",overrideHeight

	if overrideWidth == -2:
		calcwidth = 4 + dice.get(12, "Python: SmartMap: Dice: Random Width 16-64")
		print "    override width",overrideWidth
	if overrideHeight == -2:
		calcheight = 4 + dice.get(12, "Python: SmartMap: Dice: Random Height 16-64")
		print "    override height",overrideHeight
	if overrideWidth == -3:
		calcwidth = 8 + dice.get(24, "Python: SmartMap: Dice: Random Width 32-128")
		print "    override width",overrideWidth
	if overrideHeight == -3:
		calcheight = 8 + dice.get(24, "Python: SmartMap: Dice: Random Height 32-128")
		print "    override height",overrideHeight
	if overrideWidth == -4:
		calcwidth = 16 + dice.get(48, "Python: SmartMap: Dice: Random Width 64-256")
		print "    override width",overrideWidth
	if overrideHeight == -4:
		calcheight = 16 + dice.get(48, "Python: SmartMap: Dice: Random Height 64-256")
		print "    override height",overrideHeight
	
	#return the computed width and height
	print "    final width",calcwidth,"height",calcheight
	OutputMessage("Python: SmartMap: Step 1 Setting Map Size: complete")
	return (calcwidth,calcheight)

##########################################################################
#Step 1a: answer back function for deciding whether or not to wrap on x
#You must override neither or both, and SmartMap must do so because it uses
#a user-selected option to control map wrapping
def getWrapX():
	lwrapX, lwrapY, centralPole = getSelectedMapValue("Wrap:")
	print "    getWrapX: wrapX", lwrapX, "wrapY", lwrapY
	return lwrapX 

##########################################################################
#Step 1b: answer back function for deciding whether or not to wrap on y
#You must override neither or both, and SmartMap must do so because it uses
#a user-selected option to control map wrapping
def getWrapY():
	lwrapX, lwrapY, centralPole = getSelectedMapValue("Wrap:")
	print "    getWrapY: wrapX", lwrapX, "wrapY", lwrapY
	return lwrapY 

##########################################################################
#Step 2B: lay out the plot types, which is a sort of broad view of the map tiles
#basically telling you if you are looking at ocean, land, hill, or mountain
def generatePlotTypes():
	beforeStepTwo()
	global wrapX
	global wrapY
	global width
	global height
	global fillType
	global sizeSelected

	OutputMessage("Python: SmartMap: Step 2 Setting Plot Types")

	### Configuration
	print "    in step two width",width,"height",height,"wrapxy",wrapX, wrapY
	gc = CyGlobalContext()
	cymap = CyMap()
	dice = gc.getGame().getMapRand()
	widthLimit = width-1 #range of array is 0..width-1
	heightLimit = height-1 #range of array is 0..height-1
	players = gc.getGame().countCivPlayersEverAlive()
	
	#helpful information
	halfWidth = int(width/2)
	halfHeight = int(height/2)
	
	#get the user entered climate info
	climate = cymap.getClimate()
	climateArid = gc.getInfoTypeForString("CLIMATE_ARID")
	climateRocky = gc.getInfoTypeForString("CLIMATE_ROCKY")
	
	# Sea Level adjustment (from user input), transformed into width of sea separating continents
	sea = gc.getSeaLevelInfo(cymap.getSeaLevel()).getSeaLevelChange()
	seaWidth = 4 #default width for ocean separation of continents
	if (sea > 0):
		seaWidth = 7 #high sea level means extra separation
	if (sea < 0):
		seaWidth = 0 #low sea level means continents are permitted to merge
	overrideSeparation = getSelectedMapValue("Override Separation:")
	if overrideSeparation >= 0:
		seaWidth = overrideSeparation
	print "    seawidth",seaWidth

	#Number of continents
	singleTiles = False
	continentCount = getSelectedMapValue("Continents:")
	if continentCount == -1:
		continentCount = 1 + dice.get(6, "Python: SmartMap: Dice: Random Continent Count 1-6") 
	elif continentCount == -2:
		continentCount = 2 + dice.get(4, "Python: SmartMap: Dice: Random Continent Count 2-5") 
	elif continentCount == -3:
		continentCount = 3 + dice.get(6, "Python: SmartMap: Dice: Random Continent Count 3-8") 
	elif continentCount == -4:
		continentCount = 4 + dice.get(8, "Python: SmartMap: Dice: Random Continent Count 4-12") 
	elif continentCount == -5:
		continentCount = 2 + dice.get(6, "Python: SmartMap: Dice: Random Continent Count 2-7") 
	elif continentCount == -6:
		continentCount = 1 + dice.get(12, "Python: SmartMap: Dice: Random Continent Count 1-12") 
	elif continentCount == -7:
		continentCount = players
	elif continentCount == -8:
		continentCount = int(players/2)
	elif continentCount == -9:
		continentCount = int(players/3)
	elif continentCount == -10:
		continentCount = int(players/4)
	elif continentCount == -11:
		continentCount = players * 2
	elif continentCount == -12:
		continentCount = players * 3
	elif continentCount == -13:
		continentCount = players * 4
	elif continentCount == -20:
		singleTiles = True
		continentCount = int(players/2)
	if continentCount < 0:
		continentCount = 1
	print "    continentCount",continentCount

	#find out the land style
	(roundFactor,fragPPT) = getSelectedMapValue("Land Style:")
	print "    roundFactor", roundFactor
	print "    fragPPT", fragPPT

	#figure out how much of the map is supposed to be ocean
	noLakes = False
	fillType = PlotTypes.PLOT_OCEAN	
	percentOcean = getSelectedMapValue("Ocean:")
	if percentOcean == -1:
		percentOcean = 50 + dice.get(26, "Python: SmartMap: Dice: Random Ocean %")
	if percentOcean == -2:
		percentOcean = 80 + dice.get(16, "Python: SmartMap: Dice: Random Ocean %")
	if percentOcean == -3:
		percentOcean = 65 + dice.get(21, "Python: SmartMap: Dice: Random Ocean %")
	if climate == climateArid:
		percentOcean -= 10
	if climate == climateRocky:
		percentOcean -= 5
	if percentOcean == -4:
		fillType = PlotTypes.PLOT_LAND
		percentOcean = 1
	if percentOcean == -5:
		fillType = PlotTypes.PLOT_LAND
		percentOcean = 1
		noLakes = True
	
	cornerSea = False
	edgeSea = False
	centralSea = False
	greatLakes = False
	numLakes = -1
		
	if percentOcean == -6:
		fillType = PlotTypes.PLOT_LAND
		percentOcean = 14
		noLakes = False
		centralSea = True
	if percentOcean == -7:
		fillType = PlotTypes.PLOT_LAND
		percentOcean = 14
		noLakes = False
		cornerSea = True
	if percentOcean == -8:
		fillType = PlotTypes.PLOT_LAND
		percentOcean = 14
		noLakes = False
		edgeSea = True
	if percentOcean == -9:
		fillType = PlotTypes.PLOT_LAND
		percentOcean = 14
		noLakes = False
		centralSea = True
		edgeSea = True
		cornerSea = True

	if percentOcean == -10:
		fillType = PlotTypes.PLOT_LAND
		percentOcean = 12
		numLakes = 4
		greatLakes = True
	if percentOcean == -11:
		fillType = PlotTypes.PLOT_LAND
		percentOcean = 14
		numLakes = 8
		greatLakes = True
	if percentOcean == -12:
		fillType = PlotTypes.PLOT_LAND
		percentOcean = 16
		numLakes = 12
		greatLakes = True
		
		
	print "    percentOcean",percentOcean

	percentHills = getSelectedMapValue("Hills:")
	if percentHills == -1:
		percentHills = 10 + dice.get(8, "Python: SmartMap: Dice: few Random Hills %") 
	if percentHills == -2:
		percentHills = 18 + dice.get(20, "Python: SmartMap: Dice: many Random Hills %") 
	if percentHills == -3:
		percentHills = 14 + dice.get(14, "Python: SmartMap: Dice: normal Random Hills %")
	if climate == climateRocky:
		percentHills += 6
	print "    percentHills",percentHills

	percentPeaks = getSelectedMapValue("Peaks:")
	noBlockPeaks = False
	if percentPeaks == -1:
		percentPeaks = 2 + dice.get(6, "Python: SmartMap: Dice: Random Peaks %") 
	if percentPeaks == -2:
		percentPeaks = 6 + dice.get(10, "Python: SmartMap: Dice: Random Peaks %") 
	if percentPeaks == -3:
		percentPeaks = 4 + dice.get(8, "Python: SmartMap: Dice: Random Peaks %") 
	if percentPeaks == -4:
		percentPeaks = 2 + dice.get(6, "Python: SmartMap: Dice: Random Peaks %") 
		noBlockPeaks = True
	if percentPeaks == -5:
		percentPeaks = 6 + dice.get(10, "Python: SmartMap: Dice: Random Peaks %") 
		noBlockPeaks = True
	if percentPeaks == -6:
		percentPeaks = 4 + dice.get(8, "Python: SmartMap: Dice: Random Peaks %") 
		noBlockPeaks = True
	if climate == climateRocky:
		percentPeaks += 4
	print "    percentPeaks",percentPeaks
	
	(riverPct,lakePct) = getSelectedMapValue("Rivers/Lakes:")
	print "    riverPct",riverPct
	print "    lakePct",lakePct

	if percentOcean == -4:
		lakePct += 100
		lakePct *= 4
	
	### Data
	plotTypes = [fillType] * (width*height)
	ownerContinents = [-1] * (width*height)
	continentLists = []
	canBeLandLists = []
	restoreLake = 0

	#total number of useful plots on the map, this factors in staying away from the edges
	usefulPlotCount = int((width-4)*(height-4))
	extraLandFactor = 104
	if fillType == PlotTypes.PLOT_LAND:
		usefulPlotCount = width*height
		extraLandFactor = 100

	expectedLandPlotCount = int(((100-percentOcean) * usefulPlotCount) / 100)
	landToPlace = int((extraLandFactor*expectedLandPlotCount)/100)
	fragLandCount = int((landToPlace*fragPPT)/1000) 
	landToPlace -= fragLandCount
	landPlaced = 0
	if continentCount == 0:
		fragLandCount += landToPlace
		landToPlace = 0
		continentLists = []
		continentLists.append([])
		canBeLandLists = []
		canBeLandLists.append([])
		
	setupSpaceHash()
	print "    landToPlace",landToPlace,"fragLandCount",fragLandCount,"expectedLandPlotCount",expectedLandPlotCount,"from total tiles:", usefulPlotCount
	
	### Generate continents
	#make a list for each continent
	continent = 0
	while (continent < continentCount):
		#print "    add Continent List",continent
		continent += 1
		continentLists.append([])
		canBeLandLists.append([])
		
	if fillType != PlotTypes.PLOT_LAND:
		#create a starting plot for each continent
		tries = 0
		extraSpace = int( ((width*height)**0.5) / (continentCount+1))
		for contIndex in range(len(continentLists)):
			continentList = continentLists[contIndex]
			conflict = True
			while (tries < 5000 and conflict):
				tries += 1
				initX = int(width/8) + dice.get((width-int(width/8)), "Python: SmartMap: Dice: Random Continent X")
				initY = int(height/8) + dice.get((height-int(height/8)), "Python: SmartMap: Dice: Random Continent Y")
				#always grow a pangea from the center
				if continentCount == 1 and tries < 2:
					initX = int(width/2)
					initY = int(height/2)
				
				conflict = False
				for otherList in continentLists:
					if continentList != otherList:
						for otherPlot in otherList:
							diffX,diffY = getWrapDistanceXY((initX,initY),otherPlot)
							#continents must start with a certain minimum spacing
							if diffX <= seaWidth+5+extraSpace and diffY <= seaWidth+5+extraSpace:
								conflict = True
								break
				
				extraSpace -= 1
				if extraSpace < 0:
					extraSpace = 0
				if not conflict:				
					indexA = getTileIndexAtXY(initX,initY)
					if plotTypes[indexA] == PlotTypes.PLOT_OCEAN:
						continentList.append((initX,initY))
						plotTypes[indexA] = PlotTypes.PLOT_LAND
						spaceHashItem(initX,initY,contIndex)
						ownerContinents[indexA] = contIndex
						print "    initial tile addition",continentList,indexA
						for nearDir in nearDirs:
							usablePlotX,usablePlotY = getTileCoordXYWithWrap((initX,initY),nearDir)
							usableIndex = getTileIndexAtXY(usablePlotX,usablePlotY)
							if plotTypes[usableIndex] == PlotTypes.PLOT_OCEAN:
								canBeLandLists[contIndex].append((usablePlotX,usablePlotY))
								
						print "    canbeland",canBeLandLists[contIndex]
						break
							
		#remove any continent that couldn't be placed
		done = False
		while not done:
			done = True
			indexLength = len(continentLists)
			for indexB in range(0,indexLength):
				continentList = continentLists[indexB]
				if len(continentList) <= 0:
					done = False
					del continentLists[indexB]
					del canBeLandLists[indexB]
					break
		print "    placed continents", len(continentLists)
		
		#generate land plots

		#tries is used to ensure that this can't loop forever
		tries = 0
		while (landPlaced < landToPlace and tries < 7 * landToPlace):
			for contIndex in range(len(continentLists)):
				#print "cont",contIndex,"canbeland",canBeLandLists[contIndex]
				tries+=1
				if canBeLandLists[contIndex] == []:
					continue
				indexListA = dice.get(len(canBeLandLists[contIndex]), "Python: SmartMap: Dice: Random Continent listIndex backup")
				newX, newY = canBeLandLists[contIndex][indexListA]
				index = getTileIndexAtXY(newX,newY)
				if plotTypes[index] != PlotTypes.PLOT_OCEAN:
					del canBeLandLists[contIndex][indexListA]
					continue
					
				#print "cont",contIndex,"canbeland",canBeLandLists[contIndex],"considering",newX,newY
				dropProb = 0
				anyMatch = False
				
				#low probability of allowing approach to x-pole if not wrapping
				if not anyMatch: #I know this is always going to pass, but I do it for code symmetry
					if not wrapX:
						xFromEdge = newX - 2
						if newX > halfWidth:
							xFromEdge = (widthLimit - newX) - 2
						if xFromEdge < 20:
							edgeVal = dice.get(10000, "Python: SmartMap: Dice: Random Continent x-edge bending")
							if edgeVal < (25 * (20-xFromEdge) * (20-xFromEdge)):
								dropProb = (10 * (20-xFromEdge) * (20-xFromEdge))
								#print "dropProb",dropProb
								anyMatch = True
								#print "    drop due to x edge",newX,newY,edgeVal,xFromEdge,width
	
				#low probability of allowing approach to y-pole if not wrapping
				if not anyMatch:
					if not wrapY:
						yFromEdge = newY - 2
						if newY > halfHeight:
							yFromEdge = (heightLimit - newY) - 2
						if yFromEdge < 20:
							edgeVal = dice.get(10000, "Python: SmartMap: Dice: Random Continent y-edge bending")
							edgeProb = (25 * (20-yFromEdge) * (20-yFromEdge))
							#print "yFromEdge",yFromEdge,"edgeProb",edgeProb,"edgeVal",edgeVal
							if edgeVal < edgeProb:
								dropProb = (10 * (20-yFromEdge) * (20-yFromEdge))
								#print "dropProb",dropProb
								anyMatch = True
								#print "    drop due to y edge",newX,newY,edgeVal,yFromEdge,height
		
				#make sure this ocean plot doesn't touch a different continent's land
				if not anyMatch and len(continentLists) > 1 and seaWidth > 0:
					newXY = (newX,newY)
					#another method: spatial hash
					targets = listNearSpaceHash(newX, newY, seaWidth+2)
					#print "    TARGETS",len(targets)
					
					for target in targets:
						otherX, otherY, otherCont = target
						if otherCont == contIndex:
							continue
						otherIndex = getTileIndexAtXY(otherX,otherY)
						if plotTypes[otherIndex] == PlotTypes.PLOT_OCEAN:
							continue
						if ownerContinents[otherIndex] == contIndex:
							continue
						#print "found land on other continent"
						diffX, diffY = getWrapDistanceXY(newXY, (otherX,otherY))
						#this is much faster than doing an iteration through the dirs
						if (diffX+diffY) <= seaWidth or (diffX <= seaWidth and diffY <= seaWidth):
							anyMatch = True
							#print "    drop due to seawidth",newX,newY,edgeVal,"too close to cont",otherContIndex,otherList,otherPlot
							dropProb = 10001
							break
							
						#this is much faster than doing an iteration through the dirs
						separation = int(((diffX*diffX) + (diffY*diffY))**0.5)
						beyondSeaWidthTest = 2000 - (100 * (separation - seaWidth) * (separation - seaWidth))
						edgeVal = dice.get(10000, "Python: SmartMap: Dice: Random Continent sea edge bending")
						#print "near test edgeval",edgeVal,"edgeProb",edgeProb,"diffSum",diffSum,"nearVal",seaWidthNearVal
						if edgeVal < beyondSeaWidthTest:
							dropProb = max(dropProb, int(beyondSeaWidthTest/2))
							anyMatch = True
							if dropProb >= 10001:
								break
							#print "    drop due to seawidth margin",newX,newY,edgeVal,beyondSeaWidthTest,"too close to cont",otherIndex,"xy",otherX,otherY
							continue
								
				if anyMatch:
					drop = dice.get(10000, "Python: SmartMap: Dice: random canbeland drop")
					if drop < dropProb:
						del canBeLandLists[contIndex][indexListA]
					continue
				
				#this land plot is not going to touch any other continent so place				
				#print "    place land at x",newX,"y",newY,"width",width,"height",height,"index",index
				if (index < 0 or index > (width*height)-1):
					print "    possibly illegal place land at x",newX,"y",newY,"width",width,"height",height,"index",index
				plotTypes[index] = PlotTypes.PLOT_LAND
				ownerContinents[index] = contIndex
				continentLists[contIndex].append((newX,newY))
				spaceHashItem(newX, newY, contIndex)
				del canBeLandLists[contIndex][indexListA]
				landPlaced += 1
				roundChance = 1001 - roundFactor
				for nearDir in nearDirs:
					usablePlotX,usablePlotY = getTileCoordXYWithWrap((newX,newY),nearDir)
					usableIndex = getTileIndexAtXY(usablePlotX,usablePlotY)
					if plotTypes[usableIndex] == PlotTypes.PLOT_OCEAN and not ((usablePlotX,usablePlotY) in canBeLandLists[contIndex]):
						if ((usablePlotX,usablePlotY) in canBeLandLists[contIndex]):
							roundChance -= roundFactor
						roundTest = dice.get(1000, "Python: SmartMap: Dice: Random Continent growth bending")
						if roundTest < roundChance: 
							canBeLandLists[contIndex].append((usablePlotX,usablePlotY))
				#print "    land tile addition",continentList
		print "    placed",landPlaced," initial land"
			
		#fill in any one tile lakes caused by wierd growth patterns
		print "    remove single tile lakes"
		for x in range(width):
			for y in range(height):
				countNearLand = 0
				ownerCont = 0
				tileIndex = getTileIndexAtXY(x,y)
				if plotTypes[tileIndex] != PlotTypes.PLOT_OCEAN:
					continue
				for nearDir in cardinalDirs:
					ox,oy = getTileCoordXYWithWrap((x,y),nearDir)
					oIndex = getTileIndexAtXY(ox,oy)
					oType = plotTypes[oIndex]
					if ((ox != x) or (oy != y)) and oType == PlotTypes.PLOT_LAND:
						countNearLand += 1
						ownerCont = ownerContinents[oIndex]
				if countNearLand == 4:
					plotTypes[tileIndex] = PlotTypes.PLOT_LAND
					restoreLake += 1
					ownerContinents[tileIndex] = ownerCont
					continentLists[ownerCont].append((x,y))
					continue
				countNearLand = 0
				for nearDir in nearDirs:
					ox,oy = getTileCoordXYWithWrap((x,y),nearDir)
					oIndex = getTileIndexAtXY(ox,oy)
					oType = plotTypes[oIndex]
					if ((ox != x) or (oy != y)) and oType == PlotTypes.PLOT_LAND:
						countNearLand += 1
				if countNearLand >= 7:
					plotTypes[tileIndex] = PlotTypes.PLOT_LAND
					restoreLake += 1
					ownerContinents[tileIndex] = ownerCont
					continentLists[ownerCont].append((x,y))
					continue
		print "    removed",restoreLake,"single tile lakes"

		#place some random land fragments
		print "    try to place",fragLandCount,"land fragment tiles"
		#tries is used to ensure that this can't loop forever
		tries=0
		fragPlaced = 0
		usedLakeLocs = []
		fragCount = 0
		
		while (fragPlaced < fragLandCount and tries < 16 * fragLandCount):
			tries+=1
			fragX = dice.get(width, "Python: SmartMap: Dice: Random land Fragment X")
			fragY = dice.get(height, "Python: SmartMap: Dice: Random land Fragment Y")
			if tooCloseToPole(fragX,fragY,6):
				continue
			fragIndex = getTileIndexAtXY(fragX,fragY)
			if plotTypes[fragIndex] == PlotTypes.PLOT_LAND:
				continue
			
			#find nearest land plot, and consider this frag as part of that land
			spiralSearchSetup(fragX,fragY)
			landFound = False
			searchTries = 0
			listToAddIndex = 0
			while not landFound and searchTries < 400:
				searchTries += 1
				nearX, nearY = spiralSearchNextStep()
				nearIndex = getTileIndexAtXY(nearX,nearY)
				if ownerContinents[nearIndex] > -1:
					listToAddIndex = ownerContinents[nearIndex]
					landFound = True
			spiralSearchSetup(fragX,fragY)
			
			#if we didn't find a nearby land, assign to a random continent
			if not landFound:		
				listToAddIndex = dice.get(len(continentLists), "Python: SmartMap: Dice: Random Continent frag growth cont assignment")	
			continentLists[listToAddIndex].append((fragX,fragY))
			plotTypes[fragIndex] = PlotTypes.PLOT_LAND
			ownerContinents[fragIndex] = listToAddIndex
			fragCount += 1
			fragPlaced += 1
			fragSize = 1
			growFrag = dice.get(100, "Python: SmartMap: Dice: Random Continent frag growth")
			gx,gy = fragX,fragY
			passes = 0
			while growFrag < 99+int((fragPlaced*sizeSelected)/20)-int(passes/4) and fragPlaced < fragLandCount:
				passes+=1
				growFrag = dice.get(100, "Python: SmartMap: Dice: Random Continent frag growth")
				dirIndex = dice.get(len(nearDirs), "Python: SmartMap: Dice: Random Continent lake growth")
				gx, gy = getTileCoordXYWithWrap((gx,gy), nearDirs[dirIndex])
				if tooCloseToPole(gx,gy,4):
					continue
				gIndex = getTileIndexAtXY(gx,gy)
				if plotTypes[gIndex] != PlotTypes.PLOT_OCEAN:
					continue
				tries+=1
					
				#make sure this ocean plot doesn't touch a different continent's land
				newXY = (gx,gy)
				#alternative (maybe faster) method: see if any plot in -seawidth to +seawidth in x and y
				#is land and also belongs to another continent, then give further consideration
				anyMatch = False
				
				#don't have to do this logic if any other thing matched (nothing currently implemented though),
				# or if there is only one continent, or if the user didn't want a seawidth separation 
				if not anyMatch and len(continentLists) > 1 and seaWidth > 0:
					targets = listNearSpaceHash(newX, newY, seaWidth+2)
					#print "    TARGETS",len(targets)
					
					for target in targets:
						otherX, otherY, otherCont = target
						if otherCont == contIndex:
							continue
						otherIndex = getTileIndexAtXY(otherX,otherY)
						if plotTypes[otherIndex] == PlotTypes.PLOT_OCEAN:
							continue
						if ownerContinents[otherIndex] == contIndex:
							continue
						#print "found land on other continent"
						diffX, diffY = getWrapDistanceXY(newXY, (otherX,otherY))
						#this is much faster than doing an iteration through the dirs
						if (diffX+diffY) <= seaWidth or (diffX <= seaWidth and diffY <= seaWidth):
							anyMatch = True
							#print "    drop due to seawidth",newX,newY,edgeVal,"too close to cont",otherContIndex,otherList,otherPlot
							dropProb = 10001
							break
							
				if anyMatch:
					continue
					
				plotTypes[gIndex] = PlotTypes.PLOT_LAND
				ownerContinents[gIndex] = listToAddIndex
				fragPlaced += 1
				fragSize += 1
				continentLists[listToAddIndex].append((gx,gy))
			
		print "    placed",fragPlaced,"frag tiles","out of required",fragLandCount,"in",fragCount,"chunks in",tries,"tries"

	else:
		continentLists = []
		continentLists.append([])
		for x in range(width):
			for y in range(height):
				continentLists[0].append((x,y))

	#generate hills on land plots
	allLandList = getPlotsOfTypes(plotTypes, [PlotTypes.PLOT_LAND])
	hillPlotCount = int((percentHills * len(allLandList)) / 100)
	print "    hillPlotCount",hillPlotCount,"from land tiles:", len(allLandList)
	firstPlotCount = int((3*hillPlotCount)/4)
	
	#first portion of hills are placed randomly
	tries=0
	#tries is used to ensure that this can't loop forever
	hillPlaced = 0
	hillsSoFar = []
	print "      place initial hills"
	while (hillPlaced < firstPlotCount and tries < 5 * firstPlotCount):
		tries+=1
		indexListB = dice.get(len(allLandList), "Python: SmartMap: Dice: Random Continent listIndex")
		existingX, existingY = allLandList[indexListB]
		if tooCloseToPole(existingX, existingY, 2):
			continue
		index = getTileIndexAtXY(existingX,existingY)
		if plotTypes[index] != PlotTypes.PLOT_LAND:
			continue
		
		#first XXX hills are required to be well separated
		if tries < 100:
			tooClose = False
			for existingHill in hillsSoFar:
				distX,distY = getWrapDistanceXY(existingHill,(existingX, existingY))
				if (distX+distY) < 4:
					tooClose = True
					break
			if tooClose:
				continue
				
		plotTypes[index] = PlotTypes.PLOT_HILLS
		hillPlaced += 1
		hillsSoFar.append((existingX, existingY))
	print "      placed",hillPlaced," initial hills"
		

	#second portion of hills must touch a hill
	tries=0
	#tries is used to ensure that this can't loop forever
	secondPlotCount = int(1*hillPlotCount/4)
	secondHillPlaced = 0
	print "      place neighbor hills"
	while (secondHillPlaced < secondPlotCount and tries < 5 * secondPlotCount):
		tries+=1
		indexListC = dice.get(len(allLandList), "Python: SmartMap: Dice: Random Continent listIndex")
		existingX, existingY = allLandList[indexListC]
		if tooCloseToPole(existingX, existingY, 2):
			continue
		existingIndex = getTileIndexAtXY(existingX, existingY)
		if plotTypes[existingIndex] != PlotTypes.PLOT_LAND:
			continue
		for nearDir in nearDirs:
			testX,testY = getTileCoordXYWithWrap((existingX,existingY),nearDir)
			if not tooCloseToPole(testX, testY, 2):
				testIndex = getTileIndexAtXY(testX,testY)
				if plotTypes[testIndex] == PlotTypes.PLOT_HILLS:
					plotTypes[existingIndex] = PlotTypes.PLOT_HILLS
					secondHillPlaced += 1
					break #don't have to look any further
	print "      placed",secondHillPlaced," neighbor hills"


	totalHillsPlaced = hillPlaced + secondHillPlaced
	print "    placed",totalHillsPlaced,"hills out of max",hillPlotCount
	
	#generate peaks near hills
	peakPlotCount = int((percentPeaks) * len(allLandList) / 100)
	print "    peakPlotCount",peakPlotCount,"from land tiles:", len(allLandList)

	#peaks must touch a hill
	tries=0
	#tries is used to ensure that this can't loop forever
	peakPlaced = 0
	while (peakPlaced < peakPlotCount and tries < 6 * peakPlotCount):
		tries+=1
		indexListD = dice.get(len(allLandList), "Python: SmartMap: Dice: Random Continent listIndex")
		existingX, existingY = allLandList[indexListD]
		if tooCloseToPole(existingX, existingY, 2):
			continue
		existingIndex = getTileIndexAtXY(existingX,existingY)
		if plotTypes[existingIndex] != PlotTypes.PLOT_LAND and plotTypes[existingIndex] != PlotTypes.PLOT_HILLS:
			continue
		for nearDir in nearDirs:
			testX, testY = getTileCoordXYWithWrap((existingX,existingY),nearDir)
			if not tooCloseToPole(testX,testY,2):
				testIndex = getTileIndexAtXY(testX,testY)
				if plotTypes[testIndex] == PlotTypes.PLOT_HILLS: 
					plotTypes[existingIndex] = PlotTypes.PLOT_PEAK
					peakPlaced += 1
					break #don't have to consider any more dirs
	#				print "    land tile addition",continentList
	print "    placed",peakPlaced,"peaks out of max",peakPlotCount

	#place some random lakes
	lakePlotCount = int((1+int(sizeSelected)) * len(allLandList) / 600)
	lakePlotCount = 1+int((lakePlotCount*lakePct)/100)
	lakePlotCount += int(restoreLake/5)
	
	if len(allLandList) - lakePlotCount < expectedLandPlotCount:
		lakePlotCount = len(allLandList) - expectedLandPlotCount
	
	print "    try to place",lakePlotCount,"lake tiles"
	#tries is used to ensure that this can't loop forever
	tries=0
	lakePlaced = 0
	usedLakeLocs = []
	lakeCount = 0
	largestLake = 0
	if noLakes:
		lakePlotCount = 0
	while (lakePlaced < lakePlotCount and tries < 8 * lakePlotCount):
		tries+=1
		for continentList in continentLists:
			indexListE = dice.get(len(continentList), "Python: SmartMap: Dice: Random Continent listIndex")
			existingX, existingY = continentList[indexListE]
			if tooCloseToPole(existingX,existingY,4):
				continue
			existingIndex = getTileIndexAtXY(existingX,existingY)
			if plotTypes[existingIndex] == PlotTypes.PLOT_LAND:
				#is there is a lake too close already
				tooClose = False
				for usedLakeLoc in usedLakeLocs:
					distX,distY = getWrapDistanceXY((existingX, existingY),usedLakeLoc)
					if distX < 3 and distY < 3:
						tooClose = True
						break
				if tooClose:
					continue
				
				nearHillPeak = False
				allLand = True
				for nearDir in nearDirs:
					testX, testY = getTileCoordXYWithWrap((existingX,existingY),nearDir)
					if not tooCloseToPole(testX,testY,2):
						testIndex = getTileIndexAtXY(testX,testY)
						if plotTypes[testIndex] == PlotTypes.PLOT_OCEAN:
							allLand = False
							break
						if plotTypes[testIndex] == PlotTypes.PLOT_HILLS or plotTypes[testIndex] == PlotTypes.PLOT_PEAK:
							nearHillPeak = True
							break
				if allLand and nearHillPeak:
					plotTypes[existingIndex] = PlotTypes.PLOT_OCEAN
					lakeSize = 1
					lakePlaced += 1
					lakeCount += 1
					if lakeSize > largestLake:
						largestLake = lakeSize
					growLake = dice.get(100, "Python: SmartMap: Dice: Random Continent lake growth")
					lakeList = [(existingX,existingY)]
					usedLakeLocs.append((existingX,existingY))
					gx,gy = existingX,existingY
					passes = 0
					while growLake < (55 + int(lakePlaced/5) + (2*sizeSelected) - int(passes/3)) and lakePlaced < lakePlotCount:
						passes += 1
						growLake = dice.get(100, "Python: SmartMap: Dice: Random Continent lake growth")
						dirIndex = dice.get(len(cardinalDirs), "Python: SmartMap: Dice: Random Continent lake growth")
						gx, gy = getTileCoordXYWithWrap((gx,gy), cardinalDirs[dirIndex])
						if tooCloseToPole(gx,gy,4):
							continue
						gIndex = getTileIndexAtXY(gx,gy)
						if plotTypes[gIndex] != PlotTypes.PLOT_LAND:
							continue
						nearLandCount = 0
						for nearDir in nearDirs:
							testX, testY = getTileCoordXYWithWrap((gx,gy),nearDir)
							if not tooCloseToPole(testX,testY,2):
								testIndex = getTileIndexAtXY(testX,testY)
								if plotTypes[testIndex] != PlotTypes.PLOT_OCEAN:
									nearLandCount += 1
								elif (testX,testY) in lakeList:
									nearLandCount += 1
						#only grow if we are touching all land or existing parts of lake
						if nearLandCount == 8:
							plotTypes[gIndex] = PlotTypes.PLOT_OCEAN
							lakeSize += 1
							lakePlaced += 1
							lakeList.append((gx,gy))
							usedLakeLocs.append((gx,gy))
							if lakeSize > largestLake:
								largestLake = lakeSize
							#print "    lake",lakePlaced,"grew to size", lakeSize
							
	print "    placed",lakePlaced,"lake tiles","out of required",lakePlotCount,"in",lakeCount,"lakes, with largest",largestLake

	#central or other sea option placed here
	if centralSea or cornerSea or edgeSea or greatLakes:
		positionSet = []
		if centralSea or greatLakes:
			positionSet.append((int(width/2), int(height/2)))
		if cornerSea:
			positionSet.append((int(width-3), int(height-3)))
			positionSet.append((int(width-3), int(3)))
			positionSet.append((int(3), int(height-3)))
			positionSet.append((int(3), int(3)))
		if edgeSea:
			positionSet.append((int(width/2), int(3)))
			positionSet.append((int(width/2), int(height-3)))
			positionSet.append((int(3), int(height/2)))
			positionSet.append((int(width-3), int(height/2)))
			
		positionIndex = dice.get(len(positionSet), "Python: SmartMap: Dice: sea position index")
		seaPositions = [positionSet[positionIndex]]
		
		seaPlotCount = int((percentOcean*len(allLandList))/100)
		
		numGreatLakes = 1
		if greatLakes:
			numGreatLakes = numLakes + dice.get(numLakes, "Python: SmartMap: Dice: Number of great lakes")
			for lake in range(numGreatLakes):
				lx = 8+dice.get(width-8, "Python: SmartMap: Dice: position of great lakes")
				ly = 8+dice.get(height-8, "Python: SmartMap: Dice: position of great lakes")
				seaPositions.append((lx,ly))
		
		#fixme
		seaPlotCount = int(seaPlotCount / numGreatLakes)
		for seaPosition in seaPositions:
			print "    try to place central sea",seaPlotCount,"lake tiles at",seaPosition
			#tries is used to ensure that this can't loop forever
			tries=0
			seaPlaced = 0
			usedSeaLocs = [seaPosition]
			while (seaPlaced < seaPlotCount and tries < 8 * seaPlotCount):
				tries+=1
				indexListSea = dice.get(len(usedSeaLocs), "Python: SmartMap: Dice: Random Sea listIndex")
				existingX, existingY = usedSeaLocs[indexListSea]
				growthIndex = dice.get(len(nearDirs), "Python: SmartMap: Dice: Random Sea listIndex")
				growDir = nearDirs[growthIndex]
				growX, growY = getTileCoordXYWithWrap((existingX,existingY),growDir)
				growIndex = getTileIndexAtXY(growX,growY)
				if centralSea and tooCloseToPole(growX, growY, int(width/6)):
					continue
				if plotTypes[growIndex] != PlotTypes.PLOT_OCEAN:
					plotTypes[growIndex] = PlotTypes.PLOT_OCEAN
					seaPlaced += 1
					usedSeaLocs.append((growX,growY))
					
			print "    placed",seaPlaced,"sea tiles","out of required",seaPlotCount

	#correct the ocean percentage since we may have placed too many land earlier
	print "    fix ocean percentage"
	
	#recalculate land count
	allLandList = getPlotsOfTypes(plotTypes, landPlotTypes)
	currentLandPlotCount = len(allLandList)
			
	landRemovalTries = 0
	print "    remove extra land currentLandPlotCount",currentLandPlotCount,"expectedLandPlotCount",expectedLandPlotCount
	while currentLandPlotCount > expectedLandPlotCount and landRemovalTries < expectedLandPlotCount * 2048:
		landRemovalTries += 1
		rmIndex = dice.get(len(allLandList), "Python: SmartMap: Dice: Land removal X")
		rmX,rmY = allLandList[rmIndex]
		rmTileIndex = getTileIndexAtXY(rmX,rmY)
		if plotTypes[rmTileIndex] != PlotTypes.PLOT_LAND:
			continue
		for nearDir in cardinalDirs:
			ox,oy = getTileCoordXYWithWrap((rmX,rmY),nearDir)
			oIndex = getTileIndexAtXY(ox,oy)
			if plotTypes[oIndex] == fillType:
				plotTypes[rmTileIndex] = PlotTypes.PLOT_OCEAN
				allLandList.remove((rmX,rmY))
				currentLandPlotCount -= 1
				break
	print "    remove extra land completed currentLandPlotCount",currentLandPlotCount,"expectedLandPlotCount",expectedLandPlotCount
	
	if singleTiles:
		print "    make all land into single tile isles"
		print "      processing",len(allLandList),"plots of land"
		while len(allLandList) > 0:
			landIndex = dice.get(len(allLandList), "Python: SmartMap: Dice: Random Sea listIndex")
			lx, ly = allLandList[landIndex]
			tileIndex = getTileIndexAtXY(lx,ly)
			if plotTypes[tileIndex] != PlotTypes.PLOT_OCEAN:
					for xd in range(-seaWidth,seaWidth+1):
						for yd in range(-seaWidth,seaWidth+1):
							ox,oy = getTileCoordXYWithWrap((lx,ly),(xd,yd))
							if ox != lx or oy != ly:
								otherTileIndex = getTileIndexAtXY(ox,oy)
								plotTypes[otherTileIndex] = PlotTypes.PLOT_OCEAN
			allLandList.remove((lx,ly))
					
	if noBlockPeaks:				
		print "    flatten blocking peaks to hills"
		flattened = 0
		for x,y in allLandList:
			tileIndex = getTileIndexAtXY(x,y)
			if plotTypes[tileIndex] != PlotTypes.PLOT_PEAK:
				continue
			bothOcean = True
			for diff in ((-1,0),(1,0)):
				ox,oy = getTileCoordXYWithWrap((x,y),diff)
				otherIndex = getTileIndexAtXY(ox,oy)
				if plotTypes[otherIndex] != PlotTypes.PLOT_OCEAN:
					bothOcean = False
			
			if bothOcean:
				plotTypes[tileIndex] = PlotTypes.PLOT_HILLS
				flattened += 1
				continue
				
			bothOcean = True
			for diff in ((0,-1),(0,1)):
				ox,oy = getTileCoordXYWithWrap((x,y),diff)
				otherIndex = getTileIndexAtXY(ox,oy)
				if plotTypes[otherIndex] != PlotTypes.PLOT_OCEAN:
					bothOcean = False
			
			if bothOcean:
				plotTypes[tileIndex] = PlotTypes.PLOT_HILLS
				flattened += 1
				continue
		print "      flattened",flattened,"peaks"
		
	
	if not centralSea and not cornerSea and not edgeSea and not greatLakes:
		#insert a transformation to put the widest vertical ocean on the x-wrap
		transPlotTypes = plotTypes[:]			
		if wrapX:
			print "    execute x-wrap transformation"
			maxVerticalOcean = 0
			verticalOceanLocation = [1]
			for x in range(width):
				oceanForX = 0
				for xtally in (-2,-1,0,1,2):
					for y in range(height):
						checkX,checkY = getTileCoordXYWithWrap((x,y),(xtally,0))
						plotIndex = getTileIndexAtXY(checkX,checkY)
						if plotTypes[plotIndex] == PlotTypes.PLOT_OCEAN:
							oceanForX += 1
				if oceanForX > maxVerticalOcean:
					maxVerticalOcean = oceanForX
					verticalOceanLocation = [x]
				elif oceanForX == maxVerticalOcean:
					verticalOceanLocation.append(x)
					
			firstHalfLocs = []
			secondHalfLocs = []
			for loc in verticalOceanLocation:
				if loc < int(width/2):
					firstHalfLocs.append(loc)
				else:
					secondHalfLocs.append(loc)
					
			usedLocs = firstHalfLocs
			if len(secondHalfLocs) > len(usedLocs):
				usedLocs = secondHalfLocs
					
			usedLocs.sort()
			actualWrapLocation = usedLocs[int(len(usedLocs)/2)]
			actualWrapLocation += 1	
			print "    wrap X at",actualWrapLocation,"chosen from",usedLocs	
			#transform the resultplottypes by x
			for x in range(width):
				for y in range(height):
					readX, readY = getTileCoordXYWithWrap((x,y),(actualWrapLocation,0))
					readIndex = getTileIndexAtXY(readX,readY)
					writeIndex = getTileIndexAtXY(x,y)
					transPlotTypes[writeIndex] = plotTypes[readIndex]
					
		plotTypes = transPlotTypes
	
		#insert a transformation to put the widest horizontal ocean on the y-wrap
		transPlotTypes = plotTypes[:]			
		if wrapY:
			print "    execute y-wrap transformation"
			maxHorizontalOcean = 0
			horizontalOceanLocation = [1]
			for y in range(height):
				oceanForY = 0
				for ytally in (-2,-1,0,1,2):
					for x in range(width):
						checkX,checkY = getTileCoordXYWithWrap((x,y),(0,ytally))
						plotIndex = getTileIndexAtXY(checkX,checkY)
						if plotTypes[plotIndex] == PlotTypes.PLOT_OCEAN:
							oceanForY += 1
				if oceanForY > maxHorizontalOcean:
					maxHorizontalOcean = oceanForY
					horizontalOceanLocation = [y]
				elif oceanForY == maxHorizontalOcean:
					horizontalOceanLocation.append(y)
					
			firstHalfLocs = []
			secondHalfLocs = []
			for loc in horizontalOceanLocation:
				if loc < int(height/2):
					firstHalfLocs.append(loc)
				else:
					secondHalfLocs.append(loc)
					
			usedLocs = firstHalfLocs
			if len(secondHalfLocs) > len(usedLocs):
				usedLocs = secondHalfLocs
					
			usedLocs.sort()
			actualWrapLocation = usedLocs[int(len(usedLocs)/2)]
			actualWrapLocation += 1	
			print "    wrap Y at",actualWrapLocation,"chosen from",usedLocs	
			#transform the resultplottypes by y
			for x in range(width):
				for y in range(height):
					readX, readY = getTileCoordXYWithWrap((x,y),(0,actualWrapLocation))
					readIndex = getTileIndexAtXY(readX,readY)
					writeIndex = getTileIndexAtXY(x,y)
					transPlotTypes[writeIndex] = plotTypes[readIndex]
					
		plotTypes = transPlotTypes
	
	OutputMessage("Python: SmartMap: Step 2 Setting Plot Types: complete")
	# Finished generating plots!
	return plotTypes
	
##########################################################################
#Step 3: now generate the fine terrain, in particular, need to decide if 
#any land or hill tile is desert, plains, grass, tundra, or snow
def generateTerrainTypes():	
	global width
	global height	
	
	OutputMessage("Python: SmartMap: Step 3 Generate Terrain")
	gc = CyGlobalContext()
	cymap = CyMap()
	dice = gc.getGame().getMapRand()
	
	checkForBadPlots()
	
	global totalLandCount
	global landPlotTypes
	totalLandCount = 0
	for x in range(width):
		for y in range(height):
			plotType = cymap.plot(x,y).getPlotType()
			if plotType in landPlotTypes:
				totalLandCount += 1
	
	preComputeAltitude()
	global preComputedAltitude
	
	computeContinents()
	global postContinentLists
	
	#recalculate land count
	totalLandCount = 0
	totalPlotCount = 0
	for x in range(width):
		for y in range(height):
			totalPlotCount += 1
			pPlot = cymap.plot(x,y)
			if not pPlot.isWater():
				totalLandCount += 1
	pctLand = int((100*totalLandCount)/totalPlotCount)
	print "    step 3 totalLandCount",totalLandCount,"from total",totalPlotCount,"pct land",pctLand,"pct ocean",100-pctLand

	#find out the user selected terrain method
	terrainMethod = getSelectedMapValue("Terrain:")
	print "    terrainMethod",terrainMethod

	if terrainMethod == "standard":
		terraingen = TerrainGenerator()
		terrainTypes = terraingen.generateTerrain()
		OutputMessage("Python: SmartMap: Step 3 Generate Terrain: complete standard")
		return terrainTypes

	if terrainMethod == "great plains":
		terraingen = GreatPlainsTerrainGenerator()
		terrainTypes = terraingen.generateTerrain()
		OutputMessage("Python: SmartMap: Step 3 Generate Terrain: complete great plains")
		return terrainTypes

	if terrainMethod == "oasis":
		terraingen = OasisTerrainGenerator()
		terrainTypes = terraingen.generateTerrain()
		OutputMessage("Python: SmartMap: Step 3 Generate Terrain: complete oasis")
		return terrainTypes
		
	noSnow = False
	noDesert = False
	noTundra = False
	allPlains = False
	allGrass = False	
	plainsAdjust = 0
	if terrainMethod == "SmartMap no snow":
		noSnow = True
	if terrainMethod == "SmartMap no desert":
		noDesert = True
	if terrainMethod == "SmartMap no snow/desert":
		noSnow = True
		noDesert = True
	if terrainMethod == "SmartMap no snow/desert/tundra":
		noSnow = True
		noDesert = True
		noTundra = True
	if terrainMethod == "SmartMap only plains":
		allPlains = True
	if terrainMethod == "SmartMap only grass":
		allGrass = True
	if terrainMethod == "SmartMap extra grass":
		plainsAdjust = -15
	if terrainMethod == "SmartMap extra plains":
		plainsAdjust = 15

	#otherwise, smartmap
	#get the user entered climate info
	climate = cymap.getClimate()
	climateTemperate = gc.getInfoTypeForString("CLIMATE_TEMPERATE")
	climateTropical = gc.getInfoTypeForString("CLIMATE_TROPICAL")
	climateArid = gc.getInfoTypeForString("CLIMATE_ARID")
	climateCold = gc.getInfoTypeForString("CLIMATE_COLD")

	extraRiverPct = 100
	
	if climate == climateTropical:
		extraRiverPct += 36
		
	#arid games need rivers to be playable
	if climate == climateArid:
		extraRiverPct += 26
		
	(riverPct,lakePct) = getSelectedMapValue("Rivers/Lakes:")
	extraRiverPct = 1+int((extraRiverPct*riverPct)/100)

	
	print "    generate rivers"
	#first pass: place rivers.  if you don't place rivers first
	#then oasis and floodplains are impossible
	#1 in 1000 hills or 1 in 200 peaks spawns a river in a random direction	
	riverDirTypes = [CardinalDirectionTypes.CARDINALDIRECTION_NORTH,
					CardinalDirectionTypes.CARDINALDIRECTION_EAST,
					CardinalDirectionTypes.CARDINALDIRECTION_SOUTH,
					CardinalDirectionTypes.CARDINALDIRECTION_WEST]
		
	#each continent gets at least one river from its highest point			
	for continentList in postContinentLists:
		#no automatic river for continents of less than N land tiles
		if len(continentList) < int( ((totalLandCount+1) * 2) / 100):
			continue
		maxContAltitude = -1
		initialRandomIndex = dice.get(len(continentList), "Python: SmartMap: Dice: High River Initial Placement")
		cx,cy = continentList[initialRandomIndex]
		bestPlot = cymap.plot(cx,cy)
		for plot in continentList:
			plotX,plotY = plot
			pPlot = cymap.plot(plotX,plotY)
			#rivers only generate from the tallest peak or hill
			if not (pPlot.isHills() or pPlot.isPeak()):
				continue
			plotIndex = getTileIndexAtXY(plotX,plotY)
			altitude = preComputedAltitude[plotIndex]
			if altitude > maxContAltitude:
				maxContAltitude = altitude
				bestPlot = pPlot
		riverDirIndex = dice.get(len(riverDirTypes), "Python: SmartMap: Dice: High River Route")
		CyMapGenerator().doRiver(bestPlot, riverDirTypes[riverDirIndex])
		#CyEngine().addLandmark(bestPlot,"Mount Doom")

	#we need to add more rivers to maps when the peak count is low relative to the hill count
	peakCount = 0
	hillCount = 0
	for rx in range(width):
		for ry in range(height):
			pPlot = cymap.plot(rx,ry)
			if pPlot.isHills():
				hillCount += 1
			if pPlot.isPeak():
				peakCount += 1
				
	peakHillRatioPct = int(((peakCount+1)*100)/(hillCount+1))
	hillBonus = 100-peakHillRatioPct
			
	#additional random rivers			
	for rx in range(width):
		for ry in range(height):
			pPlot = cymap.plot(rx,ry)
			if pPlot.isRiver() or isCoastalPlot(rx,ry) or pPlot.isFreshWater():
				continue
			badSpot = False
			#check for another river too nearby
			for rNearDir in nearDirs:
				rnewX, rnewY = getTileCoordXYWithWrap((rx,ry), rNearDir)
				otherPlot = cymap.plot(rnewX,rnewY)
				#too close to another river
				if otherPlot.isRiver():
					badSpot = True
					break
				#too close to lake
				if otherPlot.isLake():
					badSpot = True
					break
					
			if badSpot:
				continue
				
			plotIndex = getTileIndexAtXY(rx,ry)
			plotAltitude = int(preComputedAltitude[plotIndex]/10) #range is 0 to 100
			percentFromEquatorXY = percentFromEquator(rx,ry)
			percentFromPoleXY = percentFromPole(rx,ry)
			riverOdds = dice.get(10000, "Python: SmartMap: Dice: River Placement Odds")
			if ( (pPlot.isHills() and riverOdds < int(((400+(3*hillBonus)+(5*extraRiverPct)+(2*percentFromEquatorXY)+(2*percentFromPoleXY)+(3*int(plotAltitude*plotAltitude/20)))*extraRiverPct)/100)) or 
			     (pPlot.isPeak() and riverOdds < int(((2400+extraRiverPct+(10*percentFromEquatorXY)+(10*percentFromPoleXY)+(12*int(plotAltitude*plotAltitude/20)))*extraRiverPct)/100))
			     ):
				riverDirIndex = dice.get(len(riverDirTypes), "Python: SmartMap: Dice: River Placement")
				CyMapGenerator().doRiver(pPlot, riverDirTypes[riverDirIndex])
	
	
	extraDesert = 0
	extraCold = 0
	grassFactor = 0
	
	if climate == climateArid:
		extraDesert += 20
		extraCold -= 5
		grassFactor -= 10
	if climate == climateCold:
		extraDesert -= 10
		extraCold += 20
		grassFactor -= 5
	if climate == climateTemperate:
		extraDesert -= 0
		extraCold -= 0
		grassFactor += 5
	if climate == climateTropical:
		extraDesert -= 16
		extraCold -= 16
		grassFactor += 10

	extraForest,extraJungle = getSelectedMapValue("Forest/Jungle:")
	if extraForest > 0:
		extraDesert -= 3
	if extraJungle > 0:
		extraDesert += 3
	if extraForest < 0:
		extraDesert -= 3
	if extraJungle < 0:
		extraDesert += 3

	
	print "    generate terrain"
	#default generation method: SmartMap terrain
	terrainData = [TerrainTypes.NO_TERRAIN]*(width*height)
	
	terrainDesert = gc.getInfoTypeForString("TERRAIN_DESERT")
	terrainPlains = gc.getInfoTypeForString("TERRAIN_PLAINS")
	terrainGrass = gc.getInfoTypeForString("TERRAIN_GRASS")
	terrainTundra = gc.getInfoTypeForString("TERRAIN_TUNDRA")
	terrainSnow = gc.getInfoTypeForString("TERRAIN_SNOW")
	
	preComputeWetness()
	global preComputedWetness
	
	#for every terrain we need to consider in x,y
	for x in range(width):
		for y in range(height):
			pPlot = cymap.plot(x,y)
			index = getTileIndexAtXY(x,y)
			#make sure every plot has a default set
			terrainData[index] = pPlot.getTerrainType()
			#water tiles, peaks, and hills should always take their default terrain type
			if pPlot.isWater():
				terrainData[index] = pPlot.getTerrainType()
				continue
			if pPlot.isPeak():	
				terrainData[index] = pPlot.getTerrainType()
				continue
				
			if allPlains:
				terrainData[index] = terrainPlains
				continue
			if allGrass:
				terrainData[index] = terrainGrass
				continue
			
			#land and hill tiles actually have choices in what type of tile to be
			plotType = pPlot.getPlotType()
			if plotType == PlotTypes.PLOT_LAND or plotType == PlotTypes.PLOT_HILLS:
				#note that height is at the top of the map (northmost), when you might expect bottom
				#this allows 10% snow only near the very furthest north or south strip of land
				percentFromPoleXY = percentFromPole(x,y)
				percentFromEquatorXY = percentFromEquator(x,y)
				#print "    eq%",percentFromEquatorXY,"po%",percentFromPoleXY,"x",x,"y",y
				#when very close to the pole, you get either snow, tundra, or plains
				plotAltitudeFactor = -10 + int(preComputedAltitude[index]/100) #range is -10 to 10

				#water content of this tile
				waterFactor = int(preComputedWetness[index] / 10)
				isWater = False
				if pPlot.isFreshWater() or pPlot.isRiver():
					isWater = True

				snowRand = dice.get(100, "Python: SmartMap: Dice: Random Terrain")
				tundraRand = dice.get(100, "Python: SmartMap: Dice: Random Terrain")
				if not noSnow and percentFromPoleXY<(35+extraCold+plotAltitudeFactor) and snowRand < 6*(35+extraCold-percentFromPoleXY+plotAltitudeFactor):
					terrainData[index] = terrainSnow
					continue
				if not noTundra and percentFromPoleXY<(35+extraCold+plotAltitudeFactor) and tundraRand < 70+plotAltitudeFactor:
					terrainData[index] = terrainTundra
					continue
				if percentFromPoleXY<(35+extraCold+plotAltitudeFactor):
					terrainData[index] = terrainPlains
					continue
					
				#near the poles, but further out, tundra or plains
				if not noTundra and percentFromPoleXY<(55+extraCold+plotAltitudeFactor) and tundraRand < 5*(55+extraCold-percentFromPoleXY+plotAltitudeFactor):
					terrainData[index] = terrainTundra
					continue
				if percentFromPoleXY<(55+extraCold+plotAltitudeFactor):
					terrainData[index] = terrainPlains
					continue

				#in central portion of map, get desert
				desertRand = dice.get(100, "Python: SmartMap: Dice: Random Terrain")
				plainsRand = dice.get(100, "Python: SmartMap: Dice: Random Terrain")
				desertSign = 1
				if extraDesert < 0:
					desertSign = -1
				desertWidthFactor = dice.get(abs(extraDesert)+1, "Python: SmartMap: Dice: Random Terrain")
				desertWidthFactor *= desertSign
				desertWaterFactor = waterFactor/3
				if climate == climateTropical:
					desertWaterFactor = waterFactor
				if not noDesert and (percentFromEquatorXY<(42+desertWidthFactor+plotAltitudeFactor) and 
				    desertRand < 60+plotAltitudeFactor-desertWaterFactor+extraDesert-percentFromEquatorXY+desertWidthFactor):
					terrainData[index] = terrainDesert
					continue
				if percentFromEquatorXY<(42+desertWidthFactor) and ((plainsRand<50+plainsAdjust) or ((plainsRand<80+plainsAdjust) and not isWater)):
					terrainData[index] = terrainPlains
					continue
				if percentFromEquatorXY<(42+desertWidthFactor):
					terrainData[index] = terrainGrass
					continue
					
				#grass, particularly near water
				grassRand = dice.get(100, "Python: SmartMap: Dice: Random Terrain")
				if grassRand < (50 + waterFactor + grassFactor -plainsAdjust):
					#print "    grass eq%",percentFromEquatorXY,"po%",percentFromPoleXY,"x",x,"y",y
					terrainData[index] = terrainGrass
					continue
				#everything else plains
				terrainData[index] = terrainPlains
				continue
		
	#second pass: clumpify terrain by giving each plot a chance to match its neighbors
	allLandPlots = []
	for x in range(width):
		for y in range(height):
			pPlot = cymap.plot(x,y)
			plotType = pPlot.getPlotType()
			if plotType == PlotTypes.PLOT_LAND:
				allLandPlots.append((x,y))
				
	while (len(allLandPlots) > 0):
		pickIndex = dice.get(len(allLandPlots), "Python: SmartMap: Dice: Random Terrain Plot Pick")
		(x,y) = allLandPlots[pickIndex]
		del allLandPlots[pickIndex]
		#print "flip plot",x,y
		desertN=0
		snowN=0
		tundraN=0
		grassN=0
		plainsN=0
		for nearDir in nearDirs:
			otherX, otherY = getTileCoordXYWithWrap((x,y), nearDir)
			otherPlot = cymap.plot(otherX,otherY)
			otherPlotType = otherPlot.getPlotType()
			if otherPlotType == PlotTypes.PLOT_LAND:
				otherIndex = getTileIndexAtXY(otherX,otherY)
				otherTerrain = terrainData[otherIndex]
				if otherTerrain == terrainDesert:
					desertN += 1
				elif otherTerrain == terrainPlains:
					plainsN += 1
				elif otherTerrain == terrainGrass:
					grassN += 1
				elif otherTerrain == terrainTundra:
					tundraN += 1
				elif otherTerrain == terrainSnow:
					snowN += 1
			
		scores = [desertN,snowN,tundraN,grassN,plainsN]
		maxScore = max(scores)
		if maxScore < 3:
			continue
		skipVal = dice.get(1000, "Python: SmartMap: Dice: Random Terrain Plot Pick")
		if maxScore < 6 and skipVal > 100:
			continue #only flip a certain % of plots
		if grassN == maxScore:
			terrainData[index] = terrainGrass
		elif plainsN == maxScore:
			terrainData[index] = terrainPlains
		elif desertN == maxScore:
			terrainData[index] = terrainDesert
		elif tundraN == maxScore:
			terrainData[index] = terrainTundra
		elif snowN == maxScore:
			terrainData[index] = terrainSnow
	
	#third pass: no desert within one step of fresh water on tropical
	if climate == climateTropical:
		for x in range(width):
			for y in range(height):
				pPlot = cymap.plot(x,y)
				index = getTileIndexAtXY(x,y)
				plotType = pPlot.getPlotType()
				pickedTerrainType = terrainData[index]
				if (plotType == PlotTypes.PLOT_LAND or plotType == PlotTypes.PLOT_HILLS) and pickedTerrainType == terrainDesert:
					for nearDir in nearDirs:
						otherX, otherY = getTileCoordXYWithWrap((x,y), nearDir)
						otherPlot = cymap.plot(otherX,otherY)
						if otherPlot.isFreshWater():
							terrainData[index] = terrainPlains
							break
		
				
	#fourth pass: set all hills to match most common touching terrain
	for x in range(width):
		for y in range(height):
			pPlot = cymap.plot(x,y)
			index = getTileIndexAtXY(x,y)
			plotType = pPlot.getPlotType()
			if (plotType == PlotTypes.PLOT_HILLS):
				desertN=0
				snowN=0
				tundraN=0
				grassN=0
				plainsN=0
				for nearDir in nearDirs:
					otherX, otherY = getTileCoordXYWithWrap((x,y), nearDir)
					otherPlot = cymap.plot(otherX,otherY)
					otherPlotType = otherPlot.getPlotType()
					if otherPlotType == PlotTypes.PLOT_LAND:
						otherIndex = getTileIndexAtXY(otherX,otherY)
						otherTerrain = terrainData[otherIndex]
						if otherTerrain == terrainDesert:
							desertN += 1
						elif otherTerrain == terrainPlains:
							plainsN += 1
						elif otherTerrain == terrainGrass:
							grassN += 1
						elif otherTerrain == terrainTundra:
							tundraN += 1
						elif otherTerrain == terrainSnow:
							snowN += 1
					
				scores = [desertN,snowN,tundraN,grassN,plainsN]
				maxScore = max(scores)
				if maxScore < 3:
					continue
				skipVal = dice.get(1000, "Python: SmartMap: Dice: Random Terrain Plot Pick")
				if maxScore < 5 and skipVal > 600:
					continue #only flip a certain % of plots
				if grassN == maxScore:
					terrainData[index] = terrainGrass
				elif plainsN == maxScore:
					terrainData[index] = terrainPlains
				elif desertN == maxScore:
					terrainData[index] = terrainDesert
				elif tundraN == maxScore:
					terrainData[index] = terrainTundra
				elif snowN == maxScore:
					terrainData[index] = terrainSnow
				else:
					terrainData[index] = terrainPlains
					
	#add some rivers to desert plots in temperate/tropical
	#additional random rivers			
	for rx in range(width):
		for ry in range(height):
			pPlot = cymap.plot(rx,ry)
			if pPlot.isRiver() or isCoastalPlot(rx,ry) or pPlot.isFreshWater() or not pPlot.isHills():
				continue

			badSpot = False
			nearDesert = False
			#check for another river too nearby
			for rNearDir in nearDirs:
				rnewX, rnewY = getTileCoordXYWithWrap((rx,ry), rNearDir)
				otherPlot = cymap.plot(rnewX,rnewY)
				#too close to another river
				if otherPlot.isRiver():
					badSpot = True
					break
				#too close to lake
				if otherPlot.isLake():
					badSpot = True
					break
				
				otherIndex = getTileIndexAtXY(rnewX,rnewY)
				if terrainData[otherIndex] == terrainDesert:
					nearDesert = True
					
			if badSpot or not nearDesert:
				continue
			
			plotIndex = getTileIndexAtXY(rx,ry)
			desertBonus = 0
			if terrainData[plotIndex] == terrainDesert:
				desertBonus = 1000
			plotAltitude = int(preComputedAltitude[plotIndex]/10) #range is 0 to 100
			percentFromEquatorXY = percentFromEquator(rx,ry)
			percentFromPoleXY = percentFromPole(rx,ry)
			riverOdds = dice.get(10000, "Python: SmartMap: Dice: River Placement Odds")
			if ( (pPlot.isHills() and riverOdds < int(((3500+desertBonus+(4*hillBonus)+(5*extraRiverPct)+(2*percentFromEquatorXY)+(2*percentFromPoleXY)+(3*int(plotAltitude*plotAltitude/20)))*extraRiverPct)/100))):
				riverDirIndex = dice.get(len(riverDirTypes), "Python: SmartMap: Dice: River Placement")
				CyMapGenerator().doRiver(pPlot, riverDirTypes[riverDirIndex])
	
	
	OutputMessage("Python: SmartMap: Step 3 Generate Terrain: complete")
	checkForBadPlots()
	return terrainData
	
##########################################################################
#Step 3A: add rivers.  since smartmap handles this elsewhere we must override
def addRivers():
	OutputMessage("Python: SmartMap: Step 3A rivers")
	checkForBadPlots()

	#find out the user selected terrain method
	terrainMethod = getSelectedMapValue("Terrain:")
	print "    riverMethod",terrainMethod

	standardRiverMethods = ["standard","great plains","oasis"]
	if terrainMethod in standardRiverMethods:
		CyPythonMgr().allowDefaultImpl()
		OutputMessage("Python: SmartMap: Step 3A rivers: complete standard")
		return
	checkForBadPlots()
	OutputMessage("Python: SmartMap: Step 3A rivers: complete")
	return
	
##########################################################################
#Step 3B: add lakes.  since smartmap handles this elsewhere we must override
def addLakes():
	OutputMessage("Python: SmartMap: Step 3B lakes")
	checkForBadPlots()

	#find out the user selected terrain method
	terrainMethod = getSelectedMapValue("Terrain:")
	print "    lakeMethod",terrainMethod

	standardLakeMethods = ["standard","great plains","oasis"]
	if terrainMethod in standardLakeMethods:
		CyPythonMgr().allowDefaultImpl()
		OutputMessage("Python: SmartMap: Step 3B lakes: complete standard")
		return
	checkForBadPlots()
	OutputMessage("Python: SmartMap: Step 3B lakes: complete")
	return

##########################################################################
#Step 4: add features to your terrain	
#this is responsible for things like oasis, ice floes on the ocean, floodplains
def addFeatures():
	global wrapX
	global wrapY
	global width
	global height
	OutputMessage("Python: SmartMap: Step 4 Feature Generation")
	checkForBadPlots()
	gc = CyGlobalContext()
	cymap = CyMap()
	dice = gc.getGame().getMapRand()

	#find out the user selected terrain method
	featureMethod = getSelectedMapValue("Features:")
	print "    featureMethod",featureMethod

	#we have to recompute wetness in case we used standard river or lake methods
	preComputeWetness()
	global preComputedWetness

	global featuregen
	if featureMethod == "standard":
		featuregen = FeatureGenerator()
		featuregen.addFeatures()
		OutputMessage("Python: SmartMap: Step 4 Feature Generation: complete standard")
		return 0

	if featureMethod == "great plains":
		featuregen = GreatPlainsFeatureGenerator()
		featuregen.addFeatures()
		OutputMessage("Python: SmartMap: Step 4 Feature Generation: complete great plains")
		return 0

	if featureMethod == "oasis":
		featuregen = OasisFeatureGenerator()
		featuregen.addFeatures()
		OutputMessage("Python: SmartMap: Step 4 Feature Generation: complete oasis")
		return 0
		
	#defaults to SmartMap mode
	strict = False #determines whether to allow things like floodplains/plains
	if featureMethod == "SmartMap Strict":
		strict = True
		
	#get the user entered climate info
	climate = cymap.getClimate()
	climateTemperate = gc.getInfoTypeForString("CLIMATE_TEMPERATE")
	climateTropical = gc.getInfoTypeForString("CLIMATE_TROPICAL")
	climateArid = gc.getInfoTypeForString("CLIMATE_ARID")
	climateCold = gc.getInfoTypeForString("CLIMATE_COLD")
	
	extraForest,extraJungle = getSelectedMapValue("Forest/Jungle:")
	extraCold = 0
	extraOasis = 0
	
	if climate == climateTropical:
		extraJungle += 60
		extraForest += 30
		extraOasis += 3
		
	if climate == climateTemperate:
		extraForest += 60
		extraJungle += 30

	if climate == climateCold:
		extraForest += 120
		extraCold += 2

	if climate == climateArid:
		extraJungle -= 100
		extraForest -= 50
		extraOasis -= 2
		
	featIce = gc.getInfoTypeForString("FEATURE_ICE")
	featJungle = gc.getInfoTypeForString("FEATURE_JUNGLE")
	featOasis = gc.getInfoTypeForString("FEATURE_OASIS")
	featFlood = gc.getInfoTypeForString("FEATURE_FLOOD_PLAINS")
	featForest = gc.getInfoTypeForString("FEATURE_FOREST")
	terrainDesert = gc.getInfoTypeForString("TERRAIN_DESERT")
	terrainTundra = gc.getInfoTypeForString("TERRAIN_TUNDRA")


	lastFeature = featForest
	print "    generate smartmap features"
	for x in range(width):
		for y in range(height):
			#force ice at poles
			pPlot = cymap.plot(x,y)
			if tooCloseToPole(x,y,1):
				pPlot.setFeatureType(featIce,-1)
				continue

			percentFromPoleXY = percentFromPole(x,y)
			percentFromEquatorXY = percentFromEquator(x,y)
			terrainType = pPlot.getTerrainType()
			plotIndex = getTileIndexAtXY(x,y)
			#get the water level and altitude scaled to 0..100 range
			waterFactor = int(preComputedWetness[plotIndex]/10)
			altitudeFactor = int(preComputedAltitude[plotIndex]/10)


			#ice at the poles, for whatever poles this planet has
			iceOdds = dice.get(1000, "Python: SmartMap: Dice: Feature Placement Odds")
			if pPlot.isWater() and percentFromPoleXY<=13+extraCold and pPlot.canHaveFeature(featIce) and iceOdds < 100*(13-percentFromPoleXY+extraCold):
				pPlot.setFeatureType(featIce,-1)
				lastFeature = featIce
				continue
				
			#no other features exist on water
			if pPlot.isWater():
				lastFeature = featIce
				continue
				
			#oasis don't ask if it canHaveFeature oasis because that always returns False
			oasisOdds = dice.get(1000, "Python: SmartMap: Dice: Feature Placement Odds")
			oasisBonus = 0
			if terrainType == terrainDesert:
				oasisBonus += 15
			if (pPlot.isFlatlands() and not pPlot.isFreshWater() and oasisOdds<5+extraOasis+oasisBonus and 
			    (terrainType == terrainDesert or terrainType == terrainTundra)):
				if not strict or pPlot.canHaveFeature(featOasis):
					pPlot.setFeatureType(featOasis,-1)
					lastFeature = featOasis
					continue

			#flood same issue as oasis, always False for some reason
			floodOdds = dice.get(1000, "Python: SmartMap: Dice: Feature Placement Odds")
			floodBonus = waterFactor
			floodBonus += 3*extraJungle
			if pPlot.getTerrainType() == terrainDesert:
				floodBonus += 300
			if pPlot.isFreshWater() and floodOdds<floodBonus:
				if not strict or pPlot.canHaveFeature(featFlood):
					pPlot.setFeatureType(featFlood,-1)
					lastFeature = featFlood
					continue

			#jungle
			jungleBonus = 0
			if lastFeature == featJungle:
				jungleBonus = 50
			jungleOdds = dice.get(1000, "Python: SmartMap: Dice: Feature Placement Odds")
			extraReachFactor = dice.get(int(waterFactor / 4), "Python: SmartMap: Dice: Extra Water Reach")
			if extraJungle > 0:
				extraReachFactor += dice.get(1+int(extraJungle / 4), "Python: SmartMap: Dice: Extra Forest Reach")
			if (percentFromEquatorXY < 50+extraReachFactor and pPlot.canHaveFeature(featJungle) and 
			    jungleOdds < (jungleBonus + extraJungle + 2*waterFactor + 2*(100 - altitudeFactor) + 2*(100-percentFromEquatorXY)) ):
				pPlot.setFeatureType(featJungle,-1)
				lastFeature = featJungle
				continue
	
			#forest
			forestBonus = 0
			if lastFeature == featForest:
				forestBonus = 50
			extraReachFactor = dice.get(int(waterFactor / 4), "Python: SmartMap: Dice: Extra Water Reach")
			if extraForest > 0:
				extraReachFactor += dice.get(1+int(extraForest / 4), "Python: SmartMap: Dice: Extra Forest Reach")
			forestOdds = dice.get(1000, "Python: SmartMap: Dice: Feature Placement Odds")
			if (percentFromEquatorXY > (40-extraReachFactor) and pPlot.canHaveFeature(featForest) and 
			    forestOdds < (250 + forestBonus + extraForest + waterFactor + 2*altitudeFactor + 2*percentFromEquatorXY)):
				pPlot.setFeatureType(featForest,-1)
				lastFeature = featForest
				continue
				
			lastFeature = featIce
		
	OutputMessage("Python: SmartMap: Step 4 Feature Generation: complete")
	checkForBadPlots()
	return 0
	
##########################################################################
#Step 5: add bonuses to your map
#this method puts down the tile bonuses, such as iron, whale, rice, oil, etc.	
def addBonuses():
	global width
	global height
	
	OutputMessage("Python: SmartMap: Step 5 Add bonuses")
	checkForBadPlots()

	gc = CyGlobalContext()
	cymap = CyMap()
	dice = gc.getGame().getMapRand()

	#find out the user selected terrain method
	bonusMethod = getSelectedMapValue("Bonuses:")
	print "    bonusMethod",bonusMethod

	if bonusMethod == "standard":
		CyPythonMgr().allowDefaultImpl()
		OutputMessage("Python: SmartMap: Step 5 Add bonuses: complete standard")
		return
	#Otherwise, smartmap placement
	bonusPct = 100
	clumpHappy = False
	clumpHealth = False
	strict = False
	fewStrategic = False
	if bonusMethod == "SmartMap Few":
		bonusPct = 59
		strict = True
	if bonusMethod == "SmartMap Many":
		bonusPct = 151
	if bonusMethod == "SmartMap Clump Happy":
		bonusPct = 74
		clumpHappy = True
		strict = True
	if bonusMethod == "SmartMap Clump Health":
		bonusPct = 74
		clumpHealth = True
		strict = True
	if bonusMethod == "SmartMap Clump Both":
		bonusPct = 74
		clumpHappy = True
		clumpHealth = True
		strict = True
	if bonusMethod == "SmartMap Restricted":
		bonusPct = 74
		clumpHappy = True
		clumpHealth = True
		strict = True
		fewStrategic = True
	if bonusMethod == "SmartMap Most Restricted":
		bonusPct = 59
		clumpHappy = True
		clumpHealth = True
		strict = True
		fewStrategic = True
		
	global postContinentLists
	global sizeSelected
	global totalLandCount
	global totalCoastCount
	
	#print "    totalLandCount",totalLandCount
	players = gc.getGame().countCivPlayersEverAlive()
	
	# Set plot eligibility for current bonus.
	# Begin by initiating the list of lists, into which eligible plots will be recorded.
	eligible = []
	for bonusId in range(gc.getNumBonusInfos()):
		eligible.append([])
	
	#find all eligible plots for each bonus
	for x in range(width):
		for y in range(height):
			if tooCloseToPole(x,y,3):
				continue
			pPlot = cymap.plot(x,y)
			if pPlot.isWater():
				nearestLand = pPlot.getNearestLandPlot()
				nx = nearestLand.getX()
				ny = nearestLand.getY()
				#assumption: resources cannot be more than 3 tiles from land
				if abs(nx-x) + abs(ny-y) > 6:
					continue
					
			# Check plot type and features for eligibility.
			for bonusId in range(gc.getNumBonusInfos()):
				# This method will tell us if the plot can have this bonus
				if isPlotLegalForBonus(pPlot, bonusId, strict):
					eligible[bonusId].append((x,y))
	
	bonusIdsAndOrder = []
	for bonusId in range(gc.getNumBonusInfos()):
		bonusInfoXML = gc.getBonusInfo(bonusId)
		if bonusInfoXML.getPlacementOrder() >= 0:
			bonusIdsAndOrder.append((bonusId,bonusInfoXML.getPlacementOrder()))

	bonusIdsAndOrder = sorted(bonusIdsAndOrder, key=operator.itemgetter(1))	
	#print bonusIdsAndOrder

	setupSpaceHash()

	allAddedLocs = []
	for bonusIdAndOrder in bonusIdsAndOrder:
		added = 0
		addedLocs = []
		bonusId, order = bonusIdAndOrder
		bonusInfoXML = gc.getBonusInfo(bonusId)
		type_string = bonusInfoXML.getType()
		print "    place bonus",type_string
		
		if bonusInfoXML.getHappiness() > 0 and clumpHappy:
			addBonusType([bonusId])
			print "      using standard placement due to clump happy"
			continue
		if bonusInfoXML.getHealth() > 0 and clumpHealth:
			addBonusType([bonusId])
			print "      using standard placement due to clump health"
			continue
		
		print "      using smartmap placement"
		#find out how many players in game, and make sure there are some for each
		playerFactor = int(players/6) #one per six players
		
		areaFactor = int(totalLandCount/360) #base rate of 1 of each resource for every XXX land tiles
		extraSeparation = 0

		#the bountiful ocean - I add extra resources to water tiles because there
		#is a lot of coastal land in this map design
		if type_string in resourcesWater:
			playerFactor += 1 + players
			areaFactor = int(totalCoastCount / 250)
			extraSeparation -= int((3*sizeSelected)/2)

		#some resources are meant to be a bit more common
		if type_string in resourcesCommon:
			playerFactor += 1 + int(players/3)
			areaFactor += int(totalLandCount / 200)
			extraSeparation -= sizeSelected
			
		#you get extra of the 'early' resources if you are not using one of the strict options
		if not strict and type_string in resourcesImportantEarly:
			playerFactor += 1 + int(players/4)
			areaFactor += int(totalLandCount / 300)
			
		areaFactor = int((2*areaFactor)/3) + dice.get(int(areaFactor/3), "Python: SmartMap: Dice: Bonus ExtraFactor")
		
		if fewStrategic and type_string in resourcesStrategic:
			areaFactor = int(areaFactor/2)

		#total is based on: size of map, num players, random extra per player, extra for water types
		#print "    resource",bonusId,"baseRate",baseRate,"playerExtra",playerExtra,"extraFactor",extraFactor
		count = 1 + int(((1 + playerFactor + areaFactor) * bonusPct) / 100)
		print "    for bonus type",type_string,"expecting to place a total of:",count,"of this bonus"
		
		#compute the fair shares of each resource each continent is due
		
		#first, decide which continents will participate, including only continents that are sufficiently big
		if (not type_string in resourcesWater):
			fairShareContinents = []
			for continentList in postContinentLists:
				continentLandPct = int( (100 * len(continentList)) / totalLandCount)
				if continentLandPct > 9:
					fairShareContinents.append(continentList)
				
			#if there are no continents big enough for this, skip fair sharing
			print "      fair share over", len(fairShareContinents),"continents"
			if len(fairShareContinents) > 0:	
				shares = []
				smallestFairShareContinent = len(fairShareContinents[0])
				for continentList in fairShareContinents:
					if len(continentList) < smallestFairShareContinent:
						smallestFairShareContinent = len(continentList)
						
				totalSharesNeeded = 0
				for continentList in fairShareContinents:
					share = int((len(continentList)+1) / (smallestFairShareContinent+1))
					shares.append(share)
					totalSharesNeeded += share
					
				multiplier = 1
				while totalSharesNeeded < int(count/4):
					totalSharesNeeded = (((multiplier+1)*totalSharesNeeded)/multiplier)
					multiplier += 1
				
				#if there are enough of this resource to apply the expected shares
				if totalSharesNeeded < count:
					#this tries to give each continent a fair share of each resource
					print "    attempting to fair-share",totalSharesNeeded,"out of",count,"of",type_string,"over",len(fairShareContinents),"continents"
					for contIndex in range(len(fairShareContinents)):
						contShare = multiplier * shares[contIndex]
						continentList = fairShareContinents[contIndex]
						#print "try to put",contShare,"of",type_string,"on",contIndex
						for shareNum in range(contShare):
							tries = 0
							while tries < 200:
								tries += 1
								tileIndex = dice.get(len(continentList), "Python: SmartMap: Dice: Continent Bonus Placement")
								tx,ty = continentList[tileIndex]
								pPlot = cymap.plot(tx,ty)
								#this shouldn't happen, but just in case
								if pPlot.isWater():
									continue
								if not isPlotLegalForBonus(pPlot, bonusId, strict):
									continue
								tooClose = False
								for usedLoc in addedLocs:
									distX,distY = getWrapDistanceXY((tx,ty),usedLoc)
									if distX + distY < 6 + sizeSelected + extraSeparation:
										tooClose = True
										break
								if tooClose:
									continue
									
								denseCount = 0
								toBeCounted = listNearSpaceHash(tx,ty,5 + sizeSelected + extraSeparation)
								denseCount = len(toBeCounted)
								denseTest = dice.get(1000, "Python: SmartMap: Dice: Bonus Placement Density Rule")
								density = min(65000,(denseCount*15))
								densityFactor = dice.get(density, "Python: SmartMap: Dice: Bonus Placement Density Rule")
								
								#if too dense, continue
								if denseTest < densityFactor:
									continue
									
								#print "    fair share put 1 of",type_string,"on continent",contIndex,"for shareNum",shareNum,"on try",tries
								pPlot.setBonusType(bonusId)
								added += 1
								addedLocs.append((tx,ty))
								allAddedLocs.append((tx,ty))
								spaceHashItem(tx,ty,bonusId)
								break

		# Now we assign the bonuses to eligible plots chosen completely at random
		# from among the valid plots found above

					
		#if the eligible list is too big, filter it down some
		if len(eligible[bonusId]) > 5 * (count-added):
			altEligible = []
			for sample in range(5 * (count-added)):
				subsetIndex = dice.get(len(eligible[bonusId]), "Python: SmartMap: Dice: Bonus Placement Density Rule")
				altEligible.append(eligible[bonusId][subsetIndex])
			eligible[bonusId] = altEligible
				
		#remove all used plots from eligible list
		removeUsed = True
		while removeUsed:
			removeUsed = False
			for index in range(len(eligible[bonusId])):
				px,py = eligible[bonusId][index]
				pPlot = cymap.plot(px,py)
				if pPlot.getBonusType(-1) != -1:
					del eligible[bonusId][index] # Remove this plot from the eligible list, it already had a bonus
					removeUsed = True
					break
		
		tries = 0
		print "    fair share added",added,"now place remaining",count-added,"of",type_string,"on",len(eligible[bonusId]),"tiles"
		while added < count and tries < 100:
			tries += 1
			if eligible[bonusId] == []:
				break # No eligible plots left!
				
			subsetEligible = []
			if len(eligible[bonusId]) < 12:
				subsetEligible = eligible[bonusId][:]
			else:
				for sample in range(12):
					subsetIndex = dice.get(len(eligible[bonusId]), "Python: SmartMap: Dice: Bonus Placement Density Rule")
					subsetEligible.append(eligible[bonusId][subsetIndex])
					
			lowScore = []
			lowScore.append(subsetEligible[0])
			leastScore = bonus_score(subsetEligible[0])
			for plot in subsetEligible:
				plotScore = bonus_score(plot)
				if plotScore < leastScore:
					leastScore = plotScore
					lowScore = []
					lowScore.append(plot)
				if plotScore == leastScore:
					lowScore.append(plot)
					
				
			index = dice.get(len(lowScore), "Python: SmartMap: Dice: Bonus Placement")
			px,py = lowScore[index]
			pPlot = cymap.plot(px,py)
			if pPlot.getBonusType(-1) != -1:
				eligible[bonusId].remove(lowScore[index]) # Remove this plot from the eligible list, it already had a bonus
				continue
			
			#don't place resources touching each other, not much point	
			tooClose = False
			for usedLoc in addedLocs:
				distX,distY = getWrapDistanceXY((px,py),usedLoc)
				if distX + distY < 3:
					tooClose = true
					break
			if tooClose:
				eligible[bonusId].remove(lowScore[index]) # Remove this plot from the eligible list, it was too close to another bonus
				continue

			#if there are enough remaining eligible spots, prefer not to place resources to close to others
			if (3*(count-added)) < len(eligible[bonusId]):
				bonusInfoXML = gc.getBonusInfo(bonusId)
				minAreaSize = int(bonusInfoXML.getMinAreaSize()/2)
				tooClose = False
				for usedLoc in addedLocs:
					distX,distY = getWrapDistanceXY((px,py),usedLoc)
					if distX + distY < 2 + minAreaSize + sizeSelected + extraSeparation:
						tooClose = true
						break
					
				if tooClose:
					eligible[bonusId].remove(lowScore[index]) # Remove this plot from the eligible list, it was too close to another bonus
					continue
					
				denseCount = 0
				toBeCounted = listNearSpaceHash(px,py,3 + sizeSelected + extraSeparation)
				denseCount = len(toBeCounted)
				
				denseTest = dice.get(1000, "Python: SmartMap: Dice: Bonus Placement Density Rule")
				density = min(65000,int(denseCount*denseCount*sizeSelected))
				densityFactor = dice.get(density, "Python: SmartMap: Dice: Bonus Placement Density Rule")
				#print "denseTest",denseTest,"denseCount",denseCount
				#if density factor is too high, can't place at this tile
				if denseTest < densityFactor and not pPlot.isWater():
					eligible[bonusId].remove(lowScore[index]) # Remove this plot from the eligible list, it was too close to another bonus
					continue
				
			#this actually places the bonus
			pPlot.setBonusType(bonusId)
			added += 1
			addedLocs.append((px,py))
			allAddedLocs.append((px,py))
			spaceHashItem(px,py,bonusId)

			eligible[bonusId].remove(lowScore[index]) # Remove this plot from the eligible list, we used it
		if added == 0:
			print "    failed to add type",type_string,"from count",count
		elif added > 0 and count-added > 0:
			print "    added",added,"of type",type_string,"and failed remaining",count-added
		elif added > 0:
			print "    added all",added,"of type",type_string,"successfully"
		# This bonus type is done.
	
	OutputMessage("Python: SmartMap: Step 5 Add bonuses: complete")
	checkForBadPlots()
	return 0

##########################################################################
#Step 5B: override addBonusType so that it can pass through
def addBonusType(argsList):
	"Can override to control where a type of bonus is added on the map"
	"""WARNING: If addBonuses() is overriden then this function will not automatically be called."""
	[iBonusType] = argsList
	CyPythonMgr().allowDefaultImpl()


##########################################################################
#Step 6: place goody huts
def addGoodies():
	global width
	global height
	
	OutputMessage("Python: SmartMap: Step 6 Goody Huts")
	checkForBadPlots()
	gc = CyGlobalContext()
	cymap = CyMap()
	dice = gc.getGame().getMapRand()
	
	#find out the user selected terrain method
	bonusMethod = getSelectedMapValue("Goodies:")
	print "    hutMethod (uses bonusMethod)",bonusMethod

	if bonusMethod == "standard":
		CyPythonMgr().allowDefaultImpl()
		OutputMessage("Python: SmartMap: Step 6 Goody Huts: complete standard")
		return
		
	#SmartMap goodies
	if bonusMethod == "None":
		return

	bonusPct = 100
	if bonusMethod == "SmartMap Few":
		bonusPct = 68
	if bonusMethod == "SmartMap Many":
		bonusPct = 135

	#the continent lists computed in step 5
	global postContinentLists
	#goody huts
	impGoodyHut = gc.getInfoTypeForString("IMPROVEMENT_GOODY_HUT")
	players = gc.getGame().countCivPlayersEverAlive()
	plotsForGoodies = []
	
	for px in range(width):
		for py in range(height):
			if tooCloseToPole(px,py,2):
				continue
			pPlot = cymap.plot(px,py)
			#record all plots that are potential goody locations
			if pPlot.canHaveImprovement(impGoodyHut, TeamTypes.NO_TEAM, False):
				plotsForGoodies.append(pPlot)
			

	#add randomly placed goody huts, 6 per player overall, plus some extra to make them common enough
	#if you play with only a few players on a big map
	goodyHutCount = 1+ int(((7 * players + int(len(plotsForGoodies)/40)) * bonusPct) / 100) 
	addedLocs = []
	while (goodyHutCount > 0):
		if len(plotsForGoodies) <= 0:
			break
		index = dice.get(len(plotsForGoodies), "Python: SmartMap: Dice: GoodyHut Placement")
		pPlot = plotsForGoodies[index]
		x = pPlot.getX()
		y = pPlot.getY()
		
		tooClose = False
		for usedLoc in addedLocs:
			distX,distY = getWrapDistanceXY((x,y),usedLoc)
			if distX + distY < 4:
				tooClose = true
				break
		if tooClose:
			del plotsForGoodies[index] # Remove this plot from the eligible list, it was too close to another goody
			continue
		
		
		addedLocs.append((x,y))
		pPlot.setImprovementType(impGoodyHut)
		goodyHutCount -= 1
		del plotsForGoodies[index]
		
	if goodyHutCount > 0:
		print "failed to place",goodyHutCount,"goody huts"
	OutputMessage("Python: SmartMap: Step 6 Goody Huts:complete")
	checkForBadPlots()
	return 0
		
	
##########################################################################
#Step 7: place all the players
def assignStartingPlots():
	global width
	global height
	global centralPole
	
	OutputMessage("Python: SmartMap: Step 7 Assign starting plots")
	checkForBadPlots()
	gc = CyGlobalContext()
	cymap = CyMap()
	dice = gc.getGame().getMapRand()
	
	#find out the user selected placement method
	placeMethod = getSelectedMapValue("Start Placement:")
	print "    placeMethod",placeMethod

	if placeMethod == "standard":
		CyPythonMgr().allowDefaultImpl()
		OutputMessage("Python: SmartMap: Step 7 Assign starting plots: complete standard")
		return

	newWorld = False
	playerInlandStart = False
	playerCoastalStart = False
	preferInland = False
	preferCoastal = False
	if placeMethod == "SmartMap New World":
		newWorld = True
	if placeMethod == "SmartMap New World Coastal":
		newWorld = True
		playerCoastalStart = True
		preferCoastal = True
	if placeMethod == "SmartMap New World Inland":
		newWorld = True
		playerInlandStart = True
		preferInland = True
	if placeMethod == "SmartMap Player Inland":
		playerInlandStart = True
		preferInland = True
	if placeMethod == "SmartMap Player Coastal":
		playerCoastalStart = True
		preferCoastal = True
	if placeMethod == "SmartMap All Inland":
		preferInland = True
	if placeMethod == "SmartMap All Coastal":
		preferCoastal = True

	
	#the continent lists computed in step 5
	global postContinentLists
	
	#now we know how the players get distributed per continent
	#so we just have to pick the particular starting plot on each continent per player
	
	print "    start pre-computing plot scores"
	#pre-compute plot scores for all land plots
	plotScores = [0]*(width*height)
	baseScores = [0]*(width*height)
	for px in range(width):
		for py in range(height):
			nearbyBonusTypes = []
			pickedPlot = cymap.plot(px,py)
			if pickedPlot.isWater():
				continue
			if pickedPlot.isPeak():
				continue
			pindex = getTileIndexAtXY(px,py)
			pickedScore = 0
			#the value of food etc on the actual plot tends not to matter because of the automatic
			#city minimums
			#pickedScore += 10 * int(pickedPlot.getYield(YieldTypes.YIELD_FOOD) + pickedPlot.getYield(YieldTypes.YIELD_PRODUCTION) + pickedPlot.getYield(YieldTypes.YIELD_COMMERCE))

			#other bonuses for the plot's score
			if pickedPlot.isFreshWater():
				pickedScore += 1800
			#extra points for the defensive bonus of landing on a hill
			if pickedPlot.isHills():
				pickedScore += 1200

			#a coastal start tends to be far from other players, and advantageous (can build ocean-only
			#buildings, etc), and a hill can be a big production boost
			nearOneHill = False
			nearOneOcean = False
			for nearDir in nearDirs:
				newX, newY = getTileCoordXYWithWrap((px,py), nearDir)
				if newX == px and newY == py:
					continue
				pTouchPlot = cymap.plot(newX,newY)
				
				#typically an ocean touching start is highly desireable
				if not preferInland and pTouchPlot.isWater() and not pTouchPlot.isLake() and not nearOneOcean:
					nearOneOcean = True
					pickedScore += 1800
					continue
					
				if preferCoastal and isCoastalPlot(px,py):
					pickedScore += 1800
				if preferInland and not isCoastalPlot(px,py):
					pickedScore += 1800
					
				# a lot of players like to start with a hill for production and defense
				if pTouchPlot.isHills() and not nearOneHill:
					nearOneHill = True
					pickedScore += 600
					continue
					
			#starting too close	to the pole can hamper your growth	
			poleCheck = max(int(width/8), int(height/8))
			if not tooCloseToPole(px,py,poleCheck):
				pickedScore += 2400
			if not tooCloseToPole(px,py,7):
				pickedScore += 1200
			
				
			if pickedPlot.isGoody():
				pickedScore -= 200
			bonusType = pickedPlot.getBonusType(-1)
			if bonusType != BonusTypes.NO_BONUS and bonusType not in nearbyBonusTypes:
				nearbyBonusTypes.append(bonusType)
				pickedScore += 200
				bonusInfoXML = gc.getBonusInfo(bonusType)
				type_string = bonusInfoXML.getType()
				if type_string in resourcesImportantEarly:
					pickedScore += 200
				if type_string in resourcesStrategic:
					pickedScore += 400
				
			baseScores[pindex] = pickedScore

			for nearDir in playerStartCross:
				newX, newY = getTileCoordXYWithWrap((px,py), nearDir)
				if newX == px and newY == py:
					continue
				pTouchPlot = cymap.plot(newX,newY)
				pickedScore += 12 * int(pTouchPlot.getYield(YieldTypes.YIELD_FOOD) + pTouchPlot.getYield(YieldTypes.YIELD_PRODUCTION) + pTouchPlot.getYield(YieldTypes.YIELD_COMMERCE))
				if pTouchPlot.isFreshWater():
					pickedScore += 40
					if pTouchPlot.isFlatlands():
						pickedScore += 40
				if pTouchPlot.isGoody():
					pickedScore += 200
				if pTouchPlot.isHills():
					pickedScore += 50
				if pTouchPlot.isPeak():
					pickedScore -= 80
				if pTouchPlot.isWater():
					if pTouchPlot.isLake():
						pickedScore += 40
					else:
						pickedScore -= 60
						if preferInland:
							pickedScore -= 100
					
				bonusType = pTouchPlot.getBonusType(-1)
				if bonusType != BonusTypes.NO_BONUS and bonusType not in nearbyBonusTypes:
					nearbyBonusTypes.append(bonusType)
					pickedScore += 200
					bonusInfoXML = gc.getBonusInfo(bonusType)
					type_string = bonusInfoXML.getType()
					if type_string in resourcesImportantEarly:
						pickedScore += 200
					if type_string in resourcesStrategic:
						pickedScore += 400
						
			plotScores[pindex] = pickedScore
	print "    finished pre-computing plot scores"
	
	#now for each continent, figure out the total score of all the land tiles it owns
	print "    score continents"
	postContinentScores = []
	for postContinentList in postContinentLists:
		totalYield = 0
		for plotPos in postContinentList:
			plotX, plotY = plotPos
			pindex = getTileIndexAtXY(plotX,plotY)
			totalYield += baseScores[pindex]
		postContinentScores.append(totalYield)
			
	print "    postContinentScores", postContinentScores
	
	players = gc.getGame().countCivPlayersEverAlive()
	
	#in new world mode, drop any continent list in the western hemisphere
	if newWorld:
		foundContinentToDelete = True
		while foundContinentToDelete and len(postContinentScores) > 1:
			foundContinentToDelete = False
			for index in range(len(postContinentScores)):
				conList = postContinentLists[index]
				countInEast = 0
				totalTiles = len(conList)
				for conTile in conList:
					tileX, tileY = conTile
					if tileX > int(width/2)+1:
						countInEast += 1
				if countInEast > int(totalTiles/10):
					continue
				
				del postContinentScores[index]
				del postContinentLists[index]
				foundContinentToDelete = True
				break

	#drop the lowest scoring continents over the min # of players because we won't use them
	while len(postContinentScores) > players:
		for index in range(len(postContinentScores)):
			if postContinentScores[index] == min(postContinentScores):
				del postContinentScores[index]
				del postContinentLists[index]
				break
		
	#pair continent lists with their scores, and sort by the scores		
	listsWithScores = []
	for index in range(len(postContinentLists)):
		listsWithScores.append( (postContinentLists[index], postContinentScores[index]) )
		
	listsWithScores = sorted(listsWithScores, key=operator.itemgetter(1))	
	listsWithScores.reverse()
	#print "    after drop listsWithScores", listsWithScores
	
	#assign initial players to continents, just round robbin, conveniently this will
	#automatically weight the early list entries higher
	playerContinentAssignments = []
	for continent in range(len(listsWithScores)):
		contList, score = listsWithScores[continent]
		playerContinentAssignments.append((score,0))
	playerList = []
	for playerNum in range(players):
		playerList.append(playerNum)
		
	playerContinent = 0
	for playerIndex in range(players):
		playerNum = playerList[playerIndex]
		score, numPlayers = playerContinentAssignments[playerContinent]
		contList, contScore = listsWithScores[playerContinent]
		score = contScore
		numPlayers += 1
		playerContinentAssignments[playerContinent] = (score, numPlayers)
		playerContinent += 1
		if playerContinent >= len(postContinentLists):
			playerContinent = 0
			
	print "    initial playerContinentAssignments",playerContinentAssignments
		
	#while any imbalance in the per-continent score divided by the number of 
	#players placed on that continent exists, rebalance players by moving them
	#from a lower scoring position to a higher scoring position
	anyImbalance = True
	tries = 0
	while anyImbalance and tries < (2*players+2):
		tries += 1
		anyImbalance = False
		#first step: find out the min/max existing scores
		minScore = -1
		maxScore = 0
		minScoreIndex = len(playerContinentAssignments)-1
		maxScoreIndex = 0
		for index in range(len(playerContinentAssignments)):
			score, numPlayers = playerContinentAssignments[index]
			if numPlayers > 0:
				avgScore = int(score / numPlayers)
				if avgScore < minScore or minScore == -1:
					minScore = avgScore
					minScoreIndex = index
				if avgScore > maxScore:
					maxScore = avgScore
					maxScoreIndex = index
			
		#find out if there is even possible room to place another player on the largest continent
		#we'll say that each player requires more than 4x4 tiles, at a minimum
		maxContScore, maxContPlayers = playerContinentAssignments[maxScoreIndex]
		if len(postContinentLists[maxScoreIndex]) > ((maxContPlayers+1) * 17):
			alternativeTheory = []
			altminScore = -1
			altmaxScore = 0
			for index in range(len(playerContinentAssignments)):
				score, numPlayers = playerContinentAssignments[index]
				if index == maxScoreIndex:
					numPlayers += 1  
				if index == minScoreIndex:
					numPlayers -= 1  
				if numPlayers < 0:
					numPlayers = 0
					
				if numPlayers > 0:
					avgScore = int(score / numPlayers)
					if avgScore < altminScore or altminScore == -1:
						altminScore = avgScore
					if avgScore > altmaxScore:
						altmaxScore = avgScore
					
				alternativeTheory.append((score,numPlayers))
				
			print "    minScore",minScore,"maxScore",maxScore,"theory altminScore",altminScore,"theory altmaxScore",altmaxScore
			print "    theory",alternativeTheory
			
			#if the alternative theory has better min score, or equal minscore with reduced variance (lower maxscore)
			#then we accept the alternative theory, and go through the loop again to see if we can do any better
			if altminScore > minScore:
				playerContinentAssignments = alternativeTheory
				anyImbalance = True
				print "    theory succeeds on minScore"
			elif altminScore == minScore and altmaxScore < maxScore:
				playerContinentAssignments = alternativeTheory
				anyImbalance = True
				print "    theory succeeds on maxScore"
			else:
				print "    theory fails"
			
	#now we know how many players will be placed on each continent
	print "    final playerContinentAssignments",playerContinentAssignments
	
	#knowing the continents each player will start on, we must pick a starting plot on that continent
	#for the player
	finalPlots = []
	for index in range(len(playerContinentAssignments)):
		score, numPlayers = playerContinentAssignments[index]
		if numPlayers <= 0:
			continue
		print "    find placements for continent", index, "with players", numPlayers

		#keep track of the best scoring plots (the winner we'll use)
		bestScore = 0	
		bestSeparation = 0 	
		bestPickedPlotsXY = []
		#keep track of the best separated plots, to require that the 
		#finally chosen plots must come reasonably close to maximum separation
		bestMinSeparation = 0
		continentPlotList, continentScore = listsWithScores[index]
		
		print "    estimate continent separation for continent",index
		#calculate a base expected separation
		maxPossibleSeparation = 0
		for tries in range(128 * numPlayers):
			plotIndexA = dice.get(len(continentPlotList), "Python: SmartMap: Dice: continent size estimator")
			plotIndexB = dice.get(len(continentPlotList), "Python: SmartMap: Dice: continent size estimator")
			plotA = continentPlotList[plotIndexA] 
			plotB = continentPlotList[plotIndexB] 
			distX, distY = getWrapDistanceXY(plotA,plotB)
			separation = int(((distX*distX) + (distY*distY))**0.5)
			if separation > maxPossibleSeparation:
				maxPossibleSeparation = separation
					
		demandSeparation = int(75*(maxPossibleSeparation/(numPlayers+1))/100)
		print "    continent", index, "estimated initial separation", demandSeparation
		
		#can place players on flat or hills plots 
		allStartPlotsCopy = []
		for plotA in continentPlotList:
			plotX,plotY = plotA
			if tooCloseToPole(plotX,plotY,6):
				continue
			pPlot = cymap.plot(plotX,plotY)
			if not (pPlot.isFlatlands() or pPlot.isHills()):
				continue

			plotIndex = getTileIndexAtXY(plotX,plotY)
			plotScore = plotScores[plotIndex]
			bestPlot = True
			sX = plotX - (plotX % 4)
			sY = plotY - (plotY % 4)
			for aX in range(4):
				for aY in range(4):
					nearX, nearY = getTileCoordXYWithWrap((sX,sY), (aX,aY))
					nearIndex = getTileIndexAtXY(nearX,nearY)
					nearScore = plotScores[nearIndex]
					if nearScore > plotScore:
						bestPlot = False
						break
				if not bestPlot:
					break
			if not bestPlot:
				continue
			allStartPlotsCopy.append(plotA)
			
		print "   pick start positions from possible start positions:",len(allStartPlotsCopy)	
			
		for tries in range(64):
			pickedPlotsXY = []
			allPlots = allStartPlotsCopy
			while len(allPlots) >= (numPlayers - len(pickedPlotsXY)) and len(pickedPlotsXY) < numPlayers:
				plotIndex = dice.get(len(allPlots), "Python: SmartMap: Dice: Player Placement Plot Pick")
				plotXY = allPlots[plotIndex]
				#add this randomly picked plot to the set
				pickedPlotsXY.append(allPlots[plotIndex])
				#replace list with one with only sufficiently ranged plots
				replaceList = []
				for otherPlotXY in allPlots:
					distX,distY = getWrapDistanceXY(plotXY, otherPlotXY)
					separation = int(((distX*distX) + (distY*distY))**0.5)
					if separation > demandSeparation:
						replaceList.append(otherPlotXY)
						
				allPlots = replaceList
				#print "picked",pickedPlotsXY,"remaining",allPlots
				
			#print "found",len(pickedPlotsXY),"needed",numPlayers
			if len(pickedPlotsXY) < numPlayers:
				continue
				
			altPlotsXY = []
			for plot in pickedPlotsXY:
				givenPlotX, givenPlotY = plot
				altPlotX=-1
				altPlotY=-1
				givenPlotIndex = getTileIndexAtXY(givenPlotX,givenPlotY)
				bestNearScore = plotScores[givenPlotIndex]
				for nearDir in playerStartCross:
					altX,altY = getTileCoordXYWithWrap(plot, nearDir)
					altPlot = cymap.plot(altX,altY)
					if not (altPlot.isFlatlands() or altPlot.isHills()):
						continue
					altScoreIndex = getTileIndexAtXY(altX,altY)
					if altPlotX == -1 or altPlotY == -1 or plotScores[altScoreIndex] > bestNearScore:
						bestNearScore = plotScores[altScoreIndex]
						altPlotX = altX
						altPlotY = altY
				if altPlotX == -1 or altPlotY == -1:
					altPlotsXY.append(plot)
				else:
					altPlotsXY.append((altPlotX,altPlotY))
						
			for plotSet in [pickedPlotsXY,altPlotsXY]:
				#print "    considering plots",plotSet 
				totalScore = 0
				for picked in plotSet:	
					plotX,plotY = picked
					pickedIndex = getTileIndexAtXY(plotX,plotY)
					totalScore += plotScores[pickedIndex]
					
				#find the minimum separation among the starting positions on this continent
				minSeparation = -1
				for firstIndex in range(len(plotSet)):
					firstPlot = plotSet[firstIndex]
					for secondIndex in range(firstIndex, len(plotSet)):
						if secondIndex != firstIndex:
							secondPlot = plotSet[secondIndex]
							distX, distY = getWrapDistanceXY(firstPlot,secondPlot)
							dist = int(((distX*distX) + (distY*distY))**0.5)
							if dist < minSeparation or minSeparation == -1:
								minSeparation = dist
					#this helps push starting positions away from positions on other continents			
					for anotherPlot in finalPlots:
						distX, distY = getWrapDistanceXY(firstPlot,anotherPlot)
						dist = int(160*((((distX*distX) + (distY*distY))**0.5))/100)
						if dist < minSeparation or minSeparation == -1:
							minSeparation = dist
							
				totalScore += minSeparation * int(bestScore / 100)
					
				#we require that the final placement have at least XX% of the best separation ever found
				if minSeparation > bestMinSeparation:
					bestMinSeparation = minSeparation
					demandSeparation = int ((75*bestMinSeparation)/100)	
					#print "        bestMinSeparation",bestMinSeparation,"demandSeparation",demandSeparation,"tries",tries,"for tiles",plotSet
				#if empty list, or better score exceeding demand separation, or old score doesn't exceed demand but this one does		
				if (bestPickedPlotsXY==[] or (totalScore > bestScore and minSeparation >= demandSeparation) 
					or (totalScore > bestScore and minSeparation >= bestSeparation)
					or  (bestSeparation < demandSeparation and minSeparation >= demandSeparation) ):
					bestScore = totalScore
					bestSeparation = minSeparation
					bestPickedPlotsXY = plotSet
					print "      new bestPickedPlotsXY",bestPickedPlotsXY,"sep",bestSeparation,"score",bestScore,"demand",demandSeparation,"numPlayers",numPlayers,"tries",tries
			
			
		print "    found placements for continent", index, "with players", numPlayers, "places",bestPickedPlotsXY
		for plotXY in bestPickedPlotsXY:
			finalPlots.append(plotXY)
			
	print "    tentative player Plots",finalPlots

	#now that we have very good plots, check nearby and make sure we didn't make a close miss
	altFinalPlots = []
	for plot in finalPlots:
		givenPlotX, givenPlotY = plot
		givenPlotIndex = getTileIndexAtXY(givenPlotX,givenPlotY)
		altPlotX = givenPlotX
		altPlotY = givenPlotY
		bestNearScore = plotScores[givenPlotIndex]
		for nearDir in playerStartCross:
			altX,altY = getTileCoordXYWithWrap(plot, nearDir)
			otherPlot = cymap.plot(altX,altY)
			if not (otherPlot.isFlatlands() or otherPlot.isHills()):
				continue
			scoreIndex = getTileIndexAtXY(altX,altY)
			if plotScores[scoreIndex] > bestNearScore:
				bestNearScore = plotScores[scoreIndex]
				altPlotX = altX
				altPlotY = altY
		altFinalPlots.append((altPlotX,altPlotY))
		
	#these are the finalized plots which will be assigned to the players	
	finalPlots = altFinalPlots
	print "    final player Plots",finalPlots
	
	#maybe there was an error: we didn't find enough plots, so we'll let the standard
	#placement algorithm take care of it for us
	if len(finalPlots) < players:
		CyPythonMgr().allowDefaultImpl()
		OutputMessage("Python: SmartMap: Step 7 Assign starting plots: failed smartmap so complete standard")
		return
	
	#finally, we can actually assign the final plots to the players
	for playerIndex in range(players):
		whichStartIndex = dice.get(len(finalPlots), "Python: SmartMap: Dice: Player Placement Plot Pick Place")
		#start player 0 (human in single player) near the middle IF wrapping
		if playerIndex == 0 and (wrapX or wrapY) and not centralPole:
			nearestMidScore = 0
			nearestMidIndex = 0
			for startIndex in range(len(finalPlots)):
				plotScore = 0
				startPlotX,startPlotY = finalPlots[startIndex]
				if playerCoastalStart and not isCoastalPlot(startPlotX,startPlotY):
					continue
				if playerInlandStart and isCoastalPlot(startPlotX,startPlotY):
					continue
				if wrapX:
					distFromCenterFactor = (width*width) - (abs(startPlotX-int(width/2))*abs(startPlotX-int(width/2)))
					plotScore += distFromCenterFactor
				if wrapY:
					distFromCenterFactor = (height*height) - (abs(startPlotY-int(height/2))*abs(startPlotY-int(height/2)))
					plotScore += distFromCenterFactor
				if plotScore > nearestMidScore:
					nearestMidScore = plotScore
					nearestMidIndex = startIndex
			whichStartIndex = nearestMidIndex
			
			
		#pick one of the final placement plots at random
		startX,startY = finalPlots[whichStartIndex]
		del finalPlots[whichStartIndex]
		startPlot = cymap.plot(startX,startY)
		#assign the player to it
		startPlot.setStartingPlot(True) #I believe this helps the post-processor improve map fairness
		pPlayer = gc.getPlayer(playerIndex)
		pPlayer.setStartingPlot(startPlot,True)
		playerName = pPlayer.getName()
		print "player",playerName,"placed at",startX,startY,"coastal",isCoastalPlot(startX,startY)
				
	OutputMessage("Python: SmartMap: Step 7 Assign starting plots: complete")
	checkForBadPlots()
	return 0
	
	
###########################################################################
# Stuff associated with menus
###########################################################################
selection_titles = [unicode("Continents:"),
                   unicode("Terrain:"),
                   unicode("Features:"),
                   unicode("Bonuses:"),
                   unicode("Goodies:"),
                   unicode("Ocean:"),
                   unicode("Hills:"),
                   unicode("Peaks:"),
                   unicode("Forest/Jungle:"),
                   unicode("Rivers/Lakes:"),
                   unicode("Land Style:"),
                   unicode("Start Placement:"),
                   unicode("Wrap:"),
                   unicode("Override Width:"),
                   unicode("Override Height:"),
                   unicode("Override Separation:"),
                   ]
	
#dropdown list names and corresponding values used in code
selection_names_and_values = [
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["no continents (all fragments)",0,True],
	                   ["1 pangea",1,True],
	                   ["2 huge lands",2,True],
	                   ["3 large lands",3,True],
	                   ["4 continents",4,True],
	                   ["5 continents",5,True],
	                   ["6 continents",6,True],
	                   ["7 continents",7,True],
	                   ["8 continents",8,True],
	                   ["12 islands",12,True],
	                   ["18 isles",18,True],
	                   ["24 isles",24,False],
	                   ["36 islets",36,False],
	                   ["single tile isles",-20,False],
	                   ["1 per player",-7,True],
	                   ["1 per 2 players",-8,False],
	                   ["1 per 3 players",-9,False],
	                   ["1 per 4 players",-10,False],
	                   ["2 per player",-11,True],
	                   ["3 per player",-12,False],
	                   ["4 per player",-13,False],
	                   ["random 1-6",-1,True],
	                   ["random 2-5",-2,True],
	                   ["random 3-8",-3,True],
	                   ["random 4-12",-4,True],
	                   ["random 1-12",-6,True],
	                   ["random 2-7",-5,True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["standard","standard",True],
	                   ["great plains","great plains",True],
	                   ["oasis","oasis",True],
	                   ["SmartMap no snow","SmartMap no snow",True],
	                   ["SmartMap no desert","SmartMap no desert",True],
	                   ["SmartMap no snow/desert","SmartMap no snow/desert",True],
	                   ["SmartMap no snow/desert/tundra","SmartMap no snow/desert/tundra",True],
	                   ["SmartMap only plains","SmartMap only plains",True],
	                   ["SmartMap only grass","SmartMap only grass",True],
	                   ["SmartMap extra grass","SmartMap extra grass",True],
	                   ["SmartMap extra plains","SmartMap extra plains",True],
	                   ["SmartMap","SmartMap",True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["standard","standard",True],
	                   ["great plains","great plains",True],
	                   ["oasis","oasis",True],
	                   ["SmartMap","SmartMap",False],
	                   ["SmartMap Strict","SmartMap Strict",True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["standard","standard",True],
	                   ["SmartMap Few","SmartMap Few",True],
	                   ["SmartMap Many","SmartMap Many",False],
	                   ["SmartMap Clump Happy","SmartMap Clump Happy",True],
	                   ["SmartMap Clump Health","SmartMap Clump Health",True],
	                   ["SmartMap Clump Both","SmartMap Clump Both",True],
	                   ["SmartMap Restricted","SmartMap Restricted",True],
	                   ["SmartMap Most Restricted","SmartMap Most Restricted",True],
	                   ["SmartMap Normal","SmartMap Normal",True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["standard","standard",True],
	                   ["None","None",True],
	                   ["SmartMap Few","SmartMap Few",True],
	                   ["SmartMap Many","SmartMap Many",True],
	                   ["SmartMap Normal","SmartMap Normal",True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["no ocean, no lakes",-5,False],
	                   ["no ocean",-4,False],
	                   ["central sea",-6,False],
	                   ["corner sea",-7,False],
	                   ["edge sea",-8,False],
	                   ["random sea",-9,False],
	                   ["few great lakes",-10,False],
	                   ["great lakes",-11,False],
	                   ["many great lakes",-12,False],
	                   ["minimal ocean",10,False],
	                   ["40% ocean",40,False],
	                   ["50% ocean",50,False],
	                   ["60% ocean",60,True],
	                   ["65% ocean",65,True],
	                   ["70% ocean",70,True],
	                   ["75% ocean",75,True],
	                   ["80% ocean",80,True],
	                   ["85% ocean",85,True],
	                   ["90% ocean",90,False],
	                   ["95% ocean",95,False],
	                   ["99% ocean",99,False],
	                   ["small ocean",-1,True],
	                   ["big ocean",-2,True],
	                   ["normal ocean",-3,True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["no hills",0,False],
	                   ["2% hills",2,False],
	                   ["4% hills",4,True],
	                   ["6% hills",6,True],
	                   ["8% hills",8,True],
	                   ["10% hills",10,True],
	                   ["12% hills",12,True],
	                   ["14% hills",14,True],
	                   ["16% hills",16,True],
	                   ["18% hills",18,True],
	                   ["20% hills",20,True],
	                   ["25% hills",25,False],
	                   ["30% hills",30,False],
	                   ["40% hills",40,False],
	                   ["50% hills",50,False],
	                   ["few hills",-1,True],
	                   ["many hills",-2,True],
	                   ["normal hills",-3,True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["no peaks",0,False],
	                   ["1% peaks",1,False],
	                   ["2% peaks",2,True],
	                   ["3% peaks",3,True],
	                   ["4% peaks",4,True],
	                   ["5% peaks",5,True],
	                   ["6% peaks",6,True],
	                   ["7% peaks",7,True],
	                   ["8% peaks",8,True],
	                   ["9% peaks",9,True],
	                   ["10% peaks",10,False],
	                   ["15% peaks",15,False],
	                   ["20% peaks",20,False],
	                   ["20% peaks",25,False],
	                   ["few peaks no block",-4,True],
	                   ["many peaks no block",-5,True],
	                   ["normal peaks no block",-6,True],
	                   ["few peaks",-1,True],
	                   ["many peaks",-2,True],
	                   ["normal peaks",-3,True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["light forest,normal jungle", (-200,0),True],
	                   ["light forest,heavy jungle", (-200,200),True],
	                   ["normal forest,light jungle", (0,-200),True],
	                   ["normal forest,heavy jungle", (0,200),True],
	                   ["heavy forest,light jungle", (200,-200),True],
	                   ["heavy forest,normal jungle", (200,0),True],
	                   ["light forest,light jungle", (-200,-200),True],
	                   ["heavy forest,heavy jungle", (200,200),True],
	                   ["normal forest,normal jungle", (0,0),True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["few rivers, normal lakes", (80,100),True],
	                   ["few rivers, many lakes", (80,200),True],
	                   ["normal rivers, few lakes", (100,33),True],
	                   ["normal rivers, many lakes", (100,200),True],
	                   ["many rivers, few lakes", (300,33),True],
	                   ["many rivers, normal lakes", (300,100),True],
	                   ["few rivers, few lakes", (80,33),True],
	                   ["many rivers, many lakes", (300,200),True],
	                   ["normal rivers, normal lakes", (100,100),True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["least round, extreme fragments", (-800,800),False],
	                   ["least round, heavy fragments", (-800,600),False],
	                   ["least round, many fragments", (-800,400),True],
	                   ["least round, some fragments", (-800,200),True],
	                   ["least round, no fragments", (-800,0),True],
	                   ["somewhat round, extreme fragments", (-400,800),False],
	                   ["somewhat round, heavy fragments", (-400,600),False],
	                   ["somewhat round, many fragments", (-400,400),True],
	                   ["somewhat round, some fragments", (-400,200),True],
	                   ["somewhat round, no fragments", (-400,0),True],
	                   ["very round, extreme fragments", (0,800),False],
	                   ["very round, heavy fragments", (0,600),False],
	                   ["very round, many fragments", (0,400),True],
	                   ["very round, some fragments", (0,200),True],
	                   ["very round, no fragments", (0,0),True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["standard","standard",True],
	                   ["SmartMap Player Inland","SmartMap Player Inland",True],
	                   ["SmartMap Player Coastal","SmartMap Player Coastal",True],
	                   ["SmartMap All Inland","SmartMap All Inland",True],
	                   ["SmartMap All Coastal","SmartMap All Coastal",True],
	                   ["SmartMap New World","SmartMap New World",True],
	                   ["SmartMap New World Inland","SmartMap New World Inland",True],
	                   ["SmartMap New World Coastal","SmartMap New World Coastal",True],
	                   ["SmartMap","SmartMap",True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["wrap y - central pole",(False,True,True),False],
	                   ["wrap both - central pole",(True,True,True),False],
	                   ["no wrap - central pole",(False,False,True),False],
	                   ["wrap x - central pole",(True,False,True),False],
	                   ["wrap y",(False,True,False),True],
	                   ["wrap both",(True,True,False),True],
	                   ["no wrap",(False,False,False),True],
	                   ["wrap x",(True,False,False),True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["16",4,False],
	                   ["24",6,False],
	                   ["32",8,False],
	                   ["40 (duel)",10,True],
	                   ["48",12,True],
	                   ["52 (small)",13,True],
	                   ["56",14,True],
	                   ["64 (standard)",16,True],
	                   ["72",18,True],
	                   ["80",20,True],
	                   ["84 (large)",21,False],
	                   ["100",25,False],
	                   ["104",26,False],
	                   ["120",30,False],
	                   ["128 (huge)",32,False],
	                   ["140",35,False],
	                   ["160",40,False],
	                   ["180",45,False],
	                   ["200",50,False],
	                   ["240",60,False],
	                   ["280",70,False],
	                   ["320",80,False],
	                   ["360",90,False],
	                   ["400",100,False],
	                   ["480",120,False],
	                   ["560",140,False],
	                   ["640",160,False],
	                   ["720",180,False],
	                   ["800",200,False],
	                   ["880",220,False],
	                   ["1000",250,False],
	                   ["1200",300,False],
	                   ["1400",350,False],
	                   ["1600",400,False],
	                   ["1800",450,False],
	                   ["2000",500,False],
	                   ["2400",600,False],
	                   ["2800",700,False],
	                   ["3200",800,False],
	                   ["3600",900,False],
	                   ["4000",1000,False],
	                   ["5000",1250,False],
	                   ["random 64-256",-4,False],
	                   ["random 32-128",-3,False],
	                   ["random 16-64",-2,True],
	                   ["don't override",-1,True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["16",4,False],
	                   ["24 (duel)",6,False],
	                   ["32 (tiny)",8,False],
	                   ["40 (small)",10,True],
	                   ["48",12,True],
	                   ["52 (standard)",13,True],
	                   ["56",14,True],
	                   ["64 (large)",16,True],
	                   ["72",18,True],
	                   ["80 (huge)",20,True],
	                   ["84",21,False],
	                   ["100",25,False],
	                   ["104",26,False],
	                   ["120",30,False],
	                   ["128",32,False],
	                   ["140",35,False],
	                   ["160",40,False],
	                   ["180",45,False],
	                   ["200",50,False],
	                   ["240",60,False],
	                   ["280",70,False],
	                   ["320",80,False],
	                   ["360",90,False],
	                   ["400",100,False],
	                   ["480",120,False],
	                   ["560",140,False],
	                   ["640",160,False],
	                   ["720",180,False],
	                   ["800",200,False],
	                   ["880",220,False],
	                   ["1000",250,False],
	                   ["1200",300,False],
	                   ["1400",350,False],
	                   ["1600",400,False],
	                   ["1800",450,False],
	                   ["2000",500,False],
	                   ["2400",600,False],
	                   ["2800",700,False],
	                   ["3200",800,False],
	                   ["3600",900,False],
	                   ["4000",1000,False],
	                   ["5000",1250,False],
	                   ["random 64-256",-4,False],
	                   ["random 32-128",-3,False],
	                   ["random 16-64",-2,True],
	                   ["don't override",-1,True],
	                   ],
	                  [
	                   ["smartmap random pick","smartmap random pick",False],
	                   ["0 sealevel low-no separation",0,True],
	                   ["1 single coast separation",1,True],
	                   ["2 coasts touch",2,True],
	                   ["3 frequent coast crossing",3,True],
	                   ["4 sealevel medium-rare coast crossing",4,True],
	                   ["5 no coastal crossing",5,True],
	                   ["6 developing culture touch",6,True],
	                   ["7 sealevel high",7,True],
	                   ["8 refined culture touch",8,True],
	                   ["9 refined+influential touch",9,True],
	                   ["10 influential culture touch",10,True],
	                   ["11 influential+legendary touch",11,True],
	                   ["12 legendary culture touch",12,True],
	                   ["13 no culture touch",13,True],
	                   ["don't override",-1,True],
	                   ],
	                   ]

#I don't know where this shows up!  You would think this would be
#used as the text for the Map: button, but that uses the name of
#the script instead!
def getDescription():
	#print "    getDescription"
	return unicode("SmartMap")

#return the total number of custom options for this map
def getNumCustomMapOptions():
	return len(selection_names_and_values)

#return the list of options to configure how this map is generated
def getCustomMapOptionName(argsList):
	#print "getCustomMapOptionName",argsList
	index = argsList[0]
	return selection_titles[index]

#this custom menu function returns the 'value' from a map option selected by the player
#for use by other code.
def getSelectedMapValue(optionName):
	gc = CyGlobalContext()
	cymap = CyMap()
	dice = gc.getGame().getMapRand()
	optionIndex = 0
	found = False
	for optionIndex in range(len(selection_titles)):
		title = selection_titles[optionIndex]
		if title == optionName:
			found = True
			break
			
	if not found:
		print "    !error, couldn't find entry for option named",optionName
		
	#get the user selected index (which item in the dropdown list the user selected)
	userSelectedIndex = int(cymap.getCustomMapOption(optionIndex))
	if userSelectedIndex >= len(selection_names_and_values[optionIndex]):
		userSelectedIndex = 0
		
	if userSelectedIndex == 0:
		userSelectedIndex = cachedMenuChoices[optionIndex]
		
	tries = 0
	while userSelectedIndex == 0 and tries < 100:
		userSelectedIndex = 1+ dice.get(len(selection_names_and_values[optionIndex])-1, "Python: SmartMap: Dice: Extra Water Reach")
		if not selection_names_and_values[optionIndex][userSelectedIndex][2]:
			userSelectedIndex = 0
	
	if userSelectedIndex == 0:
		done = False
		while not done:
			userSelectedIndex += 1
			if selection_names_and_values[optionIndex][userSelectedIndex][2]:
				done = True
				
	cachedMenuChoices[optionIndex] = userSelectedIndex

	#print "get userSelectedIndex", userSelectedIndex, "optionIndex",optionIndex
	#return the corresponding integer value
	return selection_names_and_values[optionIndex][userSelectedIndex][1]
	                   
#return the number of drop down entries for each custom map option
def getNumCustomMapOptionValues(argsList):
	#print "getNumCustomMapOptionValues",argsList
	index = argsList[0]
	return len(selection_names_and_values[index])
	                   
#return the description of the custom map options (for use in the dropdown)	
def getCustomMapOptionDescAt(argsList):
	#print "getCustomMapOptionDescAt",argsList
	iOption = argsList[0]
	iSelection = argsList[1]
	return unicode(selection_names_and_values[iOption][iSelection][0])
	
#return the default drop-down entry for each custom map options
def getCustomMapOptionDefault(argsList):
	#print "getCustomMapOptionDefault",argsList
	index = argsList[0]
	#this chooses the last option in every case, which i've picked out to be a good option
	result = len(selection_names_and_values[index]) - 1
	try:
		settings = open('smartmap84.cfg', 'r')
		optionsWithDefault = pickle.load(settings)
		#print optionsWithDefault
		for optionIndex in range(len(optionsWithDefault)):
			option, valueName = optionsWithDefault[index]
			if option == selection_titles[index]:
				for selectionIndex in range(len(selection_names_and_values[index])):
					selectionName, selectionValue, allowedRandomPick = selection_names_and_values[index][selectionIndex]
					if selectionName == valueName:
						result = selectionIndex
		#print "    loaded option from cfg:",option,value
		settings.close()
	except IOError:
		print "    Couldn't find smartmap84.cfg"
	except EOFError:
		print "    Bad contents in smartmap84.cfg"
		
	return result
	#replace with return -1 to set 'random' as the default choice
	#note that this is dangerous because it might pick an extremely large random choice of
	#width and height
	#return -1

#this is an advanced map, only shows up if you do custom map
def isAdvancedMap():
	"This map should not show up in simple mode"
	return 1

#this map uses the sea level option to determine continent separation minimum
def isSeaLevelMap():
	return 1
	
#this map may or may not obey the climate selector depending on which
#terrain and feature gen is chosen
def isClimateMap():
	return 1
	
###########################################################################
#Assorted helper functions
###########################################################################

def beforeInit():
	global cachedMenuChoices
	print "    before init"
	cachedMenuChoices = []
	for i in range(len(selection_names_and_values)):
		cachedMenuChoices.append((0))
		
def beforeGeneration():
	global cachedMenuChoices
	print "    before generation"
	for i in range(len(selection_names_and_values)):
		if selection_names_and_values[i][0] != "Wrap:":
			cachedMenuChoices[i] = 0


#this helper wraps the NiTextOut and also prints to the hapdebugger
def OutputMessage(message):
	NiTextOut(message)
	print message

#before step one, we must know the wrap in order to bias the width and height
def beforeStepOne():
	global wrapX
	global wrapY
	global centralPole
	wrapX, wrapY, centralPole = getSelectedMapValue("Wrap:")
	print "    before step one wrapxy", wrapX, wrapY
	#save users settings
	cymap = CyMap()
	settings = open('smartmap84.cfg', 'w')
	smoptions = []
	index = -1
	for smoption in selection_titles:
		index += 1
		selectionIndex = int(cymap.getCustomMapOption(index))
		valueName = selection_names_and_values[index][selectionIndex][0]
		smoptions.append((smoption,valueName))
		
	print "    try to store settings to smartmap84.cfg",
	print smoptions
	try:
		pickle.dump(smoptions, settings)
	except Exception, inst:
		print "    Pickling Error trying to save smartmap settings to smartmap84.cfg", inst
	settings.close()
	print "    store successful"
	
#before step two, we must know the map width and height
def beforeStepTwo():
	cymap = CyMap()
	global width
	width = cymap.getGridWidth()
	global height
	height = cymap.getGridHeight()
	global mapDimensionSize
	mapDimensionSize = int((((width*width)+(height*height))**0.5)/2)
	print "    before step two width", width, "height", height
	print "    Configuration of this mapgen:"
	valueName = ""
	for i in range(len(selection_titles)):
		value = getSelectedMapValue(selection_titles[i])
		for j in range(len(selection_names_and_values[i])):
			if value == selection_names_and_values[i][j][1]:
				valueName = selection_names_and_values[i][j][0]
		print "        ",selection_titles[i], valueName
	
#compute the continent lists (lists of land reachable by land travelers on each continent)
def computeContinents():
	global width
	global height
	OutputMessage("  Python: SmartMap: pre-compute continents")
	cymap = CyMap()
	#first collect all continents, this isn't the same as the continent lists
	#calculated during land layout because post generation touchups may have
	#changed things
	print "    start continent analysis"
	global postContinentLists
	global totalLandCount
	global totalCoastCount
	totalLandCount = 0
	totalCoastCount = 0
	checkForBadPlots()
	postContinentLists = []
	plotMarks = [0] * (width*height)
	plotMark = 0
	processPlots = []
	largestContinent = 1
	for x in range(width):
		for y in range(height):
			if tooCloseToPole(x,y,2):
				continue
				
			pPlot = cymap.plot(x,y)
			if pPlot.isFlatlands() or pPlot.isHills():
				totalLandCount += 1
			if pPlot.isWater() and pPlot.isAdjacentToLand():
				totalCoastCount += 1

			index = getTileIndexAtXY(x,y)
			if plotMarks[index] != 0:
				continue
			#if not an interesting plot, continue
			if not pPlot.isFlatlands() and not pPlot.isHills():
				continue
			postContinentList = []
			processPlots.append((x,y))
			plotMark += 1
			postContinentLists.append(postContinentList)
			
			continentSize = 1
			#recursive neighbor plot addition	
			while (len(processPlots) > 0):
				plotX,plotY = processPlots.pop(0)
				existingIndex = getTileIndexAtXY(plotX,plotY)
				if plotMarks[existingIndex] == plotMark:
					continue
				plotMarks[existingIndex] = plotMark
				postContinentList.append((plotX,plotY))
				for nearDir in nearDirs:
					newX, newY = getTileCoordXYWithWrap((plotX,plotY), nearDir)
					newIndex = getTileIndexAtXY(newX, newY)
					if plotMarks[newIndex] != 0:
						continue

					pProcPlot = cymap.plot(newX,newY)
					if not pProcPlot.isFlatlands() and not pProcPlot.isHills():
						continue
						
					continentSize += 1
					processPlots.append((newX,newY))
					
			if continentSize > largestContinent:
				largestContinent = continentSize
					#print "    processPlots",processPlots
					
	gc = CyGlobalContext()
	players = gc.getGame().countCivPlayersEverAlive()
	#remove any too-short continents from the list
	anyTooShort = True
	while anyTooShort:
		anyTooShort = False
		for listIndex in range(len(postContinentLists)):
			if (len(postContinentLists[listIndex]) < int(totalLandCount / (20+players)) and
				len(postContinentLists[listIndex]) < (largestContinent / ((players*2)+1))):
				del postContinentLists[listIndex]
				anyTooShort = True
				break
				
def bonus_score(plotCoord):
	cymap = CyMap()
	score = 0
	x, y = plotCoord
	plotIndex = getTileIndexAtXY(x,y)
	pPlot = cymap.plot(x,y)
	for nearDir in playerStartCross:
		newX, newY = getTileCoordXYWithWrap((x,y), nearDir)
		otherPlot = cymap.plot(newX,newY)
		otherPlotType = otherPlot.getPlotType()
		if otherPlot.getBonusType(-1) != -1:
			score += 1
		
	return score
	
def isCoastalPlot(px,py):
	cymap = CyMap()
	for nearDir in nearDirs:
		newX, newY = getTileCoordXYWithWrap((px,py), nearDir)
		if newX == px and newY == py:
			continue
		pTouchPlot = cymap.plot(newX,newY)
		
		#coastal is true if we touch a water that is not a lake
		if pTouchPlot.isWater() and not pTouchPlot.isLake():
			return True
	return False

					
#override this method used to compute the altitude of a plot, used to determine
#river paths 
def getRiverAltitude(argsList):
	global preComputedAltitude
	pPlot = argsList[0]
	px = pPlot.getX()
	py = pPlot.getY()
	plotIndex = getTileIndexAtXY(px,py)
	return preComputedAltitude[plotIndex]
	
#this makes a one-time computation of the altitude of all plots					
def preComputeAltitude():
	global preComputedAltitude
	global width
	global height
	OutputMessage("  Python: SmartMap: pre-compute tile altitude")
	cymap = CyMap()
	preComputedAltitude = [1]*(width*height)
	maxAltitude = 1
	for x in range(width):
		for y in range(height):
			plotIndex = getTileIndexAtXY(x,y)
			pPlot = cymap.plot(x,y)
			altitude = 1
			if pPlot.isLake():
				altitude -= 100
			#ocean is always altitude 0
			elif pPlot.isWater():
				preComputedAltitude[plotIndex] = 0
				continue
			plotType = pPlot.getPlotType()	
			if plotType == PlotTypes.PLOT_LAND:
				altitude += 100
			if plotType == PlotTypes.PLOT_HILLS:
				altitude += 500
			if plotType == PlotTypes.PLOT_PEAK:
				altitude += 1300
				
			spiralSearchSetup(x,y)
			oceanFound = False
			findOceanTries = 0
			while findOceanTries < 100:
				findOceanTries += 1
				ox,oy = spiralSearchNextStep()
				otherPlot = cymap.plot(ox,oy)
				if otherPlot.isWater() and not otherPlot.isLake():
					break
			altitude += findOceanTries * findOceanTries
			spiralSearchSetup(x,y)
					
			for nearDir in playerStartCross:
				newX, newY = getTileCoordXYWithWrap((x,y), nearDir)
				otherPlot = cymap.plot(newX,newY)
				otherPlotType = otherPlot.getPlotType()	
				if otherPlot.isLake():
					altitude -= 2
				elif otherPlot.isWater():
					altitude -= 20
				if otherPlotType == PlotTypes.PLOT_LAND:
					altitude += 5
				if otherPlotType == PlotTypes.PLOT_HILLS:
					altitude += 20
				if otherPlotType == PlotTypes.PLOT_PEAK:
					altitude += 50
			
			if altitude < 1:
				altitude = 1
				
			if altitude > maxAltitude:
				maxAltitude = altitude

			preComputedAltitude[plotIndex] = altitude
				
	#scale all altitude to 0..1000 range			
	for x in range(width):
		for y in range(height):
			plotIndex = getTileIndexAtXY(x,y)
			preComputedAltitude[plotIndex] = int((1000*preComputedAltitude[plotIndex])/maxAltitude)

#this makes a one-time computation of the wetness of all plots					
def preComputeWetness():
	global preComputedWetness
	global width
	global height
	OutputMessage("  Python: SmartMap: pre-compute tile wetness")
	cymap = CyMap()
	preComputedWetness = [1]*(width*height)
	maxWetness = 1
	for x in range(width):
		for y in range(height):
			plotIndex = getTileIndexAtXY(x,y)
			pPlot = cymap.plot(x,y)
			
			#just call oceans/lakes 1 wetness
			if pPlot.isWater():
				preComputedWetness[plotIndex] = 1
				continue
			
			wetness = 1
			if pPlot.isFreshWater():
				wetness += 48
			if pPlot.isRiver():
				wetness += 48
			
			#improved wetness calcualtor
			if pPlot.isWOfRiver():
				wetness += 32
			if pPlot.isNOfRiver():
				wetness += 32
			northX,northY = getTileCoordXYWithWrap((x,y), (0,1))
			pNorthPlot = cymap.plot(northX,northY)
			if pNorthPlot.isNOfRiver():
				wetness += 32
			westX,westY = getTileCoordXYWithWrap((x,y), (-1,0))
			pWestPlot = cymap.plot(northX,northY)
			if pWestPlot.isWOfRiver():
				wetness += 32
				
			for nearDir in playerStartCross:
				otherX, otherY = getTileCoordXYWithWrap((x,y), nearDir)
				otherPlot = cymap.plot(otherX,otherY)
				dx,dy = nearDir
				if abs(dx) < 2 and abs(dy) < 2:
					if otherPlot.isLake():
						wetness += 32
					if otherPlot.isFreshWater() or otherPlot.isRiver():
						wetness += 16
					continue
				#smaller bonuses for plots not near but in the start cross
				else:
					if otherPlot.isLake():
						wetness += 12
					if otherPlot.isFreshWater() or otherPlot.isRiver():
						wetness += 6

			preComputedWetness[plotIndex] = wetness
			if wetness > maxWetness:
				maxWetness = wetness
	#scale all wetness to 0..1000 range			
	for x in range(width):
		for y in range(height):
			plotIndex = getTileIndexAtXY(x,y)
			preComputedWetness[plotIndex] = int((1000*preComputedWetness[plotIndex])/maxWetness)

	
#this helper function returns the tile coordinate in a given direction,
# accounting for xy wrapping, note that this method may bounce back from boundaries
def getTileCoordXYWithWrap(position,direction):
	global wrapX
	global wrapY
	global width
	global height
	x,y = position
	dx,dy = direction
	x += dx
	y += dy
	#print "    wrapx",wrapX,"wrapY",wrapY,"True",True,"False",False
	if not wrapX and x >= width-3:
		x = width-3
	elif not wrapX and x <= 2:
		x = 2
	elif wrapX and x > width-1:
		x = x-width
	elif wrapX and x < 0:
		x = x+width
	if not wrapY and y >= height-3:
		y = height-3
	if not wrapY and y <= 2:
		y = 2
	elif wrapY and y > height-1:
		y = y-height
	elif wrapY and y < 0:
		y = y+height
	if x < 0:
		x = 0
	elif x > width-1:
		x = width-1
	if y < 0:
		y = 0
	elif y > height-1:
		y = height-1
	return (x,y)
	
#this helper function converts an x,y position to a tile index
#this helps to avoid ever making a mistake in computing this consistently
def getTileIndexAtXY(x,y):
	global width
	return ((y*width) + x)
	
#get the distance between two coordinates, considering that the coordinate
#system may wrap in either x or y
def getWrapDistanceXY(posA,posB):
	global wrapX
	global wrapY
	global width
	global height
	ax,ay = posA
	bx,by = posB
	xDiff = abs(ax-bx)
	yDiff = abs(ay-by)
		
	if wrapX:
		altXDiff = abs(ax-bx+width)
		if altXDiff < xDiff:
			xDiff = altXDiff
		altXDiff = abs(ax-bx-width)
		if altXDiff < xDiff:
			xDiff = altXDiff
	
	if wrapY:
		altYDiff = abs(ay-by+height)
		if altYDiff < yDiff:
			yDiff = altYDiff
		altYDiff = abs(ay-by-height)
		if altYDiff < yDiff:
			yDiff = altYDiff
			
	return (xDiff, yDiff)
	
#return true if a given point is within range coords of a pole
def tooCloseToPole(x,y,range):
	global wrapX
	global wrapY
	global centralPole
	global width
	global height
	if (not wrapX and (x<=range or x>=width-(range+1))):
		return True
	if (not wrapY and (y<=range or y>=height-(range+1))):
		return True
	if centralPole:
		xDist = abs(x-int(width/2))
		yDist = abs(x-int(height/2))
		if xDist + yDist <= range:
			return True
	return False
	
#somewhere in the default methods, an occasional bad tile is placed
#and this method corrects for that
def checkForBadPlots():
	global wrapX
	global wrapY
	global width
	global height
	global fillType

	OutputMessage("  Python: SmartMap: Check For Bad Plots")
	gc = CyGlobalContext()
	cymap = CyMap()
	featIce = gc.getInfoTypeForString("FEATURE_ICE")

	#damn, where is this happening!
	for x in range(width):
		for y in range(height):
			pPlot = cymap.plot(x,y)
			plotType = pPlot.getPlotType()
			if tooCloseToPole(x,y,1):
				if plotType != fillType:
					#print "bad xy",x,y,plotType
					pPlot.setPlotType(fillType,True,True)
					pPlot.setFeatureType(featIce,-1)
					pPlot.setBonusType(-1)
	
#this helper function will tell you what % the plot is from a pole	
def percentFromPole(x,y):
	global wrapX
	global wrapY
	global width
	global height
	global centralPole
	global mapDimensionSize
	halfX = int(width/2)
	halfY = int(height/2)
	
	xPercentFromPole = 100
	if not wrapX:
		xFromPole = x-2
		if x > halfX:
			xFromPole = (width-3) - x
		xPercentFromPole = int((xFromPole*100)/halfX)

	yPercentFromPole = 100
	if not wrapY:
		yFromPole = y-2
		if y > halfY:
			yFromPole = (height-3) - y
		yPercentFromPole = int((yFromPole*100)/halfY)
		
	standardPolePct = min(xPercentFromPole, yPercentFromPole)
	
	centralPct = 100	
	if centralPole:
		xDist = abs(x-halfX)
		yDist = abs(y-halfY)
		totalDist = int(((xDist*xDist)+(yDist*yDist))**0.5)
		centralPct = int((totalDist*100)/mapDimensionSize)
		
	return min(standardPolePct, centralPct)
	
#this helper function will tell you what % the plot is from the equator	
def percentFromEquator(x,y):
	global wrapX
	global wrapY
	global width
	global height
	global centralPole
	global mapDimensionSize

	xBands = [int(width/2)]
	yBands = [int(height/2)]
	divFactor = 2
	if centralPole:
		xBands = [int(width/4), int(3*width/4)]
		yBands = [int(height/4), int(3*height/4)]
		divFactor = 4
	
	xPercentFromEquator = 100
	if not wrapX:
		for band in xBands:
			bandDist = abs(x - band)
			xPercentFromEquator = min(xPercentFromEquator, int((bandDist*100)/(width/divFactor)))

	yPercentFromEquator = 100
	if not wrapY:
		for band in yBands:
			bandDist = abs(y - band)
			yPercentFromEquator = min(yPercentFromEquator, int((bandDist*100)/(height/divFactor)))

	if centralPole and wrapX and wrapY:
		return 2*abs(50 - percentFromPole(x,y))
		
	elif wrapX and wrapY:
		halfX = int(width/2)
		halfY = int(height/2)
		xDist = abs(x-halfX)
		yDist = abs(y-halfY)
		totalDist = int(((xDist*xDist)+(yDist*yDist))**0.5)
		centralPct = int((totalDist*100)/mapDimensionSize)
		centralPct = 2*abs(50 - centralPct)
		return centralPct
		
	return min(xPercentFromEquator, yPercentFromEquator)

def spiralSearchSetup(startX, startY):
	global spiralCenterX
	global spiralCenterY
	global spiralStep
	global spiralSign
	global spiralX
	global spiralY
	global spiralQueue
	spiralCenterX = startX
	spiralCenterY = startY
	spiralStep = 0
	spiralSign = 1
	spiralX = 0
	spiralY = 0
	spiralQueue = []
	
def spiralSearchNextStep():
	global spiralCenterX
	global spiralCenterY
	global spiralStep
	global spiralSign
	global spiralX
	global spiralY
	global spiralQueue
	
	if len(spiralQueue) > 0:
		return spiralQueue.pop()
		
	for x in range(spiralStep):
		spiralX += spiralSign
		spiralQueue.append(getTileCoordXYWithWrap((spiralCenterX,spiralCenterY),(spiralX,spiralY)))
	for y in range(spiralStep):
		spiralY += spiralSign
		spiralQueue.append(getTileCoordXYWithWrap((spiralCenterX,spiralCenterY),(spiralX,spiralY)))
	spiralSign *= -1
	spiralStep += 1
	if len(spiralQueue) > 0:
		return spiralQueue.pop()
	return (spiralCenterX,spiralCenterY)

def isPlotLegalForBonus(plot,bonus,strict):
	gc = CyGlobalContext()
	bonusInfoXML = gc.getBonusInfo(bonus)
	type_string = bonusInfoXML.getType()
	
	if plot.getBonusType(-1) != -1:
		return False
	if plot.getFeatureType() == gc.getInfoTypeForString("FEATURE_OASIS"):
		return False # Soren wants no bonuses in oasis plots. So mote it be.

	#for whatever reason, these resources never validate as canHaveBonus so we override
	#and place them on any empty hill, unless strict mode is set
	resourcesEmptyHill = ('BONUS_ALUMINUM','BONUS_GOLD','BONUS_SILVER')
	if not strict:
		if type_string in resourcesEmptyHill and plot.isHills():
			return True

	if not plot.canHaveBonus(bonus,True):
		return False

	if strict:
		plotLatitude = plot.getLatitude()
		minLatitude = bonusInfoXML.getMinLatitude()
		maxLatitude = bonusInfoXML.getMaxLatitude()
		if minLatitude > 0 and plotLatitude < minLatitude:
			return False
		if maxLatitude < 90 and plotLatitude > maxLatitude:
			return False

	return True

#a spatial hash for finding things near a given position
spaceHash = []
def setupSpaceHash():
	global width
	global height
	global spaceHash
	global partitionWidth
	global partitionHeight
	spaceHash = []
	partitionWidth = 1 + int(width/20)
	partitionHeight = 1 + int(height/20)
	for x in range(21):
		spaceHash.append([])
		for y in range(21):
			spaceHash[x].append([])

def spaceHashItem(x,y,item):
	global spaceHash
	global partitionWidth
	global partitionHeight
	xh = int(x/partitionWidth)
	yh = int(y/partitionHeight)
	#print "    space hash item ",x,y,xh,yh,item
	spaceHash[xh][yh].append((x,y,item))
	
def listNearSpaceHash(x,y,searchRange):
	global spaceHash
	global partitionWidth
	global partitionHeight
	result = []
	hashes = []
	
	xstep = partitionWidth - 1
	ystep = partitionHeight - 1
	xd = -(searchRange+1)
	while xd <= (searchRange + 1):
		yd = -(searchRange+1)
		while yd <= (searchRange + 1):
			xp, yp = getTileCoordXYWithWrap((x,y),(xd,yd))
			xh = int(xp/partitionWidth)
			yh = int(yp/partitionHeight)
			if not (xh,yh) in hashes:
				hashes.append((xh,yh))
			yd += ystep
		xd += xstep
				
	for hash in hashes:
		xh,yh = hash
		for entry in spaceHash[xh][yh]:
			result.append(entry)
	return result
					
def getPlotsOfTypes(tiles,types):
	global width
	global height
	result = []
	for x in range(width):
		for y in range(height):
			index = getTileIndexAtXY(x,y)
			if tiles[index] in types:
				result.append((x,y))
	return result
	

##########################################################################
##########################################################################
##########################################################################
##########################################################################
#code below here is ripped from the standard map scripts, to allow the selection
#of other script behaviors on some of the options
#some needed classes
# subclass TerrainGenerator to redefine everything. This is a regional map.
class GreatPlainsTerrainGenerator(CvMapGeneratorUtil.TerrainGenerator):
	def __init__(self, iRockyDesertPercent=50, iRockyPlainsPercent=30, 
	             iGrassPercent=17, iDesertPercent=8, iTexDesertPercent=20,
	             iEastDesertPercent=2, iEastPlainsPercent=23, 
	             fWestLongitude=0.15, fEastLongitude=0.65, fTexLat=0.37,
	             fTexEast=0.55, fracXExp=-1, fracYExp=-1, grain_amount=4):
		 #Note: If you change longitude values here, then you will...
		 #...need to change them elsewhere in the script, as well.
		self.gc = CyGlobalContext()
		self.map = CyMap()

		self.grain_amount = grain_amount + self.gc.getWorldInfo(self.map.getWorldSize()).getTerrainGrainChange()

		self.iWidth = self.map.getGridWidth()
		self.iHeight = self.map.getGridHeight()

		self.mapRand = self.gc.getGame().getMapRand()

		self.iFlags = 0  # Disallow FRAC_POLAR flag, to prevent "zero row" problems.

		self.rocky=CyFractal()
		self.plains=CyFractal()
		self.east=CyFractal()
		self.variation=CyFractal()

		self.iRockyDTopPercent = 100
		self.iRockyDBottomPercent = max(0,int(100-iRockyDesertPercent))
		self.iRockyPTopPercent = int(100 - iRockyDesertPercent)
		self.iRockyPBottomPercent = max(0,int(100-iRockyDesertPercent-iRockyPlainsPercent))
		self.iDesertTopPercent = 100
		self.iDesertBottomPercent = max(0,int(100-iDesertPercent))
		self.iTexDesertTopPercent = 70
		self.iTexDesertBottomPercent = max(0,int(70-iTexDesertPercent))
		self.iGrassTopPercent = int(iGrassPercent)
		self.iGrassBottomPercent = 0
		self.iTexGrassBottomPercent = 6
		self.iEastDTopPercent = 100
		self.iEastDBottomPercent = max(0,int(100-iEastDesertPercent))
		self.iEastPTopPercent = int(100 - iEastDesertPercent)
		self.iEastPBottomPercent = max(0,int(100-iEastDesertPercent-iEastPlainsPercent))
		self.iMountainTopPercent = 75
		self.iMountainBottomPercent = 60

		self.fWestLongitude = fWestLongitude
		self.fEastLongitude = fEastLongitude
		self.fTexLat = fTexLat
		self.fTexEast = fTexEast

		self.iGrassPercent = iGrassPercent
		self.iDesertPercent = iDesertPercent
		self.iTexDesertPercent = iTexDesertPercent
		self.iEastDesertPercent = iEastDesertPercent
		self.iEastPlainsPercent = iEastPlainsPercent
		self.iRockyDesertPercent = iRockyDesertPercent
		self.iRockyPlainsPercent = iRockyPlainsPercent

		self.fracXExp = fracXExp
		self.fracYExp = fracYExp

		self.initFractals()
		
	def initFractals(self):
		self.rocky.fracInit(self.iWidth, self.iHeight, self.grain_amount, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)
		self.iRockyDTop = self.rocky.getHeightFromPercent(self.iRockyDTopPercent)
		self.iRockyDBottom = self.rocky.getHeightFromPercent(self.iRockyDBottomPercent)
		self.iRockyPTop = self.rocky.getHeightFromPercent(self.iRockyPTopPercent)
		self.iRockyPBottom = self.rocky.getHeightFromPercent(self.iRockyPBottomPercent)

		self.plains.fracInit(self.iWidth, self.iHeight, self.grain_amount+1, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)
		self.iDesertTop = self.plains.getHeightFromPercent(self.iDesertTopPercent)
		self.iDesertBottom = self.plains.getHeightFromPercent(self.iDesertBottomPercent)
		self.iTexDesertTop = self.plains.getHeightFromPercent(self.iTexDesertTopPercent)
		self.iTexDesertBottom = self.plains.getHeightFromPercent(self.iTexDesertBottomPercent)
		self.iGrassTop = self.plains.getHeightFromPercent(self.iGrassTopPercent)
		self.iGrassBottom = self.plains.getHeightFromPercent(self.iGrassBottomPercent)
		self.iTexGrassBottom = self.plains.getHeightFromPercent(self.iTexGrassBottomPercent)

		self.east.fracInit(self.iWidth, self.iHeight, self.grain_amount, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)
		self.iEastDTop = self.east.getHeightFromPercent(self.iEastDTopPercent)
		self.iEastDBottom = self.east.getHeightFromPercent(self.iEastDBottomPercent)
		self.iEastPTop = self.east.getHeightFromPercent(self.iEastPTopPercent)
		self.iEastPBottom = self.east.getHeightFromPercent(self.iEastPBottomPercent)

		self.variation.fracInit(self.iWidth, self.iHeight, self.grain_amount, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)

		self.terrainDesert = self.gc.getInfoTypeForString("TERRAIN_DESERT")
		self.terrainPlains = self.gc.getInfoTypeForString("TERRAIN_PLAINS")
		self.terrainGrass = self.gc.getInfoTypeForString("TERRAIN_GRASS")

	def getLatitudeAtPlot(self, iX, iY):
		lat = iX/float(self.iWidth) # 0.0 = west

		# Adjust latitude using self.variation fractal, to mix things up:
		lat += (128 - self.variation.getHeight(iX, iY))/(255.0 * 5.0)

		# Limit to the range [0, 1]:
		if lat < 0:
			lat = 0.0
		if lat > 1:
			lat = 1.0

		return lat

	def generateTerrain(self):		
		terrainData = [0]*(self.iWidth*self.iHeight)
		for x in range(self.iWidth):
			for y in range(self.iHeight):
				iI = y*self.iWidth + x
				terrain = self.generateTerrainAtPlot(x, y)
				terrainData[iI] = terrain
		return terrainData

	def generateTerrainAtPlot(self,iX,iY):
		lat = self.getLatitudeAtPlot(iX,iY)

		if (self.map.plot(iX, iY).isWater()):
			return self.map.plot(iX, iY).getTerrainType()

		if lat <= self.fWestLongitude:
			val = self.rocky.getHeight(iX, iY)
			if val >= self.iRockyDBottom and val <= self.iRockyDTop:
				terrainVal = self.terrainDesert
			elif val >= self.iRockyPBottom and val <= self.iRockyPTop:
				terrainVal = self.terrainPlains
			else:
				long = iY/float(self.iHeight)
				if long > 0.23:
					terrainVal = self.terrainGrass
				else:
					terrainVal = self.terrainDesert
		elif lat > self.fEastLongitude:
			val = self.east.getHeight(iX, iY)
			if val >= self.iEastDBottom and val <= self.iEastDTop:
				terrainVal = self.terrainDesert
			elif val >= self.iEastPBottom and val <= self.iEastPTop:
				terrainVal = self.terrainPlains
			else:
				terrainVal = self.terrainGrass
		elif lat > self.fWestLongitude and lat <= self.fTexEast and iY/float(self.iHeight) <= self.fTexLat:
			# More desert added to Texas region at Soren's request.
			val = self.east.getHeight(iX, iY)
			if val >= self.iDesertBottom and val <= self.iDesertTop:
				terrainVal = self.terrainDesert
			elif val >= self.iTexDesertBottom and val <= self.iTexDesertTop:
				terrainVal = self.terrainDesert
			elif val >= self.iTexGrassBottom and val <= self.iGrassTop:
				terrainVal = self.terrainGrass
			else:
				terrainVal = self.terrainPlains
		else:
			val = self.plains.getHeight(iX, iY)
			if val >= self.iDesertBottom and val <= self.iDesertTop:
				terrainVal = self.terrainDesert
			elif val >= self.iGrassBottom and val <= self.iGrassTop:
				terrainVal = self.terrainGrass
			else:
				terrainVal = self.terrainPlains

		if (terrainVal == TerrainTypes.NO_TERRAIN):
			return self.map.plot(iX, iY).getTerrainType()

		return terrainVal


# subclass TerrainGenerator to redefine everything. This is a regional map. Ice need not apply!
# Latitudes, ratios, the works... It's all rewired. - Sirian June 20, 2005

class OasisTerrainGenerator(CvMapGeneratorUtil.TerrainGenerator):
	def __init__(self, iGrassPercent=50, iPlainsPercent=35,
	             iNorthernPlainsPercent=40, iOasisGrassPercent=9,
	             iOasisPlainsPercent=16, iOasisTopLatitude=0.69,
	             iJungleLatitude=0.14, iOasisBottomLatitude=0.3,
	             fracXExp=-1, fracYExp=-1, grain_amount=4):
		
		self.grain_amount = grain_amount
		
		self.gc = CyGlobalContext()
		self.map = CyMap()

		self.iWidth = self.map.getGridWidth()
		self.iHeight = self.map.getGridHeight()

		self.mapRand = self.gc.getGame().getMapRand()
		self.iFlags = 0

		self.grass=CyFractal()
		self.Oasisgrass=CyFractal()
		self.plains=CyFractal()
		self.Oasisplains=CyFractal()
		self.northernplains=CyFractal()
		self.variation=CyFractal()

		self.iGrassTopPercent = 100
		self.iGrassBottomPercent = max(0,int(100-iGrassPercent))
		self.iPlainsTopPercent = 100
		self.iPlainsBottomPercent = max(0,int(100-iGrassPercent-iPlainsPercent))
		self.iOasisGrassTopPercent = 100
		self.iOasisGrassBottomPercent = max(0,int(100-iOasisGrassPercent))
		self.iOasisPlainsTopPercent = 100
		self.iOasisPlainsBottomPercent = max(0,int(100-iOasisGrassPercent-iOasisPlainsPercent))
		self.iNorthernPlainsBottomPercent = max(0,int(100-iNorthernPlainsPercent))
		self.iMountainTopPercent = 75
		self.iMountainBottomPercent = 60
		
		self.iOasisBottomLatitude = iOasisBottomLatitude
		self.iOasisTopLatitude = iOasisTopLatitude
		self.iJungleLatitude = iJungleLatitude
		
		self.iGrassPercent = iGrassPercent
		self.iPlainsPercent = iPlainsPercent
		self.iOasisGrassPercent = iOasisGrassPercent
		self.iOasisPlainsPercent = iOasisPlainsPercent
		self.iNorthernPlainsPercent = iNorthernPlainsPercent
		
		self.fracXExp = fracXExp
		self.fracYExp = fracYExp

		self.initFractals()
		
	def initFractals(self):
		self.grass.fracInit(self.iWidth, self.iHeight, self.grain_amount, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)
		self.iGrassTop = self.grass.getHeightFromPercent(self.iGrassTopPercent)
		self.iGrassBottom = self.grass.getHeightFromPercent(self.iGrassBottomPercent)

		self.plains.fracInit(self.iWidth, self.iHeight, self.grain_amount+1, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)
		self.iPlainsTop = self.plains.getHeightFromPercent(self.iPlainsTopPercent)
		self.iPlainsBottom = self.plains.getHeightFromPercent(self.iPlainsBottomPercent)

		self.Oasisgrass.fracInit(self.iWidth, self.iHeight, self.grain_amount, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)
		self.iOasisGrassTop = self.grass.getHeightFromPercent(self.iOasisGrassTopPercent)
		self.iOasisGrassBottom = self.grass.getHeightFromPercent(self.iOasisGrassBottomPercent)

		self.Oasisplains.fracInit(self.iWidth, self.iHeight, self.grain_amount+1, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)
		self.iOasisPlainsTop = self.plains.getHeightFromPercent(self.iOasisPlainsTopPercent)
		self.iOasisPlainsBottom = self.plains.getHeightFromPercent(self.iOasisPlainsBottomPercent)

		self.northernplains.fracInit(self.iWidth, self.iHeight, self.grain_amount+1, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)
		self.iNorthernPlainsBottom = self.plains.getHeightFromPercent(self.iNorthernPlainsBottomPercent)

		self.variation.fracInit(self.iWidth, self.iHeight, self.grain_amount, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)

		self.terrainDesert = self.gc.getInfoTypeForString("TERRAIN_DESERT")
		self.terrainPlains = self.gc.getInfoTypeForString("TERRAIN_PLAINS")
		self.terrainGrass = self.gc.getInfoTypeForString("TERRAIN_GRASS")

	def getLatitudeAtPlot(self, iX, iY):
		lat = iY/float(self.iHeight) # 0.0 = south edge, 1.0 = north edge

		 #Adjust latitude using self.variation fractal, to mix things up:
		lat += (128 - self.variation.getHeight(iX, iY))/(255.0 * 5.0)

		# Limit to the range [0, 1]:
		if lat < 0:
			lat = 0.0
		if lat > 1:
			lat = 1.0

		return lat

	def generateTerrain(self):		
		terrainData = [0]*(self.iWidth*self.iHeight)
		for x in range(self.iWidth):
			for y in range(self.iHeight):
				iI = y*self.iWidth + x
				terrain = self.generateTerrainAtPlot(x, y)
				terrainData[iI] = terrain
		return terrainData

	def generateTerrainAtPlot(self,iX,iY):
		lat = self.getLatitudeAtPlot(iX,iY)

		if (self.map.plot(iX, iY).isWater()):
			return self.map.plot(iX, iY).getTerrainType()

		terrainVal = self.terrainDesert

		if lat > self.iOasisTopLatitude:
			plainsVal = self.plains.getHeight(iX, iY)
			if plainsVal >= self.iNorthernPlainsBottom:
				terrainVal = self.terrainPlains
			else:
				terrainVal = self.terrainGrass
		elif lat < self.iJungleLatitude:
			terrainVal = self.terrainGrass
		elif lat < self.iOasisBottomLatitude and lat >= self.iJungleLatitude:
			grassVal = self.grass.getHeight(iX, iY)
			plainsVal = self.plains.getHeight(iX, iY)
			if ((grassVal >= self.iGrassBottom) and (grassVal <= self.iGrassTop)):
				terrainVal = self.terrainGrass
			elif ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)):
				terrainVal = self.terrainPlains
		else:
			OasisgrassVal = self.grass.getHeight(iX, iY)
			OasisplainsVal = self.plains.getHeight(iX, iY)
			if ((OasisgrassVal >= self.iOasisGrassBottom) and (OasisgrassVal <= self.iOasisGrassTop)):
				terrainVal = self.terrainGrass
			elif ((OasisplainsVal >= self.iOasisPlainsBottom) and (OasisplainsVal <= self.iOasisPlainsTop)):
				terrainVal = self.terrainPlains

		if (terrainVal == TerrainTypes.NO_TERRAIN):
			return self.map.plot(iX, iY).getTerrainType()

		return terrainVal

class GreatPlainsFeatureGenerator(CvMapGeneratorUtil.FeatureGenerator):
	def __init__(self, iJunglePercent=40, iEastForestPercent=45, 
	             iForestPercent=8, iRockyForestPercent=55, 
	             forest_grain=6, fracXExp=-1, fracYExp=-1):
		self.gc = CyGlobalContext()
		self.map = CyMap()
		self.mapRand = self.gc.getGame().getMapRand()
		self.forests = CyFractal()
		
		self.iFlags = 0  # Disallow FRAC_POLAR flag, to prevent "zero row" problems.

		self.iGridW = self.map.getGridWidth()
		self.iGridH = self.map.getGridHeight()
		
		self.iJunglePercent = iJunglePercent
		self.iForestPercent = iForestPercent
		self.iEastForestPercent = iEastForestPercent
		self.iRockyForestPercent = iRockyForestPercent
		
		self.forest_grain = forest_grain + self.gc.getWorldInfo(self.map.getWorldSize()).getFeatureGrainChange()

		self.fracXExp = fracXExp
		self.fracYExp = fracYExp

		self.__initFractals()
		self.__initFeatureTypes()
	
	def __initFractals(self):
		self.forests.fracInit(self.iGridW, self.iGridH, self.forest_grain, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)
		
		self.iJungleLevel = self.forests.getHeightFromPercent(100 - self.iJunglePercent)
		self.iForestLevel = self.forests.getHeightFromPercent(self.iForestPercent)
		self.iEastForestLevel = self.forests.getHeightFromPercent(self.iEastForestPercent)
		self.iRockyForestLevel = self.forests.getHeightFromPercent(self.iRockyForestPercent)
		
	def __initFeatureTypes(self):
		self.featureJungle = self.gc.getInfoTypeForString("FEATURE_JUNGLE")
		self.featureForest = self.gc.getInfoTypeForString("FEATURE_FOREST")
		self.featureOasis = self.gc.getInfoTypeForString("FEATURE_OASIS")

	def getLatitudeAtPlot(self, iX, iY):
		 #0.0 = bottom edge, 1.0 = top edge.
		return iX/float(self.iGridW)

	def addFeaturesAtPlot(self, iX, iY):
		#"adds any appropriate features at the plot (iX, iY) where (0,0) is in the SW"
		long = iX/float(self.iGridW)
		lat = iY/float(self.iGridH)
		pPlot = self.map.sPlot(iX, iY)

		for iI in range(self.gc.getNumFeatureInfos()):
			if pPlot.canHaveFeature(iI):
				if self.mapRand.get(10000, "Add Feature PYTHON") < self.gc.getFeatureInfo(iI).getAppearanceProbability():
					pPlot.setFeatureType(iI, -1)

		if (pPlot.getFeatureType() == FeatureTypes.NO_FEATURE):
			# Jungles only in Louisiana Wetlands!
			if long > 0.65 and lat < 0.45:
				self.addJunglesAtPlot(pPlot, iX, iY, lat)
			
		if (pPlot.getFeatureType() == FeatureTypes.NO_FEATURE):
			self.addForestsAtPlot(pPlot, iX, iY, lat, long)
		
	def addIceAtPlot(self, pPlot, iX, iY, lat):
		 #We don' need no steeking ice. M'kay? Alrighty then.
		ice = 0
	
	def addJunglesAtPlot(self, pPlot, iX, iY, lat):
		 #Warning: this version of JunglesAtPlot is using the forest fractal!
		if pPlot.canHaveFeature(self.featureJungle):
			if (self.forests.getHeight(iX, iY) >= self.iJungleLevel):
				pPlot.setFeatureType(self.featureJungle, -1)

	def addForestsAtPlot(self, pPlot, iX, iY, lat, long):
		 #Evergreens in the Rockies. (Pinion Pines! Juniper! Spruce!)
		 #Deciduous trees elsewhere.
		if (long < 0.16 and lat > 0.23) and (pPlot.isFlatlands() or pPlot.isHills()):
			if self.forests.getHeight(iX, iY) <= self.iRockyForestLevel:
				pPlot.setFeatureType(self.featureForest, 2)
		elif long > 0.72 and pPlot.canHaveFeature(self.featureForest):
			if self.forests.getHeight(iX, iY) <= self.iEastForestLevel:
				pPlot.setFeatureType(self.featureForest, 0)
		else:
			if pPlot.canHaveFeature(self.featureForest):
				if self.forests.getHeight(iX, iY) <= self.iForestLevel:
					pPlot.setFeatureType(self.featureForest, 0)

class OasisFeatureGenerator(CvMapGeneratorUtil.FeatureGenerator):
	def __init__(self, iJunglePercent=40, iForestPercent=45,
                     jungle_grain=5, forest_grain=6, fracXExp=-1, fracYExp=-1):
		self.gc = CyGlobalContext()
		self.map = CyMap()
		self.mapRand = self.gc.getGame().getMapRand()
		self.jungles = CyFractal()
		self.forests = CyFractal()
		self.iFlags = 0
		self.iGridW = self.map.getGridWidth()
		self.iGridH = self.map.getGridHeight()
		
		self.iJunglePercent = iJunglePercent
		self.iForestPercent = iForestPercent
		
		self.jungle_grain = jungle_grain
		self.forest_grain = forest_grain

		self.fracXExp = fracXExp
		self.fracYExp = fracYExp

		self.__initFractals()
		self.__initFeatureTypes()
	
	def __initFractals(self):
		self.jungles.fracInit(self.iGridW+1, self.iGridH+1, self.jungle_grain, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)
		self.forests.fracInit(self.iGridW+1, self.iGridH+1, self.forest_grain, self.mapRand, self.iFlags, self.fracXExp, self.fracYExp)
		
		self.iJungleLevel = self.jungles.getHeightFromPercent(self.iJunglePercent)
		self.iForestLevel = self.forests.getHeightFromPercent(self.iForestPercent)
		
	def __initFeatureTypes(self):
		self.featureJungle = self.gc.getInfoTypeForString("FEATURE_JUNGLE")
		self.featureForest = self.gc.getInfoTypeForString("FEATURE_FOREST")
		self.featureOasis = self.gc.getInfoTypeForString("FEATURE_OASIS")

	def getLatitudeAtPlot(self, iX, iY):
		 #0.0 = bottom edge, 1.0 = top edge.
		return iY/float(self.iGridH)

	def addFeaturesAtPlot(self, iX, iY):
		#"adds any appropriate features at the plot (iX, iY) where (0,0) is in the SW"
		lat = self.getLatitudeAtPlot(iX, iY)
		pPlot = self.map.sPlot(iX, iY)

		for iI in range(self.gc.getNumFeatureInfos()):
			if pPlot.canHaveFeature(iI):
				if self.mapRand.get(10000, "Add Feature PYTHON") < self.gc.getFeatureInfo(iI).getAppearanceProbability():
					pPlot.setFeatureType(iI, -1)

		if (pPlot.getFeatureType() == FeatureTypes.NO_FEATURE):
			# Jungles only in the deep south or in the Oasis!
			if lat < 0.16:
				self.addJunglesAtPlot(pPlot, iX, iY, lat)
			elif lat > 0.32 and lat < 0.65 and (pPlot.getTerrainType() == self.gc.getInfoTypeForString("TERRAIN_GRASS")):
				pPlot.setFeatureType(self.featureJungle, -1)

		if (pPlot.getFeatureType() == FeatureTypes.NO_FEATURE):
			# No forests in the Oasis!
			if lat > 0.71 or lat < 0.3:
				self.addForestsAtPlot(pPlot, iX, iY, lat)

		if pPlot.isFlatlands() and (pPlot.getFeatureType() == FeatureTypes.NO_FEATURE) and (pPlot.getTerrainType() == self.gc.getInfoTypeForString("TERRAIN_DESERT")):
			# Add more Oases!
			if (lat < 0.71 and lat > 0.3) and self.mapRand.get(9, "Add Extra Oases PYTHON") == 0:
				pPlot.setFeatureType(self.featureOasis, -1)
                                		
	def addIceAtPlot(self, pPlot, iX, iY, lat):
		 #We don' need no steeking ice. M'kay? Alrighty then.
		ice = 0
	
	def addJunglesAtPlot(self, pPlot, iX, iY, lat):
		if pPlot.canHaveFeature(self.featureJungle):
			if self.jungles.getHeight(iX+1, iY+1) <= self.iJungleLevel:
				pPlot.setFeatureType(self.featureJungle, -1)

	def addForestsAtPlot(self, pPlot, iX, iY, lat):
		 #No evergreens.
		if pPlot.canHaveFeature(self.featureForest):
			if self.forests.getHeight(iX+1, iY+1) <= self.iForestLevel:
				pPlot.setFeatureType(self.featureForest, 0)

# 8.6
#     Correct the override width/height entries to reflect the 'standard' 
#       width and height numbers (previously this was rough, now it is exact).
#       So these numbers now match what scripts that use the base XML produce.
#     Added extra plains / extra grass options to the terrain options, which
#       will produce more of that type of terrain.
#     Added 'new world' coastal/inland start options, which will place start 
#       positions in the east (as new world) but also favor coastal or inland
#       starts.
#     Added 'no block' peak options.  This forbids peaks from having ocean
#       on both sides either vertically of horizontally.  This prevents certain
#       common cases of resource blocking that some people found annoying.
#       Any such peak is reduced to a hill, as a result you will have less total
#       peaks on your map.
#     Improved performance of land placement in low sea width (removed unneeded checks).
#     Improved performance of land placement in high sea width by spatially
#       hashing the previously placed lands, and reducing the search for conflicting
#       lands to nearby hash entries.  This makes a huge difference on big maps with
#       high sea levels selected (>50% faster).
#     Improved distribution of hills and peaks so that small continents don't get
#       a disproportionate amount of them.
#     Tweaked up likelihood of floodplains slightly.  You may see a bit more desert
#       in temperate maps.  If you don't like desert much you can of course turn it 
#       off or switch to tropical which has much less desert.
#     Improved performance for placing resources on large maps with lots of resources.
#     Overall performance improvements will no generate the most difficult type of map
#       (huge, land heavy, high sea level map with 2 or more continents) in less than
#       a minute on my p4m-1.8ghz.
# 8.5
#     Performance improvements.
#     Improved hill placement for a huge/65ocean/30hills map from 12 seconds to 2.
#     Improved huge/65ocean/many resources from 90 seconds to 30.
# 8.4A
#     Quick Fix for problem with placing players on pangea maps introduced in 8.4
#       (would cause an endless loop, freezing civ iv on pangea and certain other
#        maps where it couldn't decide the best continental distribution of players)
# 8.4 
#     Tweaked resource placement to obey minlatitude and maxlatitude for the stricter
#       resource placement options (SmartMap few, and all clumping options).
#     Tweaked resource rates some more based on forum feedback.
#     Made some other adjustments to the fair resource distribution to simplify it, and
#       to distribute resources a little more evenly in certain cases.  Also improved
#       chances that all expected resources will be placed.
#     Added terrain options to place no desert, no snow, no snow/desert, 
#       no snow/desert/tundra, only plains, or only grass.  If you wan't minimal
#       snow/desert/tundra, play on tropical, and turn jungles to light if necessary.
#     Added 'SmartMap restricted / most restricted' option to resource placement, which 
#       enforces every placement rule I understand, clumps both happy and health resources,
#       and offers fewer (restricted)/much fewer (most restricted) resources in general.
#     Move the version history for prior versions to end of file
#     Added player/all inland/coastal placement options.  Player inland will do its best
#       to give the first player an inland start.  Coastal prefers a coastal start.  All
#       attempts to do the same for all players (less likely to succeed, based on available
#       map positions).
#     Add single tile isles option.  This make for a very wierd game in generic civiv, but
#       was requested by people developing space/planet based variants.
#     Fixed bug in override separation that caused it to behave always as if low sea level
#       was selected instead.
#     Added 'central pole' variation on all wrap options.  This creates an icy pole in the
#       center of the map, as if you were looking down on earth from above the north pole,
#       also known as polar projection.  This now allows you to have 0,1,2,3 poled maps,
#       which you can envision as different ways to unwrap a spherical world onto a flat
#       rectangular map.
#     Fixed a small math error in determining the distance from the pole/equator which
#       became obvious when working with the polar projection maps.  This may cause you
#       to see a little more snow/tundra than previously.
#
#       
# 8.3
#     Tweak tundra, snow, desert picking process to make more likely on small maps.
#     Add new separation override to allow absolute control of continent separation.
#     More rivers sourced from hills if there are few peaks.
#     Added options for central, edge, and corner sea to the ocean selections to
#       generate that style of map.  Particularly, if you enjoy the 'central sea'
#       map style, you can now generate that with SmartMap.
#     Tweak resource rates to make whales not quite so common, and to address other
#       resource rate comments reported.
#     Place resources in placement order to try to prevent good map positions from
#       being taken by resources placed in wrong order.
#     Added SmartMap New World placement option.  This will prefer to place civs
#       only on continents in the eastern hemisphere.  This option will default
#       to ordinary smartmap placement if there are not at least 2 continents, or if there 
#       is not a continent at least 90% of which is located in the western hemisphere.
# 8.2
#     Add option for 'no ocean' that will place nearly all land.
#       Note: takes longer to generate high land count maps.  Be patient.
#     Add option for 'no ocean, no lakes' that will place all land.
#     Improve the placement performance of smartmap placement on high land count
#       maps by ruling out very low scoring placement positions earlier.
#       This will improve overall map generation performance slightly.
#     Small improvements to forest and lake placement, attempt to cure the
#       rare 'all forest tiles' on normal forest problem (I've never reproduced
#       this, but the chance that any such bug survived should be low.)
#     Added 'smartmap random pick' option to all (smartmap controlled)
#       menus, this makes a choice from among the more typically desirable
#       options, and also stays random the next time you create a map,
#       unlike the standard 'random' option, and even randomly re-picks if you
#       choose to 'regenerate map'.  I highly recommend using this in place of
#       the standard 'random' option.
# 8.1
#     Add Clump Health and Clump Both resource placement options.
#     Clean up land placement code organization a little bit.
#     Force release of spiral search memory, and reduce how far spiral search
#       searches to improve performance.
#     Fix the defaults for override with and height that were causing people
#       to get out of memory errors.
# 8.0
#     Use pickle to store last preferences.
#     Tweaked the adjustment to rivers and lakes so that many rivers will be more
#       noticeable, and few lakes will really cut down the amount of lakes.
#     Fix resource rates for smaller maps.
#     Allow start placement to place settlers on hills as well as flatlands.
#     Weight hills for start placement appropriately to the defensive bonus.
#     Add spiral search feature, this may be of use to other mapmakers, used in
#       my script to assign fragments as belonging to the nearest continent.  Spiral
#       search is the most efficient method of doing a 'find nearest X' on any grid
#       based map, assuming you don't have a list of all X available.
#     On higher river levels, make hills more likely to start rivers.
#     Make rivers a little more common in general.  Make the automatic river for each
#       sufficiently large continent always start from a hill or peak if one is
#       available, or a random tile if no hill or peak is found.
#     Improve river paths by using spiral search to measure distance to nearest ocean
#     Don't flatten peaks on coast, as this can create interesting strategic issues,
#       and also tended to leave far fewer than the expected number of peaks.
#     Fix a bug in the start placement picker that was giving extra scores for the 
#       start plot when it was supposed to be scoring nearby plots.
#     Adjust how fragments are generated, and the number of fragments for few/many,
#       to make maps using smaller numbers of fragments look less chaotic.  Fragments
#       also now sort-of obey the sealevel rules: they are allowed to start anywhere,
#       but can only grow into one neighboring continent, so they are more likely to
#       create an archipelago type of continental extension, rather than so frequently
#       causing continents to merge.
#     New SmartMap Clump Happy resources option which distributes all happiness 
#       resources like the standard generator (in clumps so trade is more likely to be
#       necessary, hence 'clump happy'), but other resources in the smartmap style so
#       that the distribution is fairer.
#     Corrected a small land counting problem that sometimes gave too much land for
#       a given ocean percentage.
#     Added a few new options to the override width and height to allow for generating
#       random ranges of sizes, as a couple of people had requested that.
# 7.2
#     Separate GoodyHut option from Bonuses, and add support for
#       no/low/normal/high rates of goody huts.
#     Add support for low/normal/high levels of resources (with low = 2/3rds
#       and high = 4/3rds normal, same for goody huts).
#     Rearrange all of the options to put the less commonly used ones at the end.
#     Tweak up the default hills % slightly.
#     Tweak up the river rate slightly.
#     Make broad empty plains a little less likely by distributing hills better.
#     Add some extra continent options, 1 per 2 players, 1 per 3, 1 per 4
#       these may be easier for some to think about than picking an exact number.
#     Added continent options 2 per player, 3 per, 4 per.  These are good to use
#       if you're using low sea levels to allow land blobs to merge.
#     Improve the likelyhood that small numbers of continents are well separated.
#     Improve the chance that player start plots will neighbor at least one hill.
#     Added option to adjust the river and lake frequencies.
#     Tweaked the width/height calculation so that super wide maps aren't as likely.
#     Tweaked hills placement to make big empty plains less likely.
#     Improved the resource distribution to get better strategic resource separation.
#     Added no-continents option, which will result in all land being placed in
#       fragments instead (looks suspiciously like 'archipelago').
#     Fixed a bug in the counting of land resulting in fewer peaks than expected.
#     
# 7.1
#     After reading that Atlas can generate 1600x1600 maps (not that civiv can
#       load them) decided to bump up the maximum width/height overrides so
#       that if, in the future, civiv can handle such maps, SmartMap will be
#       ready.  Can now, hypothetically, generate 5000x5000 maps.
#     Also, converted the units to map tiles, since that is probably less confusing
#       for non map-scripters.
#     Bump up the size of duel, tiny, small due to our allowance for more ocean
#       and also because the actual chosen dimensions tend to result in a map with
#       slightly smaller than the expected area.
#         Areas     Duel Tiny Small Standard Large Huge
#         SmartMap: 80   140  220   360      560   860
#         Standard: 60   104  160   273      416   640
#     Make SmartMap strict default for feature generation.
#     Tweak up the default ocean size due to the larger map size.
# 7.0
#     Add new generation option: Land Style.  Land style allows you to opt
#       for less roundish continents, and also to allow for small independent
#       land fragments separate from the main continents.  When land fragments
#       are generated, they may be generated anywhere, and may violate the 
#       sea level strategy guarantees, so use at your own risk.  The less rounded
#       continent options without land fragments will obey the sea level rules,
#       but may be less fair.
#     Based on a report of it being annoying to have your capital placed on
#       the wrap-edge of the minimap, I now rotate the initial land placement
#       to put the minimum number possible of land near the map wrap edge.
#       While this does not guarantee this will not happen, it makes it highly
#       unlikely.
#     Further, made the first player always pick the center most start position
#       if wrap is enabled.  This should make single player games always have
#       the human player start fairly well centered.  This should absolutely
#       guarantee that a single human player will not start with their capital
#       on a map wrap edge.
#     Fixed a couple of places where I did not consider y-wrapping correctly.
#       But I guess no one plays y-wrap (or both wrap), or they'd have surely noticed.
#     Introduce option 'SmartMap Strict' for feature generation.  This will only place
#       floodplains & tundra where the normal scripts would.
#     Made floodplains on anything other than desert less likely, and oasis on tundra
#       less likely for the non-strict feature generator.
#     Enhanced the plot wetness calculator so plot wetness can be factored into more
#       of the generation process.
#     Improved the player placement rules to prefer not to place players initially
#       on a stretch of land reaching into the ocean.  
#     Added ReadMeSmartMap.txt
# 6.1
#     Tweak up number of hills for 'many hills'.
# 6.0
#     Improved lake generation system, grows more lakes beyond one plot
#     Numerous terrain relevance improvements.  More desert&snow at high altitude,
#       less desert near water, etc.
#     Changed continent computation to more accurately determine the large
#       continents, which should as a side effect improve the fair resource
#       distribution.  Removed tiny continents which will never be player
#       starts from the fair resource distribution.
#     Improved tile wetness calculator to consider number of rivers and lakes
#       touched by tile (a tile touched by fresh water on multiple sides is
#       now more likely to shift from desert->plains->grass).
#     Added a river originating at the highest point of each continent, which
#       greatly improves the fairness of river distribution overall.
#     Changed forest/jungle options to allow independent choice of forest and
#       jungle coverage (light,normal,heavy independently selectable for each)
#     General code cleanup, reorganization, commenting.
#     Ran code through pychecker and cleaned all warnings, pychecker found a
#       couple of legitimate bugs that may have improved the resulting maps.
# 5.1
#     Improvements to fair bonus distribution algorithm.
#     Fixed a typo bug causing python error on undefined edgeprob.
#     Tweak desert rate down slightly.
# 5.0
#     Work on refining the continent boundaries so water straights are not so
#       obvious when you see the resulting cymap.
#     Rewrite river generation, move it before terrain generation so terrains
#       can consider rivers (allow grasslands and other 'wetter' terrain to be
#       more common around rivers, as well as forests/jungle).
#     Replace the standard river altitude function with one that seems to yeild
#       better river paths.
#     Assorted terrain and feature improvements.  Restored possibility of flood
#       plains in desert, and made terrain tend to blob together more rather
#       than appearing so completely random.  Allowed hills to have terrain rather
#       than always defaulting to grass (so desert/hills are now possible).  I can't
#       believe no one complained about that one.
#     Found a major logic error in the player placement algorithm that was unfairly
#       favoring near-lake placements (I has assumed CyPlot.isCoastalLand() was equivalent
#       to asking: can I build a lighthouse?, this turned out not to be true).
#       As a result, initial placement is much more likely to be coastal now, and 
#       this has a nice side effect of improving the average separation of players.
#     Tweaked resource rates a little more based on various feedback.
#     Replace standard lake addition function so we don't get lakes in odd places.
#     Futher performance optimization.  The low and medium sea level options are
#       now more than twice as fast, and the high sea level option is about 50%
#       faster.  I can now generate a huge map with low sea level in under a minute.
#     After discovering that the map sizes in great plains are way off, resized my 
#       maps to the following settings:
#                   Duel Tiny Small Standard Large Huge
#         SmartMap: 60   100  180   320      560   1000
#         Standard: 60   104  160   273      416   640
#       This is a bump up for everything except huge, which becomes slightly smaller,
#       and thus hopefully usable by more people.
#     Added related option to override the map width and height.  You can now select
#       your map size fairly precisely, including the ability to pick map sizes that
#       are so large they may cause civiv to have problems.  Up to 256x256 map positions,
#       which would be a 1024x1024 map, with an area of over 1 million (compare to huge at
#       an area of 16 thousand!)  If you are able to generate a map of even 128x128 I'll be
#       shocked.  Nevertheless, the option is there.  So no complaining that I made huge
#       smaller!  Just set your game to huge, then override the size to whatever you want.
#       Or for a bizarre experience, set your map to duel, and override the size to whatever
#       you'd like.  For kicks I generated a 256x8 map, and that was just barely startable
#       with my 1 Gig of memory.  Minimap looks ridiculous.
# 4.2
#     Increase goody hut count slightly, particularly on maps with lots of land.
# 4.1
#     Slightly refined smartmap resource placement to make distribution
#       fairer between continents, and not to allow so many duped resources
#       within one city radius.
#     Small upward tweak to resource rates.
# 4.0
#     Overall performance improvement:
#        Now down to about: 
#        <5 seconds for duel,tiny,small
#        10 seconds for standard,
#        40 seconds for large,
#        ~2 minutes on huge
#        On a 1.8Ghz Pentium-M with 1Gig.
#     Optimize land placement speed by remembering all potential growth positions
#       rather than looking for them randomly.
#     Optimized and improved player placement function.  No longer makes certain
#       rare placement errors, and the speed is dramatically better, particularly
#       for continents with a large number of players to be placed (pangea).
#     Optimized bonus placement.  This used to get very slow on large/huge maps due
#       to making a pass through all tiles for each bonus type.  Now makes one pass
#       through all tiles and caches the result.
#     Changed the player option read function to take a string rather than an index
#       to make it clearer in code, and also make it impossible to break when
#       changing the order of options in the list.
#     Land areas are 20,40,90,210,500,1200 (duel,tiny,small,standard,large,huge)
#       Compared to  15,24,48,88 ,154,252 for great plains.
#       Smartmap rounds down calculated width and height,
#       so the areas are not quite as big as they seem,
#       but large and huge are noticeably bigger.  This allows for more flexible
#       use of the ocean coverage option.  Note to firaxis: it would be a better
#       design if in the expansion pack or a patch you passed the expected land
#       area to the grid size function from the current mod!  Then maps could
#       dynamically size themselves to the expected land area and provide fair
#       gameplay regardless of other chosen settings (and be compatible with mods
#       that make use of different sized maps ... no more hardcoding like in
#       great plains)!
#       Note: I can just barely start a huge game with 1Gig memory.  This implies
#       firaxis is using somewhere in the neighborhood of 32k memory per tile.
#       Good grief, what on earth are they storing!
# 3.2
#     Add options for specifying weight of initial jungle/forest cover.
#     More performance optimizations.
# 3.1 
#     Fix rare issue with placements landing too close together.
#     Slight tweaks to placement scoring to value early resources higher.
#     Fix issue with placements on separate continents too close together.
# 3.0 
#     Re-enable climate selection, which may or may not be honored.
#     Restore 36-continent option, accidental deletion.
#     Create a better default option for each selector.
#     Un-smooth edges near poles.
#     Properly handle allowing continents to wrap on wrap axis.
#     Performance optimizations by reducing function call counts.
#     Custom player placement rules.
#     Complete replacement of bonus system.
#     Separate out goody hut placement and put in correct order.
#     Cleaner debug print messages.
# 2.0 
#     Reduce bonus percentage for SmartMap bonus.
#     Allow standard bonus distribution as an option.
#     Bias width,height towards wrap direction.
#     un-straight-edge where continents meet.
#     SmartMap terrainGen allows more desert.
#     SmartMap featureGen suggests more forest and jungle.
#     Support for random ranges in number of continents.
# 1.0 
#     Initial version.
