# FILE: FfH_flavor_mapmod.py 
# VERSION: 3.00
# AUTHOR: David Reichert (dreiche2)
# PURPOSE: For the Fall from Heaven (FfH) mod: 
#		   Have civs start in regions that match their flavor.
# DESCRIPTION: 
# This module can be used in conjunction with other map scripts. It redistributes
# the original starting locations among the civs according to FfH flavor, i.e.
# Ljosalfar (elves) get a location with many forests, Lanun ("pirates") get a coastal
# start etc.

# Note that this module does not change the map itself, and only uses the starting locations
# assigned by the base map script. Hence, if there are no forests on the map at all, the elves
# won't get any, either. So there will only be a tendency (albeit a clear one), no guarantee 
# that all the relevant civs have 'flavory' staring positions. An advantage is that this module
# minimally interfers with what the map scripts does.


# USAGE:
# It is very simple to use this module in a map script of your choice (e.g. the default continents 
# script). Just copy this file into the <civ4>/PublicMaps folder, and then open the map script file 
# (e.g. Continents.py) with a text editor and copypaste the small code block below this paragraph into
# the script (preferably at the top), and uncomment the 'import' line by removing the '#'. Of course 
# you're advised to backup the map script and rename your modified version to something like 
# "FfH_whatever.py" so that you don't confuse the various versions. You can modify the flavor settings 
# to your liking, see SETTINGS below.

## copy below, remove '#' in 'from...import' line.

###############################################################################
# This map script was modified to use the Fall from Heaven flavor map module. #
# See FfH_flavor_mapmod.py for details.                                       #
###############################################################################

#from FfH_flavor_mapmod import normalizeStartingPlotLocations

###############################################################################

## copy above





# COMPATIBILITY
# The flavor module should work with most map scripts. It uses the normalizeStartingPlotLocations()
# function, hence civ does not use its default implementation. That means that members of a team
# are no longer placed close together as by default.

# A conflict occurs if the map script itselfs implements normalizeStartingPlotLocations(), though
# only few do. If that's the case you can simply remove/comment out the relevant code in the map
# script, although of course you remove any functionality that was implemented there.


# SETTINGS
# List of civs with preferred starting spots:
#
# Malakim: desert
# Lanun: coastal start
# Ljosalfar: forest
# Luchuirp and Khazad: hills
# Doviello: tundra
# Illians: tundra (low priority)
# Hippus: like: grass, plains; dislike somewhat: hills, forests

# Note: It should be easy to modify the settings to your liking, and add new flavor civs.
# Just look at the code below. Everything relevant happens in normalizeStartingPlotLocations().




# VERSION HISTORY

# 3.0
# Put the flavor code into a module. Prior versions had all the code inside the modified map scripts.
# Should be straight forward now for people to make their own modified scripts.
#
# Also removed everything save the core flavor feature (i.e. removed tower placement, map checker).
# Now the flavor module really concentrates on its main purpose.

# 2.0
# 	*flavor
# 	 *Starting locations determined by the base script are redistributed to have flavor civs start
# 	 *in matching surroundings, if possible
# 	 
# 	- added CivFlavorInfo class to store civ relates infos in (see normalizeStartingPlotLocations() 
# 	  to change civ details).
# 	- added priority to change the order in which flavor civs get their starting locations. 
# 	  Possible values: 1,2,3. 1 is highest priority.
# 	- civs can have any combination of flavor terrains, features, boni and plotTypes (e.g. Hills),
# 	  each with its own value, which counts toward the total score of a starting location.
# 	- regionSize, closeRegionSize and closeFactor can be changed for each civ independently.

# 	*FfH map checker
# 	 *Usually runs in FfH on game start, adds necessary resources if they are scarce. Added to
# 	 *script so that map checker also works if map is regenerated.
# 	 
# 	- added map checker from FfH CvEventManager.py and CustomFunctions.py.

# 	*Sentry Tower redistribution
# 	 *choke points are plots with blocking plots around and some non-blocking land in between,
# 	 *basically they are a passage through impassible terrain (peaks, water). Only plots adjacent
# 	 *to possible choke plot are considered. This is rather a proof of concept. Sureshot from the
#	 *FfH forums made the AI build forts on towers and defend them. Hence, this would result
#	 *in guarded forts on choke points...
# 	 
# 	- number of sentry towers is increased by 50% from the number created by the base script
# 	- half of towers is relocated to choke points.
# 	  		  
# VERSION 2.0 civs:
# List of civs with preferred starting spots:
#
# Malakim: desert
# Lanun: coastal start
# Ljosalfar: forest
# Luchuirp and Khazad: hills
# Doviello: tundra
# Illians: tundra (low priority)
# Hippus: like: grass, plains; dislike somewhat: hills, forests



# VERSION 1.05 description

# This is the result of 1 1/2 days of learning Python and Civ modding, so bear with me :)
#
# This map script is a modified version of the original script to be used
# with the Fall from Heaven (FfH) mod. The purpose of the modification is to make civs start in 
# a region which matches their specific flavour, i.e. Ljosalfar (the elves) should
# start in a region with many forests, etc. (see list below).
# 
# I didn't want to mess with the algorithm that generates the starting locations in 
# the first place, so all I do is to take the starting positions the base script 
# created and redistribute them among the civs as good as possible. For example, for the
# elves I go through the list of locations and select the one with the most forests in
# the surrounding region. Of course, this method will only create a tendency and will
# not always result in flavour starts. In particularly:
#
# 1. It doesn't change the map itself. If there are no forests whatsoever then the 
# elves will have to live without them.
#
# 2.The greater the number of players in the game, the greater the number of generated 
# starting positions, and hence the higher the probability that there is some nice spot 
# available for the flavour civs.
#
# 3. I only overrode normalizeStartingPlotLocations(), which seems to be the only thing
# that is called after the starting locations have been assigned. The good news is that
# it seems like most scripts don't mess with this function, so it should be easy to 
# add the functionality of this script to other scripts simply by adding my implementation 
# of normalizeStartingPlotLocations() (together with some helper functions).
#
# 4. Players with lower id's get to pick their sweet spot first, i.e. in particularly 
# the human player in single player games.
#
# 5. Finally, the purpose of this script is flavour, not balance ;)
#
# Have fun! David (dreiche2)
#



from CvPythonExtensions import *
import CvUtil
import CvMapGeneratorUtil


###################### some parameters ########################
#
# these are default settings. They can be changed for single civs
# in normalizeStartingPlotLocations(), e.g. InfoMalakim.regionSize = 4


def regionSize():
	return 10 #region in which to look is a square reaching regionSize tiles in any direction
			 #from the center tile

def closeRegionSize():
	return 4 #closer squares have larger impact
	
def closeFactor():
	return 1 #weight for closer squares

###############################################################

FFH_DEBUG=False


#helper function
#unfortunately there seems to be no inverse procedure in general
#could have stored strings in the dictionary instead of the type ints :/
def IT(string):
	cgc = CyGlobalContext()
	return cgc.getInfoTypeForString(string)

	

	
############################################################################
# in this function the locatons are redistributed. Make your changes here...

def normalizeStartingPlotLocations():

	#creating flavor civs

	#possible parameters to change:
	#.terrains[]
	#.boni[]
	#.plotTypes[]
	#.features[]
	#.SetPriority()
	#.coastalStartValue
	#.regionSize   (etc.)
	#
	#examples see below, and see the civFlavorInfo class 
	#after normalizeStartingPlotLocations()
	
			
	civInfos = {} #hurray for dictionaries!
	#info = CivFlavorInfo(IT("CIVILIZATION_LUMPLAND"))
	#infoLumpland.terrains[IT("TERRAIN_TUNDRA")] = 1
	#infoLumpland.setPriority(3) # lower priority compared to Doviello, which use the default value 
							   # of 2, because the Illians don't get boni from tundra up to now; 
							   # default priority is 2, possible 1-3
	#infoLumpland.closeFactor = 1  # don't value close tiles stronger as by default
	#civInfos[IT("CIVILIZATION_LUMPLAND")] = infoLumpland
	################


#EURO-COLD - tundra
	infoRussia = CivFlavorInfo(IT("CIVILIZATION_RUSSIA"))
	infoRussia.setPriority(1)
	infoRussia.terrains[IT("TERRAIN_TUNDRA")] = 1
	infoRussia.features[IT("FEATURE_ICE")] = 1
	infoRussia.boni[IT("BONUS_SHEEP")] = 1
	infoRussia.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoRussia.terrains[IT("TERRAIN_DESERT")] = -1
	infoRussia.regionSize = 10
	civInfos[IT("CIVILIZATION_RUSSIA")] = infoRussia
	#################
	infoViking = CivFlavorInfo(IT("CIVILIZATION_VIKING"))
	infoViking.setPriority(1)
	infoViking.terrains[IT("TERRAIN_TUNDRA")] = 1
	infoViking.features[IT("FEATURE_ICE")] = 1
	infoViking.boni[IT("BONUS_SHEEP")] = 1
	infoViking.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoViking.terrains[IT("TERRAIN_DESERT")] = -1
	infoViking.regionSize = 10
	civInfos[IT("CIVILIZATION_VIKING")] = infoViking
	#################

#EURO-TEMPERATE - sheep
	infoCelt = CivFlavorInfo(IT("CIVILIZATION_CELT"))
	infoCelt.setPriority(1)
	infoCelt.boni[IT("BONUS_SHEEP")] = 1
	infoCelt.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoCelt.terrains[IT("TERRAIN_DESERT")] = -1
	infoCelt.regionSize = 10
	civInfos[IT("CIVILIZATION_CELT")] = infoCelt
	#################
	infoEngland = CivFlavorInfo(IT("CIVILIZATION_ENGLAND"))
	infoEngland.setPriority(1)
	infoEngland.boni[IT("BONUS_SHEEP")] = 1
	infoEngland.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoEngland.terrains[IT("TERRAIN_DESERT")] = -1
	infoEngland.regionSize = 10
	civInfos[IT("CIVILIZATION_ENGLAND")] = infoEngland
	#################
	infoFrance = CivFlavorInfo(IT("CIVILIZATION_FRANCE"))
	infoFrance.setPriority(1)
	infoFrance.boni[IT("BONUS_SHEEP")] = 1
	infoFrance.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoFrance.terrains[IT("TERRAIN_DESERT")] = -1
	infoFrance.regionSize = 10
	civInfos[IT("CIVILIZATION_FRANCE")] = infoFrance
	#################
	infoGermany = CivFlavorInfo(IT("CIVILIZATION_GERMANY"))
	infoGermany.setPriority(1)
	infoGermany.boni[IT("BONUS_SHEEP")] = 1
	infoGermany.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoGermany.terrains[IT("TERRAIN_DESERT")] = -1
	infoGermany.regionSize = 10
	civInfos[IT("CIVILIZATION_GERMANY")] = infoGermany
	#################
	infoHolyRomanEmpire = CivFlavorInfo(IT("CIVILIZATION_HOLY_ROMAN_EMPIRE"))
	infoHolyRomanEmpire.setPriority(1)
	infoHolyRomanEmpire.boni[IT("BONUS_SHEEP")] = 1
	infoHolyRomanEmpire.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoHolyRomanEmpire.terrains[IT("TERRAIN_DESERT")] = -1
	infoHolyRomanEmpire.regionSize = 10
	civInfos[IT("CIVILIZATION_HOLY_ROMAN_EMPIRE")] = infoHolyRomanEmpire
	#################
	infoNetherlands = CivFlavorInfo(IT("CIVILIZATION_NETHERLANDS"))
	infoNetherlands.setPriority(1)
	infoNetherlands.boni[IT("BONUS_SHEEP")] = 1
	infoNetherlands.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoNetherlands.terrains[IT("TERRAIN_DESERT")] = -1
	infoNetherlands.regionSize = 10
	civInfos[IT("CIVILIZATION_NETHERLANDS")] = infoNetherlands
	#################

#EURO-MEDITERRANEAN - sheep, cow
	infoByzantium = CivFlavorInfo(IT("CIVILIZATION_BYZANTIUM"))
	infoByzantium.setPriority(1)
	infoByzantium.boni[IT("BONUS_SHEEP")] = 1
	infoByzantium.boni[IT("BONUS_COW")] = 1
	infoByzantium.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoByzantium.regionSize = 10
	civInfos[IT("CIVILIZATION_BYZANTIUM")] = infoByzantium
	#################
	infoGreece = CivFlavorInfo(IT("CIVILIZATION_GREECE"))
	infoGreece.setPriority(1)
	infoGreece.boni[IT("BONUS_SHEEP")] = 1
	infoGreece.boni[IT("BONUS_COW")] = 1
	infoGreece.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoGreece.regionSize = 10
	civInfos[IT("CIVILIZATION_GREECE")] = infoGreece
	#################
	infoPortugal = CivFlavorInfo(IT("CIVILIZATION_PORTUGAL"))
	infoPortugal.setPriority(1)
	infoPortugal.boni[IT("BONUS_SHEEP")] = 1
	infoPortugal.boni[IT("BONUS_COW")] = 1
	infoPortugal.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoPortugal.regionSize = 10	
	civInfos[IT("CIVILIZATION_PORTUGAL")] = infoPortugal
	#################
	infoRome = CivFlavorInfo(IT("CIVILIZATION_ROME"))
	infoRome.setPriority(1)
	infoRome.boni[IT("BONUS_SHEEP")] = 1
	infoRome.boni[IT("BONUS_COW")] = 1
	infoRome.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoRome.regionSize = 10	
	civInfos[IT("CIVILIZATION_ROME")] = infoRome
	#################
	infoSpain = CivFlavorInfo(IT("CIVILIZATION_SPAIN"))
	infoSpain.setPriority(1)
	infoSpain.boni[IT("BONUS_SHEEP")] = 1
	infoSpain.boni[IT("BONUS_COW")] = 1
	infoSpain.terrains[IT("TERRAIN_JUNGLE")] = -1
	infoSpain.regionSize = 10
	civInfos[IT("CIVILIZATION_SPAIN")] = infoSpain
	#################

#ARID MIDDLE EAST/NORTH AFRICA - desert
	infoArabia = CivFlavorInfo(IT("CIVILIZATION_ARABIA"))
	infoArabia.setPriority(3)
	infoArabia.terrains[IT("TERRAIN_DESERT")] = 1
	infoArabia.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoArabia.features[IT("FEATURE_ICE")] = -1
	infoArabia.regionSize = 10
	civInfos[IT("CIVILIZATION_ARABIA")] = infoArabia
	#################
	infoBabylon = CivFlavorInfo(IT("CIVILIZATION_BABYLON"))
	infoBabylon.setPriority(3)
	infoBabylon.terrains[IT("TERRAIN_DESERT")] = 1
	infoBabylon.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoBabylon.features[IT("FEATURE_ICE")] = -1
	infoBabylon.regionSize = 10
	civInfos[IT("CIVILIZATION_BABYLON")] = infoBabylon
	#################	
	infoCarthage = CivFlavorInfo(IT("CIVILIZATION_CARTHAGE"))
	infoCarthage.setPriority(3)
	infoCarthage.terrains[IT("TERRAIN_DESERT")] = 1
	infoCarthage.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoCarthage.features[IT("FEATURE_ICE")] = -1	
	infoCarthage.regionSize = 10
	civInfos[IT("CIVILIZATION_CARTHAGE")] = infoCarthage
	#################
	infoEgypt = CivFlavorInfo(IT("CIVILIZATION_EGYPT"))
	infoEgypt.setPriority(3)
	infoEgypt.terrains[IT("TERRAIN_DESERT")] = 1
	infoEgypt.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoEgypt.features[IT("FEATURE_ICE")] = -1
	infoEgypt.regionSize = 10
	civInfos[IT("CIVILIZATION_EGYPT")] = infoEgypt
	#################
	infoOttoman = CivFlavorInfo(IT("CIVILIZATION_OTTOMAN"))
	infoOttoman.setPriority(3)
	infoOttoman.terrains[IT("TERRAIN_DESERT")] = 1
	infoOttoman.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoOttoman.features[IT("FEATURE_ICE")] = -1
	infoOttoman.regionSize = 10
	civInfos[IT("CIVILIZATION_OTTOMAN")] = infoOttoman
	#################
	infoPersia = CivFlavorInfo(IT("CIVILIZATION_PERSIA"))
	infoPersia.setPriority(3)
	infoPersia.terrains[IT("TERRAIN_DESERT")] = 1
	infoPersia.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoPersia.features[IT("FEATURE_ICE")] = -1
	infoPersia.regionSize = 10
	civInfos[IT("CIVILIZATION_PERSIA")] = infoPersia
	#################
	infoSumeria = CivFlavorInfo(IT("CIVILIZATION_SUMERIA"))
	infoSumeria.setPriority(3)
	infoSumeria.terrains[IT("TERRAIN_DESERT")] = 1
	infoSumeria.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoSumeria.features[IT("FEATURE_ICE")] = -1
	infoSumeria.regionSize = 10
	civInfos[IT("CIVILIZATION_SUMERIA")] = infoSumeria
	#################

#TROPICAL AFRICA - jungle
	infoEthiopia = CivFlavorInfo(IT("CIVILIZATION_ETHIOPIA"))
	infoEthiopia.setPriority(3)
	infoEthiopia.terrains[IT("TERRAIN_JUNGLE")] = 1
	infoEthiopia.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoEthiopia.features[IT("FEATURE_ICE")] = -1
	infoEthiopia.boni[IT("BONUS_SHEEP")] = -1
	infoEthiopia.regionSize = 10
	civInfos[IT("CIVILIZATION_ETHIOPIA")] = infoEthiopia
	#################
	infoMali = CivFlavorInfo(IT("CIVILIZATION_MALI"))
	infoMali.setPriority(3)
	infoMali.boni[IT("TERRAIN_JUNGLE")] = 1
	infoMali.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoMali.features[IT("FEATURE_ICE")] = -1
	infoMali.boni[IT("BONUS_SHEEP")] = -1
	infoMali.regionSize = 10
	civInfos[IT("CIVILIZATION_MALI")] = infoMali
	#################
	infoZulu = CivFlavorInfo(IT("CIVILIZATION_ZULU"))
	infoZulu.setPriority(3)
	infoZulu.boni[IT("TERRAIN_JUNGLE")] = 1
	infoZulu.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoZulu.features[IT("FEATURE_ICE")] = -1
	infoZulu.boni[IT("BONUS_SHEEP")] = -1
	infoZulu.regionSize = 10
	civInfos[IT("CIVILIZATION_ZULU")] = infoZulu
	#################

#SOUTH ASIA - jungle
	infoIndia = CivFlavorInfo(IT("CIVILIZATION_INDIA"))
	infoIndia.setPriority(3)
	infoIndia.terrains[IT("TERRAIN_JUNGLE")] = 1
	infoIndia.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoIndia.features[IT("FEATURE_ICE")] = -1
	infoIndia.boni[IT("BONUS_SHEEP")] = -1
	infoIndia.regionSize = 10
	civInfos[IT("CIVILIZATION_INDIA")] = infoIndia
	#################

#EAST ASIA
	#################
	infoKhmer = CivFlavorInfo(IT("CIVILIZATION_KHMER"))
	infoKhmer.setPriority(2)
	infoKhmer.boni[IT("BONUS_PIG")] = 1
	infoKhmer.terrains[IT("TERRAIN_JUNGLE")] = 1
	infoKhmer.terrains[IT("TERRAIN_TUNDRA")] = -1
	infoKhmer.features[IT("FEATURE_ICE")] = -1
	infoKhmer.boni[IT("BONUS_SHEEP")] = -1
	infoKhmer.regionSize = 10
	civInfos[IT("CIVILIZATION_KHMER")] = infoKhmer
	#################
	civInfos[IT("CIVILIZATION_KHMER")] = infoKhmer
	infoKhmer.setPriority(2)
	infoChina = CivFlavorInfo(IT("CIVILIZATION_CHINA"))
	infoChina.boni[IT("BONUS_PIG")] = 1
	infoChina.boni[IT("BONUS_SHEEP")] = -1
	infoChina.regionSize = 10
	civInfos[IT("CIVILIZATION_CHINA")] = infoChina
	#################
	infoJapan = CivFlavorInfo(IT("CIVILIZATION_JAPAN"))
	infoJapan.setPriority(2)
	#infoJapan.boni[IT("BONUS_PIG")] = 1
	infoJapan.boni[IT("BONUS_SHEEP")] = -1
	infoJapan.regionSize = 10
	civInfos[IT("CIVILIZATION_JAPAN")] = infoJapan
	#################
	infoKorea = CivFlavorInfo(IT("CIVILIZATION_KOREA"))
	infoKorea.setPriority(2)
	#infoKorea.boni[IT("BONUS_PIG")] = 1
	infoKorea.boni[IT("BONUS_SHEEP")] = -1
	infoKorea.regionSize = 10
	civInfos[IT("CIVILIZATION_KOREA")] = infoKorea
	#################
	infoMongol = CivFlavorInfo(IT("CIVILIZATION_MONGOL"))
	infoMongol.setPriority(2)
	#infoMongol.boni[IT("BONUS_PIG")] = 1
	infoMongol.boni[IT("BONUS_SHEEP")] = -1
	info.regionSize = 10
	civInfos[IT("CIVILIZATION_MONGOL")] = infoMongol
	#################

#WARM AMERICAS - dye
	infoAmerica = CivFlavorInfo(IT("CIVILIZATION_AMERICA"))
	infoAmerica.setPriority(3)
	#infoAmerica.boni[IT("BONUS_CORN")] = 1
	infoAmerica.regionSize = 10
	civInfos[IT("CIVILIZATION_AMERICA")] = infoAmerica
	#################
	infoNativeAmerica = CivFlavorInfo(IT("CIVILIZATION_NATIVE_AMERICA"))
	infoNativeAmerica.setPriority(3)
	#infoNativeAmerica.boni[IT("BONUS_CORN")] = 1
	infoNativeAmerica.regionSize = 10
	civInfos[IT("CIVILIZATION_NATIVE_AMERICA")] = infoNativeAmerica
	#################

#TROPICAL AMERICAS - jungle, dye
	infoAztec = CivFlavorInfo(IT("CIVILIZATION_AZTEC"))
	infoAztec.setPriority(3)
	#infoAztec.boni[IT("BONUS_CORN")] = 1
	#infoAztec.terrains[IT("TERRAIN_JUNGLE")] = 1
	infoAztec.regionSize = 10
	civInfos[IT("CIVILIZATION_AZTEC")] = infoAztec
	#################
	infoInca = CivFlavorInfo(IT("CIVILIZATION_INCA"))
	infoInca.setPriority(3)
	#infoInca.boni[IT("BONUS_CORN")] = 1
	#infoInca.terrains[IT("TERRAIN_JUNGLE")] = 1
	infoInca.regionSize = 10
	civInfos[IT("CIVILIZATION_INCA")] = infoInca
	#################
	infoMaya = CivFlavorInfo(IT("CIVILIZATION_MAYA"))
	infoMaya.setPriority(3)
	#infoMaya.boni[IT("BONUS_CORN")] = 1
	#infoMaya.terrains[IT("TERRAIN_JUNGLE")] = 1
	infoMaya.regionSize = 10
	civInfos[IT("CIVILIZATION_MAYA")] = infoMaya
	#################

	
	debugOut("flavor civs: ")
	debugOut(civInfos.keys())
	
	#reallocation
	NiTextOut("Python: Redistributing starting positions (FfH_script_version) ...")	
	cgc = CyGlobalContext()
	dice = cgc.getGame().getMapRand()
	cymap = CyMap() 
	width = cymap.getGridWidth()
	height = cymap.getGridHeight()
	iPlayers = cgc.getGame().countCivPlayersEverAlive()

	startingPlots = []
	playersLeft = range(iPlayers)
	playersToRemove = []

		
	#creating list of starting plots
	for playerIndex in range(iPlayers):
		pPlayer = cgc.getPlayer(playerIndex)
		pStartPlot = pPlayer.getStartingPlot()	  
		startingPlots.append(pStartPlot)
		

	#reallocation...
	for priorityCount in [1,2,3]:
		debugOut("priority count: " + repr(priorityCount))
		for pRem in playersToRemove:
			playersLeft.remove(pRem)
		playersToRemove = []
		debugOut("players left:")
		debugOut(playersLeft)
		for playerIndex in playersLeft:
			player=cgc.getPlayer(playerIndex)
			civName=player.getCivilizationType()
			debugOut("checking the following civ: " + repr(civName))
			if civName in civInfos.keys() and  civInfos[civName]._priority == priorityCount:
				debugOut("ok found a candidate")
				bestIndex=findBestPlot(civInfos[civName], startingPlots)
				player.setStartingPlot(startingPlots[bestIndex],True)	
				del startingPlots[bestIndex]	
				playersToRemove.append(playerIndex)	#ok some code could be more elegant						
	
	#remaining players	
	for Index in range(len(playersLeft)):
		player=cgc.getPlayer(playersLeft[Index])
		player.setStartingPlot(startingPlots[Index],True)	


	return None

	
	

################# for reallocation: helper functions & classes below #######################
		



class CivFlavorInfo: # use setter function to change member values with underscore

	def __init__(self, name):
		self._name = name
		self._priority = 2 
		self.terrains = {}
		self.features = {}
		self.boni = {}
		self.plotTypes = {}
		self.regionSize = regionSize()
		self.closeRegionSize = closeRegionSize()
		self.closeFactor = closeFactor()
		self.coastalStartValue = 0
	
	def setPriority(self,priority): # the lower the earlier
		if priority < 1:
			priority = 1
		if priority > 3:
			priority = 3
		self.priority_ = priority


# returns index of best plot for the civ
def findBestPlot(civInfo, plotList):
	debugOut("finding now best spot for: " + repr(civInfo._name))
	debugOut("civ info:")
	debugOut(civInfo.terrains)
	debugOut(civInfo.plotTypes)
	debugOut(civInfo.features)
	debugOut(civInfo.boni)
	debugOut("coastalStartValue:" + repr(civInfo.coastalStartValue))
	cymap = CyMap() 
	bestIndex=0
	foundMax=0

	for plotIndex in range(len(plotList)):
		startPlot=plotList[plotIndex]
		x0=startPlot.getX()
		y0=startPlot.getY()	
		# print "start tile at ", x0," ", y0
		debugOut("checking start at " +repr(x0) +" " + repr(y0))

		score=0
		if isCoast(startPlot):
			score+=civInfo.coastalStartValue


		for x in range(x0-regionSize(),x0+regionSize()+1):
			for y in range(y0-regionSize(),y0+regionSize()+1):
				plotValue=0
				plot=cymap.plot(x,y)

				terrain = plot.getTerrainType()
				plotType = plot.getPlotType()				
				feature = plot.getFeatureType()				
				bonus = plot.getBonusType(-1)				

				if terrain in civInfo.terrains.keys():
					plotValue+=civInfo.terrains[terrain]
				if plotType in civInfo.plotTypes.keys():
					plotValue+=civInfo.plotTypes[plotType]
				if feature in civInfo.features.keys():
					plotValue+=civInfo.features[feature]
				if bonus in civInfo.boni.keys():
					plotValue+=civInfo.boni[bonus]

				if x >= x0-closeRegionSize() and x <= x0+closeRegionSize() \
and y >= y0-closeRegionSize() and y <= y0+closeRegionSize():
					plotValue*=closeFactor()

				score+=plotValue
						
		debugOut("found here: " + repr(score))
		if foundMax<score:
			foundMax=score
			bestIndex=plotIndex				
	debugOut("found max: " +repr(foundMax))
	
	return bestIndex	
	



# coastal plot?
def isCoast(plot):
	WaterArea = plot.waterArea()
	if not WaterArea.isNone():
		if not WaterArea.isLake():
			return	True
	return False
		

def debugOut(output):
	if FFH_DEBUG:
		print(output)
	return	

