################################################################################
## File: World Piece Vanila v1.0
## Author: Keith Sponburgh (Seven05)
##
## Substantial portions of this code taken from:
## PerfectWorld.py (Rich Marinaccio aka cephalo)
##
## Without PerfectWorld this map script wouldn't exist.  The base code used is
## the PerfectWorld.py script, my changes are simply minor modifications to that
## code.
##
## Additional Credits:
## SevenSpirits (Starting plot modifications)
## GRM7584 (Earth2 plot generator)
##
## The only change in this mapscript from the one used in my World Piece mod is
## in the generateTerrainTypes() function where all of the climate specific
## terrains have been replaced by the default terrains.  This occurs after the
## world and climate have both been generated so the various biome types are
## supported but lack the visual differences seen in World Piece.
##
################################################################################
## GLOBAL TUNING VARIABLES: Change these to customize the map results

#How much land versus total map squares. Keep in mind that values higher than
#0.25 will greatly minimize the chance of a significant 'new world'.
LandPercent = 0.25

#How many map squares will be above peak threshold and thus 'peaks'.
PeakPercent = 0.03

#How many map squares will be above hill threshold and thus 'hills' unless
#they are also above peak threshold in which case they will be 'peaks'.
HillPercent = 0.12

#How many map squares will be below desert rainfall threshold. In this case,
#rain levels close to zero are very likely to be desert, while rain levels close
#to the desert threshold will more likely be plains.
DesertPercent = 0.075

#How many map squares will be below plains rainfall threshold. Rain levels close
#to the desert threshold are likely to be plains, while those close to the plains
#threshold are likely to be grassland. Any rain above plains threshold will be
#grassland.
PlainsPercent = 0.13

#---The following variables are not based on percentages. Because temperature
#---is so strongly tied to latitude, using percentages for things like ice and
#---tundra leads to very strange results if most of the worlds land lies near
#---the equator

#What temperature will be considered cold enough to be ice. Temperatures range
#from coldest 0.0 to hottest 1.0.
IceTemp = .15

#What temperature will be considered cold enough to be tundra. Temperatures range
#from coldest 0.0 to hottest 1.0.
TundraTemp = .30

#Hotter than this temperature will be considered deciduous forest, colder will
#be evergreen forest.Temperatures range from coldest 0.0 to hottest 1.0.
ForestTemp = .45

#What temperature will be considered hot enough to be jungle. Temperatures range
#from coldest 0.0 to hottest 1.0.
JungleTemp = .65

#---The following variables modify various map details.

#The percent chance that an oasis may appear in desert. A tile must be desert and
#surrounded on all sides by desert.
OasisChance = .06

#How many squares are added to a lake for each unit of river length flowing
#into it.
LakeSizePerRiverLength = .8

#This value modifies LakeSizePerRiverLength when a lake begins in desert
DesertLakeModifier = .4

#This value is used to decide if enough water has accumulated to form a river.
#A higher value creates more rivers over the entire map.
RiverThreshold = 0.3

#This value is multiplied by the river threshold to determine if there is
#enough water to flood a river bank and create a floodplain. A value of 1.0
#will cause every river to flood, higher values will cause fewer river tiles
#to flood.
FloodPlainThreshold = 1.0

#---I experimented alot with these rain smoothing variables, and decided that
#---they do not improve the map. Therefore, I have set them to zero. Go ahead
#---and play with these if you want, but their main effect is to diminish
#---all the hard work that went into simulating climate.

#How much noise to add to the rainfall map. This helps break up harsh lines
#near the rainy latitudes.
RainNoise = .1

#Smoothing factor for the rain map. Works with RainNoise to avoid harsh
#artifacts due to rainfall wind zone changes, especially near the equator.
RainSmoothingFactor = .02

#---These values are for evaluating starting locations

#The following values are used for assigning starting locations
CommerceValue = 9
ProductionValue = 11
FoodValue = 10

#Coastal/River cities are important, how important is determined by this
#value.
CoastalCityValueBonus = 1.2
FreshWaterValueBonus = 1
RiverCityValueBonus = 1.2

#Decides whether to use the Python random generator or the one that is
#intended for use with civ maps. The Python random has much higher precision
#than the civ one. 53 bits for Python result versus 16 for getMapRand. The
#rand they use is actually 32 bits, but they shorten the result to 16 bits.
#However, the problem with using the Python random is that it may create
#syncing issues for multi-player now or in the future, therefore it must
#be optional.
UsePythonRandom = True

from CvPythonExtensions import *
import CvUtil
import CvMapGeneratorUtil


from array import array
from random import random,randint,seed
import math
import sys
import operator

class PerfectWorldRandom :
	def __init__(self):
		return
	def seed(self):
		#Python randoms are not usable in network games.
		if UsePythonRandom:
			self.usePR = True
		else:
			self.usePR = False
		if self.usePR and CyGame().isNetworkMultiPlayer():
			print "Detecting network game. Setting UsePythonRandom to False."
			self.usePR = False
		if self.usePR:
			# Python 'long' has unlimited precision, while the random generator
			# has 53 bits of precision, so I'm using a 53 bit integer to seed the map!
			seed() #Start with system time
			seedValue = randint(0,9007199254740991)
			seed(seedValue)
			print "Random seed (Using Pyhon rands) for this map is %(s)20d" % {"s":seedValue}
		else:
			gc = CyGlobalContext()
			self.mapRand = gc.getGame().getMapRand()

			seedValue = self.mapRand.get(65535,"Seeding mapRand - PerfectWorld.py")
			self.mapRand.init(seedValue)
			print "Random seed (Using getMapRand) for this map is %(s)20d" % {"s":seedValue}
		return
	def random(self):
		if self.usePR:
			return random()
		else:
			#This formula is identical to the getFloat function in CvRandom. It
			#is not exposed to Python so I have to recreate it.
			fResult = float(self.mapRand.get(65535,"Getting float -PerfectWorld.py"))/float(65535)
			return fResult
	def randint(self,rMin,rMax):
		if self.usePR:
			return randint(rMin,rMax)
		else:
			iResult = rMin + self.mapRand.get(rMax - rMin,"Getting a randint - PerfectWorld.py")
			return iResult
#Set up random number system for global access
PWRand = PerfectWorldRandom()

class Heightmap :
	def __init__(self):
		return
	def GenerateMap(self,mapWidth,mapHeight,landPercent):
		if landPercent < 0 or landPercent > 1:
			raise ValueError, "landPercent must be between 0.0 and 1.0"

		#This is necessary because the doRiver function requires one recursion
		#level for each unit of river length. Since we don't want to limit our
		#river length to 25, we must set the recursion limit somewhat longer
		sys.setrecursionlimit( 100)

		# Map options
		global AllowNewWorld
		global AllowPangaea
		global LandmassStyle

		self.mapHeight = mapHeight
		self.mapWidth = mapWidth
		print "Map dimensions are %(h)2d x %(w)2d" % {"h":mapHeight,"w":mapWidth}
		self.chunkSize = self.FindIdealChunkSize(mapWidth,mapHeight)
		self.landPercent = landPercent

		#initial level, this will be changed to fit map
		self.seaLevel = .45

		#Polar Gap pushes land away from the poles
		self.polarGap = 0.25

		self.map = array('d')
		#initialize map with zeros
		for i in range(0,mapHeight*mapWidth):
			self.map.append(0.0)

		self.GenerateHeightField()
		self.NormalizeHeightMap()

		if AllowNewWorld == 1:
			if AllowPangaea == 0 and LandmassStyle != 1 and LandmassStyle != 3:
				self.FinalizeHeightField()

		self.PerformTectonics()
		self.NormalizeHeightMap()

		self.FindSeaLevel()
		self.GenerateFinalPlotTypes()
		self.FillInLakes()
		self.ShiftLandToMiddle()

	def FindIdealChunkSize(self,Width,Height):
		#The ideal chunksize is the largest power of 2 that is smaller than
		#Width/4, such that the Height is also evenly divisible by chunkSize.
		#If the chunkSize is too small (< Width/16) the map will be dotted with
		#tiny islands which looks bad and is not the intent of this map script
		#so an exception is thrown.
		targetSize = Width/4
		s = targetSize
		while s >= Width/16:
			p = 1
			while 2**p < targetSize:
				if (2**p)*2 > s:
					if Width % (2**p) == 0 and Height % (2**p) == 0:
						print "Final chunk size is " + str(2**p)
						return 2**p
				p += 1
			s = s/2
		errorString = \
		"This map script has fairly strict size requirements. " + \
		"Dimensions that are powers of 2 should guarantee compliance. " + \
		"Specifically, map must be evenly divisible, in Width and in " + \
		"Height, by a power of two that is at least mapWidth/16."

		raise ValueError,errorString


		return
	def GenerateHeightField(self):
		#get number of peaks needed to approximate land percentage
		numberOfPeaks = int(round((self.mapHeight/self.chunkSize) * (self.mapWidth/self.chunkSize) * self.landPercent))

		# Map options
		global LandmassStyle

		#randomly place initial peaks at beggining of each chunk until the
		#the exact desired amount is reached.
		topGap = self.mapHeight - self.chunkSize
		botGap = self.chunkSize

		allPeaksPlaced = False
		peaksPlaced = 0
		while allPeaksPlaced == False:
			#Western Region
			iSouthY = int(botGap)
			iNorthY = int(topGap) - 1
			iWestX = int(self.mapWidth * 0.1)
			iEastX = int(self.mapWidth * 0.4) - 1

			for y in range(iSouthY,iNorthY,self.chunkSize):
				if not allPeaksPlaced:
					for x in range(iWestX,iEastX,self.chunkSize):
						i = self.GetIndex(x,y)
						if PWRand.random() < self.landPercent * 0.8:
							self.map[i] = 1.0
							peaksPlaced += 1
							if peaksPlaced == numberOfPeaks:
								allPeaksPlaced = True
								break

			#Eastern Region
			iSouthY = int(botGap)
			iNorthY = int(topGap) - 1
			iWestX = int(self.mapWidth * 0.6)
			iEastX = int(self.mapWidth * 0.9) - 1

			for y in range(iSouthY,iNorthY,self.chunkSize):
				if not allPeaksPlaced:
					for x in range(iWestX,iEastX,self.chunkSize):
						i = self.GetIndex(x,y)
						if PWRand.random() < self.landPercent:
							self.map[i] = 1.0
							peaksPlaced += 1
							if peaksPlaced == numberOfPeaks:
								allPeaksPlaced = True
								break

		#now generate rest of map using midpoint displacement. For rectangular
		#heightmaps this requires two passes
		currentGrain = float(self.chunkSize)
		while currentGrain > 1:
			for y in range(0,self.mapHeight,int(currentGrain)):
				if y < botGap * 1.25 or y > topGap * 0.75:
					for x in range(0,self.mapWidth,int(currentGrain)):
						#h is scalar for random displacement
						h = currentGrain/float(self.chunkSize)
						#calculate points one currentGrain/2 to the right and
						#also one down
						mid = self.GetIndex(x + int(currentGrain/2),y)
						left = self.GetIndex(x,y)
						right = self.GetIndex(x + int(currentGrain),y)
						self.map[mid] = ((self.map[left] + self.map[right])/2 \
							+ (h*PWRand.random() - h/2.0)) * 0.5
						#also add heuristic to left and right to help reduce artifacts
						self.map[left] += (h*PWRand.random() - h/2.0) * 0.5
						self.map[right] += (h*PWRand.random() - h/2.0) * 0.5
						#now do down
						mid = self.GetIndex(x,y + int(currentGrain/2))
						top = self.GetIndex(x,y)
						bottom = self.GetIndex(x,y + int(currentGrain))
						self.map[mid] = ((self.map[top] + self.map[bottom])/2 \
							+ (h*PWRand.random() - h/2.0)) * 0.5
						#add heuristic to bottom, top is handled above (top = left)
						self.map[bottom] += (h*PWRand.random() - h/2.0) * 0.5
				else:
					for x in range(0,self.mapWidth,int(currentGrain)):
						#h is scalar for random displacement
						h = currentGrain/float(self.chunkSize)
						#calculate points one currentGrain/2 to the right and
						#also one down
						mid = self.GetIndex(x + int(currentGrain/2),y)
						left = self.GetIndex(x,y)
						right = self.GetIndex(x + int(currentGrain),y)
						self.map[mid] = (self.map[left] + self.map[right])/2 \
							+ (h*PWRand.random() - h/2.0)
						#also add heuristic to left and right to help reduce artifacts
						self.map[left] += (h*PWRand.random() - h/2.0)
						self.map[right] += (h*PWRand.random() - h/2.0)
						#now do down
						mid = self.GetIndex(x,y + int(currentGrain/2))
						top = self.GetIndex(x,y)
						bottom = self.GetIndex(x,y + int(currentGrain))
						self.map[mid] = (self.map[top] + self.map[bottom])/2 \
							+ (h*PWRand.random() - h/2.0)
						#add heuristic to bottom, top is handled above (top = left)
						self.map[bottom] += (h*PWRand.random() - h/2.0)

			for y in range(0,self.mapHeight,int(currentGrain)):
				if y < botGap * 1.24 or y > topGap * 0.75:
					for x in range(0,self.mapWidth,int(currentGrain)):
						#h is scalar for random displacement
						h = (currentGrain/2)/float(self.chunkSize)
						#calculate points one currentGrain/2 to the right and
						#also one down
						mid = self.GetIndex(x + int(currentGrain/2),y + int(currentGrain/2))
						top = self.GetIndex(x + int(currentGrain/2),y)
						bottom = self.GetIndex(x + int(currentGrain/2),y + int(currentGrain))
						left = self.GetIndex(x,y + int(currentGrain/2))
						right = self.GetIndex(x+int(currentGrain),y + int(currentGrain/2))
						self.map[mid] = ((self.map[left] + self.map[right] \
							+ self.map[top] + self.map[bottom])/4 \
							+ (h*PWRand.random() - h/2.0)) * 0.5
				else:
					for x in range(0,self.mapWidth,int(currentGrain)):
						#h is scalar for random displacement
						h = (currentGrain/2)/float(self.chunkSize)
						#calculate points one currentGrain/2 to the right and
						#also one down
						mid = self.GetIndex(x + int(currentGrain/2),y + int(currentGrain/2))
						top = self.GetIndex(x + int(currentGrain/2),y)
						bottom = self.GetIndex(x + int(currentGrain/2),y + int(currentGrain))
						left = self.GetIndex(x,y + int(currentGrain/2))
						right = self.GetIndex(x+int(currentGrain),y + int(currentGrain/2))
						self.map[mid] = (self.map[left] + self.map[right] \
							+ self.map[top] + self.map[bottom])/4 \
							+ (h*PWRand.random() - h/2.0)
			currentGrain = currentGrain/2

		if LandmassStyle == 1: #Pangaea
			self.terraMap = array('d')
			plotgen = PangaeaMultilayeredFractal()
			self.terraMap = plotgen.generatePlotsByRegion()

			for i in range(self.mapWidth * self.mapHeight):
				if self.terraMap[i] == PlotTypes.PLOT_OCEAN:
					self.map[i] -= 0.50
				else:
					self.map[i] += 0.50
		elif LandmassStyle == 2: #Terra
			self.terraMap = array('d')
			plotgen = TerraMultilayeredFractal()
			self.terraMap = plotgen.generatePlotsByRegion()

			for i in range(self.mapWidth * self.mapHeight):
				if self.terraMap[i] == PlotTypes.PLOT_OCEAN:
					self.map[i] -= 0.50
				else:
					self.map[i] += 0.50
		elif LandmassStyle == 3: #Earth2 Continents, very strong influence for this one
			self.terraMap = array('d')
			plotgen = EarthMultilayeredFractal()
			self.terraMap = plotgen.generatePlotsByRegion()

			for i in range(self.mapWidth * self.mapHeight):
				if self.terraMap[i] == PlotTypes.PLOT_OCEAN:
					self.map[i] -= 1.0
				else:
					self.map[i] += 1.0

		return

	def FinalizeHeightField(self):
		topGap = self.mapHeight - self.chunkSize
		botGap = self.chunkSize
		#Island region(s)
		regionSelect = PWRand.randint(1,6)
		iSouthY = 0
		iNorthY = 0
		iWestX = 0
		iEastX = 0
		if regionSelect == 1:
			#Northwest quadrant
			print "Island region defined - Northwest"
			iSouthY = int(self.mapHeight * 0.5)
			iNorthY = int(topGap) - 1
			iWestX = int(self.mapWidth * 0.1)
			iEastX = int(self.mapWidth * 0.4) - 1
		elif regionSelect == 2:
			#Southwest quadrant
			print "Island region defined - Southwest"
			iSouthY = int(botGap)
			iNorthY = int(self.mapHeight * 0.5) - 1
			iWestX = int(self.mapWidth * 0.1)
			iEastX = int(self.mapWidth * 0.4) - 1
		elif regionSelect == 3:
			#Northeast quadrant
			print "Island region defined - Northeast"
			iSouthY = int(self.mapHeight * 0.5)
			iNorthY = int(topGap) - 1
			iWestX = int(self.mapWidth * 0.6)
			iEastX = int(self.mapWidth * 0.9) - 1
		elif regionSelect == 4:
			#Southeast quadrant
			print "Island region defined - Southeast"
			iSouthY = int(botGap)
			iNorthY = int(self.mapHeight * 0.5) - 1
			iWestX = int(self.mapWidth * 0.6)
			iEastX = int(self.mapWidth * 0.9) - 1
		else:
			print "No island region defined"

		long1 = int((PWRand.random() * 0.25 + 0.5) * self.mapWidth)
		long2 = int(long1 + (self.mapWidth * 0.5))

		if long2 > self.mapWidth:
			long2 = long2 - self.mapWidth

		print "Digging oceanic trench."
		for y in range(0,self.mapHeight,1):
			biasPole = abs((self.mapHeight * 0.5) - y) / (self.mapHeight * 0.5)
			long1 = PWRand.randint(long1 - 1,long1 + 1)
			long2 = PWRand.randint(long2 - 1,long2 + 1)
			if y < self.chunkSize:
				reducePole = 1 - y / self.chunkSize
			elif abs(y - self.mapHeight) < self.chunkSize:
				reducePole = 1 - abs(y - self.mapHeight) / self.chunkSize
			for x in range(0,self.mapWidth,1):
				i = self.GetIndex(x,y)
				trenchWidth = PWRand.randint(max(4,int(self.mapWidth/10)),max(6,int(self.mapWidth/8)))

				#Island region noise
				if y > iSouthY and y < iNorthY and PWRand.random() < 0.75:
					if x > iWestX and x < iEastX:
						self.map[i] *= PWRand.random() * 0.75 + 0.5

				#if PWRand.random() < biasPole:
				#	self.map[i] *= PWRand.random() * 0.5 + 0.5

				if x == long1 or x == long2:
					for xx in range(x - trenchWidth,x + trenchWidth,1):
						#Subtractive noise along simulated oceanic trench
						ii = self.GetIndex(xx,y)
						if abs(xx - x) < 2:
							self.map[ii] = 0
						elif abs(xx - x) < trenchWidth/2:
							self.map[ii] *= PWRand.random() * 0.25 + 0.25
						else:
							self.map[ii] *= PWRand.random() * 0.5 + 0.5
				self.map[i] *= reducePole
		return
	def GetIndex(self,x,y):
		#wrap in both directions for the purpose of generating heights. This
		#allows us to continue using even numbered sizes for maps, and should
		#not impact the desired result.
		dy = y; dx = x
		if y >= self.mapHeight:
			dy = y - self.mapHeight
		elif y < 0:
			dy = self.mapHeight + y
		if x >= self.mapWidth:
			dx = x - self.mapWidth
		elif x < 0:
			dx = self.mapWidth + x

		index = dy*self.mapWidth + dx
		return index

	def PerformTectonics(self):
		self.plateMap = array('i')
		doneMap = array('b') #use a little extra memory to speed this operation up
		#initialize map with zeros
		for i in range(0,self.mapHeight*self.mapWidth):
			self.plateMap.append(0)
			doneMap.append(False)

		#seed plates
		randomRegions = False
		randomPlates = PWRand.randint(2,4)
		numberOfPlates = self.mapWidth / randomPlates

		if randomPlates < 4:
			randomRegions = True

		topGap = self.mapHeight - self.chunkSize
		botGap = self.chunkSize

		#The regions work by defining their portion of the map and removing plates
		#in those regions.

		#Default plate placement
		print "Seeding plates..."
		for x in range(1,numberOfPlates + 1):
			i = PWRand.randint(0,self.mapWidth*self.mapHeight - 1)
			if self.plateMap[i] == 0:
				self.plateMap[i] = x

		if randomRegions == True:
			print "Generating plate regions (random regions)"
			regionSelect = PWRand.randint(1,2)
			if regionSelect == 1:
				#Western Region
				print "Western region selected."
				iSouthY = int(botGap)
				iNorthY = int(topGap) - 1
				iHeight = iNorthY - iSouthY + 1
				iWestX = int(self.mapWidth * 0.1)
				iEastX = int(self.mapWidth * 0.4) - 1
				iWidth = iEastX - iWestX + 1
			else:
				#Eastern Region
				print "Eastern region selected."
				iSouthY = int(botGap)
				iNorthY = int(topGap) - 1
				iHeight = iNorthY - iSouthY + 1
				iWestX = int(self.mapWidth * 0.6)
				iEastX = int(self.mapWidth * 0.9) - 1
				iWidth = iEastX - iWestX + 1

			for y in range(iSouthY,iNorthY):
				for x in range(iWestX,iEastX):
					i = self.GetIndex(x,y)
					if self.plateMap[i] != 0 and PWRand.random() < 0.25:
						self.plateMap[i] = 0

		#grow plates
		doneMapFull = False
		while not doneMapFull:
			doneMapFull = True
			for y in range(0,self.mapHeight):
				for x in range(0,self.mapWidth):
					i = self.GetIndex(x,y)
					if self.plateMap[i] != 0 and doneMap[i] == False:
						up = self.GetIndex(x,y-1)
						down = self.GetIndex(x,y+1)
						left = self.GetIndex(x-1,y)
						right = self.GetIndex(x+1,y)
						done = True
						if self.plateMap[up] == 0:
							done = False
							if PWRand.random() < .5:
								self.plateMap[up] = self.plateMap[i]
						if self.plateMap[down] == 0:
							done = False
							if PWRand.random() < .5:
								self.plateMap[down] = self.plateMap[i]
						if self.plateMap[left] == 0:
							done = False
							if PWRand.random() < .5:
								self.plateMap[left] = self.plateMap[i]
						if self.plateMap[right] == 0:
							done = False
							if PWRand.random() < .5:
								self.plateMap[right] = self.plateMap[i]
						if done == True:
							doneMap[i] = True
						else:
							doneMapFull = False

		#Stagger the plates somewhat to add interest
		for iPlate in range(1,numberOfPlates + 1):
			polarPlate = False

			#Bottom of map
			for y in range (0,botGap):
				for x in range(0,self.mapWidth):
					i = self.GetIndex(x,y)
					if self.plateMap[i] == iPlate:
						polarPlate = True

			#Top of map
			for y in range (0,topGap):
				for x in range(0,self.mapWidth):
					i = self.GetIndex(x,y)
					if self.plateMap[i] == iPlate:
						polarPlate = True

			global LandmassStyle
			adjPlate = PWRand.random() * 0.5 + 0.5

			if LandmassStyle == 3: #Earth continents
				adjPlate = 0.1
			elif LandmassStyle != 0: #Pangaea or Terra
				adjPlate = PWRand.random() * 0.5

			for i in range(0,self.mapHeight*self.mapWidth):
				if self.plateMap[i] == iPlate:
					if polarPlate == False and PWRand.random() < 0.5:
						self.map[i] += adjPlate
					else:
						self.map[i] -= adjPlate

		#Now that the plates are grown, we must decide how the plate boundaries
		#will effect our map. Basically, plots away from plate boundaries will
		#cause some sinking, while plots closer to them while rise. After
		#multiplying, the main heightmap will be re-normalized.
		sinkValue = 1.0
		peakValue = 1.5
		searchRadius = 6

		for y in range(self.mapHeight):
			for x in range(self.mapWidth):
				distance = self.FindDistanceToPlateBoundary(x,y,searchRadius)
				i = self.GetIndex(x,y)
				if distance != 0.0:
					self.map[i] *= self.GetInfluFromDistance(sinkValue,peakValue,searchRadius,distance)

		return
	def GetInfluFromDistance(self,sinkValue,peakValue,searchRadius,distance):
		influence = peakValue
		maxDistance = math.sqrt(pow(float(searchRadius),2) + pow(float(searchRadius),2))
		#minDistance = 1.0
		influence -= ((peakValue - sinkValue)* (distance - 1.0))/(maxDistance - 1.0)
		return influence
	def FindDistanceToPlateBoundary(self,x,y,searchRadius):
		minDistance = 10.0
		i = self.GetIndex(x,y)
		for yy in range(y - searchRadius,y + searchRadius):
			for xx in range(x - searchRadius,x + searchRadius):
				ii = self.GetIndex(xx,yy)
				if self.plateMap[i] != self.plateMap[ii]:
					distance = math.sqrt(pow(float(xx-x),2) + pow(float(yy-y),2))
					if distance < minDistance:
						minDistance = distance

		if minDistance == 10.0:
			return 0.0

		return minDistance
	def FindSeaLevel(self):
		flatCount = 0
		seaTolerance = 0.025
		totalCount = self.mapHeight*self.mapWidth
		seaInTolerance = False
		seaLevelChange = 0.2

		print "Finding sea level."

		iterations = 0
		while not seaInTolerance:
			iterations += 1
			flatCount = 0

			for i in range(self.mapHeight*self.mapWidth):
				if self.map[i] > self.seaLevel:
					flatCount += 1

			currentLandPercent = float(flatCount)/float(totalCount)
			#First pass is handled differently to minimize the number of passes needed
			if iterations == 1:
				differencePercent = float(abs(currentLandPercent - self.landPercent))
				seaLevelChange = differencePercent
				differencePercentOut = int(differencePercent * 100)
				print "First pass difference is %(d)2d." % {"d":differencePercentOut}
			else:
				if currentLandPercent < self.landPercent + seaTolerance and \
				   currentLandPercent > self.landPercent - seaTolerance:
					seaInTolerance = True
				elif currentLandPercent < self.landPercent:
					self.seaLevel -= seaLevelChange
					currentLandPercentOut = int(currentLandPercent * 100)
					print "Pass %(p)2d, lowering sea level. Current land is %(c)2d percent." % {"p":iterations,"c":currentLandPercentOut}
				elif currentLandPercent > self.landPercent:
					self.seaLevel += seaLevelChange
					currentLandPercentOut = int(currentLandPercent * 100)
					print "Pass %(p)2d, raising sea level. Current land is %(c)2d percent." % {"p":iterations,"c":currentLandPercentOut}

				if iterations > 19 or seaLevelChange < 0.01: #Close enough
					seaInTolerance = True

				seaLevelChange *= 0.66

		finalLandPercent = int(currentLandPercent * 100)
		print "Sea level defined in %(p)2d attempts, flat count is %(f)2d or %(l)2d percent." % {"p":iterations,"f":flatCount,"l":finalLandPercent}
		return
	def NormalizeHeightMap(self):
		#find highest and lowest points
		maxAlt = 0.0
		minAlt = 0.0
		for y in range(self.mapHeight):
			for x in range(self.mapWidth):
				plot = self.map[self.GetIndex(x,y)]
				if plot > maxAlt:
					maxAlt = plot
				if plot < minAlt:
					minAlt = plot
		#normalize map so that all altitudes are between 1 and 0
		#first add minAlt to all values if necessary
		if minAlt < 0.0:
			for y in range(self.mapHeight):
				for x in range(self.mapWidth):
					self.map[self.GetIndex(x,y)] -= minAlt
		#add minAlt to maxAlt also before scaling entire map
		maxAlt -= minAlt
		scaler = 1.0/maxAlt
		for y in range(self.mapHeight):
			for x in range(self.mapWidth):
				self.map[self.GetIndex(x,y)] = self.map[self.GetIndex(x,y)] * scaler
		return
	def GenerateFinalPlotTypes(self):
		self.OCEAN = 0
		self.LAND = 1
		self.HILLS = 2
		self.PEAK = 3

		diffMap = array('d')
		self.plotMap = array('i')

		for i in range(0,self.mapHeight*self.mapWidth):
			diffMap.append(0.0)
			self.plotMap.append(self.OCEAN)

		for y in range(self.mapHeight):
			for x in range(self.mapWidth):
				i = self.GetIndex(x,y)
				myAlt = self.map[i]
				if myAlt > self.seaLevel:
					minAlt = self.GetLowestNeighbor(x,y)
					diffMap[i] = myAlt - minAlt

		peakHeight = FindValueFromPercent(diffMap,PeakPercent,0.001,True)
		hillHeight = FindValueFromPercent(diffMap,HillPercent,0.001,True)

		for y in range(self.mapHeight):
			for x in range(self.mapWidth):
				i = self.GetIndex(x,y)
				myAlt = self.map[i]
				if myAlt > self.seaLevel:
					minAlt = self.GetLowestNeighbor(x,y)
					if minAlt < myAlt-peakHeight:
						self.plotMap[i] = self.PEAK
					elif minAlt < myAlt-hillHeight:
						self.plotMap[i] = self.HILLS
					else:
						self.plotMap[i] = self.LAND

		return
	def GetLowestNeighbor(self,x,y):
		minAlt = 1.0
		for yy in range(y-1,y+1):
			for xx in range(x-1,x+1):
				i = self.GetIndex(xx,yy)
				myAlt = self.map[i]
				if myAlt < self.seaLevel: #seaLevel is minimum consideration
					myAlt = self.seaLevel
				if(minAlt > myAlt):
					minAlt = myAlt
		return minAlt
	def FillInLakes(self):
		#smaller lakes need to be filled in for now. The river system will
		#most likely recreate them later due to drainage calculation
		#according to certain rules. This makes the lakes look much better
		#and more sensible.
		am = Areamap(self.mapWidth,self.mapHeight)
		am.defineAreas(False)
##        am.PrintAreaMap()
		oceanID = am.getOceanID()
		for i in range(self.mapHeight*self.mapWidth):
			if self.plotMap[i] == self.OCEAN and am.areaMap[i] != oceanID:
				#check the size of this body of water, if too small,
				#change to land
				for a in am.areaList:
					if a.ID == am.areaMap[i] and a.size < 35:
						self.plotMap[i] = self.LAND

		return
	def ShiftLandToMiddle(self):
		bestColumn = -1
		bestColumnSeaTileCount = -1
		for x in range(self.mapWidth):
			columnSeaTileCount = 0
			for y in range(self.mapHeight):
				i = self.GetIndex(x,y)
				if self.plotMap[i] == self.OCEAN:
					columnSeaTileCount += 1
			if columnSeaTileCount > bestColumnSeaTileCount:
				bestColumnSeaTileCount = columnSeaTileCount
				bestColumn = x
			elif (columnSeaTileCount == bestColumnSeaTileCount) and (PWRand.random() < .25):
				bestColumn = x
		if bestColumn <= 0:
			return
		newPlotMap = array('i')
		#initialize map with 0CEAN
		for i in range(0,self.mapHeight*self.mapWidth):
			newPlotMap.append(self.OCEAN)
		newMap = array('d')
		#initialize map with zeros
		for i in range(0,self.mapHeight*self.mapWidth):
			newMap.append(0.0)
		for x in range(self.mapWidth):
			xprime = x - bestColumn - 1
			for y in range(self.mapHeight):
				i = self.GetIndex(x,y)
				iprime = self.GetIndex(xprime,y)
				newPlotMap[iprime] = self.plotMap[i]
				newMap[iprime] = self.map[i]
		self.map = newMap
		self.plotMap = newPlotMap

		return
#global access
print "Initializing heightmap..."
hm = Heightmap()

##This function is a general purpose value tuner. It finds a value that will be greater
##than or less than the desired percent of a whole map within a given tolerance. Map values
##should be between 0 and 1.
def FindValueFromPercent(mmap,percent,tolerance,greaterThan):
	inTolerance = False
	#to speed things up a little, lets take some time to find the middle value
	#in the dataset and use that to begin our search
	minV = 100.0
	maxV = 0.0
	for i in range(hm.mapHeight*hm.mapWidth):
		if minV > mmap[i]:
			minV = mmap[i]
		if maxV < mmap[i]:
			maxV = mmap[i]
	mid = (maxV - minV)/2.0 + minV
##    print "starting threshold = "
##    print mid
##    print "desired percent",percent
	threshold = mid
	thresholdChange = mid
	iterations = 0
	while not inTolerance:
		iterations += 1
		if(iterations > 500):
			print "can't find value within tolerance, end value = "
			print threshold, thresholdChange
			break #close enough
		matchCount = 0
##        print "threshold",threshold
		for i in range(hm.mapHeight*hm.mapWidth):
			if mmap[i] != 0.0:
				if greaterThan == True:
					if(mmap[i] > threshold):
						matchCount += 1
				else:
					if(mmap[i] < threshold):
						matchCount += 1
##        print "matchCount",matchCount
		currentPercent = float(matchCount)/float(hm.mapHeight*hm.mapWidth)
##        print "currentPercent",currentPercent
		if currentPercent < percent + tolerance and \
		   currentPercent > percent - tolerance:
			inTolerance = True
		elif greaterThan == True:
			if currentPercent < percent - tolerance:
				threshold -= thresholdChange
			else:
				threshold += thresholdChange
		else:
			if currentPercent > percent + tolerance:
				threshold -= thresholdChange
			else:
				threshold += thresholdChange
		if not inTolerance:
#                print "sealevel = " + str(self.seaLevel), "change = " + str(seaLevelChange)
			thresholdChange = thresholdChange/2.0

		#at this point value should be in tolerance or close to it
	return threshold
##############################################################################
## Custom Area tracker allows for considering coast as land to find
## 'New World' type continents that cannot be reached with galleys.
## Also used for filling in lakes since the required code is similar
##############################################################################
class Areamap :
	def __init__(self,width,height):
		self.mapWidth = width
		self.mapHeight = height
		self.areaMap = array('i')
		#initialize map with zeros
		for i in range(0,self.mapHeight*self.mapWidth):
			self.areaMap.append(0)
		return
	def defineAreas(self,coastIsLand):
		#coastIsLand = True means that we are trying to find continents that
		#are not connected by coasts to the main landmasses, allowing us to
		#find continents suitable as a 'New World'. Otherwise, we
		#are just looking to fill in lakes and coast needs to be considered
		#water in that case
#        self.areaSizes = array('i')
		self.areaList = list()
		areaID = 0
		#make sure map is erased in case it is used multiple times
		for i in range(0,self.mapHeight*self.mapWidth):
			self.areaMap[i] = 0
#        for i in range(0,1):
		for i in range(0,self.mapHeight*self.mapWidth):
			if self.areaMap[i] == 0: #not assigned to an area yet
				areaID += 1
				areaSize,isWater = self.fillArea(i,areaID,coastIsLand)
				area = Area(areaID,areaSize,isWater)
				self.areaList.append(area)
		return

	def isWater(self,x,y,coastIsLand):
		#coastIsLand = True means that we are trying to find continents that
		#are not connected by coasts to the main landmasses, allowing us to
		#find continents suitable as a 'New World'. Otherwise, we
		#are just looking to fill in lakes and coast needs to be considered
		#water in that case
		ii = self.getIndex(x,y)
		if coastIsLand:
			if hm.plotMap[ii] == hm.OCEAN and terr.terrainMap[ii] != terr.COAST:
				return True
			else:
				return False
		else:
			if hm.plotMap[ii] == hm.OCEAN:
				return True
			else:
				return False

		return False
	def getAreaByID(self,areaID):
		for i in range(len(self.areaList)):
			if self.areaList[i].ID == areaID:
				return self.areaList[i]
		return None
	def getOceanID(self):
#        self.areaList.sort(key=operator.attrgetter('size'),reverse=True)
		self.areaList.sort(lambda x,y:cmp(x.size,y.size))
		self.areaList.reverse()
		for a in self.areaList:
			if a.water == True:
				return a.ID

	def getNewWorldID(self):
		# Map options
		global AllowNewWorld

		nID = 0
		continentList = list()
		for a in self.areaList:
			if a.water == False:
				continentList.append(a)

		totalLand = 0
		for c in continentList:
			totalLand += c.size

		print totalLand

		#sort all the continents by size, largest first
		continentList.sort(lambda x,y:cmp(x.size,y.size))
		continentList.reverse()

		print ''
		print "All continents"
		print self.PrintList(continentList)

		#now remove a percentage of the landmass to be considered 'Old World'
		oldWorldSize = 0
		#biggest continent is automatically 'Old World'
		oldWorldSize += continentList[0].size
		del continentList[0]

		if AllowNewWorld == 1:
		#get the next largest continent and temporarily remove from list
		#add it back later and is automatically 'New World'
		#If new worlds aren't allowed we still use the new world code
		#to ensure starting areas are on the larger landmasses.
			biggestNewWorld = continentList[0]
			del continentList[0]

			#sort list by ID rather than size
			continentList.sort(lambda x,y:cmp(x.ID,y.ID))
			continentList.reverse()

			oldWorldMinSize = max(12,totalLand * 0.025)
			idealOldWorldSize = int(totalLand * 0.60)
		else:
			oldWorldMinSize = max(12,totalLand * 0.03)
			idealOldWorldSize = int(totalLand * 0.75)

		oldWorldPercent = float(oldWorldSize)/float(totalLand)
		print "Ideal Old World total size is %(w)2d" % {"w":idealOldWorldSize}
		print "Current Old World size is %(w)2d" % {"w":oldWorldSize}

		while oldWorldSize < idealOldWorldSize:
			print "Minimum Old World continent size is %(w)2d" % {"w":oldWorldMinSize}
			print "There are %(c)2d continents in the list." % {"c":len(continentList)}
			o = len(continentList) - 1
			n = 0
			while n < o:
				if oldWorldPercent > 0.60:
					print "Old world continent selection complete, total size is %(t)2d" % {"t":oldWorldSize}
					break
				print "Evaluating continent ID %(d)2d, size is %(s)2d." % {"d":continentList[n].ID,"s":continentList[n].size}
				if continentList[n].size > oldWorldMinSize:
					oldWorldSize += continentList[n].size
					oldWorldPercent = float(oldWorldSize)/float(totalLand)
					print "Adding continent %(c)2d to old world, total size is now %(t)2d" % {"c":continentList[n].ID,"t":oldWorldSize}
					del continentList[n]
					#decrease the list length (o) by one to prevent going out of bounds
					#do not incriment the current count (n) here since that list item was removed
					o -= 1
				else: #incriment the counter if no continent was removed from the list
					n += 1
			oldWorldMinSize *= 0.75
			idealOldWorldSize *= 0.75

		if AllowNewWorld == 1:
		#add back the biggestNewWorld continent if it was removed earlier
			continentList.append(biggestNewWorld)

		#what remains in the list will be considered 'New World'
		print "New World Continents"
		print self.PrintList(continentList)

		#get ID for the next continent, we will use this ID for 'New World'
		#designation
		nID = continentList[0].ID
		del continentList[0] #delete to avoid unnecessary overwrite

		#now change all the remaining continents to also have nID as their ID
		for i in range(self.mapHeight*self.mapWidth):
			for c in continentList:
				if c.ID == self.areaMap[i]:
					self.areaMap[i] = nID

		return nID

	def getIndex(self,x,y):
		#wrap in X direction only.
		dx = x
		if y >= self.mapHeight or y < 0:
			return -1

		if x >= self.mapWidth:
			dx = x % self.mapWidth
		elif x < 0:
			dx = x % self.mapWidth

		index = y*self.mapWidth + dx
		return index

	def fillArea(self,index,areaID,coastIsLand):
		#first divide index into x and y
		y = index/self.mapWidth
		x = index%self.mapWidth
		#We check 8 neigbors for land,but 4 for water. This is because
		#the game connects land squares diagonally across water, but
		#water squares are not passable diagonally across land
		self.segStack = list()
		self.size = 0
		isAreaWater = self.isWater(x,y,coastIsLand)
		#place seed on stack for both directions
		seg = LineSegment(y,x,x,1)
		self.segStack.append(seg)
		seg = LineSegment(y+1,x,x,-1)
		self.segStack.append(seg)
		while(len(self.segStack) > 0):
			seg = self.segStack.pop()
			self.scanAndFillLine(seg,areaID,isAreaWater,coastIsLand)
##            if (seg.y < 8 and seg.y > 4) or (seg.y < 70 and seg.y > 64):
##            if (areaID == 4
##                PrintPlotMap(hm)
##                self.PrintAreaMap()

		return self.size,self.isWater(x,y,coastIsLand)
	def scanAndFillLine(self,seg,areaID,isWater,coastIsLand):
		#check for y + dy being off map
		i = self.getIndex(seg.xLeft,seg.y + seg.dy)
		if i < 0:
##            print "scanLine off map ignoring",str(seg)
			return
		debugReport = False
##        if (seg.y < 8 and seg.y > 4) or (seg.y < 70 and seg.y > 64):
##        if (areaID == 4):
##            debugReport = True
		#for land tiles we must look one past the x extents to include
		#8-connected neighbors
		if isWater:
			landOffset = 0
		else:
			landOffset = 1
		lineFound = False
		#first scan and fill any left overhang
		if debugReport:
			print ""
			print "areaID = %(a)4d" % {"a":areaID}
			print "isWater = %(w)2d, landOffset = %(l)2d" % {"w":isWater,"l":landOffset}
			print str(seg)
			print "Going left"
		for xLeftExtreme in range(seg.xLeft - landOffset,0 - (self.mapWidth*20),-1):
			i = self.getIndex(xLeftExtreme,seg.y + seg.dy)
			if debugReport:
				print "xLeftExtreme = %(xl)4d" % {'xl':xLeftExtreme}
			if self.areaMap[i] == 0 and isWater == self.isWater(xLeftExtreme,seg.y + seg.dy,coastIsLand):
				self.areaMap[i] = areaID
				self.size += 1
				lineFound = True
			else:
				#if no line was found, then xLeftExtreme is fine, but if
				#a line was found going left, then we need to increment
				#xLeftExtreme to represent the inclusive end of the line
				if lineFound:
					xLeftExtreme += 1
				break
		if debugReport:
			print "xLeftExtreme finally = %(xl)4d" % {'xl':xLeftExtreme}
			print "Going Right"
		#now scan right to find extreme right, place each found segment on stack
#        xRightExtreme = seg.xLeft - landOffset #needed sometimes? one time it was not initialized before use.
		xRightExtreme = seg.xLeft #needed sometimes? one time it was not initialized before use.
		for xRightExtreme in range(seg.xLeft,self.mapWidth*20,1):
			if debugReport:
				print "xRightExtreme = %(xr)4d" % {'xr':xRightExtreme}
			i = self.getIndex(xRightExtreme,seg.y + seg.dy)
			if self.areaMap[i] == 0 and isWater == self.isWater(xRightExtreme,seg.y + seg.dy,coastIsLand):
				self.areaMap[i] = areaID
				self.size += 1
				if lineFound == False:
					lineFound = True
					xLeftExtreme = xRightExtreme #starting new line
					if debugReport:
						print "starting new line at xLeftExtreme= %(xl)4d" % {'xl':xLeftExtreme}
			elif lineFound == True: #found the right end of a line segment!
				lineFound = False
				#put same direction on stack
				newSeg = LineSegment(seg.y + seg.dy,xLeftExtreme,xRightExtreme - 1,seg.dy)
				self.segStack.append(newSeg)
				if debugReport:
					print "same direction to stack",str(newSeg)
				#determine if we must put reverse direction on stack
				if xLeftExtreme < seg.xLeft or xRightExtreme >= seg.xRight:
					#out of shadow so put reverse direction on stack also
					newSeg = LineSegment(seg.y + seg.dy,xLeftExtreme,xRightExtreme - 1,-seg.dy)
					self.segStack.append(newSeg)
					if debugReport:
						print "opposite direction to stack",str(newSeg)
				if xRightExtreme >= seg.xRight + landOffset:
					if debugReport:
						print "finished with line"
					break; #past the end of the parent line and this line ends
			elif lineFound == False and xRightExtreme >= seg.xRight + landOffset:
				if debugReport:
					print "no additional lines found"
				break; #past the end of the parent line and no line found
			else:
				continue #keep looking for more line segments
		if lineFound == True: #still a line needing to be put on stack
			if debugReport:
				print "still needing to stack some segs"
			lineFound = False
			#put same direction on stack
			newSeg = LineSegment(seg.y + seg.dy,xLeftExtreme,xRightExtreme - 1,seg.dy)
			self.segStack.append(newSeg)
			if debugReport:
				print str(newSeg)
			#determine if we must put reverse direction on stack
			if xLeftExtreme < seg.xLeft or xRightExtreme - 1 > seg.xRight:
				#out of shadow so put reverse direction on stack also
				newSeg = LineSegment(seg.y + seg.dy,xLeftExtreme,xRightExtreme - 1,-seg.dy)
				self.segStack.append(newSeg)
				if debugReport:
					print str(newSeg)

		return
	#for debugging
	def PrintAreaMap(self):

		print "Area Map"
		for y in range(hm.mapHeight - 1,-1,-1):
			lineString = ""
			for x in range(hm.mapWidth):
				mapLoc = self.areaMap[hm.GetIndex(x,y)]
				if mapLoc + 34 > 127:
					mapLoc = 127 - 34
				lineString += chr(mapLoc + 34)
			lineString += "-" + str(y)
			print lineString
		oid = self.getOceanID()
		if oid == None or oid + 34 > 255:
			print "Ocean ID is unknown"
		else:
			print "Ocean ID is %(oid)4d or %(c)s" % {'oid':oid,'c':chr(oid + 34)}
		lineString = " "
		print lineString

		return
	def PrintList(self,s):
		for a in s:
			char = chr(a.ID + 34)
			lineString = str(a) + ' ' + char
			print lineString

class LineSegment :
	def __init__(self,y,xLeft,xRight,dy):
		self.y = y
		self.xLeft = xLeft
		self.xRight = xRight
		self.dy = dy
	def __str__ (self):
		string = "y = %(y)3d, xLeft = %(xl)3d, xRight = %(xr)3d, dy = %(dy)2d" % \
		{'y':self.y,'xl':self.xLeft,'xr':self.xRight,'dy':self.dy}
		return string

class Area :
	def __init__(self,iD,size,water):
		self.ID = iD
		self.size = size
		self.water = water

	def __str__(self):
		string = "{ID = %(i)4d, size = %(s)4d, water = %(w)1d}" % \
		{'i':self.ID,'s':self.size,'w':self.water}
		return string

##########################################################################
##Global Normalized map function
def NormalizeMap(Width,Height,fMap):
	#find highest and lowest points
	maxAlt = 0.0
	minAlt = 0.0
	for y in range(Height):
		for x in range(Width):
			plot = fMap[hm.GetIndex(x,y)]
			if plot > maxAlt:
				maxAlt = plot
			if plot < minAlt:
				minAlt = plot
	#normalize map so that all altitudes are between 1 and 0
	#first add minAlt to all values if necessary
	if minAlt < 0.0:
		for y in range(Height):
			for x in range(Width):
				fMap[hm.GetIndex(x,y)] -= minAlt
	#add minAlt to maxAlt also before scaling entire map
	maxAlt -= minAlt
	scaler = 1.0/maxAlt
	for y in range(Height):
		for x in range(Width):
			fMap[hm.GetIndex(x,y)] = fMap[hm.GetIndex(x,y)] * scaler
	return

##############################################################################
## Terrain generation
##############################################################################
'''
This class is used by the TemperatureMap and RainfallMap classes to govern the
general behavior of winds. Prevailing winds are the most important factor in
distributing temperature and rainfall, but they tend to fluctuate a great deal
depending on the time of year and the surrounding geology. Therefore, prevailing
surface winds are simulated in this script by a right-triangle shaped area
with it's base as the source of distribution. It can be thought of as a
'shotgun blast' eminating from the source map square.
^     7
|    /
|   /
|  /
| /
S----------->
The influence of the source square decreases as the distance from it increases
'''
class WindZones :
	def __init__(self,mapHeight,topLat,botLat):
		self.NPOLAR = 0
		self.NTEMPERATE = 1
		self.NEQUATOR = 2
		self.SEQUATOR = 3
		self.STEMPERATE = 4
		self.SPOLAR = 5
		self.NOZONE = 99
		self.mapHeight = mapHeight
		self.topLat = topLat
		self.botLat = botLat
	def GetZone(self,y):
		if y < 0 or y >= self.mapHeight:
			return self.NOZONE
		lat = self.GetLatitude(y)
		if lat > 60:
			return self.NPOLAR
		elif lat > 25:
			return self.NTEMPERATE
		elif lat > 0:
			return self.NEQUATOR
		elif lat > -25:
			return self.SEQUATOR
		elif lat > -60:
			return self.STEMPERATE
		else:
			return self.SPOLAR
		return
	def GetZoneName(self,zone):
		if zone == self.NPOLAR:
			return "NPOLAR"
		elif zone == self.NTEMPERATE:
			return "NTEMPERATE"
		elif zone == self.NEQUATOR:
			return "NEQUATOR"
		elif zone == self.SEQUATOR:
			return "SEQUATOR"
		elif zone == self.STEMPERATE:
			return "STEMPERATE"
		else:
			return "SPOLAR"
		return
	def GetYFromZone(self,zone,bTop):
		if bTop:
			for y in range(hm.mapHeight - 1,-1,-1):
				if zone == self.GetZone(y):
					return y
		else:
			for y in range(hm.mapHeight):
				if zone == self.GetZone(y):
					return y
		return
	def GetBlastBoxHeight(self,y):
		zone = self.GetZone(y)
		we,ns = self.GetWindDirections(y)
		if ns < 0:
			maxY = self.GetYFromZone(zone,True) - 1#zone *before* this is max
		else:
			maxY = self.GetYFromZone(zone,False) + 1 #zone *after* this is max
		height = maxY - y
		return height
	def GetZoneSize(self):
		latitudeRange = self.topLat - self.botLat
		degreesPerDY = float(latitudeRange)/float(self.mapHeight)
		size = 30.0/degreesPerDY
		return size
	def GetLatitude(self,y):
		latitudeRange = self.topLat - self.botLat
		degreesPerDY = float(latitudeRange)/float(self.mapHeight)
		latitude = (self.topLat - (int(round(float(y)* degreesPerDY)))) * -1
		return latitude
	def GetWindDirections(self,y):
		z = self.GetZone(y)
		#get x,y directions
		return self.GetWindDirectionsInZone(z)
	def GetWindDirectionsInZone(self,z):
		#get x,y directions
		if z == self.NPOLAR:
			return (-1,-1)
		elif z == self.NTEMPERATE:
			return (1,1)
		elif z == self.NEQUATOR:
			return (-1,-1)
		elif z == self.SEQUATOR:
			return (-1,1)
		elif z == self.STEMPERATE:
			return (1,-1)
		elif z == self.SPOLAR:
			return (-1,1)
		return (0,0)
'''
This class generates and provides access to a temperature map of the world.
Temperatures range from 0.0(coldest) to 1.0(hottest)
'''
class TemperatureMap :
	def __init__(self):
		#To provide global access without allocating alot of resources for
		#nothing, object initializer must be empty
		return
	def GenerateTempMap(self,topLatitude,bottomLatitude):
		self.tempMap = array('d')
		self.sunMap = array('d')
		#initialize map with zeros
		for i in range(0,hm.mapHeight*hm.mapWidth):
			self.tempMap.append(0.0)
			self.sunMap.append(0.0)

		self.wz = WindZones(hm.mapHeight,topLatitude,bottomLatitude)
		for y in range(hm.mapHeight):
			for x in range(hm.mapWidth):
				self.SetInitialTemperature(x,y)
		NormalizeMap(hm.mapWidth,hm.mapHeight,self.sunMap)
		NormalizeMap(hm.mapWidth,hm.mapHeight,self.tempMap)
		for z in range(self.wz.NPOLAR,self.wz.SPOLAR + 1):
			we,ns = self.wz.GetWindDirectionsInZone(z)
			if ns < 0:
				yStart = self.wz.GetYFromZone(z,False)
				yStop = self.wz.GetYFromZone(z,True) + ns
			else:
				yStart = self.wz.GetYFromZone(z,True)
				yStop = self.wz.GetYFromZone(z,False) + ns
			for y in range(yStart,yStop,ns):
				#flip we for the x loop so that it goes opposite of wind
				#direction
				if we > 0:
					xStart = hm.mapWidth - 1
					xStop = -1
				else:
					xStart = 0
					xStop = hm.mapWidth
				for x in range(xStart,xStop,(-1)*we):
					self.DistributeTemp(x,y,we,ns)

		return
	def SetInitialTemperature(self,x,y):
		#this procedure decides how much heat is present at a location
		#before any heat transfer process takes place.
		oceanPoleTemp = 0.0
		lat = abs(float(self.wz.GetLatitude(y)))

		i = hm.GetIndex(x,y)
		if hm.plotMap[i] == hm.OCEAN:
			oceanTempRange = 1.0 - oceanPoleTemp
			latRange = 90.0 #always
			tempPerLatChange = oceanTempRange/latRange
			temp = (latRange - lat) * tempPerLatChange + oceanPoleTemp
			self.sunMap[i] = temp
			self.tempMap[i] = temp
		else:
			latRange = 90.0
			tempPerLatChange = 1.0/latRange
			temp = (latRange - lat) * tempPerLatChange
			self.sunMap[i] = temp
			self.tempMap[i] = temp

		return
	def DistributeTemp(self,x,y,we,ns):
		bDebug = False
		xStop = x + (hm.mapWidth/6 * we)
		i = hm.GetIndex(x,y)
		xWeight = 0.40
		yWeight = 0.35
		if self.wz.GetZone(y) == self.wz.NPOLAR:
			pass
		for xx in range(x,xStop,we):
			i = hm.GetIndex(xx,y)
			xBack = hm.GetIndex(xx - we,y)
			sourceTemp = self.tempMap[xBack]
			if bDebug:
				print xx,y
				print self.tempMap[i], sourceTemp
			self.tempMap[i] = self.tempMap[i] * (1.0 - xWeight) + sourceTemp  * xWeight
			if bDebug:
				print self.tempMap[i]
			if self.wz.GetZone(y) == self.wz.GetZone(y - ns):
				yBack = hm.GetIndex(xx,y - ns)
				sourceTemp = self.tempMap[yBack]
				if bDebug:
					print self.tempMap[i],sourceTemp
				self.tempMap[i] = self.tempMap[i] * (1.0 - yWeight) + sourceTemp * yWeight
				if bDebug:
					print self.tempMap[i]
			xWeight = xWeight*xWeight
			yWeight = yWeight*yWeight

		return
#global access
print "Initializing temperature map..."
tm = TemperatureMap()

class RainfallMap :
	def __init__(self):
		#To provide global access without allocating alot of resources for
		#nothing, object initializer must be empty
		return
	def GenerateRainMap(self,topLatitude,bottomLatitude):
		self.rainMap = array('d')
		smoothMap = array('d')
		self.distHeight = hm.mapWidth/4 #yes, mapWIDTH /4 the dist maps are square
		self.distWidth = hm.mapWidth/4
		rainDumped = array('d')#records rainfall from one square overwriting each time
		rainPotential = array('d')
		#initialize map with zeros or tempMap value if ocean
		for i in range(0,hm.mapHeight*hm.mapWidth):
			smoothMap.append(0.0)
			if hm.plotMap[i] == hm.OCEAN:
				self.rainMap.append(tm.tempMap[i])
			else:
				self.rainMap.append(0.1)
		#initialize local distribution maps
		for i in range(0,self.distHeight * self.distWidth):
			rainDumped.append(0.0)
			rainPotential.append(0.0)
		#step through each windzone
		self.wz = WindZones(hm.mapHeight,topLatitude,bottomLatitude)
		for z in range(self.wz.NPOLAR,self.wz.SPOLAR + 1):
			we,ns = self.wz.GetWindDirectionsInZone(z)
			self.printDetails = False
			print we,ns,PrintWindDirection(we,ns)
			if ns > 0:
				yStart = self.wz.GetYFromZone(z,False)
				yStop = self.wz.GetYFromZone(z,True) + ns
			else:
				yStart = self.wz.GetYFromZone(z,True)
				yStop = self.wz.GetYFromZone(z,False) + ns
			for y in range(yStart,yStop,ns):
				if we < 0:
					xStart = hm.mapWidth - 1
					xStop = -1
				else:
					xStart = 0
					xStop = hm.mapWidth
				for x in range(xStart,xStop,we):
					i = hm.GetIndex(x,y)
					if(hm.plotMap[i] == hm.OCEAN):
						self.DistributeRain(rainDumped,rainPotential,x,y,yStop,we,ns)
						self.printDetails = False
		#Add some smoothing and noise to the rain map to avoid harsh
		#looking artifacts
		for y in range(hm.mapHeight):
			for x in range(hm.mapWidth):
				i = hm.GetIndex(x,y)
				avg = 0.0
				for yy in range(y-1,y+2):
					for xx in range(x-1,x+2):
						ii = hm.GetIndex(x,y)
						avg += self.rainMap[ii]
				avg = avg/9.0
				smoothMap[i] = (self.rainMap[i] + (avg * RainSmoothingFactor) + (PWRand.random()* RainNoise * self.rainMap[i]))/(1 + RainSmoothingFactor + (RainNoise * self.rainMap[i]))
		for i in range(0,hm.mapHeight*hm.mapWidth):
			self.rainMap[i] = smoothMap[i]

		#We dont care about ocean rainfall, so set those to zero before
		#normalization
		for y in range(hm.mapHeight):
			for x in range(hm.mapWidth):
				i = hm.GetIndex(x,y)
				if self.rainMap[i] < 0.0:
					raise ValueError,"negative values detected"
				if hm.plotMap[i] == hm.OCEAN:
					self.rainMap[i] = 0.0

		NormalizeMap(hm.mapWidth,hm.mapHeight,self.rainMap)

		return
	def DistributeRain(self,rd,rp,x,y,oldYStop,we,ns):
		#set up ranges and offsets
		if ns < 0:
			yStart = self.distHeight - 1
			yStop = yStart - (y - oldYStop)
			offsetY = y - yStart
		else:
			yStart = 0
			yStop = oldYStop - y
			offsetY = y
		if we > 0:
			xStart = 0
			xStop = self.distWidth
			offsetX = x
		else:
			xStart = self.distWidth - 1
			xStop = -1
			offsetX = x - (self.distWidth - 1)

		#seed corner with tempMap(x,y)
		i = hm.GetIndex(x,y)
		ii = yStart * self.distWidth + xStart
		rp[ii] = tm.tempMap[i]
		lastRow = False
		for yy in range(yStart,yStop,ns):
			for xx in range(xStart,xStop,we):
				if yy == yStop - ns:
					lastRow = True
				self.SplashRain(rd,rp,offsetX,offsetY,xx,yy,ns,we,lastRow)

		for yy in range(yStart,yStop,ns):
			for xx in range(xStart,xStop,we):
				i = hm.GetIndex(xx + offsetX,yy + offsetY)
				ii = (yy * self.distWidth + xx)
				self.rainMap[i] += rd[ii]

		return
	def SplashRain(self,rd,rp,offsetX,offsetY,x,y,ns,we,lastRow):
		#first pull rain potential from the y direction and dump the excess there,
		#then push rain potential to the x direction and dump the excess here.
		#This mimicks the operation of spreading the rain in two (NE etc.)
		#directions simultaniously for each square
		xDamper = 0.120#These dampers diminish the amount of moisture loss coming
		yDamper = 0.160#from temperature changes
		i = y * self.distWidth + x
		imap = hm.GetIndex(x + offsetX,y + offsetY)
		yy = y - ns
		if(yy >= 0 and yy < self.distHeight): #dont access over the dist maps
			ii = yy * self.distWidth + x #ii = source address
			#first calculate temp difference. temp from 1.0 to 0 will dump all
			#the rain, from 0.0 to 1.0 will dump none. Dump the rain at the
			#source square, then subtract rain dumped from current rain
			#potential and pass on to destination
			iimap = hm.GetIndex(x + offsetX,yy + offsetY)
			tempDifference = tm.tempMap[iimap] - tm.tempMap[imap]
			tempFactor = tempDifference * 0.5 + 0.5 #convert -1 to 1 range to 0 to 1
			tempFactor *= yDamper #diminish effect so rain doesn't all dump on the coast
			rainToDump = rp[ii] * 0.5 * tempFactor
			rainToPassOn = rp[ii] * 0.5 - (rainToDump * (1 - (hm.map[imap] - hm.seaLevel)))
			#rainToPassOn = rp[ii] * 0.5 - rainToDump
			rd[ii] += rainToDump
			if((x == 0 and we > 0) or (x == self.distWidth - 1 and we < 0)):
				rp[i] = rainToPassOn #overwrite on first column
			else:
				rp[i] += rainToPassOn

		#now do essentially the same thing in the x direction except the
		#source and destination are switched. Also, as each square is initially
		#approached from the x direction first, we overwrite, rather than adding
		#to existing rain dump
		xx = x + we
		if(xx >= 0 and xx < self.distWidth):
			ii = y * self.distWidth + xx #ii = destination address
			iimap = hm.GetIndex(xx + offsetX,y + offsetY)
			tempDifference = tm.tempMap[imap] - tm.tempMap[iimap]
			tempFactor = tempDifference * 0.5 + 0.5 #convert -1 to 1 range to 0 to 1
			tempFactor *= xDamper #diminish effect so rain doesn't all dump on the coast
			rainToDump = rp[i] * 0.5 * tempFactor
			rainToPassOn = rp[i] * 0.5 - (rainToDump * (1 - (hm.map[imap] - hm.seaLevel)))
			#rainToPassOn = rp[i] * 0.5 - rainToDump
			if lastRow == True:
				rd[i] = rp[i] * xDamper #windzone boundary dump it all!
			else:
				rd[i] = rainToDump #overwriting old data, no +=
			rp[ii] = rainToPassOn #overwriting old data
		else:
			#overwrite rainToDump on last call in x direction
			rd[i] = 0.0
		return

##Global access
print "Initializing rainfall map..."
rm = RainfallMap()

class TerrainMap :
	def __init__(self):
		#To provide global access without allocating alot of resources for
		#nothing, object initializer must be empty
		return
	def GenerateTerrainMap(self):
		self.DESERT = 0
		self.PLAINS = 1
		self.ICE = 2
		self.TUNDRA = 3
		self.GRASS = 4
		self.HILL = 5
		self.COAST = 6
		self.OCEAN = 7
		self.PEAK = 8
		self.COOLGRASS = 10
		self.COOLPLAINS = 11
		self.HOTDESERT = 12
		self.HOTGRASS = 13
		self.HOTPLAINS = 14
		self.DEEPOCEAN = 15
		self.HOTSAVANNA = 16
		self.DIRT = 17
		self.HOTDIRT = 18
		self.DIRTTUNDRA = 19
		self.SAVANNA = 20
		self.COOLSAVANNA = 21
		self.terrainMap = array('i')

		PolarLand = False
		global DefaultBiome

		minRain = 10.0
		for i in range(hm.mapWidth*hm.mapHeight):
			#Initialize terrainMap with ocean
			self.terrainMap.append(self.DEEPOCEAN)
			#Remove polar land
			if tm.tempMap[i] < IceTemp + (PWRand.random() * 0.1) and PolarLand == False:
				hm.plotMap[i] = hm.OCEAN
			#Find minimum rainfall on land
			if hm.plotMap[i] != hm.OCEAN:
				if rm.rainMap[i] < minRain:
					minRain = rm.rainMap[i]

		print "Generating alternate biomes..."
# Generate area list for regional texture placement, each are is a landmass and coast does NOT connect the
# landmasses here.
		am = Areamap(hm.mapWidth,hm.mapHeight)
		am.defineAreas(False)
		totalLand = 0

		oceanID = am.getOceanID()
		for a in am.areaList:
			if a.ID != oceanID:
				totalLand += a.size

# 'Australian' Biome
# 20% increase in temperature, 20% decrease in rainfall.
# Red dirt region with replacement desert, hot desert and hot plains.  Medium landmasses only with restricted
# placement once it occupies more than 10% of the total land area.
# Desert -> Dirt
# HotDesert -> HotDirt
# HotPlains -> Savanna
# Plains -> DirtPlains
# Cool Plains -> Cool DirtPlains
# Tundar -> DirtTundra
		bFlip = False
		redSandList = list()
		redSandSize = 0
		redSandMin = totalLand * 0.05
		redSandMax = totalLand * 0.30
		for a in am.areaList:
			if a.ID != oceanID:
				if a.size > redSandMin and a.size < redSandMax:
					if bFlip == False:
						aID = a.ID
						redSandSize += a.size
						redSandList.append(aID)
						bFlip = True
					elif redSandSize/totalLand < 0.15:
						bFlip = False
		print "%(a)2d areas in the Australian Biome" % {"a":len(redSandList)}

# Great Plains Biome
# Modified thresholds to increase likelyhood of plains and decrease likelyhood of deserts with a 20% decrease in temperature.
		bFlip = False
		greatPlainsList = list()
		greatPlainsSize = 0
		greatPlainsMin = totalLand * 0.10
		greatPlainsMax = totalLand * 0.45
		for a in am.areaList:
			if a.ID != oceanID:
				if a.size > greatPlainsMin and a.size < greatPlainsMax:
					if bFlip == False:
						aID = a.ID
						greatPlainsSize += a.size
						greatPlainsList.append(aID)
						bFlip = True
					elif greatPlainsSize/totalLand < 0.30:
						bFlip = False
		print "%(a)2d areas in the Great Plains Biome" % {"a":len(greatPlainsList)}

		self.desertThreshold = FindValueFromPercent(rm.rainMap,DesertPercent,.0001,False)
		self.plainsThreshold = FindValueFromPercent(rm.rainMap,PlainsPercent,.0001,False)
		for y in range(hm.mapHeight):
			for x in range(hm.mapWidth):
				i = hm.GetIndex(x,y)

				#Altitude modifier for rainfall
				altRange = 1.0 - hm.seaLevel
				altMod = (hm.map[i] - hm.seaLevel) * float(1.0/altRange)
				altRain = (rm.rainMap[i] * (1.0 - altMod) + rm.rainMap[i]) / 2
				altTemp = tm.tempMap[i]
				altDesertThreshold = self.desertThreshold
				altPlainsThreshold = self.plainsThreshold

				# Biome setup
				redSand = False
				greatPlains = False
				if am.areaMap[i] in redSandList:
					altRain *= 0.75
					altTemp *= 1.25
					altDesertThreshold *= 1.25
					altPlainsThreshold *= 1.25
					redSand = True
				elif am.areaMap[i] in greatPlainsList:
					altTemp *= 0.8
					altDesertThreshold *= 0.95
					altPlainsThreshold *= 1.25
					greatPlains = True
				else:
					if DefaultBiome == 1: #Tropical
						altRain *= 1.2
						altTemp *= 1.2
						altDesertThreshold *= 0.9
						altPlainsThreshold *= 0.8
					elif DefaultBiome == 2: #Arid
						altRain *= 0.8
						altTemp *= 1.2
						altDesertThreshold *= 1.1
						altPlainsThreshold *= 1.2
					elif DefaultBiome == 3: #Cool
						altRain *= 0.8
						altTemp *= 0.8

				if hm.plotMap[i] == hm.OCEAN:
					if hm.map[i] > hm.seaLevel * 0.66:
						self.terrainMap[i] = self.OCEAN
					for yy in range(y - 3,y + 4,1):
						for xx in range(x - 3,x + 4,1):
							ii = hm.GetIndex(xx,yy)
							if hm.plotMap[ii] != hm.OCEAN and self.terrainMap[i] == self.DEEPOCEAN:
								self.terrainMap[i] = self.OCEAN
							if abs(x - xx) < 2 and abs(y - yy) < 2:
								if hm.plotMap[ii] != hm.OCEAN:
									self.terrainMap[i] = self.COAST

				#instead of harsh thresholds, allow a random deviation chance
				#based on how close to the threshold the rainfall is
				elif rm.rainMap[i] < altDesertThreshold:
					if tm.tempMap[i] < TundraTemp - (PWRand.random() * 0.1):
						self.terrainMap[i] = self.ICE
					elif tm.tempMap[i] < TundraTemp + (PWRand.random() * 0.1):
						if redSand == True:
							self.terrainMap[i] = self.DIRTTUNDRA
						else:
							self.terrainMap[i] = self.TUNDRA
					else:
						if altRain < (PWRand.random() * (altDesertThreshold - minRain) + altDesertThreshold - minRain)/2.0 + minRain:
							if altTemp < ForestTemp - (PWRand.random() * 0.1):
								self.terrainMap[i] = self.TUNDRA
							elif altTemp > JungleTemp + (PWRand.random() * 0.1):
								# Reduce some peaks & hills on hot desert terrain
								if hm.plotMap[i] == hm.PEAK and PWRand.random() < 0.33:
									hm.plotMap[i] = hm.HILLS
								elif hm.plotMap[i] == hm.HILLS and PWRand.random() < 0.66:
									hm.plotMap[i] = hm.LAND

								if redSand == True:
									self.terrainMap[i] = self.HOTDIRT
								else:
									self.terrainMap[i] = self.HOTDESERT
							else:
								# Reduce some peaks & hills on desert terrain
								if hm.plotMap[i] == hm.PEAK and PWRand.random() < 0.25:
									hm.plotMap[i] = hm.HILLS
								elif hm.plotMap[i] == hm.HILLS and PWRand.random() < 0.50:
									hm.plotMap[i] = hm.LAND

								if redSand == True:
									self.terrainMap[i] = self.DIRT
								else:
									self.terrainMap[i] = self.DESERT
						else:
							if altTemp < ForestTemp + (PWRand.random() * 0.1):
								if redSand == True:
									self.terrainMap[i] = self.COOLSAVANNA
								else:
									self.terrainMap[i] = self.COOLPLAINS
							elif altTemp > JungleTemp + (PWRand.random() * 0.1):
								if redSand == True:
									self.terrainMap[i] = self.HOTSAVANNA
								else:
									self.terrainMap[i] = self.HOTPLAINS
							else:
								if redSand == True:
									self.terrainMap[i] = self.SAVANNA
								else:
									self.terrainMap[i] = self.PLAINS
				elif rm.rainMap[i] < altPlainsThreshold:
					if tm.tempMap[i] < IceTemp + (PWRand.random() * 0.1):
						self.terrainMap[i] = self.ICE
					elif tm.tempMap[i] < TundraTemp + (PWRand.random() * 0.1):
						if redSand == True:
							self.terrainMap[i] = self.DIRTTUNDRA
						else:
							self.terrainMap[i] = self.TUNDRA
					else:
						if altRain < ((PWRand.random() * (altPlainsThreshold - altDesertThreshold) + altPlainsThreshold - altDesertThreshold))/2.0 + altDesertThreshold:
							if altTemp < ForestTemp + (PWRand.random() * 0.1):
								if redSand == True:
									self.terrainMap[i] = self.COOLSAVANNA
								else:
									self.terrainMap[i] = self.COOLPLAINS
							elif altTemp > JungleTemp + (PWRand.random() * 0.1):
								if redSand == True:
									self.terrainMap[i] = self.HOTSAVANNA
								else:
									self.terrainMap[i] = self.HOTPLAINS
							else:
								if redSand == True:
									self.terrainMap[i] = self.SAVANNA
								else:
									self.terrainMap[i] = self.PLAINS
						else:
							if altTemp < ForestTemp + (PWRand.random() * 0.1):
								self.terrainMap[i] = self.COOLGRASS
							elif altTemp > JungleTemp:
								self.terrainMap[i] = self.HOTGRASS
							else:
								self.terrainMap[i] = self.GRASS
				else:
					if tm.tempMap[i] < IceTemp + (PWRand.random() * 0.1):
						self.terrainMap[i] = self.ICE
					elif tm.tempMap[i] < TundraTemp + (PWRand.random() * 0.1):
						if redSand == True:
							self.terrainMap[i] = self.DIRTTUNDRA
						else:
							self.terrainMap[i] = self.TUNDRA
					else:
						if altTemp < ForestTemp + (PWRand.random() * 0.1):
							self.terrainMap[i] = self.COOLGRASS
						elif altTemp > JungleTemp:
							self.terrainMap[i] = self.HOTGRASS
						else:
							self.terrainMap[i] = self.GRASS
		return
	def generateContinentMap(self):
		self.cm = Areamap(hm.mapWidth,hm.mapHeight)
		self.cm.defineAreas(True)
		self.newWorldID = self.cm.getNewWorldID()
		return
##Global access
print "Initializing terrain map..."
terr = TerrainMap()

#OK! now that directions N,S,E,W are important, we have to keep in mind that
#the map plots are ordered from 0,0 in the SOUTH west corner! NOT the northwest
#corner! That means that Y increases as you go north.
class RiverMap :
	def __init__(self):
		#To provide global access without allocating alot of resources for
		#nothing, object initializer must be empty
		return
	def GenerateRiverMap(self):
		self.L = 0 #also denotes a 'pit' or 'flat'
		self.N = 1
		self.S = 2
		self.E = 3
		self.W = 4
		self.O = 5 #used for ocean or land without a river

		#averageHeightMap, flowMap, averageRainfallMap and drainageMap are offset from the other maps such that
		#each element coincides with a four tile intersection on the game map
		self.averageHeightMap = array('d')
		self.flowMap = array('i')
		self.averageRainfallMap = array('d')
		self.drainageMap = array('d')
		self.riverMap = array('i')
		#initialize maps with zeros
		for i in range(0,hm.mapHeight*hm.mapWidth):
			self.averageHeightMap.append(0.0)
			self.flowMap.append(0)
			self.averageRainfallMap.append(0.0)
			self.drainageMap.append(0.0)
			self.riverMap.append(self.O)
		#Get highest intersection neighbor
		for y in range(hm.mapHeight):
			for x in range(hm.mapWidth):
				i = hm.GetIndex(x,y)
				maxHeight = 0.0;
				for yy in range(y,y-2,-1):
					for xx in range(x,x+2):
						ii = hm.GetIndex(xx,yy)
						#use an average hight of <0 to denote an ocean border
						#this will save processing time later
						if(hm.plotMap[ii] == hm.OCEAN):
							maxHeight = -100.0
						elif maxHeight < hm.map[ii] and maxHeight >= 0:
							maxHeight = hm.map[ii]
				self.averageHeightMap[i] = maxHeight
		#create flowMap by checking for the lowest of each 4 connected
		#neighbor plus self
		for y in range(hm.mapHeight):
			for x in range(hm.mapWidth):
				i = hm.GetIndex(x,y)
				lowestAlt = self.averageHeightMap[i]
				if(lowestAlt < 0.0):
					#if height is <0 then that means this intersection is
					#adjacent to an ocean and has no flow
					self.flowMap[i] = self.O
				else:
					#First assume this place is lowest, like a 'pit'. Then
					#for each place that is lower, add it to a list to be
					#randomly chosen as the drainage path
					drainList = list()
					self.flowMap[i] = self.L
					ii = hm.GetIndex(x,y+1)
					#in the y direction, avoid wrapping
					if(y > 0 and self.averageHeightMap[ii] < lowestAlt):
						drainList.append(self.N)
					ii = hm.GetIndex(x,y-1)
					if(y < hm.mapHeight - 1 and self.averageHeightMap[ii] < lowestAlt):
						drainList.append(self.S)
					ii = hm.GetIndex(x-1,y)
					if(self.averageHeightMap[ii] < lowestAlt):
						drainList.append(self.W)
					ii = hm.GetIndex(x+1,y)
					if(self.averageHeightMap[ii] < lowestAlt):
						drainList.append(self.E)
					count = len(drainList)
					if count > 0:
						choice = int(PWRand.random()*count)
						self.flowMap[i] = drainList[choice]

		#Create average rainfall map so that each intersection is an average
		#of the rainfall from rm.rainMap
		for y in range(hm.mapHeight):
			for x in range(hm.mapWidth):
				i = hm.GetIndex(x,y)
				avg = 0.0;
				for yy in range(y,y-2,-1):
					for xx in range(x,x+2):
						ii = hm.GetIndex(xx,yy)
						avg += rm.rainMap[ii]
				avg = avg/4.0
				self.averageRainfallMap[i] = avg * avg

		#Now use the flowMap as a guide to distribute average rainfall.
		#Wherever the most rainfall ends up is where the rivers will be.
		for y in range(hm.mapHeight):
			for x in range(hm.mapWidth):
				i = hm.GetIndex(x,y)
				flow = self.flowMap[i]
				rainFall = self.averageRainfallMap[i]
				xx = x
				yy = y
				while(flow != self.L and flow != self.O):
					if(flow == self.N):
						yy += 1
					elif(flow == self.S):
						yy -= 1
					elif(flow == self.W):
						xx -= 1
					elif(flow == self.E):
						xx += 1
					#wrap
					if(xx < 0):
						xx = hm.mapWidth - 1
					elif(xx >= hm.mapWidth):
						xx = 0
					if(yy < 0):
						yy = hm.mapHeight - 1
					elif(yy >= hm.mapHeight):
						yy = 0
					#dump rainfall here
					ii = hm.GetIndex(xx,yy)
					self.drainageMap[ii] += rainFall
					#reset flow
					flow = self.flowMap[ii]

		riverThreshold = terr.plainsThreshold * RiverThreshold
		for i in range(hm.mapHeight*hm.mapWidth):
			if(self.drainageMap[i] > riverThreshold):
				self.riverMap[i] = self.flowMap[i]
			else:
				self.riverMap[i] = self.O

		return
	def isFloodPlainThreshold(self,x,y):
		riverThreshold = terr.plainsThreshold * RiverThreshold
		maxRain = 0
		#Loops are opposite of above examples because here we are going
		#from intersection to square instead of vice versa.
		for yy in range(y,y+2):
			for xx in range(x,x-2,-1):
				ii = hm.GetIndex(xx,yy)
				if maxRain < self.drainageMap[ii]:
					maxRain = self.drainageMap[ii]
		if maxRain > riverThreshold * FloodPlainThreshold:
			return True
		return False
##Global access
print "Initializing river map..."
rivers = RiverMap()

class BonusPlacer :
	def __init__(self):
		return
	def AddBonuses(self):
		gc = CyGlobalContext()
		gameMap = CyMap()
		gameMap.recalculateAreas()
		self.AssignBonusAreas()
		numBonuses = gc.getNumBonusInfos()
		for order in range(numBonuses):
			for i in range(numBonuses):
				bonusInfo = gc.getBonusInfo(self.bonusList[i].eBonus)
				if bonusInfo.getPlacementOrder() == order:
					self.AddBonusType(self.bonusList[i].eBonus)#Both unique and non-unique bonuses
		return
	def AddBonusType(self,eBonus):
		gc = CyGlobalContext()
		gameMap = CyMap()
		#first get bonus/area link
		for i in range(gc.getNumBonusInfos()):
			if self.bonusList[i].eBonus == eBonus:
				bonus = self.bonusList[i]
		if bonus.desiredBonusCount <= 0:#Non map bonuses?
			return
		bonusCount = bonus.desiredBonusCount
		#Create a list of map indices and shuffle them
		preshuffle = list()
		plotIndexList = list()
		for i in range(hm.mapWidth*hm.mapHeight):
			preshuffle.append(i)
		for i in range(hm.mapWidth*hm.mapHeight):
			randIndex = PWRand.randint(0,len(preshuffle)-1)
			if randIndex < 0 or randIndex >= len(preshuffle):
				raise ValueError, " bad index shuffling plot list randIndex=%(r)d listLength=%(l)d, mapSize=%(s)d" % {"r":randIndex,"l":len(preshuffle),"s":hm.mapWidth*hm.mapHeight}
			plotIndexList.append(preshuffle[randIndex])
			del preshuffle[randIndex]
			if len(preshuffle) == 0:
				break
		bonusInfo = gc.getBonusInfo(eBonus)
		print "Trying to place %(n)d of %(bt)s" % {"n":bonusCount,"bt":bonusInfo.getType()}
		#now add bonuses
		for i in range(len(plotIndexList)):
			index = plotIndexList[i]
			plot = gameMap.plotByIndex(index)
			x = plot.getX()
			y = plot.getY()
			if self.CanPlaceBonusAt(plot,eBonus,False,False) and y > 1 and y < hm.mapHeight - 2:
				#temporarily remove any feature
				featureEnum = plot.getFeatureType()
				if featureEnum != FeatureTypes.NO_FEATURE:
					featureVariety = plot.getFeatureVariety()
					plot.setFeatureType(FeatureTypes.NO_FEATURE,-1)
				#place bonus
				plot.setBonusType(eBonus)
				bonusCount -= 1
				#restore the feature if possible
				if featureEnum != FeatureTypes.NO_FEATURE:
					if bonusInfo == None or bonusInfo.isFeature(featureEnum):
						plot.setFeatureType(featureEnum,featureVariety)
				groupRange = bonusInfo.getGroupRange()
				for dx in range(-groupRange,groupRange + 1):
					for dy in range(-groupRange,groupRange + 1):
						if bonusCount > 0:
							loopPlot = self.plotXY(x,y,dx,dy)
							if loopPlot != None:
								if loopPlot.getX() == -1:
									raise ValueError, "plotXY returns invalid plots plot= %(x)d, %(y)d" % {"x":x,"y":y}
								if self.CanPlaceBonusAt(loopPlot,eBonus,False,False):
									if PWRand.randint(0,99) < bonusInfo.getGroupRand():
										#temporarily remove any feature
										featureEnum = loopPlot.getFeatureType()
										if featureEnum != FeatureTypes.NO_FEATURE:
											featureVariety = loopPlot.getFeatureVariety()
											loopPlot.setFeatureType(FeatureTypes.NO_FEATURE,-1)
										#place bonus
										loopPlot.setBonusType(eBonus)
										bonusCount -= 1
										#restore the feature if possible
										if featureEnum != FeatureTypes.NO_FEATURE:
											if bonusInfo == None or bonusInfo.isFeature(featureEnum):
												loopPlot.setFeatureType(featureEnum,featureVariety)
			if bonusCount == 0:
				break
		if bonusCount > 0:
			print "*** Could not place %(n)d of %(bt)s ***" % {"n":bonusCount,"bt":bonusInfo.getType()}
		else:
			print "Successfully placed %(n)d of %(bt)s" % {"n":bonus.desiredBonusCount,"bt":bonusInfo.getType()}
		return
	def plotXY(self,x,y,dx,dy):
		gameMap = CyMap()
		#The one that civ uses will return junk so I have to make one
		#that will not
		x = (x + dx) % hm.mapWidth
		y = y + dy
		if y < 0 or y > hm.mapHeight - 1:
			return None
		return gameMap.plot(x,y)

	def AssignBonusAreas(self):
		gc = CyGlobalContext()
		self.areas = CvMapGeneratorUtil.getAreas()
		gameMap = CyMap()
		preShuffledBonusList = list()
		self.bonusList = list()
		#Create and shuffle the bonus list and keep tally on
		#one-area bonuses and find the smallest min area requirement
		#among those
		numUniqueBonuses = 0
		minLandAreaSize = -1
		for i in range(gc.getNumBonusInfos()):
			bonus = BonusArea()
			bonus.eBonus = i
			preShuffledBonusList.append(bonus)
			bonusInfo = gc.getBonusInfo(i)
			if bonusInfo.isOneArea() == True:
				numUniqueBonuses += 1
				minAreaSize = bonusInfo.getMinAreaSize()
				if (minLandAreaSize == -1 or minLandAreaSize > minAreaSize) and \
				minAreaSize > 0:
					minLandAreaSize = minAreaSize
		for i in range(gc.getNumBonusInfos()):
			iChooseBonus = PWRand.randint(0,len(preShuffledBonusList)-1)
			self.bonusList.append(preShuffledBonusList[iChooseBonus])
			del preShuffledBonusList[iChooseBonus]
		numBonuses = gc.getNumBonusInfos()
		for i in range(numBonuses):
			self.bonusList[i].desiredBonusCount = self.CalculateNumBonusesToAdd(self.bonusList[i].eBonus)
			bonusInfo = gc.getBonusInfo(self.bonusList[i].eBonus)
			eBonus = self.bonusList[i].eBonus
			if bonusInfo.isOneArea() == False:
				continue #Only assign areas to area bonuses
			areaSuitabilityList = list()
			for area in self.areas:
				if area.getNumTiles() >= minLandAreaSize:
					aS = AreaSuitability(area.getID())
					aS.suitability,aS.numPossible = self.CalculateAreaSuitability(area,eBonus)
					areaSuitabilityList.append(aS)
			#Calculate how many areas to assign (numUniqueBonuses will be > 0 if we get here)
			areasPerBonus =  1
			#Sort areaSuitabilityList best first
			areaSuitabilityList.sort(lambda x,y:cmp(x.numPossible,y.numPossible))
			areaSuitabilityList.reverse()
			#assign the best areas to this bonus
			for n in range(areasPerBonus):
				self.bonusList[i].areaList.append(areaSuitabilityList[n].areaID)
			#assign areas that have a high suitability also
			for n in range(areasPerBonus,len(areaSuitabilityList)):
				if areaSuitabilityList[n].suitability > 0.3:
					self.bonusList[i].areaList.append(areaSuitabilityList[n].areaID)
		return
	def CanPlaceBonusAt(self,plot,eBonus,bIgnoreLatitude,bIgnoreArea):
		gc = CyGlobalContext()
		gameMap = CyMap()
		x = plot.getX()
		y = plot.getY()
		areaID = plot.getArea()
		if self.PlotCanHaveBonus(plot,eBonus,bIgnoreLatitude,bIgnoreArea) == False:
			return False
		for i in range(DirectionTypes.NUM_DIRECTION_TYPES):
			loopPlot = plotDirection(x,y,DirectionTypes(i))
			if loopPlot.getBonusType(TeamTypes.NO_TEAM) != BonusTypes.NO_BONUS and loopPlot.getBonusType(TeamTypes.NO_TEAM) != eBonus:
			   return False

		bonusInfo = gc.getBonusInfo(eBonus)
		classInfo = gc.getBonusClassInfo(bonusInfo.getBonusClassType())
		if plot.isWater() == True:
			if gameMap.getNumBonusesOnLand(eBonus) * 100/(gameMap.getNumBonuses(eBonus) + 1) < bonusInfo.getMinLandPercent():
				return False
		#Make sure there are no bonuses of the same class (but a different type) nearby:
		if classInfo != None:
			iRange = classInfo.getUniqueRange()
			for dx in range(-iRange,iRange+1):
				for dy in range(-iRange,iRange+1):
					loopPlot = self.plotXY(x,y,dx,dy)
					if loopPlot != None:
						if areaID == loopPlot.getArea():
							if plotDistance(x, y, loopPlot.getX(), loopPlot.getY()) <= iRange:
								eOtherBonus = loopPlot.getBonusType(TeamTypes.NO_TEAM)
								if eOtherBonus != BonusTypes.NO_BONUS:
									if gc.getBonusInfo(eOtherBonus).getBonusClassType() == bonusInfo.getBonusClassType():
										return False
		#Make sure there are no bonuses of the same type nearby:
		iRange = bonusInfo.getUniqueRange()
		for dx in range(-iRange,iRange+1):
			for dy in range(-iRange,iRange+1):
				loopPlot = self.plotXY(x,y,dx,dy)
				if loopPlot != None:
					if areaID == loopPlot.getArea():
						if plotDistance(x, y, loopPlot.getX(), loopPlot.getY()) <= iRange:
							eOtherBonus = loopPlot.getBonusType(TeamTypes.NO_TEAM)
							if eOtherBonus != BonusTypes.NO_BONUS:
								if eOtherBonus == eBonus:
									return False

		return True
	def PlotCanHaveBonus(self,plot,eBonus,bIgnoreLatitude,bIgnoreArea):
		#This function is like CvPlot::canHaveBonus but will
		#ignore blocking features and checks for a valid area.
		gc = CyGlobalContext()
		if eBonus == BonusTypes.NO_BONUS:
			return True
		if plot.getBonusType(TeamTypes.NO_TEAM) != BonusTypes.NO_BONUS:
			return False
		if plot.isPeak() == True:
			return False
		bonusInfo = gc.getBonusInfo(eBonus)
		if plot.getFeatureType() != FeatureTypes.NO_FEATURE:
			#Here is the change from canHaveBonus. Feature must not block unless
			#there is a different feature requirement for the bonus. On this
			#map bonuses trump features.
			featureRequired = False
			for i in range(gc.getNumFeatureInfos()):
				if bonusInfo.isFeature(FeatureTypes(i)) == True:
					featureRequired = True
					break
			if featureRequired == True and bonusInfo.isTerrain(plot.getTerrainType()) == False:
				if bonusInfo.isFeature(plot.getFeatureType()) == False:
					return False #Wrong feature
				if bonusInfo.isFeatureTerrain(plot.getTerrainType()) == False:
					return False
			elif bonusInfo.isTerrain(plot.getTerrainType()) == False:
				return False

		else:
			if bonusInfo.isTerrain(plot.getTerrainType()) == False:
				return False
		if plot.isHills() == True:
			if bonusInfo.isHills() == False:
				return False
		if plot.isFlatlands() == True:
			if bonusInfo.isFlatlands() == False:
				return False
		if bonusInfo.isNoRiverSide() == True:
			if plot.isRiverSide() == True:
				return False
		if bonusInfo.getMinAreaSize() != -1:
			if plot.area().getNumTiles() < bonusInfo.getMinAreaSize():
				return False
		if bIgnoreLatitude == False:
			if plot.getLatitude() > bonusInfo.getMaxLatitude():
				return False
			if plot.getLatitude() < bonusInfo.getMinLatitude():
				return False
		if plot.isPotentialCityWork() == False:
			return False

		if bIgnoreArea == False and bonusInfo.isOneArea() == True:
			areaID = plot.getArea()
			areaFound = False
			for i in range(len(self.bonusList)):
				if self.bonusList[i].eBonus == eBonus:
					areaList = self.bonusList[i].areaList
					for n in range(len(areaList)):
						if areaList[n] == areaID:
							areaFound = True
							break
					if areaFound == True:
						break
			if areaFound == False:
				return False

		return True
	def CalculateNumBonusesToAdd(self,eBonus):
		#This is like the function in CvMapGenerator except it uses
		#self.PlotCanHaveBonus instead of CvPlot::canHaveBonus
		gc = CyGlobalContext()
		gameMap = CyMap()
		game = CyGame()
		bonusInfo = gc.getBonusInfo(eBonus)
		rand1 = PWRand.randint(0,bonusInfo.getRandAppearance1())
		rand2 = PWRand.randint(0,bonusInfo.getRandAppearance2())
		rand3 = PWRand.randint(0,bonusInfo.getRandAppearance3())
		rand4 = PWRand.randint(0,bonusInfo.getRandAppearance4())
		baseCount = bonusInfo.getConstAppearance() + rand1 + rand2 + rand3 + rand4

		bIgnoreLatitude = False
		bIgnoreArea = True
		landTiles = 0
		numPossible = 0
		if bonusInfo.getTilesPer() > 0:
			for i in range(hm.mapWidth*hm.mapHeight):
				plot = gameMap.plotByIndex(i)
				if self.PlotCanHaveBonus(plot,eBonus,bIgnoreLatitude,bIgnoreArea):
					numPossible += 1
			landTiles += numPossible/bonusInfo.getTilesPer()
		players = game.countCivPlayersAlive() * bonusInfo.getPercentPerPlayer()/100
		bonusCount = baseCount * (landTiles + players)/100
		bonusCount = max(0,bonusCount)

		return bonusCount
	def GetUniqueBonusTypeCountInArea(self,area):
		gc = CyGlobalContext()
		areaID = area.getID()
		uniqueBonusCount = 0
		for i in range(len(self.bonusList)):
			areaList = self.bonusList[i].areaList
			bonusInfo = gc.getBonusInfo(self.bonusList[i].eBonus)
			if bonusInfo.isOneArea() == False:
				continue
			for n in range(len(areaList)):
				if areaList[n] == areaID:
					uniqueBonusCount += 1
					break

		return uniqueBonusCount
	def GetSameClassTypeCountInArea(self,area,eBonus):
		gc = CyGlobalContext()
		areaID = area.getID()
		uniqueBonusCount = 0
		bonusInfo = gc.getBonusInfo(eBonus)
		eClass = bonusInfo.getBonusClassType()
		if eClass == BonusClassTypes.NO_BONUSCLASS:
			return 0
		classInfo = gc.getBonusClassInfo(eClass)
		if classInfo == None:
			return 0
		uRange = classInfo.getUniqueRange()
		for i in range(len(self.bonusList)):
			areaList = self.bonusList[i].areaList
			bonusInfo = gc.getBonusInfo(self.bonusList[i].eBonus)
			if bonusInfo.isOneArea() == False:
				continue
			if bonusInfo.getBonusClassType() != eClass:
				continue
			for n in range(len(areaList)):
				if areaList[n] == areaID:
					uniqueBonusCount += 1
					break
		#Same class types tend to really crowd out any bonus
		#types that are placed later. A single cow can block
		#5 * 5 squares of pig territory for example. Probably
		#shouldn't place them on the same area at all, but
		#sometimes it might be necessary.
		return uniqueBonusCount * uRange * uRange
	def CalculateAreaSuitability(self,area,eBonus):
		gc = CyGlobalContext()
		gameMap = CyMap()
		areaID = area.getID()
		uniqueTypesInArea = self.GetUniqueBonusTypeCountInArea(area)
		sameClassTypesInArea = self.GetSameClassTypeCountInArea(area,eBonus)
		#Get the raw number of suitable tiles
		numPossible = 0
		for i in range(hm.mapWidth*hm.mapHeight):
			plot = gameMap.plotByIndex(i)
			if plot.getArea() == areaID:
				if self.PlotCanHaveBonus(plot,eBonus,False,True):
					numPossible += 1
		numPossible = numPossible/(uniqueTypesInArea + sameClassTypesInArea + 1)
		suitability = float(numPossible)/float(area.getNumTiles())
		return suitability,numPossible
#Global Access
print "Initializing bonus placement..."
bp = BonusPlacer()

class BonusArea :
	def __init__(self):
		self.eBonus = -1
		self.desiredBonusCount = -1
		self.areaList = list()

class AreaSuitability :
	def __init__(self,areaID):
		self.areaID = areaID
		self.suitability = 0
		self.numPossible = 0

class StartingPlotFinder :
	def __init__(self):
		return
	def SetStartingPlots(self):
		gc = CyGlobalContext()
		gameMap = CyMap()
		iPlayers = gc.getGame().countCivPlayersEverAlive()
		gameMap.recalculateAreas()
		areas = CvMapGeneratorUtil.getAreas()

		#get old/new world status
		areaOldWorld = self.setupOldWorldAreaList()

		#Shuffle players so the same player doesn't always get the first pick.
		#lifted from Highlands.py that ships with Civ.
		player_list = []
		for plrCheckLoop in range(gc.getMAX_CIV_PLAYERS()):
				if CyGlobalContext().getPlayer(plrCheckLoop).isEverAlive():
						player_list.append(plrCheckLoop)
		shuffledPlayers = []
		for playerLoop in range(iPlayers):
				iChoosePlayer = PWRand.randint(0,len(player_list)-1)
				shuffledPlayers.append(player_list[iChoosePlayer])
				del player_list[iChoosePlayer]

		self.startingAreaList = list()
		for i in range(len(areas)):
			if areaOldWorld[i] == True and areas[i].getNumTiles() > 5:
				startArea = StartingArea(areas[i].getID())
				self.startingAreaList.append(startArea)

		#Get the value of the whole old world
		oldWorldValue = 0
		for i in range(len(self.startingAreaList)):
			oldWorldValue += self.startingAreaList[i].rawValue

		#calulate value per player of old world
		oldWorldValuePerPlayer = oldWorldValue/len(shuffledPlayers)

		#Sort startingAreaList by rawValue
		self.startingAreaList.sort(lambda x, y: cmp(x.rawValue, y.rawValue))

		#Get rid of areas that have less value than oldWorldValuePerPlayer
		#as they are too small to put a player on, however leave at least
		#half as many continents as there are players, just in case the
		#continents are *all* quite small.
		numAreas = max(1,len(self.startingAreaList) - len(shuffledPlayers)/2)
		for i in range(numAreas):
			if self.startingAreaList[0].rawValue < oldWorldValuePerPlayer:
				del self.startingAreaList[0]
			else:
				break #All remaining should be big enough

		#Recalculate the value of the whole old world
		oldWorldValue = 0
		for i in range(len(self.startingAreaList)):
			oldWorldValue += self.startingAreaList[i].rawValue

		#Recalulate value per player of old world so we are starting more
		#accurately
		oldWorldValuePerPlayer = oldWorldValue/len(shuffledPlayers)

		#Record the ideal number of players for each continent
		for startingArea in self.startingAreaList:
			startingArea.idealNumberOfPlayers = int(round(float(startingArea.rawValue)/float(oldWorldValuePerPlayer)))

		#Now we want best first
		self.startingAreaList.reverse()
		print "Number of starting areas is %(s)3d" % {"s":len(self.startingAreaList)}

		iterations = 0
		while True:
			iterations += 1
			if iterations > 20:
				raise ValueError, "Too many iterations in starting area choosing loop."
			chosenStartingAreas = list()
			playersPlaced = 0
			#add up idealNumbers
			idealNumbers = 0
			for startingArea in self.startingAreaList:
				idealNumbers += startingArea.idealNumberOfPlayers
			if idealNumbers < len(shuffledPlayers):
				self.startingAreaList[0].idealNumberOfPlayers += 1
			elif idealNumbers > len(shuffledPlayers):
				self.startingAreaList[0].idealNumberOfPlayers -= 1

			#Choose areas
			for startingArea in self.startingAreaList:
				if startingArea.idealNumberOfPlayers + playersPlaced <= len(shuffledPlayers):
					chosenStartingAreas.append(startingArea)
					playersPlaced += startingArea.idealNumberOfPlayers

			#add up idealNumbers again
			idealNumbers = 0
			for startingArea in chosenStartingAreas:
				idealNumbers += startingArea.idealNumberOfPlayers
			if idealNumbers == len(shuffledPlayers):
				break
		for startingArea in chosenStartingAreas:
			for i in range(startingArea.idealNumberOfPlayers):
				startingArea.playerList.append(shuffledPlayers[0])
				del shuffledPlayers[0]
			startingArea.FindStartingPlots()
		if len(shuffledPlayers) > 0:
			raise ValueError,"Some players not placed in starting plot finder!"

		#Now set up for normalization
		self.plotList = list()
		for startingArea in self.startingAreaList:
			for i in range(len(startingArea.plotList)):
				self.plotList.append(startingArea.plotList[i])

		#Remove bad features. (Jungle in the case of standard game)
		for i in range(len(self.plotList)):
			if self.plotList[i].vacant == True:
				continue
			plot = gameMap.plot(self.plotList[i].x,self.plotList[i].y)
			featureInfo = gc.getFeatureInfo(plot.getFeatureType())
			if featureInfo != None:
				totalYield = 0
				for y in range(YieldTypes.NUM_YIELD_TYPES):
					totalYield += featureInfo.getYieldChange(YieldTypes(y))
				if totalYield <= 0:#bad feature
					plot.setFeatureType(FeatureTypes.NO_FEATURE,-1)

			for n in range(DirectionTypes.NUM_DIRECTION_TYPES):
				loopPlot = plotDirection(self.plotList[i].x,self.plotList[i].y,DirectionTypes(n))
				if loopPlot == None:
					continue
				featureInfo = gc.getFeatureInfo(loopPlot.getFeatureType())
				if featureInfo != None:
					totalYield = 0
					for y in range(YieldTypes.NUM_YIELD_TYPES):
						totalYield += featureInfo.getYieldChange(YieldTypes(y))
					if totalYield <= 0:#bad feature
						loopPlot.setFeatureType(FeatureTypes.NO_FEATURE,-1)

		#find the best totalValue
		self.plotList.sort(lambda x,y:cmp(x.totalValue,y.totalValue))
		self.plotList.reverse()
		bestTotalValue = self.plotList[0].totalValue
		print "bestTotalValue = %(b)6d" % {"b":bestTotalValue}
		for i in range(len(self.plotList)):
			if self.plotList[i].vacant == True:
				continue
			currentTotalValue = self.plotList[i].totalValue
			percentLacking = 1.0 - (float(currentTotalValue)/float(bestTotalValue))
			if percentLacking > 0:
				bonuses = min(3,int(percentLacking/0.24))
				self.boostCityPlotValue(self.plotList[i].x,self.plotList[i].y,bonuses)

		return
	def setupOldWorldAreaList(self):
		gc = CyGlobalContext()
		gameMap = CyMap()
		#get official areas and make corresponding lists that determines old
		#world vs. new and also the pre-settled value.
		areas = CvMapGeneratorUtil.getAreas()
		areaOldWorld = list()

		for i in range(len(areas)):
			for pI in range(hm.mapHeight*hm.mapWidth):
				plot = gameMap.plotByIndex(pI)
				if plot.getArea() == areas[i].getID():
					if terr.cm.areaMap[pI] == terr.newWorldID:
						areaOldWorld.append(False)#new world true = old world false
					else:
						areaOldWorld.append(True)
					break

		return areaOldWorld
	def getCityPotentialValue(self,x,y):
		gc = CyGlobalContext()
		gameMap = CyMap()
		value = 0
		plot = gameMap.plot(x,y)
		sPlot = StartPlot(x,y,0)
		crapCount = 0

		if plot.isWater() == True:
			return 0
		for i in range(gc.getNUM_CITY_PLOTS()):
			plot = plotCity(x,y,i)
			dval = self.getPlotPotentialValue(plot.getX(),plot.getY(),sPlot.isCoast())
			value += dval
			if dval == 0:
				crapCount += 1

		value /= max(1, crapCount - 4)

		if sPlot.isCoast():
			value = int(float(value)*CoastalCityValueBonus)
		if sPlot.plot().isRiver():
			value = int(float(value)*RiverCityValueBonus)
		if sPlot.plot().waterArea().isLake() or sPlot.plot().isRiver():
			value = int(float(value)*FreshWaterValueBonus)

		return value
	def getPlotPotentialValue(self,x,y,seaSide):
		gc = CyGlobalContext()
		gameMap = CyMap()
		game = gc.getGame()
		debugOut = False
		value = 0
		commerce = 0
		food = 0
		production = 0

		plot = gameMap.plot(x,y)
		if debugOut: print "Evaluating plot x = %(x)d, y = %(y)d" % {"x":x,"y":y}

		commerce += plot.calculateBestNatureYield(YieldTypes.YIELD_COMMERCE,TeamTypes.NO_TEAM)
		food += plot.calculateBestNatureYield(YieldTypes.YIELD_FOOD,TeamTypes.NO_TEAM)
		if seaSide:
			if plot.getPlotType() == PlotTypes.PLOT_OCEAN:
				commerce += 1
		production += plot.calculateBestNatureYield(YieldTypes.YIELD_PRODUCTION,TeamTypes.NO_TEAM)

		if debugOut: print "Natural yields. Food=%(f)d, Production=%(p)d, Commerce=%(c)d" % {"f":food,"p":production,"c":commerce}

		#Get best bonus improvement score. Test tachnology era of bonus
		#first, then test each improvement
		bestImp = None
		bonusEnum = plot.getBonusType(TeamTypes.NO_TEAM)
		bonusInfo = gc.getBonusInfo(bonusEnum)

		if bonusInfo != None and (gc.getTechInfo(bonusInfo.getTechCityTrade()) == None or \
		gc.getTechInfo(bonusInfo.getTechCityTrade()).getEra() <= game.getStartEra()) and \
		(gc.getTechInfo(bonusInfo.getTechReveal()) == None or \
		gc.getTechInfo(bonusInfo.getTechReveal()).getEra() <= game.getStartEra()):
			if bonusInfo == None:
				if debugOut: print "Bonus Type = None"
			else:
				if debugOut: print "Bonus Type = %(b)s <------------------------------------------------------------" % {"b":bonusInfo.getType()}
			commerce += bonusInfo.getYieldChange(YieldTypes.YIELD_COMMERCE)
			food += bonusInfo.getYieldChange(YieldTypes.YIELD_FOOD)
			production += bonusInfo.getYieldChange(YieldTypes.YIELD_PRODUCTION)
			if debugOut: print "Bonus yields. Food=%(f)d, Production=%(p)d, Commerce=%(c)d" % \
			{"f":food,"p":production,"c":commerce}
		else:
			bonusEnum = -1
			bonusInfo = None

		improvementList = list()

		for n in range(gc.getNumBuildInfos()):
			#Test for improvement validity
			buildInfo = gc.getBuildInfo(n)
			impEnum = buildInfo.getImprovement()
			impInfo = gc.getImprovementInfo(impEnum)
			if impInfo == None:
				continue
			if gc.getTechInfo(buildInfo.getTechPrereq()).getEra() > game.getStartEra():
				if debugOut: print "Tech era not high enough for %(s)s" % {"s":impInfo.getType()}
				continue
			else:
				if debugOut: print "Tech is high enough for %(s)s" % {"s":impInfo.getType()}

			if plot.canHaveImprovement(impEnum,TeamTypes.NO_TEAM,True) == True:
				if debugOut: print "Plot can have %(s)s" % {"s":impInfo.getType()}
				#This function will not find bonus yield changes for NO_PLAYER much to my annoyance
				impCommerce = plot.calculateImprovementYieldChange(impEnum,YieldTypes.YIELD_COMMERCE,PlayerTypes.NO_PLAYER,False)
				impFood = plot.calculateImprovementYieldChange(impEnum,YieldTypes.YIELD_FOOD,PlayerTypes.NO_PLAYER,False)
				impProduction = plot.calculateImprovementYieldChange(impEnum,YieldTypes.YIELD_PRODUCTION,PlayerTypes.NO_PLAYER,False)
				impValue = 0

				if bonusEnum != -1:
					impCommerce += impInfo.getImprovementBonusYield(bonusEnum,YieldTypes.YIELD_COMMERCE)
					impFood += impInfo.getImprovementBonusYield(bonusEnum,YieldTypes.YIELD_FOOD)
					impProduction += impInfo.getImprovementBonusYield(bonusEnum,YieldTypes.YIELD_PRODUCTION)

				#See if feature is removed, if so we must subtract the added yield
				#from that feature
				featureEnum = plot.getFeatureType()
				if featureEnum != -1 and buildInfo.isFeatureRemove(featureEnum) == True:
					featureInfo = gc.getFeatureInfo(featureEnum)
					impCommerceChange = (featureInfo.getYieldChange(YieldTypes.YIELD_COMMERCE) + \
					featureInfo.getRiverYieldChange(YieldTypes.YIELD_COMMERCE) + \
					featureInfo.getHillsYieldChange(YieldTypes.YIELD_COMMERCE))

					impFoodChange = (featureInfo.getYieldChange(YieldTypes.YIELD_FOOD) + \
					featureInfo.getRiverYieldChange(YieldTypes.YIELD_FOOD) + \
					featureInfo.getHillsYieldChange(YieldTypes.YIELD_FOOD))

					impProductionChange = (featureInfo.getYieldChange(YieldTypes.YIELD_PRODUCTION) + \
					featureInfo.getRiverYieldChange(YieldTypes.YIELD_PRODUCTION) + \
					featureInfo.getHillsYieldChange(YieldTypes.YIELD_PRODUCTION))

					impCommerce -= impCommerceChange
					impFood -= impFoodChange
					impProduction -= impProductionChange

					#This basically makes forests better and jungles worse
					impValue -= (impFoodChange * FoodValue + impProductionChange * ProductionValue + impCommerceChange * CommerceValue) / 3

				imp = Improvement(impEnum,impFood,impProduction,impCommerce,impValue)
				improvementList.append(imp)
				if debugOut:
					print "Improv yields. Food=%(f)d, Production=%(p)d, Commerce=%(c)d" % {"f":impFood,"p":impProduction,"c":impCommerce}
			else:
				if debugOut:
					print "Plot can not have %(s)s" % {"s":impInfo.getType()}

		for i in range(len(improvementList)):
			impCommerce = improvementList[i].commerce + commerce
			impFood = improvementList[i].food + food
			impProduction = improvementList[i].production + production
			impValue = impCommerce * CommerceValue + impFood * FoodValue + impProduction * ProductionValue
			improvementList[i].value += impValue

		if len(improvementList) > 0:
			#sort all allowed improvement values to find the best
			improvementList.sort(lambda x,y:cmp(x.value,y.value))
			improvementList.reverse()
			bestImp = improvementList[0]
			if debugOut:
				print "bestImp.value=%(b)d" % {"b":bestImp.value}

			commerce += bestImp.commerce
			food += bestImp.food
			production += bestImp.production
		else:
			if debugOut: print "no improvement possible here"

		value = commerce * CommerceValue + food * FoodValue + production * ProductionValue
		if debugOut:
			print "Evaluating. Food=%(f)d, Production=%(p)d, Commerce=%(c)d" % {"f":food,"p":production,"c":commerce}

		#penalty for crappy tiles you won't work, because they have pathetic base values
		baseVal = plot.calculateBestNatureYield(YieldTypes.YIELD_FOOD,TeamTypes.NO_TEAM) * FoodValue \
		+ plot.calculateBestNatureYield(YieldTypes.YIELD_PRODUCTION,TeamTypes.NO_TEAM) * ProductionValue \
		+ plot.calculateBestNatureYield(YieldTypes.YIELD_COMMERCE,TeamTypes.NO_TEAM) * CommerceValue

		#ignore forests here - tundra forests still suck
		featureEnum = plot.getFeatureType()
		if featureEnum == gc.getInfoTypeForString("FEATURE_FOREST"):
			baseVal -= ProductionValue

		#count lighthouses
		if seaSide:
			if plot.getPlotType() == PlotTypes.PLOT_OCEAN:
				baseVal += FoodValue - CommerceValue - 1
				if debugOut:
					print "Seasideness helps here."

		#shaft the crap tiles!
		if baseVal < FoodValue + ProductionValue:
			value *= 0.7
			if debugOut:
				print "%(v)d is worse than a plains!" % {"v":baseVal}

		if baseVal < FoodValue + 2 * CommerceValue:
			value *= 0.4
			if debugOut:
				print "%(v)d is worse than a coast without lighthouse!" % {"v":baseVal}

		#resources are very good. This section is to reward the happy/health/power they give
		bonus = plot.getBonusType(TeamTypes.NO_TEAM)
		if bonus != BonusTypes.NO_BONUS:
			if debugOut:
				print "has a bonus"
			tech = gc.getBonusInfo(bonus).getTechCityTrade()

			#resources you can use early are the most important for a capital
			if tech != TechTypes.NO_TECH and gc.getTechInfo(tech) != None and gc.getTechInfo(tech).getEra() > game.getStartEra():
				value += FoodValue * 6 / (gc.getTechInfo(tech).getEra() - game.getStartEra())
			else:
				value += FoodValue * 17

			#sea resources when you can't build boats are very sad
			if plot.getPlotType() == PlotTypes.PLOT_OCEAN and seaSide == False:
				value /= 3

		#penalize crap tiles more
		#basically, if there is a crappy tile with no resource, it has no value.
		#setting the value to zero lets us use that later to count the number of really crappy tiles.
		if value <= 2 * FoodValue:
			value = 0

		#finally, coast isn't total crap, but it can't be improved
		if plot.getPlotType() == PlotTypes.PLOT_OCEAN:
			value /= 2

		if debugOut: print "Final Value = %(v)d" % {"v":value}
		if debugOut: print "****************************************************************"

		return value
	def boostCityPlotValue(self,x,y,bonuses):
		mapGen = CyMapGenerator()
		value = self.getCityPotentialValue(x,y)
		debugOut = False

		gc = CyGlobalContext()
		gameMap = CyMap()
		game = gc.getGame()
		#Shuffle the bonus order so that different cities have different preferences
		#for bonuses
		bonusList = list()
		numBonuses = gc.getNumBonusInfos()
		for i in range(numBonuses):
			bonusList.append(i)
		shuffledBonuses = list()
		for i in range(numBonuses):
			n = PWRand.randint(0,len(bonusList) - 1)
			shuffledBonuses.append(bonusList[n])
			del bonusList[n]

		if len(shuffledBonuses) != numBonuses:
			raise ValueError, "Bad bonus shuffle. Learn 2 shuffle."

		bonusCount = 0

		yields = []
		yields.append(YieldTypes.YIELD_PRODUCTION)
		yields.append(YieldTypes.YIELD_FOOD)
		yields.append(YieldTypes.YIELD_COMMERCE)
		yields.append(YieldTypes.YIELD_COMMERCE)
		yields.append(YieldTypes.YIELD_FOOD)
		yields.append(YieldTypes.YIELD_PRODUCTION)
		yields.append(YieldTypes.YIELD_FOOD)
		yields.append(YieldTypes.YIELD_COMMERCE)
		yields.append(YieldTypes.YIELD_PRODUCTION)

		for n in range(len(yields)):
			for i in range(gc.getNUM_CITY_PLOTS()):
				value = self.getCityPotentialValue(x,y)
				if debugOut: print "value now at %(v)d" % {"v":value}
				if bonusCount >= bonuses:
					if debugOut: print "Placed all bonuses."
					if debugOut: print "****************************************************"
					return
				plot = plotCity(x,y,i)
				if plot.getX() == x and plot.getY() == y:
					continue
				if plot.getBonusType(TeamTypes.NO_TEAM) != BonusTypes.NO_BONUS:
					continue
				#temporarily remove any feature
				featureEnum = plot.getFeatureType()
				if featureEnum != FeatureTypes.NO_FEATURE:
					featureVariety = plot.getFeatureVariety()
					plot.setFeatureType(FeatureTypes.NO_FEATURE,-1)
				for b in range(gc.getNumBonusInfos()):
					bonusEnum = shuffledBonuses[b]
					bonusInfo = gc.getBonusInfo(bonusEnum)
					if bonusInfo.isNormalize() == False:
						continue
					if bonusInfo.getYieldChange(yields[n]) < 1:
						continue
					if bonusInfo.getTechCityTrade() == TechTypes.NO_TECH or \
					gc.getTechInfo(bonusInfo.getTechCityTrade()).getEra() <= game.getStartEra():
						if bp.PlotCanHaveBonus(plot,bonusEnum,False,False) == False:
							continue

						plot.setBonusType(bonusEnum)
						bonusCount += 1
						break
				#restore the feature if possible
				if featureEnum != FeatureTypes.NO_FEATURE:
					bonusInfo = gc.getBonusInfo(plot.getBonusType(TeamTypes.NO_TEAM))
					if bonusInfo == None or bonusInfo.isFeature(featureEnum):
						plot.setFeatureType(featureEnum,featureVariety)

				if bonusCount > n:
					break

		return
#Global access
print "Initializing starting plot finder..."
spf = StartingPlotFinder()

class Improvement :
	def __init__(self,e,food,production,commerce,value):
		self.e = e
		self.food = food
		self.production = production
		self.commerce = commerce
		self.value = value

class StartingArea :
	def __init__(self,areaID):
		self.areaID = areaID
		self.playerList = list()
		self.plotList = list()
		self.distanceTable = array('i')
		self.rawValue = 0
		self.CalculatePlotList()
		self.idealNumberOfPlayers = 0
		return
	def CalculatePlotList(self):
		gc = CyGlobalContext()
		gameMap = CyMap()

		for y in range(hm.mapHeight):
			for x in range(hm.mapWidth):
				plot = gameMap.plot(x,y)
				if plot.getArea() == self.areaID:
					#don't place a city on top of a bonus
					if plot.getBonusType(TeamTypes.NO_TEAM) != BonusTypes.NO_BONUS:
						continue
					value = spf.getCityPotentialValue(x,y)
					if value > 0:
						startPlot = StartPlot(x,y,value)
						if plot.isWater() == True:
							raise ValueError, "potential start plot is water!"
						self.plotList.append(startPlot)
		#Sort plots by local value
		self.plotList.sort(lambda x, y: cmp(x.localValue, y.localValue))

		#To save time and space let's get rid of some of the lesser plots
		if len(self.plotList) == 0:
			return

		topPlotsLen = (len(self.plotList) * 1) / 8 + 2
		if topPlotsLen < len(self.plotList):
			self.plotList.reverse()
			bestVal = 0
			count  = 0
			for i in range(topPlotsLen):
				bestVal += self.plotList[i].localValue
				count += 1
			bestVal /= count
			self.plotList.reverse()
			while (self.plotList[0].localValue * 4 < bestVal):
				del self.plotList[0]

		#You now should be able to eliminate more plots by sorting high to low
		#and having the best plot eat plots within 3 squares, then same for next,
		#etc.
		self.plotList.reverse()
		numPlots = len(self.plotList)
		for n in range(numPlots):
			#At some point the length of plot list will be much shorter than at
			#the beginning of the loop, so it can never end normally
			if n >= len(self.plotList) - 1:
				break
			y = self.plotList[n].y
			x = self.plotList[n].x
			tmp = range(n,len(self.plotList))
			tmp.reverse()
			for m in tmp:
				dy = abs(self.plotList[m].y - y)
				dx = abs(self.plotList[m].x - x)
				if dx > hm.mapWidth / 2:
					dx = hm.mapWidth - dx
				if (m != n) & (dx <= 4) & (dy <= 4):
					del self.plotList[m]

		#At this point we should have a list of the very best places
		#to build cities on this continent. Now we need a table with
		#the distance from each city to every other city

		#Create distance table
		for i in range(len(self.plotList)*len(self.plotList)):
			self.distanceTable.append(-11)
		#Fill distance table
		for n in range(len(self.plotList)):
			#While were already looping lets calculate the raw value
			self.rawValue += self.plotList[n].localValue
			avgDistance = 0
			for m in range(n,len(self.plotList)):
				nPlot = gameMap.plot(self.plotList[n].x,self.plotList[n].y)
				mPlot = gameMap.plot(self.plotList[m].x,self.plotList[m].y)
				distance = self.getDistance(nPlot.getX(),nPlot.getY(),mPlot.getX(),mPlot.getY())
				self.distanceTable[n*len(self.plotList) + m] = distance
				self.distanceTable[m*len(self.plotList) + n] = distance
				avgDistance += distance
			self.plotList[n].avgDistance = avgDistance

		return
	def FindStartingPlots(self):
		gc = CyGlobalContext()
		gameMap = CyMap()
		numPlayers = len(self.playerList)
		if numPlayers <= 0:
			return

		###
		###NOTE: Added a complicated system where we loop through cities from best to worst,
		###and each city casts a vote against the worst city that is within some distance of it.
		###Two votes kills a city. The voting radius grows each iteration. Loop until only the
		###best remain.
		###This change kind of ruined the subsystem where cities get extra points for "controlling"
		###other cities, but it's totally worth it.
		numPlots = len(self.plotList)
		if(numPlots == 0):
			return

		alives = range(numPlots) #if it's equal to its index, it's alive. -1 it's dead. other number it's "on the block"
		wasVote = True
		countdown = range(len(self.plotList))
		countdown.reverse()
		votes = 0
		deaths = 0
		time = 0
		distance = 3
		leave = False
		if (numPlayers >= numPlots) | (numPlayers == 1):
			leave = True
		while (leave==False):
			distance += 1
			time += 1
			if time > numPlots + 10000:
				raise ValueError, "numPlots: "+str(numPlots)+"  deaths: "+str(deaths)+"  votes: "+str(votes)
			wasVote = False
			for n in range(numPlots):
				if leave:
					break
				if alives[n] == -1:
					continue
				if n >= len(self.plotList) - 1:
					break
				y = self.plotList[n].y
				x = self.plotList[n].x
				for m in countdown:
					if leave:
						break
					if alives[m] == -1:
						continue
					dy = abs(self.plotList[m].y - y)
					dx = abs(self.plotList[m].x - x)
					if dx > hm.mapWidth / 2:
						dx = hm.mapWidth - dx
					if (m != n) & (dx <= distance) & (dy <= distance):
						wasVote = True
						if alives[m] == m:
							alives[m] = n
							votes += 1
						else:
							deaths += 1
							alives[m] = -1
							if numPlayers + deaths >= numPlots:
								leave = True
								if numPlayers + deaths > numPlots:
									raise ValueError, "Not enough plots left!   numPlots: "+str(numPlots)+"  deaths: "+str(deaths)+"  votes: "+str(votes)+"  players: "+str(numPlayers)
						break

		#Now place all starting positions
		n = 0
		for m in range(len(self.plotList)):
			###
			if n >= numPlayers:
				break
			if (alives[m] != -1):
			###
			###if self.plotList[m].vacant == False:
				sPlot = gameMap.plot(self.plotList[m].x,self.plotList[m].y)
				if sPlot.isWater() == True:
					raise ValueError, "Start plot is water!"
				playerID = self.playerList[n]
				player = gc.getPlayer(playerID)
				sPlot.setStartingPlot(True)
				player.setStartingPlot(sPlot,True)
				###
				self.plotList[m].vacant = False
				###
				n += 1

		self.CalculateStartingPlotValues()

		return
	def CalculateStartingPlotValues(self):
		gameMap = CyMap()
		numPlots = len(self.plotList)

		for n in range(numPlots):
			self.plotList[n].owner = -1
			self.plotList[n].totalValue = 0

		for n in range(numPlots):
			if self.plotList[n].vacant == True:
				continue
			self.plotList[n].totalValue = self.plotList[n].localValue
			self.plotList[n].numberOfOwnedCities = 0
			for m in range(numPlots):
				i = n * numPlots + m
				nDistance = self.distanceTable[i]
				if nDistance == -1:
					leastDistance = False
				else:
					leastDistance = True #Being optimistic, prove me wrong
				for p in range(numPlots):
					if p == n or self.plotList[p].vacant == True:
						continue
					ii = p * numPlots + m
					pDistance = self.distanceTable[ii]
					if pDistance > -1 and pDistance <= nDistance:
						leastDistance = False #Proven wrong
						break

				if leastDistance == True:
					self.plotList[n].totalValue += self.plotList[m].localValue / 23
					self.plotList[m].owner = self.plotList[n]
					self.plotList[m].distanceToOwner = nDistance
					self.plotList[n].numberOfOwnedCities += 1

		return
	def getDistance(self,x,y,dx,dy):
		xx = x - dx
		if abs(xx) > hm.mapWidth/2:
			xx = hm.mapWidth - abs(xx)

		distance = max(abs(xx),abs(y - dy))
		return distance
	def PrintPlotMap(self):
		gameMap = CyMap()
		print "Starting Plot Map"
		for y in range(hm.mapHeight - 1,-1,-1):
			lineString = ""
			for x in range(hm.mapWidth):
				inList = False
				for n in range(len(self.plotList)):
					if self.plotList[n].x == x and self.plotList[n].y == y:
						if self.plotList[n].plot().isWater() == True:
							if self.plotList[n].vacant == True:
								lineString += 'VV'
							else:
								lineString += 'OO'
						else:
							if self.plotList[n].vacant == True:
								lineString += 'vv'
							else:
								lineString += 'oo'
						inList = True
						break
				if inList == False:
					plot = gameMap.plot(x,y)
					if plot.isWater() == True:
						lineString += '.;'
					else:
						lineString += '[]'

			lineString += "-" + str(y)
			print lineString

		lineString = " "
		print lineString

		return
	def PrintPlotList(self):
		for n in range(len(self.plotList)):
			print str(n) + ' ' + str(self.plotList[n])
		return

	def PrintDistanceTable(self):
		print "Distance Table"
		lineString = "%(n)05d" % {"n":0} + ' '
		for n in range(len(self.plotList)):
			lineString += "%(n)05d" % {"n":n} + ' '
		print lineString
		lineString = ""
		for n in range(len(self.plotList)):
			lineString = "%(n)05d" % {"n":n} + ' '
			for m in range(len(self.plotList)):
				i = n * len(self.plotList) + m
				lineString += "%(d)05d" % {"d":self.distanceTable[i]} + ' '
			print lineString
		return

class StartPlot :
	def __init__(self,x,y,localValue):
		self.x = x
		self.y = y
		self.localValue = localValue
		self.totalValue = 0
		self.numberOfOwnedCities = 0
		self.distanceToOwner = -1
		self.nearestStart = -1
		self.vacant = True
		self.owner = None
		self.avgDistance = 0
		return
	def isCoast(self):
		gameMap = CyMap()
		plot = gameMap.plot(self.x,self.y)
		waterArea = plot.waterArea()
		if waterArea.isNone() == True or waterArea.isLake() == True:
			return False
		return True
	def plot(self):
		gameMap = CyMap()
		return gameMap.plot(self.x,self.y)
	def copy(self):
		cp = StartPlot(self.x,self.y,self.localValue)
		cp.totalValue = self.totalValue
		cp.numberOfOwnedCities = self.numberOfOwnedCities
		cp.distanceToOwner = self.distanceToOwner
		cp.nearestStart = self.nearestStart
		cp.vacant = self.vacant
		cp.owner = self.owner
		cp.avgDistance = self.avgDistance
		return cp
	def __str__(self):
		linestring = "x=%(x)3d,y=%(y)3d,localValue=%(lv)6d,totalValue =%(tv)6d, nearestStart=%(ad)6d" % \
		{"x":self.x,"y":self.y,"lv":self.localValue,"tv":self.totalValue,"ad":self.nearestStart}
		return linestring
'''
This function examines a lake area and removes ugly surrounding rivers. Any
river that is flowing away from the lake, or alongside the lake will be
removed. This function also returns a list of riverID's that flow into the
lake.
'''
def cleanUpLake(x,y):
	gc = CyGlobalContext()
	mmap = gc.getMap()
	riversIntoLake = list()
	plot = mmap.plot(x,y+1)#North
	if plot != 0 and plot.isNOfRiver() == True:
		plot.setNOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	if plot != 0 and plot.isWOfRiver() == True:
		if plot.getRiverNSDirection() == CardinalDirectionTypes.CARDINALDIRECTION_SOUTH:
			riversIntoLake.append(plot.getRiverID())
		else:
			plot.setWOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	plot = mmap.plot(x - 1,y)#West
	if plot != 0 and plot.isWOfRiver() == True:
		plot.setWOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	if plot != 0 and plot.isNOfRiver() == True:
		if plot.getRiverWEDirection() == CardinalDirectionTypes.CARDINALDIRECTION_EAST:
			riversIntoLake.append(plot.getRiverID())
		else:
			plot.setNOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	plot = mmap.plot(x + 1,y)#East
	if plot != 0 and plot.isNOfRiver() == True:
		if plot.getRiverWEDirection() == CardinalDirectionTypes.CARDINALDIRECTION_WEST:
			riversIntoLake.append(plot.getRiverID())
		else:
			plot.setNOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	plot = mmap.plot(x,y-1)#South
	if plot != 0 and plot.isWOfRiver() == True:
		if plot.getRiverNSDirection() == CardinalDirectionTypes.CARDINALDIRECTION_NORTH:
			riversIntoLake.append(plot.getRiverID())
		else:
			plot.setWOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	plot = mmap.plot(x-1,y+1)#Northwest
	if plot != 0 and plot.isWOfRiver() == True:
		if plot.getRiverNSDirection() == CardinalDirectionTypes.CARDINALDIRECTION_SOUTH:
			riversIntoLake.append(plot.getRiverID())
		else:
			plot.setWOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	if plot != 0 and plot.isNOfRiver() == True:
		if plot.getRiverWEDirection() == CardinalDirectionTypes.CARDINALDIRECTION_EAST:
			riversIntoLake.append(plot.getRiverID())
		else:
			plot.setNOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	plot = mmap.plot(x+1,y+1)#Northeast
	if plot != 0 and plot.isNOfRiver() == True:
		if plot.getRiverWEDirection() == CardinalDirectionTypes.CARDINALDIRECTION_WEST:
			riversIntoLake.append(plot.getRiverID())
		else:
			plot.setNOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	plot = mmap.plot(x-1,y-1)#Southhwest
	if plot != 0 and plot.isWOfRiver() == True:
		if plot.getRiverNSDirection() == CardinalDirectionTypes.CARDINALDIRECTION_NORTH:
			riversIntoLake.append(plot.getRiverID())
		else:
			plot.setWOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	#Southeast plot is not relevant

	return riversIntoLake

'''
It looks bad to have a lake, fed by a river, sitting right next to the coast.
This function tries to minimize that occurance by replacing it with a
natural harbor, which looks much better.
'''
def makeHarbor(x,y,oceanMap):
	oceanID = oceanMap.getOceanID()
	i = oceanMap.getIndex(x,y)
	if oceanMap.areaMap[i] != oceanID:
		return
	#N
	xx = x
	yy = y + 2
	ii = oceanMap.getIndex(xx,yy)
	if ii > -1 and \
	oceanMap.getAreaByID(oceanMap.areaMap[ii]).water == True and \
	oceanMap.areaMap[ii] != oceanID:
		makeChannel(x,y + 1)
		oceanMap.defineAreas(False)
		oceanID = oceanMap.getOceanID()
	#S
	xx = x
	yy = y - 2
	ii = oceanMap.getIndex(xx,yy)
	if ii > -1 and \
	oceanMap.getAreaByID(oceanMap.areaMap[ii]).water == True and \
	oceanMap.areaMap[ii] != oceanID:
		makeChannel(x,y - 1)
		oceanMap.defineAreas(False)
		oceanID = oceanMap.getOceanID()
	#E
	xx = x + 2
	yy = y
	ii = oceanMap.getIndex(xx,yy)
	if ii > -1 and \
	oceanMap.getAreaByID(oceanMap.areaMap[ii]).water == True and \
	oceanMap.areaMap[ii] != oceanID:
		makeChannel(x + 1,y)
		oceanMap.defineAreas(False)
		oceanID = oceanMap.getOceanID()
	#W
	xx = x - 2
	yy = y
	ii = oceanMap.getIndex(xx,yy)
	if ii > -1 and \
	oceanMap.getAreaByID(oceanMap.areaMap[ii]).water == True and \
	oceanMap.areaMap[ii] != oceanID:
		makeChannel(x - 1,y)
		oceanMap.defineAreas(False)
		oceanID = oceanMap.getOceanID()
	#NW
	xx = x - 1
	yy = y + 1
	ii = oceanMap.getIndex(xx,yy)
	if ii > -1 and \
	oceanMap.getAreaByID(oceanMap.areaMap[ii]).water == True and \
	oceanMap.areaMap[ii] != oceanID:
		makeChannel(x - 1,y)
		oceanMap.defineAreas(False)
		oceanID = oceanMap.getOceanID()
	#NE
	xx = x + 1
	yy = y + 1
	ii = oceanMap.getIndex(xx,yy)
	if ii > -1 and \
	oceanMap.getAreaByID(oceanMap.areaMap[ii]).water == True and \
	oceanMap.areaMap[ii] != oceanID:
		makeChannel(x + 1,y)
		oceanMap.defineAreas(False)
		oceanID = oceanMap.getOceanID()
	#SW
	xx = x - 1
	yy = y - 1
	ii = oceanMap.getIndex(xx,yy)
	if ii > -1 and \
	oceanMap.getAreaByID(oceanMap.areaMap[ii]).water == True and \
	oceanMap.areaMap[ii] != oceanID:
		makeChannel(x ,y - 1)
		oceanMap.defineAreas(False)
		oceanID = oceanMap.getOceanID()
	#NW
	xx = x - 1
	yy = y + 1
	ii = oceanMap.getIndex(xx,yy)
	if ii > -1 and \
	oceanMap.getAreaByID(oceanMap.areaMap[ii]).water == True and \
	oceanMap.areaMap[ii] != oceanID:
		makeChannel(x,y + 1)
		oceanMap.defineAreas(False)
		oceanID = oceanMap.getOceanID()
	return
def makeChannel(x,y):
	gc = CyGlobalContext()
	mmap = gc.getMap()
	terrainCoast = gc.getInfoTypeForString("TERRAIN_COAST")
	cleanUpLake(x,y)
	plot = mmap.plot(x,y)
	plot.setRiverID(-1)
	plot.setNOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	plot.setWOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	plot.setTerrainType(terrainCoast,True,True)
	i = hm.GetIndex(x,y)
	hm.plotMap[i] = hm.OCEAN
	return
def expandLake(x,y,riversIntoLake,oceanMap):
	class LakePlot :
		def __init__(self,x,y,altitude):
			self.x = x
			self.y = y
			self.altitude = altitude
	gc = CyGlobalContext()
	mmap = gc.getMap()
	terrainCoast = gc.getInfoTypeForString("TERRAIN_COAST")
	lakePlots = list()
	lakeNeighbors = list()
	lakeSize = 2
	for i in riversIntoLake:
		lakeSize += riverList[i]
	i = oceanMap.getIndex(x,y)
	desertModifier = 1.0
	if terr.terrainMap[i] == terr.DESERT or terr.terrainMap[i] == terr.HOTDESERT:
		desertModifier = DesertLakeModifier
	lakeSize = int(lakeSize * LakeSizePerRiverLength * desertModifier + 0.9999)
	start = LakePlot(x,y,hm.map[i])
	lakeNeighbors.append(start)
#    print "lakeSize",lakeSize
	while lakeSize > 0 and len(lakeNeighbors) > 0:
#        lakeNeighbors.sort(key=operator.attrgetter('altitude'),reverse=False)
		lakeNeighbors.sort(lambda x,y:cmp(x.altitude,y.altitude))
		currentLakePlot = lakeNeighbors[0]
		del lakeNeighbors[0]
		lakePlots.append(currentLakePlot)
		plot = mmap.plot(currentLakePlot.x,currentLakePlot.y)
		#if you are erasing a river to make a lake, make the lake smaller
		if plot.isNOfRiver() == True or plot.isWOfRiver() == True:
			lakeSize -= 1
		makeChannel(currentLakePlot.x,currentLakePlot.y)
		#Add valid neighbors to lakeNeighbors
		for n in range(4):
			if n == 0:#N
				xx = currentLakePlot.x
				yy = currentLakePlot.y + 1
				ii = oceanMap.getIndex(xx,yy)
			elif n == 1:#S
				xx = currentLakePlot.x
				yy = currentLakePlot.y - 1
				ii = oceanMap.getIndex(xx,yy)
			elif n == 2:#E
				xx = currentLakePlot.x + 1
				yy = currentLakePlot.y
				ii = oceanMap.getIndex(xx,yy)
			elif n == 3:#W
				xx = currentLakePlot.x - 1
				yy = currentLakePlot.y
				ii = oceanMap.getIndex(xx,yy)
			else:
				raise ValueError, "too many cardinal directions"
			if ii != -1:
				#if this neighbor is in water area, then quit
				areaID = oceanMap.areaMap[ii]
				if areaID == 0:
					raise ValueError, "areaID = 0 while generating lakes. This is a bug"
				for n in range(len(oceanMap.areaList)):
					if oceanMap.areaList[n].ID == areaID:
						if oceanMap.areaList[n].water == True:
#                            print "lake touched waterID = %(id)3d with %(ls)3d squares unused" % {'id':areaID,'ls':lakeSize}
#                            print "n = %(n)3d" % {"n":n}
#                            print str(oceanMap.areaList[n])
							return
				if rivers.riverMap[ii] != rivers.L and mmap.plot(xx,yy).isWater() == False:
					lakeNeighbors.append(LakePlot(xx,yy,hm.map[ii]))

		lakeSize -= 1
#    print "lake finished normally at %(x)2d,%(y)2d" % {"x":x,"y":y}
	return
###############################################################################
#functions that civ is looking for
###############################################################################
def getDescription():
		"""
		A map's Description is displayed in the main menu when players go to begin a game.
		For no description return an empty string.
		"""
		return "Random map that simulates earth-like plate tectonics, " +\
		"geostrophic winds and rainfall. Also, this map has a high likelyhood " +\
		"of 'New World' type lands only reachable on the open ocean. Be warned " +\
		"that these new lands are often home to terrifying hordes of " +\
		"mace-wielding barbarians. Any explorers you send here should " +\
		"be thoroughly briefed on the risks involved prior to embarcation. " +\
		"Use of the 'buddy system' is encouraged."

#This doesn't work with my river system so it is disabled. Some civs
#might start without a river. Boo hoo.
def normalizeAddRiver():
	return
def normalizeAddLakes():
	return
def normalizeAddGoodTerrain():
	return
def normalizeRemoveBadTerrain():
	return
def normalizeRemoveBadFeatures():
	return
def normalizeAddFoodBonuses():
	return
def normalizeAddExtras():
	return
def addLakes():
	print "Adding Lakes"
	gc = CyGlobalContext()
	mmap = gc.getMap()
	terrainCoast = gc.getInfoTypeForString("TERRAIN_COAST")
#    PrintFlowMap()
	oceanMap = Areamap(hm.mapWidth,hm.mapHeight)
	oceanMap.defineAreas(False)
##    oceanMap.PrintList(oceanMap.areaList)
##    oceanMap.PrintAreaMap()
	for y in range(hm.mapHeight):
		for x in range(hm.mapWidth):
			i = hm.GetIndex(x,y)
			if rivers.flowMap[i] == rivers.L:
				riversIntoLake = cleanUpLake(x,y)
				plot = mmap.plot(x,y)
				if len(riversIntoLake) > 0:
##                    plot.setRiverID(-1)
##                    plot.setNOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
##                    plot.setWOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
##                    #plot.setPlotType(PlotTypes.PLOT_OCEAN,True,True) setTerrain handles this already
##                    plot.setTerrainType(terrainCoast,True,True)
					expandLake(x,y,riversIntoLake,oceanMap)
				else:
					#no lake here, but in that case there should be no rivers either
					plot.setRiverID(-1)
					plot.setNOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
					plot.setWOfRiver(False,CardinalDirectionTypes.NO_CARDINALDIRECTION)
	oceanMap.defineAreas(False)
##    oceanMap.PrintList(oceanMap.areaList)
##    oceanMap.PrintAreaMap()
	for y in range(hm.mapHeight):
		for x in range(hm.mapWidth):
			i = hm.GetIndex(x,y)
			makeHarbor(x,y,oceanMap)
	return
def isAdvancedMap():
	"""
	Advanced maps only show up in the map script pulldown on the advanced menu.
	Return 0 if you want your map to show up in the simple singleplayer menu
	"""
	return 0
def isClimateMap():
	"""
	Uses the Climate options
	"""
	return 0

def isSeaLevelMap():
	"""
	Uses the Sea Level options
	"""
	return 0
def getTopLatitude():
	"Default is 90. 75 is past the Arctic Circle"
	return 80

def getBottomLatitude():
	"Default is -90. -75 is past the Antartic Circle"
	return -90
def getGridSize(argsList):
	"Adjust grid sizes for optimum results"
	grid_sizes = {
		WorldSizeTypes.WORLDSIZE_DUEL:		(12,8),
		WorldSizeTypes.WORLDSIZE_TINY:		(16,10),
		WorldSizeTypes.WORLDSIZE_SMALL:		(22,14),
		WorldSizeTypes.WORLDSIZE_STANDARD:	(26,16),
		WorldSizeTypes.WORLDSIZE_LARGE:		(32,20),
		WorldSizeTypes.WORLDSIZE_HUGE:		(36,24)
	}
	if (argsList[0] == -1): # (-1,) is passed to function on loads
		return []
	[eWorldSize] = argsList
	return grid_sizes[eWorldSize]

def generatePlotTypes():
	gc = CyGlobalContext()
	mmap = gc.getMap()
	iNumPlotsX = mmap.getGridWidth()
	iNumPlotsY = mmap.getGridHeight()
	PWRand.seed()

	# Map options
	global AllowNewWorld
	global DefaultBiome
	global AllowPangaea
	global LandmassStyle

	AllowNewWorld = CyMap().getCustomMapOption(0)
	DefaultBiome = CyMap().getCustomMapOption(1)
	AllowPangaea = CyMap().getCustomMapOption(2)
	LandmassStyle = CyMap().getCustomMapOption(3)

	hm.GenerateMap(iNumPlotsX,iNumPlotsY,LandPercent)

	tm.GenerateTempMap(getTopLatitude(),getBottomLatitude())
	rm.GenerateRainMap(getTopLatitude(),getBottomLatitude())
	terr.GenerateTerrainMap()

	plotTypes = [PlotTypes.PLOT_OCEAN] * (iNumPlotsX*iNumPlotsY)

	for i in range(iNumPlotsX*iNumPlotsY):
		mapLoc = hm.plotMap[i]
		if mapLoc == hm.PEAK:
			plotTypes[i] = PlotTypes.PLOT_PEAK
		elif mapLoc == hm.HILLS:
			plotTypes[i] = PlotTypes.PLOT_HILLS
		elif mapLoc == hm.LAND:
			plotTypes[i] = PlotTypes.PLOT_LAND
		else:
			plotTypes[i] = PlotTypes.PLOT_OCEAN
	print "Finished generating plot types."
	return plotTypes
def generateTerrainTypes():
	NiTextOut("Generating Terrain  ...")
	print "Adding Terrain"
	gc = CyGlobalContext()
	terrainDesert = gc.getInfoTypeForString("TERRAIN_DESERT")
	terrainHotDesert = gc.getInfoTypeForString("TERRAIN_DESERT")
	terrainPlains = gc.getInfoTypeForString("TERRAIN_PLAINS")
	terrainCoolPlains = gc.getInfoTypeForString("TERRAIN_PLAINS")
	terrainHotPlains = gc.getInfoTypeForString("TERRAIN_PLAINS")
	terrainIce = gc.getInfoTypeForString("TERRAIN_SNOW")
	terrainTundra = gc.getInfoTypeForString("TERRAIN_TUNDRA")
	terrainGrass = gc.getInfoTypeForString("TERRAIN_GRASS")
	terrainCoolGrass = gc.getInfoTypeForString("TERRAIN_GRASS")
	terrainHotGrass = gc.getInfoTypeForString("TERRAIN_GRASS")
	terrainHill = gc.getInfoTypeForString("TERRAIN_HILL")
	terrainCoast = gc.getInfoTypeForString("TERRAIN_COAST")
	terrainOcean = gc.getInfoTypeForString("TERRAIN_OCEAN")
	terrainDeepOcean = gc.getInfoTypeForString("TERRAIN_OCEAN")
	terrainPeak = gc.getInfoTypeForString("TERRAIN_PEAK")
	terrainDirt = gc.getInfoTypeForString("TERRAIN_DESERT")
	terrainHotDirt = gc.getInfoTypeForString("TERRAIN_DESERT")
	terrainHotSavanna = gc.getInfoTypeForString("TERRAIN_PLAINS")
	terrainDirtTundra = gc.getInfoTypeForString("TERRAIN_TUNDRA")
	terrainSavanna = gc.getInfoTypeForString("TERRAIN_PLAINS")
	terrainCoolSavanna = gc.getInfoTypeForString("TERRAIN_PLAINS")

	#tm.GenerateTempMap(80,-80)
	#rm.GenerateRainMap(80,-80)
	#terr.GenerateTerrainMap()
	terrainTypes = [0]*(hm.mapWidth*hm.mapHeight)
	for i in range(hm.mapWidth*hm.mapHeight):
		if terr.terrainMap[i] == terr.OCEAN:
			terrainTypes[i] = terrainOcean
		elif terr.terrainMap[i] == terr.DEEPOCEAN:
			terrainTypes[i] = terrainDeepOcean
		elif terr.terrainMap[i] == terr.COAST:
			terrainTypes[i] = terrainCoast
		elif terr.terrainMap[i] == terr.DESERT:
			terrainTypes[i] = terrainDesert
		elif terr.terrainMap[i] == terr.HOTDESERT:
			terrainTypes[i] = terrainHotDesert
		elif terr.terrainMap[i] == terr.PLAINS:
			terrainTypes[i] = terrainPlains
		elif terr.terrainMap[i] == terr.COOLPLAINS:
			terrainTypes[i] = terrainCoolPlains
		elif terr.terrainMap[i] == terr.HOTPLAINS:
			terrainTypes[i] = terrainHotPlains
		elif terr.terrainMap[i] == terr.GRASS:
			terrainTypes[i] = terrainGrass
		elif terr.terrainMap[i] == terr.COOLGRASS:
			terrainTypes[i] = terrainCoolGrass
		elif terr.terrainMap[i] == terr.HOTGRASS:
			terrainTypes[i] = terrainHotGrass
		elif terr.terrainMap[i] == terr.TUNDRA:
			terrainTypes[i] = terrainTundra
		elif terr.terrainMap[i] == terr.ICE:
			terrainTypes[i] = terrainIce
		elif terr.terrainMap[i] == terr.DIRT:
			terrainTypes[i] = terrainDirt
		elif terr.terrainMap[i] == terr.HOTDIRT:
			terrainTypes[i] = terrainHotDirt
		elif terr.terrainMap[i] == terr.HOTSAVANNA:
			terrainTypes[i] = terrainHotSavanna
		elif terr.terrainMap[i] == terr.SAVANNA:
			terrainTypes[i] = terrainSavanna
		elif terr.terrainMap[i] == terr.COOLSAVANNA:
			terrainTypes[i] = terrainCoolSavanna
		elif terr.terrainMap[i] == terr.DIRTTUNDRA:
			terrainTypes[i] = terrainDirtTundra
	print "Finished generating terrain types."
	terr.generateContinentMap()
	return terrainTypes
def doRiver(x,y,direction,riverID,riverLength):
	gc = CyGlobalContext()
	pmap = gc.getMap()
	startPlot = pmap.plot(x,y)
	riverPlot = 0
	adjPlot = 0
	if startPlot.getRiverID() == -1:
		startPlot.setRiverID(riverID)
	#meeting another river is the end of this river
	elif startPlot.getRiverID() != riverID:
		return riverLength
	if direction == rivers.N:
		riverPlot = startPlot
		adjPlot = pmap.plot(x+1,y)#East
		if adjPlot == 0:
			return riverLength
		if adjPlot.isWater() == True or riverPlot.isWater() == True or \
		riverPlot.isWOfRiver == True:
			return riverLength
		startPlot.setRiverID(riverID)
		riverPlot.setWOfRiver(True,CardinalDirectionTypes.CARDINALDIRECTION_NORTH)
		riverPlot = pmap.plot(x,y + 1)#North
	elif direction == rivers.E:
		riverPlot = pmap.plot(x+1,y)#East
		if riverPlot == 0:
			return riverLength
		adjPlot = pmap.plot(x+1,y-1)#SouthEast
		if adjPlot == 0:
			return riverLength
		if adjPlot.isWater() == True or riverPlot.isWater() == True or \
		riverPlot.isNOfRiver == True:
			return riverLength
		startPlot.setRiverID(riverID)
		riverPlot.setNOfRiver(True,CardinalDirectionTypes.CARDINALDIRECTION_EAST)
	elif direction == rivers.S:
		riverPlot = pmap.plot(x,y-1)#South
		if riverPlot == 0:
			return riverLength
		adjPlot = pmap.plot(x+1,y-1)#SouthEast
		if adjPlot == 0:
			return riverLength
		if riverPlot.isWOfRiver() == True or adjPlot.isWater() == True or \
		riverPlot.isWater() == True:
			return riverLength
		startPlot.setRiverID(riverID)
		riverPlot.setWOfRiver(True,CardinalDirectionTypes.CARDINALDIRECTION_SOUTH)
	elif direction == rivers.W:
		riverPlot = startPlot
		adjPlot = pmap.plot(x,y-1)#South
		if adjPlot == 0:
			return riverLength
		if riverPlot.isNOfRiver == True or adjPlot.isWater() == True or \
		riverPlot.isWater() == True:
			return riverLength
		startPlot.setRiverID(riverID)
		riverPlot.setNOfRiver(True,CardinalDirectionTypes.CARDINALDIRECTION_WEST)
		riverPlot = pmap.plot(x - 1,y)#West

	if riverPlot == 0:
		return riverLength#River off the edge of map
	#Get next direction rivers.riverMap should cooincide with the SE corner of plot
	x = riverPlot.getX()
	y = riverPlot.getY()
	sePlot = pmap.plot(x+1,y-1)#SE
	if sePlot != 0 and sePlot.isWater() == True:
		return riverLength#River into the ocean
	i = hm.GetIndex(x,y)
	direction = rivers.riverMap[i]
	if direction != rivers.L or direction != rivers.O:
		riverLength = doRiver(x,y,direction,riverID,riverLength + 1)
	return riverLength

riverList = list()
def addRivers():
	NiTextOut("Adding Rivers....")
	print "Adding Rivers"
	gc = CyGlobalContext()
	pmap = gc.getMap()
	nextRiverID = 0
	rivers.GenerateRiverMap()
##    PrintRiverAndTerrainAlign()
	for y in range(hm.mapHeight):
		for x in range(hm.mapWidth):
			i = hm.GetIndex(x,y)
			#Check for a river source that has a direction and also that
			#has 4 cardinal neighbors that have no river or a river that
			#leads away
			source = True #until proven wrong
			if rivers.riverMap[i] != rivers.O and rivers.riverMap[i] != rivers.L:
				xx = x
				yy = y + 1
				ii = hm.GetIndex(xx,yy)
				if rivers.riverMap[ii] == rivers.S:
					source = False
				xx = x
				yy = y - 1
				ii = hm.GetIndex(xx,yy)
				if rivers.riverMap[ii] == rivers.N:
					source = False
				xx = x - 1
				yy = y
				ii = hm.GetIndex(xx,yy)
				if rivers.riverMap[ii] == rivers.E:
					source = False
				xx = x + 1
				yy = y
				ii = hm.GetIndex(xx,yy)
				if rivers.riverMap[ii] == rivers.W:
					source = False
			else:
				source = False
			if source == True:
				riverLength = doRiver(x,y,rivers.riverMap[i],nextRiverID,0)
				riverList.append(riverLength)
				nextRiverID += 1

	#peaks and rivers don't always mix well graphically, so lets eliminate
	#these potential glitches. Basically if there are adjacent peaks on both
	#sides of a river, either in a cardinal direction or diagonally, they
	#will look bad.
	for y in range(hm.mapHeight):
		for x in range(hm.mapWidth):
			plot = pmap.plot(x,y)
			if plot.isPeak() == True:
				if plot.isNOfRiver() == True:
					for xx in range(x - 1,x + 2,1):
						yy = y - 1
						if yy < 0:
							break
						#wrap in x direction
						if xx == -1:
							xx = hm.mapWidth - 1
						elif xx == hm.mapWidth:
							xx = 0
						newPlot = pmap.plot(xx,yy)
						if newPlot.isPeak():
							plot.setPlotType(PlotTypes.PLOT_HILLS,True,True)
							break
			#possibly changed so checked again
			if plot.isPeak() == True:
				if plot.isWOfRiver() == True:
					for yy in range(y - 1,y + 2,1):
						xx = x + 1
						if xx == hm.mapWidth:
							xx = 0
						#do not wrap in y direction
						if yy == -1:
							continue
						elif yy == hm.mapHeight:
							continue
						newPlot = pmap.plot(xx,yy)
						if newPlot.isPeak():
							plot.setPlotType(PlotTypes.PLOT_HILLS,True,True)
							break

	print nextRiverID," rivers generated."

def addFeatures():
	NiTextOut("Generating Features  ...")
	print "Adding Features"
	gc = CyGlobalContext()
	mmap = gc.getMap()
	featureIce = gc.getInfoTypeForString("FEATURE_ICE")
	featureJungle = gc.getInfoTypeForString("FEATURE_JUNGLE")
	featureOasis = gc.getInfoTypeForString("FEATURE_OASIS")
	featureFloodPlains = gc.getInfoTypeForString("FEATURE_FLOOD_PLAINS")
	featureForest = gc.getInfoTypeForString("FEATURE_FOREST")
	FORESTLEAFY = 0
	FORESTEVERGREEN = 1
	FORESTSNOWY = 2
	#First create ice.
	#100% chance for top and bottom row, higher chance on coast tiles
	iceChance = 1.5
	for y in range(0,hm.chunkSize,1):
		for x in range(hm.mapWidth):
			plot = mmap.plot(x,y)
			i = hm.GetIndex(x,y)
			if plot != 0 and plot.isWater() == True:
				iceProb = iceChance
				if tm.tempMap[i] > TundraTemp:
					iceProb * 0.25
				if terr.terrainMap[i] == terr.DEEPOCEAN:
					iceProb *= 0.75
				elif terr.terrainMap[i] == terr.COAST:
					iceProb *= 2

				if PWRand.random() < iceProb:
					plot.setFeatureType(featureIce,0)
		iceChance *= 0.66

	iceChance = 1.5
	for y in range(hm.mapHeight - 1,hm.mapHeight - hm.chunkSize,-1):
		for x in range(hm.mapWidth):
			plot = mmap.plot(x,y)
			i = hm.GetIndex(x,y)
			if plot != 0 and plot.isWater() == True:
				iceProb = iceChance
				if tm.tempMap[i] > TundraTemp:
					iceProb * 0.25
				if terr.terrainMap[i] == terr.DEEPOCEAN:
					iceProb *= 0.75
				elif terr.terrainMap[i] == terr.COAST:
					iceProb *= 2

				if PWRand.random() < iceProb:
					plot.setFeatureType(featureIce,0)
		iceChance *= 0.66

	#Now plant forest or jungle and place floodplains and oasis
	for y in range(hm.mapHeight):
		for x in range(hm.mapWidth):
			i = hm.GetIndex(x,y)
			plot = mmap.plot(x,y)

			altRain = rm.rainMap[i]
			#if CyMap().getCustomMapOption(1) == 1: #Tropical
			#	altRain *= 1.1
			#elif CyMap().getCustomMapOption(1) == 2: #Arid
			#	altRain *= 0.9

			#Floodplains
			#We also force forests/jungles on riversides if no floodplain is created.
			if plot.isRiver() == True and hm.plotMap[i] != hm.PEAK and plot.isWater() == False:
				if hm.plotMap[i] != hm.HILLS and rivers.isFloodPlainThreshold(x,y) == True and\
				(terr.terrainMap[i] == terr.DESERT or terr.terrainMap[i] == terr.HOTDESERT):
					plot.setFeatureType(featureFloodPlains,0)
				elif hm.plotMap[i] != hm.HILLS and rivers.isFloodPlainThreshold(x,y) == True and\
				(terr.terrainMap[i] == terr.DIRT or terr.terrainMap[i] == terr.HOTDIRT):
					plot.setFeatureType(featureFloodPlains,0)
				else:
					if terr.terrainMap[i] == terr.HOTGRASS:
						plot.setFeatureType(featureJungle,0)
					elif terr.terrainMap[i] == terr.GRASS or terr.terrainMap[i] == terr.PLAINS or\
					terr.terrainMap[i] == terr.HOTPLAINS or terr.terrainMap[i] == terr.HOTSAVANNA or\
					terr.terrainMap[i] == terr.SAVANNA:
						plot.setFeatureType(featureForest,FORESTLEAFY)
					elif terr.terrainMap[i] == terr.COOLGRASS or terr.terrainMap[i] == terr.COOLPLAINS or\
					terr.terrainMap[i] == terr.COOLSAVANNA:
						plot.setFeatureType(featureForest,FORESTEVERGREEN)
					elif terr.terrainMap[i] == terr.TUNDRA or terr.terrainMap[i] == terr.DIRTTUNDRA:
						plot.setFeatureType(featureForest,FORESTSNOWY)
			#forest and jungle
			elif plot.isWater() == False and hm.plotMap[i] != hm.PEAK:
				if altRain > terr.plainsThreshold*1.5:#jungle
					if terr.terrainMap[i] == terr.HOTGRASS:
						plot.setFeatureType(featureJungle,0)
					elif terr.terrainMap[i] == terr.GRASS:
						plot.setFeatureType(featureForest,FORESTLEAFY)
					elif terr.terrainMap[i] == terr.COOLGRASS:
						plot.setFeatureType(featureForest,FORESTEVERGREEN)
					elif terr.terrainMap[i] == terr.TUNDRA or terr.terrainMap[i] == terr.DIRTTUNDRA:
						plot.setFeatureType(featureForest,FORESTSNOWY)
				elif altRain > terr.desertThreshold:#forest
					if altRain > PWRand.random() * terr.plainsThreshold * 2:
						if terr.terrainMap[i] == terr.COOLSAVANNA and PWRand.random() < 0.5:
							plot.setFeatureType(featureForest,FORESTLEAFY)
						elif terr.terrainMap[i] == terr.HOTPLAINS and PWRand.random() < 0.25:
							plot.setFeatureType(featureForest,FORESTLEAFY)
						elif terr.terrainMap[i] == terr.GRASS or terr.terrainMap[i] == terr.PLAINS:
							plot.setFeatureType(featureForest,FORESTLEAFY)
						elif terr.terrainMap[i] == terr.COOLGRASS or terr.terrainMap[i] == terr.COOLPLAINS:
							plot.setFeatureType(featureForest,FORESTEVERGREEN)
						elif terr.terrainMap[i] == terr.TUNDRA or terr.terrainMap[i] == terr.DIRTTUNDRA:
							plot.setFeatureType(featureForest,FORESTSNOWY)
			#Oasis
			elif (terr.terrainMap[i] == terr.DESERT or terr.terrainMap[i] == terr.HOTDESERT) and\
			hm.plotMap[i] != hm.HILLS and hm.plotMap[i] != hm.PEAK and plot.isWater() == False:
				if plot.isRiver() == false:
					#is this square surrounded by desert?
					foundNonDesert = False
					#print "trying to place oasis"
					for yy in range(y - 1,y + 2):
						for xx in range(x - 1,x + 2):
							ii = hm.GetIndex(xx,yy)
							surPlot = mmap.plot(xx,yy)
							if terr.terrainMap[ii] != terr.DESERT or terr.terrainMap[ii] != terr.HOTDESERT and\
							hm.plotMap[ii] != hm.PEAK:
								foundNonDesert = True
							elif surPlot == 0:
								foundNonDesert = True
							elif surPlot.isWater() == True:
								foundNonDesert = True
							elif surPlot.getFeatureType() == featureOasis:
								foundNonDesert = True
					if foundNonDesert == False:
						if PWRand.random() < OasisChance:
							plot.setFeatureType(featureOasis,0)

	return
def addBonuses():
	bp.AddBonuses()
	return
def assignStartingPlots():
	gc = CyGlobalContext()
	gameMap = CyMap()
	iPlayers = gc.getGame().countCivPlayersEverAlive()
	spf.SetStartingPlots()


##############################################################################
#debugging stuff
##############################################################################
def PrintWindDirection(we,ns):
	dString = "-"
	if(ns == 1):
		dString += '^'
	else:
		dString += 'v'
	if(we == 1):
		dString += '>'
	else:
		dString += '<'
	return dString

def PrintDistMap(distMap,distWidth,distHeight):
	print "Local Distribution Map"
	for y in range(distHeight - 1,-1,-1):
		lineString = ""
		for x in range(distWidth):
			mapLoc = distMap[y * distWidth + x]
			lineString += '%(mapLoc)0.2f ' % {'mapLoc' : mapLoc}
		print lineString

	lineString = " "
	print lineString
	return
def PrintHeightMap(hm):
	print "Height Map"
	for y in range(hm.mapHeight):
		lineString = ""
		for x in range(hm.mapWidth):
			mapLoc = hm.map[hm.GetIndex(x,y)]
			if mapLoc > hm.hillLevel:
				lineString += 'P'
			elif mapLoc > hm.flatLevel:
				lineString += 'H'
			elif mapLoc > hm.seaLevel:
				lineString += 'F'
			else:
				lineString += 'O'
		print lineString
	lineString = " "
	print lineString

	return
def PrintPlateMap(hm):
	print "Plate Map"
	for y in range(hm.mapHeight):
		lineString = ""
		for x in range(hm.mapWidth):
			mapLoc = hm.plateMap[hm.GetIndex(x,y)]
			lineString += chr(mapLoc + 33)
		print lineString
	lineString = " "
	print lineString

	return
def PrintInfluMap(hm):
	print "Influence Map"
	for y in range(hm.mapHeight):
		lineString = ""
		for x in range(hm.mapWidth):
			mapLoc = hm.influMap[hm.GetIndex(x,y)]
			lineString += chr(int(mapLoc*10) - 8 + 48)
		print lineString

	lineString = " "
	print lineString
	return
def PrintPlotMap(hm):
	print "Plot Map"
	for y in range(hm.mapHeight - 1,-1,-1):
		lineString = ""
		for x in range(hm.mapWidth):
			mapLoc = hm.plotMap[hm.GetIndex(x,y)]
			if mapLoc == hm.PEAK:
				lineString += 'P'
			elif mapLoc == hm.HILLS:
				lineString += 'H'
			elif mapLoc == hm.LAND:
				lineString += 'F'
			else:
				lineString += '.'
		print lineString
	lineString = " "
	print lineString

	return
def PrintTempMap(tm,tMap):
	print "Temperature Map"
	for y in range(hm.mapHeight - 1,-1,-1):
		lineString = ""
		for x in range(hm.mapWidth):
			mapLoc = tMap[hm.GetIndex(x,y)]
			i = hm.GetIndex(x,y)
			if hm.plotMap[i] == hm.OCEAN:
				lineString += '.'
			else:
				lineString += chr(int(mapLoc*10) + 48)
		lineString += "-" + rm.wz.GetZoneName(tm.wz.GetZone(y))
		print lineString

	lineString = " "
	print lineString
	return
def PrintRainMap(rm,rMap,bOcean):
	print "Rain Map"
	for y in range(hm.mapHeight - 1,-1,-1):
		lineString = ""
		for x in range(hm.mapWidth):
			if bOcean:
				if hm.plotMap[hm.GetIndex(x,y)] == hm.OCEAN:
					mapLoc = rMap[hm.GetIndex(x,y)]
					lineString += chr(int(mapLoc*10) + 48)
				else:
					lineString += 'X'
			else:
				if hm.plotMap[hm.GetIndex(x,y)] == hm.OCEAN:
					lineString += '.'
				else:
					mapLoc = rMap[hm.GetIndex(x,y)]
					lineString += chr(int(mapLoc*10) + 48)
		we,ns = rm.wz.GetWindDirectionsInZone(rm.wz.GetZone(y))
		lineString += "-" + rm.wz.GetZoneName(rm.wz.GetZone(y)) + PrintWindDirection(we,ns)
		print lineString

	lineString = " "
	print lineString
	return
def PrintTruncRainMap(rm,rMap,bOcean):
	print "Truncated Rain Map"
	minV = 1.0
	for i in range(hm.mapWidth*hm.mapHeight):
		if rMap[i] > 0.0 and rMap[i] < minV:
			minV = rMap[i]
	print "min value",minV
	for y in range(hm.mapHeight - 1,-1,-1):
		lineString = ""
		for x in range(hm.mapWidth):
			if bOcean:
				if hm.plotMap[hm.GetIndex(x,y)] == hm.OCEAN:
					mapLoc = rMap[hm.GetIndex(x,y)]
					lineString += chr(int(mapLoc*10) + 48)
				else:
					lineString += 'X'
			else:
				if hm.plotMap[hm.GetIndex(x,y)] == hm.OCEAN:
					lineString += '.'
				else:
					mapLoc = rMap[hm.GetIndex(x,y)]
					if mapLoc == minV:
						lineString += 'X'
					else:
						mapLoc *= 10
						if mapLoc > .95:
							mapLoc = .95
						lineString += chr(int(mapLoc*10) + 48)
		we,ns = rm.wz.GetWindDirectionsInZone(rm.wz.GetZone(y))
		lineString += "-" + rm.wz.GetZoneName(rm.wz.GetZone(y)) + PrintWindDirection(we,ns)
		print lineString

	lineString = " "
	print lineString
	return
def PrintTerrainMap():
	print "Terrain Map"
	for y in range(hm.mapHeight - 1,-1,-1):
		lineString = ""
		for x in range(hm.mapWidth):
			mapLoc = terr.terrainMap[hm.GetIndex(x,y)]
			if mapLoc == terr.OCEAN:
				lineString += '.'
			elif mapLoc == terr.COAST:
				lineString += '.'
			elif mapLoc == terr.DESERT:
				lineString += 'D'
			elif mapLoc == terr.GRASS:
				lineString += 'R'
			elif mapLoc == terr.PLAINS:
				lineString += 'P'
			elif mapLoc == terr.TUNDRA:
				lineString += 'T'
			elif mapLoc == terr.ICE:
				lineString += 'I'
		lineString += "-" + rm.wz.GetZoneName(rm.wz.GetZone(y))
		print lineString
	lineString = " "
	print lineString

def PrintRiverMap():
	print "River Map"
	for y in range(hm.mapHeight - 1,-1,-1):
		lineString = ""
		for x in range(hm.mapWidth):
			mapLoc = rivers.riverMap[hm.GetIndex(x,y)]
			if mapLoc == rivers.O:
				lineString += '.'
			elif mapLoc == rivers.L:
				lineString += 'L'
			elif mapLoc == rivers.N:
				lineString += 'N'
			elif mapLoc == rivers.S:
				lineString += 'S'
			elif mapLoc == rivers.E:
				lineString += 'E'
			elif mapLoc == rivers.W:
				lineString += 'W'
		lineString += "-" + rm.wz.GetZoneName(rm.wz.GetZone(y))
		print lineString
	lineString = " "
	print lineString

def PrintFlowMap():
	print "Flow Map"
	for y in range(hm.mapHeight - 1,-1,-1):
		lineString = ""
		for x in range(hm.mapWidth):
			mapLoc = rivers.flowMap[hm.GetIndex(x,y)]
			if mapLoc == rivers.O:
				lineString += '.'
			elif mapLoc == rivers.L:
				lineString += 'L'
			elif mapLoc == rivers.N:
				lineString += 'N'
			elif mapLoc == rivers.S:
				lineString += 'S'
			elif mapLoc == rivers.E:
				lineString += 'E'
			elif mapLoc == rivers.W:
				lineString += 'W'
		lineString += "-" + rm.wz.GetZoneName(rm.wz.GetZone(y))
		print lineString
	lineString = " "
	print lineString
def PrintRiverAndTerrainAlign():
	print "River Alignment Check"
	for y in range(hm.mapHeight - 1,-1,-1):
		lineString1 = ""
		lineString2 = ""
		for x in range(hm.mapWidth):
			mapLoc = terr.terrainMap[hm.GetIndex(x,y)]
			if mapLoc == terr.OCEAN:
				lineString1 += 'O.'
			elif mapLoc == terr.COAST:
				lineString1 += 'O.'
			elif mapLoc == terr.DESERT:
				lineString1 += 'D.'
			elif mapLoc == terr.GRASS:
				lineString1 += 'R.'
			elif mapLoc == terr.PLAINS:
				lineString1 += 'P.'
			elif mapLoc == terr.TUNDRA:
				lineString1 += 'T.'
			elif mapLoc == terr.ICE:
				lineString1 += 'I.'
			mapLoc = rivers.riverMap[hm.GetIndex(x,y)]
			if mapLoc == rivers.O:
				lineString2 += '..'
			elif mapLoc == rivers.L:
				lineString2 += '.L'
			elif mapLoc == rivers.N:
				lineString2 += '.N'
			elif mapLoc == rivers.S:
				lineString2 += '.S'
			elif mapLoc == rivers.E:
				lineString2 += '.E'
			elif mapLoc == rivers.W:
				lineString2 += '.W'
		lineString1 += "-" + rm.wz.GetZoneName(rm.wz.GetZone(y))
		lineString2 += "-" + rm.wz.GetZoneName(rm.wz.GetZone(y))
		print lineString1
		print lineString2
	lineString1 = " "
	print lineString1

def PrintAvgRainMap():
	print "Avg Rain Map"
	for y in range(hm.mapHeight):
		lineString = ""
		for x in range(hm.mapWidth):
				if hm.plotMap[hm.GetIndex(x,y)] == hm.OCEAN:
					lineString += '.'
				else:
					mapLoc = rivers.averageRainfallMap[hm.GetIndex(x,y)]
					lineString += chr(int(mapLoc*10) + 48)
		lineString += "-" + rm.wz.GetZoneName(rm.wz.GetZone(y))
		print lineString

	lineString = " "
	print lineString
	return
def getNumCustomMapOptions():
	return 4

def getCustomMapOptionName(argsList):
	[iOption] = argsList
	option_names = {
		0:	"TXT_KEY_MAP_SCRIPT_NEW_WORLD",
		1:	"TXT_KEY_MAP_SCRIPT_DEFAULT_BIOME",
		2:	"TXT_KEY_MAP_SCRIPT_SUPPRESS_PANGAEA",
		3:	"TXT_KEY_MAP_SCRIPT_LANDMASS_STYLE"
		}
	translated_text = unicode(CyTranslator().getText(option_names[iOption], ()))
	return translated_text

def getNumCustomMapOptionValues(argsList):
	[iOption] = argsList
	option_values = {
		0:	2,
		1:	4,
		2:	2,
		3:	4
		}
	return option_values[iOption]

def getCustomMapOptionDescAt(argsList):
	[iOption, iSelection] = argsList
	selection_names = {
		0:	{
			0: "TXT_KEY_MAP_SCRIPT_NEW_WORLD_NO",
			1: "TXT_KEY_MAP_SCRIPT_NEW_WORLD_YES"
			},
		1:	{
			0: "TXT_KEY_MAP_SCRIPT_BIOME_NATURAL",
			1: "TXT_KEY_MAP_SCRIPT_BIOME_TROPICAL",
			2: "TXT_KEY_MAP_SCRIPT_BIOME_ARID",
			3: "TXT_KEY_MAP_SCRIPT_BIOME_COOL"
			},
		2:	{
			0: "TXT_KEY_MAP_SCRIPT_SUPPRESS_PANGAEA_YES",
			1: "TXT_KEY_MAP_SCRIPT_SUPPRESS_PANGAEA_NO"
			},
		3:	{
			0: "TXT_KEY_MAP_SCRIPT_LANDMASS_NATURAL",
			1: "TXT_KEY_MAP_SCRIPT_LANDMASS_PANGAEA",
			2: "TXT_KEY_MAP_SCRIPT_LANDMASS_TERRA",
			3: "TXT_KEY_MAP_SCRIPT_LANDMASS_EARTHLY"
			}
		}
	translated_text = unicode(CyTranslator().getText(selection_names[iOption][iSelection], ()))
	return translated_text

def getCustomMapOptionDefault(argsList):
	[iOption] = argsList
	option_defaults = {
		0:	1,
		1:	0,
		2:	0,
		3:	0
		}
	return option_defaults[iOption]

class PangaeaMultilayeredFractal(CvMapGeneratorUtil.MultilayeredFractal):
	# Subclass. Only the controlling function overridden in this case.
	def generatePlotsByRegion(self):
		# Sirian's MultilayeredFractal class, controlling function.
		# You -MUST- customize this function for each use of the class.
		#
		# The following grain matrix is specific to Pangaea.py
		sizekey = self.map.getWorldSize()
		sizevalues = {
			WorldSizeTypes.WORLDSIZE_DUEL:      3,
			WorldSizeTypes.WORLDSIZE_TINY:      3,
			WorldSizeTypes.WORLDSIZE_SMALL:     4,
			WorldSizeTypes.WORLDSIZE_STANDARD:  4,
			WorldSizeTypes.WORLDSIZE_LARGE:     4,
			WorldSizeTypes.WORLDSIZE_HUGE:      5
			}
		grain = sizevalues[sizekey]

		# Sea Level adjustment (from user input), limited to value of 5%.
		sea = self.gc.getSeaLevelInfo(self.map.getSeaLevel()).getSeaLevelChange()
		sea = min(sea, 5)
		sea = max(sea, -5)

		# The following regions are specific to Pangaea.py
		mainWestLon = 0.2
		mainEastLon = 0.8
		mainSouthLat = 0.2
		mainNorthLat = 0.8
		subcontinentDimension = 0.4
		bSouthwardShift = False

		subcontinentDimension = 0.3
		# Shift mainland north or south?
		global shiftRoll
		shiftRoll = self.dice.get(2, "Shift - Pangaea PYTHON")
		if shiftRoll == 1:
			mainNorthLat += 0.075
			mainSouthLat += 0.075
		else:
			mainNorthLat -= 0.075
			mainSouthLat -= 0.075
			bSouthwardShift = True
		# Define potential subcontinent slots (regional definitions).
		numSubcontinents = 4 + self.dice.get(3, "Subcontinents - Pangaea PYTHON")
		# List values: [westLon, southLat, vertRange, horzRange, southShift]
		scValues = [[0.05, 0.575, 0.0, 0.0, 0.15],
			    [0.05, 0.275, 0.0, 0.0, 0.15],
			    [0.2, 0.175, 0.0, 0.0, 0.15],
			    [0.5, 0.175, 0.0, 0.0, 0.15],
			    [0.65, 0.575, 0.0, 0.0, 0.15],
			    [0.65, 0.275, 0.0, 0.0, 0.15],
			    [0.2, 0.675, 0.0, 0.0, 0.15],
			    [0.5, 0.675, 0.0, 0.0, 0.15]
			    ]

		# Generate the main land mass, first pass (to vary shape).
		mainWestX = int(self.iW * mainWestLon)
		mainEastX = int(self.iW * mainEastLon)
		mainNorthY = int(self.iH * mainNorthLat)
		mainSouthY = int(self.iH * mainSouthLat)
		mainWidth = mainEastX - mainWestX + 1
		mainHeight = mainNorthY - mainSouthY + 1

		mainWater = 55+sea

		self.generatePlotsInRegion(mainWater,
		                           mainWidth, mainHeight,
		                           mainWestX, mainSouthY,
		                           2, grain,
		                           self.iHorzFlags, self.iTerrainFlags,
		                           -1, -1,
		                           True, 15,
		                           2, False,
		                           False
		                           )

		# Second pass (to ensure cohesion).
		second_layerHeight = mainHeight/2
		second_layerWestX = mainWestX + mainWidth/10
		second_layerEastX = mainEastX - mainWidth/10
		second_layerWidth = second_layerEastX - second_layerWestX + 1
		second_layerNorthY = mainNorthY - mainHeight/4
		second_layerSouthY = mainSouthY + mainHeight/4

		second_layerWater = 60+sea

		self.generatePlotsInRegion(second_layerWater,
		                           second_layerWidth, second_layerHeight,
		                           second_layerWestX, second_layerSouthY,
		                           1, grain,
		                           self.iHorzFlags, self.iTerrainFlags,
		                           -1, -1,
		                           True, 15,
		                           2, False,
		                           False
		                           )

		# Add subcontinents.
		# Subcontinents can be akin to India/Alaska, Europe, or the East Indies.
		while numSubcontinents > 0:
			# Choose a slot for this subcontinent.
			if len(scValues) > 1:
				scIndex = self.dice.get(len(scValues), "Subcontinent Placement - Pangaea PYTHON")
			else:
				scIndex = 0
			[scWestLon, scSouthLat, scVertRange, scHorzRange, scSouthShift] = scValues[scIndex]
			scWidth = int(subcontinentDimension * self.iW)
			scHeight = int(subcontinentDimension * self.iH)
			scHorzShift = 0; scVertShift = 0
			if scHorzRange > 0.0:
				scHorzShift = self.dice.get(int(self.iW * scHorzRange), "Subcontinent Variance - Terra PYTHON")
			if scVertRange > 0.0:
				scVertShift = self.dice.get(int(self.iW * scVertRange), "Subcontinent Variance - Terra PYTHON")
			scWestX = int(self.iW * scWestLon) + scHorzShift
			scEastX = scWestX + scWidth
			if scEastX >= self.iW: # Trouble! Off the right hand edge!
				while scEastX >= self.iW:
					scWidth -= 1
					scEastX = scWestX + scWidth
			scSouthY = int(self.iH * scSouthLat) + scVertShift
			# Check for southward shift.
			if bSouthwardShift:
				scSouthY -= int(self.iH * scSouthShift)
			scNorthY = scSouthY + scHeight
			if scNorthY >= self.iH: # Trouble! Off the top edge!
				while scNorthY >= self.iH:
					scHeight -= 1
					scNorthY = scSouthY + scHeight

			scShape = self.dice.get(5, "Subcontinent Shape - Terra PYTHON")
			if scShape > 1: # Regular subcontinent.
				scWater = 55+sea; scGrain = 1; scRift = -1
			elif scShape == 1: # Irregular subcontinent.
				scWater = 66+sea; scGrain = 2; scRift = 2
			else: # scShape == 0, Archipelago subcontinent.
				scWater = 77+sea; scGrain = grain; scRift = -1

			self.generatePlotsInRegion(scWater,
			                           scWidth, scHeight,
			                           scWestX, scSouthY,
			                           scGrain, grain,
			                           self.iRoundFlags, self.iTerrainFlags,
			                           6, 6,
			                           True, 7,
			                           scRift, False,
			                           False
			                           )

			del scValues[scIndex]
			numSubcontinents -= 1

		# All regions have been processed. Plot Type generation completed.
		return self.wholeworldPlotTypes

class TerraMultilayeredFractal(CvMapGeneratorUtil.MultilayeredFractal):
	# Subclass. Only the controlling function overridden in this case.
	def generatePlotsByRegion(self):
		# Sirian's MultilayeredFractal class, controlling function.
		# You -MUST- customize this function for each use of the class.
		#
		# The following grain matrix is specific to Terra.py
		sizekey = self.map.getWorldSize()
		sizevalues = {
			WorldSizeTypes.WORLDSIZE_DUEL:      (3,2,1,2),
			WorldSizeTypes.WORLDSIZE_TINY:      (3,2,1,2),
			WorldSizeTypes.WORLDSIZE_SMALL:     (4,2,1,2),
			WorldSizeTypes.WORLDSIZE_STANDARD:  (4,2,1,2),
			WorldSizeTypes.WORLDSIZE_LARGE:     (4,2,1,2),
			WorldSizeTypes.WORLDSIZE_HUGE:      (5,2,1,2)
			}
		(archGrain, contGrain, gaeaGrain, eurasiaGrain) = sizevalues[sizekey]

		# Sea Level adjustment (from user input), limited to value of 5%.
		sea = self.gc.getSeaLevelInfo(self.map.getSeaLevel()).getSeaLevelChange()
		sea = min(sea, 5)
		sea = max(sea, -5)

		# The following regions are specific to Terra.py
		newworldWestLon = 0.05
		newworldEastLon = 0.35
		eurasiaWestLon = 0.45
		eurasiaEastLon = 0.95
		eurasiaNorthLat = 0.95
		eurasiaSouthLat = 0.45
		thirdworldDimension = 0.125
		thirdworldNorthLat = 0.35
		thirdworldSouthLat = 0.05
		subcontinentLargeHorz = 0.2
		subcontinentLargeVert = 0.32
		subcontinentLargeNorthLat = 0.6
		subcontinentLargeSouthLat = 0.28
		subcontinentSmallDimension = 0.125
		subcontinentSmallNorthLat = 0.525
		subcontinentSmallSouthLat = 0.4

		# Dice rolls to randomize the quadrants (specific to Terra.py's regions)
		roll1 = self.dice.get(2, "Eurasian Hemisphere N/S - Terra PYTHON")
		if roll1 == 1:
			eurasiaNorthLat -= 0.4; eurasiaSouthLat -= 0.4
			thirdworldNorthLat += 0.6; thirdworldSouthLat += 0.6
			subcontinentLargeNorthLat += 0.12; subcontinentLargeSouthLat += 0.12
			subcontinentSmallNorthLat += 0.075; subcontinentSmallSouthLat += 0.075
		roll2 = self.dice.get(2, "Eurasian Hemisphere E/W - Terra PYTHON")
		if roll2 == 1:
			newworldWestLon += 0.6; newworldEastLon += 0.6
			eurasiaWestLon -= 0.4; eurasiaEastLon -= 0.4

		# Simulate the Old World - a large continent akin to Earth's Eurasia.
		NiTextOut("Generating the Old World (Python Terra) ...")
		# Set dimensions of the Old World region (specific to Terra.py)
		eurasiaWestX = int(self.iW * eurasiaWestLon)
		eurasiaEastX = int(self.iW * eurasiaEastLon)
		eurasiaNorthY = int(self.iH * eurasiaNorthLat)
		eurasiaSouthY = int(self.iH * eurasiaSouthLat)
		eurasiaWidth = eurasiaEastX - eurasiaWestX + 1
		eurasiaHeight = eurasiaNorthY - eurasiaSouthY + 1

		eurasiaWater = 55+sea

		self.generatePlotsInRegion(eurasiaWater,
		                           eurasiaWidth, eurasiaHeight,
		                           eurasiaWestX, eurasiaSouthY,
		                           eurasiaGrain, archGrain,
		                           self.iHorzFlags, self.iTerrainFlags,
		                           -1, -1,
		                           True, 11,
		                           2, False,
		                           False
		                           )

		# Eurasia, second layer (to increase pangaea-like cohesion).
		twHeight = eurasiaHeight/2
		twWestX = eurasiaWestX + eurasiaWidth/10
		twEastX = eurasiaEastX - eurasiaWidth/10
		twWidth = twEastX - twWestX + 1
		twNorthY = eurasiaNorthY - eurasiaHeight/4
		twSouthY = eurasiaSouthY + eurasiaHeight/4

		twWater = 60+sea; twGrain = 1; twRift = 2

		self.generatePlotsInRegion(twWater,
		                           twWidth, twHeight,
		                           twWestX, twSouthY,
		                           twGrain, archGrain,
		                           self.iHorzFlags, self.iTerrainFlags,
		                           -1, -1,
		                           True, 11,
		                           twRift, False,
		                           False
		                           )

		# Simulate the New World - land masses akin to Earth's American continents.
		# First simulate North America
		NiTextOut("Generating the New World (Python Terra) ...")
		nwWestX = int(self.iW * newworldWestLon)
		nwEastX = int(self.iW * newworldEastLon)
		nwNorthY = int(self.iH * 0.85)
		nwSouthY = int(self.iH * 0.52)
		nwWidth = nwEastX - nwWestX + 1
		nwHeight = nwNorthY - nwSouthY + 1

		nwWater = 61+sea; nwGrain = 1; nwRift = -1

		self.generatePlotsInRegion(nwWater,
		                           nwWidth, nwHeight,
		                           nwWestX, nwSouthY,
		                           nwGrain, archGrain,
		                           self.iVertFlags, self.iTerrainFlags,
		                           6, 6,
		                           True, 7,
		                           nwRift, False,
		                           False
		                           )

		# Now simulate South America
		nwsRoll = self.dice.get(2, "New World South E/W - Terra PYTHON")
		nwsVar = 0.0
		if nwsRoll == 1: nwsVar = 0.05
		nwsWestX = nwWestX + int(self.iW * (0.08 - nwsVar)) # Not as wide as the north
		nwsEastX = nwEastX - int(self.iW * (0.03 + nwsVar))
		nwsNorthY = int(self.iH * 0.47)
		nwsSouthY = int(self.iH * 0.25)
		nwsWidth = nwsEastX - nwsWestX + 1
		nwsHeight = nwsNorthY - nwsSouthY + 1

		nwsWater = 55+sea; nwsGrain = 1; nwsRift = -1

		self.generatePlotsInRegion(nwsWater,
		                           nwsWidth, nwsHeight,
		                           nwsWestX, nwsSouthY,
		                           nwsGrain, archGrain,
		                           self.iRoundFlags, self.iTerrainFlags,
		                           6, 6,
		                           True, 5,
		                           nwsRift, False,
		                           False
		                           )

		nwpWestX = nwWestX + int(self.iW * (0.1 - nwsVar)) # Not as wide as the north
		nwpEastX = nwEastX - int(self.iW * (0.07 + nwsVar))
		nwpNorthY = int(self.iH * 0.3)
		nwpSouthY = int(self.iH * 0.18)
		nwpWidth = nwpEastX - nwpWestX + 1
		nwpHeight = nwpNorthY - nwpSouthY + 1

		nwpWater = 67+sea; nwpGrain = 1; nwpRift = -1

		self.generatePlotsInRegion(nwpWater,
		                           nwpWidth, nwpHeight,
		                           nwpWestX, nwpSouthY,
		                           nwpGrain, archGrain,
		                           self.iVertFlags, self.iTerrainFlags,
		                           6, 5,
		                           True, 3,
		                           nwpRift, False,
		                           False
		                           )

		# Now the Yukon
		twWidth = int(self.iW * 0.15)
		twWestX = nwWestX
		boreal = self.dice.get(2, "New World North E/W - Terra PYTHON")
		if boreal == 1: twWestX += int(self.iW * 0.15)
		twEastX = twWestX + twWidth
		twNorthY = int(self.iH * 0.93)
		twSouthY = int(self.iH * 0.75)
		twHeight = twNorthY - twSouthY + 1

		twWater = 68+sea; twGrain = 2; twRift = -1

		self.generatePlotsInRegion(twWater,
		                           twWidth, twHeight,
		                           twWestX, twSouthY,
		                           twGrain, archGrain,
		                           self.iRoundFlags, self.iTerrainFlags,
		                           6, 5,
		                           True, 5,
		                           twRift, False,
		                           False
		                           )

		# Now add a random region of arctic islands
		twWidth = int(thirdworldDimension * self.iW)
		twHeight = int(thirdworldDimension * self.iH)
		if boreal == 0:
			twEastX = nwEastX
			twWestX = twEastX - twWidth
		else:
			twWestX = nwWestX
			twEastX = twWestX + twWidth
		twNorthY = int(self.iH * 0.975)
		twSouthY = int(self.iH * 0.85)

		twWater = 76+sea; twGrain = archGrain; twRift = -1

		self.generatePlotsInRegion(twWater,
		                           twWidth, twHeight,
		                           twWestX, twSouthY,
		                           twGrain, archGrain,
		                           self.iHorzFlags, self.iTerrainFlags,
		                           6, 5,
		                           True, 5,
		                           twRift, False,
		                           False
		                           )

		# Now simulate Central America
		nwcVar = 0.0
		if nwsRoll == 1: nwcVar = 0.04
		nwcWidth = int(self.iW * 0.06)
		nwcRoll = self.dice.get(2, "Central America and Carribean Placement - Terra PYTHON")
		nwcWestX = nwWestX + int(self.iW * (0.1 + nwcVar))
		nwcEastX = nwcWestX + nwcWidth
		nwcNorthY = int(self.iH * 0.6)
		nwcSouthY = int(self.iH * 0.42)
		nwcHeight = nwcNorthY - nwcSouthY + 1

		nwcWater = 60+sea; nwcGrain = 1; nwcRift = -1

		self.generatePlotsInRegion(nwcWater,
		                           nwcWidth, nwcHeight,
		                           nwcWestX, nwcSouthY,
		                           nwcGrain, archGrain,
		                           self.iVertFlags, self.iTerrainFlags,
		                           6, 5,
		                           True, 5,
		                           nwcRift, False,
		                           False
		                           )

		# Now the Carribean islands
		carVar = 0.0
		if nwsRoll == 1: carVar = 0.15
		twWidth = int(0.15 * self.iW)
		twEastX = nwEastX - int(carVar * self.iW)
		twWestX = twEastX - twWidth
		twNorthY = int(self.iH * 0.55)
		twSouthY = int(self.iH * 0.47)
		twHeight = twNorthY - twSouthY + 1

		twWater = 75+sea; twGrain = archGrain + 1; twRift = -1

		self.generatePlotsInRegion(twWater,
		                           twWidth, twHeight,
		                           twWestX, twSouthY,
		                           twGrain, archGrain,
		                           0, self.iTerrainFlags,
		                           6, 5,
		                           True, 3,
		                           twRift, False,
		                           False
		                           )

		# Add subcontinents to the Old World, one large, one small. (Terra.py)
		# Subcontinents can be akin to pangaea, continents, or archipelago.
		# The large adds an amount of land akin to subSaharan Africa.
		# The small adds an amount of land akin to South Pacific islands.
		NiTextOut("Generating the Third World (Python Terra) ...")
		scLargeWidth = int(subcontinentLargeHorz * self.iW)
		scLargeHeight = int(subcontinentLargeVert * self.iH)
		scRoll = self.dice.get((eurasiaWidth - scLargeWidth), "Large Subcontinent Placement - Terra PYTHON")
		scWestX = eurasiaWestX + scRoll
		scEastX = scWestX + scLargeWidth
		scNorthY = int(self.iH * subcontinentLargeNorthLat)
		scSouthY = int(self.iH * subcontinentLargeSouthLat)

		scShape = self.dice.get(4, "Large Subcontinent Shape - Terra PYTHON")
		if scShape > 1: # Massive subcontinent! (Africa style)
			scWater = 55+sea; scGrain = 1; scRift = 2
		elif scShape == 1: # Standard subcontinent.
			scWater = 66+sea; scGrain = 2; scRift = 2
		else: # scShape == 0, Archipelago subcontinent.
			scWater = 77+sea; scGrain = archGrain; scRift = -1

		self.generatePlotsInRegion(scWater,
		                           scLargeWidth, scLargeHeight,
		                           scWestX, scSouthY,
		                           scGrain, archGrain,
		                           self.iRoundFlags, self.iTerrainFlags,
		                           6, 6,
		                           True, 7,
		                           scRift, False,
		                           False
		                           )

		scSmallWidth = int(subcontinentSmallDimension * self.iW)
		scSmallHeight = int(subcontinentSmallDimension * self.iH)
		endless = 1
		while endless == 1: # Prevent excessive overlap of the two subcontinents.
			scsRoll = self.dice.get((eurasiaWidth - scSmallWidth), "Small Subcontinent Placement - Terra PYTHON")
			scsWestX = eurasiaWestX + scsRoll
			if abs((scsWestX + self.iW/12) - scWestX) > self.iW/8: break
		scsEastX = scsWestX + scSmallWidth
		scsNorthY = int(self.iH * subcontinentSmallNorthLat)
		scsSouthY = int(self.iH * subcontinentSmallSouthLat)

		scsShape = self.dice.get(4, "Small Subcontinent Shape - Terra PYTHON")
		if scsShape == 2: # Massive subcontinent!
			scsWater = 55+sea; scsGrain = 1; scsRift = 2
		elif scsShape == 1: # Standard subcontinent. (India style).
			scsWater = 66+sea; scsGrain = 2; scsRift = 2
		else: # scsShape == 0 or 3, Archipelago subcontinent (East Indies style).
			scsWater = 77+sea; scsGrain = archGrain; scsRift = -1

		self.generatePlotsInRegion(scsWater,
		                           scSmallWidth, scSmallHeight,
		                           scsWestX, scsSouthY,
		                           scsGrain, archGrain,
		                           self.iHorzFlags, self.iTerrainFlags,
		                           6, 5,
		                           True, 5,
		                           scsRift, False,
		                           False
		                           )

		# Now simulate random lands akin to Australia and Antarctica
		extras = 2 + self.dice.get(3, "Number of Minor Regions - Terra PYTHON")
		for loop in range(extras):
			# Two to four of these regions.
			twWidth = int(thirdworldDimension * self.iW)
			twHeight = int(thirdworldDimension * self.iH)
			twVertRange = int(0.3 * self.iH) - twHeight
			twRoll = self.dice.get((eurasiaWidth - twWidth), "Minor Region Placement - Terra PYTHON")
			twWestX = eurasiaWestX + twRoll
			twEastX = scWestX + scLargeWidth
			twVertRoll = self.dice.get(twVertRange, "Minor Region Placement - Terra PYTHON")
			twNorthY = int(self.iH * thirdworldNorthLat) + twVertRoll
			twSouthY = int(self.iH * thirdworldSouthLat) + twVertRoll

			twShape = self.dice.get(3, "Minor Region Shape - Terra PYTHON")
			if twShape == 2: # Massive subcontinent!
				twWater = 60+sea; twGrain = 1; twRift = 2
			elif twShape == 1: # Standard subcontinent.
				twWater = 65+sea; twGrain = 2; twRift = 2
			else: # twShape == 0, Archipelago subcontinent.
				twWater = 70+sea; twGrain = archGrain; twRift = -1

			self.generatePlotsInRegion(twWater,
			                           twWidth, twHeight,
			                           twWestX, twSouthY,
			                           twGrain, archGrain,
			                           self.iHorzFlags, self.iTerrainFlags,
			                           6, 5,
			                           True, 5,
			                           twRift, False,
			                           False
			                           )

		# All regions have been processed. Plot Type generation completed.
		return self.wholeworldPlotTypes

class EarthMultilayeredFractal(CvMapGeneratorUtil.MultilayeredFractal):
	# Subclass. Only the controlling function overridden in this case.
	def generatePlotsByRegion(self):
		global AllowNewWorld
		AllowNewWorld = self.map.getCustomMapOption(0)
		# Sirian's MultilayeredFractal class, controlling function.
		# You -MUST- customize this function for each use of the class.
		#
		# The following grain matrix is specific to Earth2.py
		sizekey = self.map.getWorldSize()
		sizevalues = {
			WorldSizeTypes.WORLDSIZE_DUEL:      (3,2,1),
			WorldSizeTypes.WORLDSIZE_TINY:      (3,2,1),
			WorldSizeTypes.WORLDSIZE_SMALL:     (4,2,1),
			WorldSizeTypes.WORLDSIZE_STANDARD:  (4,2,1),
			WorldSizeTypes.WORLDSIZE_LARGE:     (4,2,1),
			WorldSizeTypes.WORLDSIZE_HUGE:      (5,2,1)
			}
		(ScatterGrain, BalanceGrain, GatherGrain) = sizevalues[sizekey]

		# Sea Level adjustment (from user input), limited to value of 5%.
		sea = self.gc.getSeaLevelInfo(self.map.getSeaLevel()).getSeaLevelChange()
		sea = min(sea, 5)
		sea = max(sea, -5)

		# The following regions are specific to Earth2.py

		NATundraWestLon = 0.05
		NATundraEastLon = 0.21
		GreenlandWestLon = 0.26
		GreenlandEastLon = 0.39
		NAmericasWestLon = 0.10
		NAmericasEastLon = 0.29
		FloridaWestLon = 0.28
		FloridaEastLon = 0.30
		MexicoWestLon = 0.12
		MexicoEastLon = 0.20
		CAmericasWestLon = 0.13
		CAmericasEastLon = 0.25
		PanamaWestLon = 0.21
		PanamaEastLon = 0.25
		CaribWestLon = 0.17
		CaribEastLon = 0.35
		SAmericasWestLon = 0.19
		SAmericasEastLon = 0.33
		BrazilWestLon = 0.21
		BrazilEastLon = 0.37
		AndesWestLon = 0.23
		AndesEastLon = 0.26
		EmeraldWestLon = 0.54
		EmeraldEastLon = 0.58
		NEuropeWestLon = 0.60
		NEuropeEastLon = 0.68
		CEuropeWestLon = 0.54
		CEuropeEastLon = 0.68
		IberiaWestLon = 0.51
		IberiaEastLon = 0.55
		MediWestLon = 0.54
		MediEastLon = 0.68
		NumidiaWestLon = 0.50
		NumidiaEastLon = 0.58
		AfricaWestLon = 0.50
		AfricaEastLon = 0.68
		CAfricaWestLon = 0.58
		CAfricaEastLon = 0.68
		SAfricaWestLon = 0.60
		SAfricaEastLon = 0.66
		SiberiaWestLon = 0.67
		SiberiaEastLon = 0.95
		SteppeWestLon = 0.67
		SteppeEastLon = 0.92
		NearEastWestLon = 0.67
		NearEastEastLon = 0.75
		ArabiaWestLon = 0.68
		ArabiaEastLon = 0.73
		IndiaWestLon = 0.73
		IndiaEastLon = 0.81
		ChinaWestLon = 0.80
		ChinaEastLon = 0.89
		IndoChinaWestLon = 0.80
		IndoChinaEastLon = 0.94
		JapanWestLon = 0.91
		JapanEastLon = 0.94
		AustraliaWestLon = 0.84
		AustraliaEastLon = 0.96
		SouthPacificWestLon = 0.01
		SouthPacificEastLon = 0.20
		NATundraNorthLat = 0.94
		NATundraSouthLat = 0.80
		GreenlandNorthLat = 0.90
		GreenlandSouthLat = 0.78
		NAmericasNorthLat = 0.82
		NAmericasSouthLat = 0.65
		FloridaNorthLat = 0.66
		FloridaSouthLat = 0.64
		MexicoNorthLat = 0.70
		MexicoSouthLat = 0.60
		CAmericasNorthLat = 0.65
		CAmericasSouthLat = 0.54
		PanamaNorthLat = 0.55
		PanamaSouthLat = 0.54
		CaribNorthLat = 0.66
		CaribSouthLat = 0.50
		SAmericasNorthLat = 0.55
		SAmericasSouthLat = 0.25
		BrazilNorthLat = 0.50
		BrazilSouthLat = 0.40
		AndesNorthLat = 0.56
		AndesSouthLat = 0.14
		EmeraldNorthLat = 0.86
		EmeraldSouthLat = 0.81
		NEuropeNorthLat = 0.92
		NEuropeSouthLat = 0.80
		CEuropeNorthLat = 0.80
		CEuropeSouthLat = 0.72
		IberiaNorthLat = 0.74
		IberiaSouthLat = 0.68
		MediNorthLat = 0.75
		MediSouthLat = 0.60
		AfricaNorthLat = 0.60
		AfricaSouthLat = 0.44
		NumidiaNorthLat = 0.64
		NumidiaSouthLat = 0.55
		CAfricaNorthLat = 0.46
		CAfricaSouthLat = 0.25
		SAfricaNorthLat = 0.30
		SAfricaSouthLat = 0.18
		SiberiaNorthLat = 0.94
		SiberiaSouthLat = 0.82
		SteppeNorthLat = 0.85
		SteppeSouthLat = 0.70
		NearEastNorthLat = 0.75
		NearEastSouthLat = 0.59
		ArabiaNorthLat = 0.60
		ArabiaSouthLat = 0.55
		IndiaNorthLat = 0.75
		IndiaSouthLat = 0.47
		ChinaNorthLat = 0.75
		ChinaSouthLat = 0.50
		IndoChinaNorthLat = 0.55
		IndoChinaSouthLat = 0.32
		JapanNorthLat = 0.75
		JapanSouthLat = 0.65
		AustraliaNorthLat = 0.30
		AustraliaSouthLat = 0.15
		SouthPacificNorthLat = 0.45
		SouthPacificSouthLat = 0.15

		# Simulate the Western Hemisphere - North American Tundra.
		NiTextOut("Generating North America (Python Earth2) ...")
		# Set dimensions of North American Tundra
		NATundraWestX = int(self.iW * NATundraWestLon)
		NATundraEastX = int(self.iW * NATundraEastLon)
		NATundraNorthY = int(self.iH * NATundraNorthLat)
		NATundraSouthY = int(self.iH * NATundraSouthLat)
		NATundraWidth = NATundraEastX - NATundraWestX + 1
		NATundraHeight = NATundraNorthY - NATundraSouthY + 1

		NATundraWater = 35+sea

		self.generatePlotsInRegion(NATundraWater,
				   NATundraWidth, NATundraHeight,
				   NATundraWestX, NATundraSouthY,
				   BalanceGrain, GatherGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Western Hemisphere - Greenland.
		NiTextOut("Generating North America (Python Earth2) ...")
		# Set dimensions of Greenland
		GreenlandWestX = int(self.iW * GreenlandWestLon)
		GreenlandEastX = int(self.iW * GreenlandEastLon)
		GreenlandNorthY = int(self.iH * GreenlandNorthLat)
		GreenlandSouthY = int(self.iH * GreenlandSouthLat)
		GreenlandWidth = GreenlandEastX - GreenlandWestX + 1
		GreenlandHeight = GreenlandNorthY - GreenlandSouthY + 1

		GreenlandWater = 70+sea

		self.generatePlotsInRegion(GreenlandWater,
				   GreenlandWidth, GreenlandHeight,
				   GreenlandWestX, GreenlandSouthY,
				   ScatterGrain, GatherGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   1, True,
				   False
				   )

		self.generatePlotsInRegion(GreenlandWater,
				   GreenlandWidth, GreenlandHeight,
				   GreenlandWestX, GreenlandSouthY,
				   ScatterGrain, GatherGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   1, True,
				   False
				   )

		# Simulate the Western Hemisphere - North America Mainland.
		NiTextOut("Generating North America (Python Earth2) ...")
		# Set dimensions of North America Mainland
		NAmericasWestX = int(self.iW * NAmericasWestLon)
		NAmericasEastX = int(self.iW * NAmericasEastLon)
		NAmericasNorthY = int(self.iH * NAmericasNorthLat)
		NAmericasSouthY = int(self.iH * NAmericasSouthLat)
		NAmericasWidth = NAmericasEastX - NAmericasWestX + 1
		NAmericasHeight = NAmericasNorthY - NAmericasSouthY + 1

		NAmericasWater = 40+sea

		self.generatePlotsInRegion(NAmericasWater,
				   NAmericasWidth, NAmericasHeight,
				   NAmericasWestX, NAmericasSouthY,
				   GatherGrain, BalanceGrain,
				   self.iVertFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(NAmericasWater,
				   NAmericasWidth, NAmericasHeight,
				   NAmericasWestX, NAmericasSouthY,
				   GatherGrain, BalanceGrain,
				   self.iVertFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Western Hemisphere - Florida.
		NiTextOut("Generating North America (Python Earth2) ...")
		# Set dimensions of Florida
		FloridaWestX = int(self.iW * FloridaWestLon)
		FloridaEastX = int(self.iW * FloridaEastLon)
		FloridaNorthY = int(self.iH * FloridaNorthLat)
		FloridaSouthY = int(self.iH * FloridaSouthLat)
		FloridaWidth = FloridaEastX - FloridaWestX + 1
		FloridaHeight = FloridaNorthY - FloridaSouthY + 1

		FloridaWater = 40+sea

		self.generatePlotsInRegion(FloridaWater,
				   FloridaWidth, FloridaHeight,
				   FloridaWestX, FloridaSouthY,
				   GatherGrain, ScatterGrain,
				   self.iVertFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Western Hemisphere - Mexico.
		NiTextOut("Generating Central America (Python Earth2) ...")
		# Set dimensions of Mexico
		MexicoWestX = int(self.iW * MexicoWestLon)
		MexicoEastX = int(self.iW * MexicoEastLon)
		MexicoNorthY = int(self.iH * MexicoNorthLat)
		MexicoSouthY = int(self.iH * MexicoSouthLat)
		MexicoWidth = MexicoEastX - MexicoWestX + 1
		MexicoHeight = MexicoNorthY - MexicoSouthY + 1

		MexicoWater = 30+sea

		self.generatePlotsInRegion(MexicoWater,
				   MexicoWidth, MexicoHeight,
				   MexicoWestX, MexicoSouthY,
				   GatherGrain, GatherGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Western Hemisphere - Central America.
		NiTextOut("Generating Central America (Python Earth2) ...")
		# Set dimensions of Central America
		CAmericasWestX = int(self.iW * CAmericasWestLon)
		CAmericasEastX = int(self.iW * CAmericasEastLon)
		CAmericasNorthY = int(self.iH * CAmericasNorthLat)
		CAmericasSouthY = int(self.iH * CAmericasSouthLat)
		CAmericasWidth = CAmericasEastX - CAmericasWestX + 1
		CAmericasHeight = CAmericasNorthY - CAmericasSouthY + 1

		CAmericasWater = 80+sea

		self.generatePlotsInRegion(CAmericasWater,
				   CAmericasWidth, CAmericasHeight,
				   CAmericasWestX, CAmericasSouthY,
				   GatherGrain, GatherGrain,
				   self.iVertFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Western Hemisphere - Panama.
		NiTextOut("Generating Central America (Python Earth2) ...")
		# Set dimensions of Panama
		PanamaWestX = int(self.iW * PanamaWestLon)
		PanamaEastX = int(self.iW * PanamaEastLon)
		PanamaNorthY = int(self.iH * PanamaNorthLat)
		PanamaSouthY = int(self.iH * PanamaSouthLat)
		PanamaWidth = PanamaEastX - PanamaWestX + 1
		PanamaHeight = PanamaNorthY - PanamaSouthY + 1

		PanamaWater = 85+sea

		self.generatePlotsInRegion(PanamaWater,
				   PanamaWidth, PanamaHeight,
				   PanamaWestX, PanamaSouthY,
				   GatherGrain, GatherGrain,
				   self.iHorzFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Western Hemisphere - The Caribbean.
		NiTextOut("Generating Central America (Python Earth2) ...")
		# Set dimensions of The Caribbean
		CaribWestX = int(self.iW * CaribWestLon)
		CaribEastX = int(self.iW * CaribEastLon)
		CaribNorthY = int(self.iH * CaribNorthLat)
		CaribSouthY = int(self.iH * CaribSouthLat)
		CaribWidth = CaribEastX - CaribWestX + 1
		CaribHeight = CaribNorthY - CaribSouthY + 1

		CaribWater = 90+sea

		self.generatePlotsInRegion(CaribWater,
				   CaribWidth, CaribHeight,
				   CaribWestX, CaribSouthY,
				   ScatterGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Western Hemisphere - South America.
		NiTextOut("Generating South America (Python Earth2) ...")
		# Set dimensions of South America
		SAmericasWestX = int(self.iW * SAmericasWestLon)
		SAmericasEastX = int(self.iW * SAmericasEastLon)
		SAmericasNorthY = int(self.iH * SAmericasNorthLat)
		SAmericasSouthY = int(self.iH * SAmericasSouthLat)
		SAmericasWidth = SAmericasEastX - SAmericasWestX + 1
		SAmericasHeight = SAmericasNorthY - SAmericasSouthY + 1

		SAmericasWater = 65+sea

		self.generatePlotsInRegion(SAmericasWater,
				   SAmericasWidth, SAmericasHeight,
				   SAmericasWestX, SAmericasSouthY,
				   GatherGrain, GatherGrain,
				   self.iVertFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(SAmericasWater,
				   SAmericasWidth, SAmericasHeight,
				   SAmericasWestX, SAmericasSouthY,
				   GatherGrain, GatherGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Western Hemisphere - Brazil.
		NiTextOut("Generating South America (Python Earth2) ...")
		# Set dimensions of Brazil
		BrazilWestX = int(self.iW * BrazilWestLon)
		BrazilEastX = int(self.iW * BrazilEastLon)
		BrazilNorthY = int(self.iH * BrazilNorthLat)
		BrazilSouthY = int(self.iH * BrazilSouthLat)
		BrazilWidth = BrazilEastX - BrazilWestX + 1
		BrazilHeight = BrazilNorthY - BrazilSouthY + 1

		BrazilWater = 45+sea

		self.generatePlotsInRegion(BrazilWater,
				   BrazilWidth, BrazilHeight,
				   BrazilWestX, BrazilSouthY,
				   GatherGrain, GatherGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(BrazilWater,
				   BrazilWidth, BrazilHeight,
				   BrazilWestX, BrazilSouthY,
				   GatherGrain, GatherGrain,
				   self.iHorzFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Western Hemisphere - Andes.
		NiTextOut("Generating South America (Python Earth2) ...")
		# Set dimensions of Andes
		AndesWestX = int(self.iW * AndesWestLon)
		AndesEastX = int(self.iW * AndesEastLon)
		AndesNorthY = int(self.iH * AndesNorthLat)
		AndesSouthY = int(self.iH * AndesSouthLat)
		AndesWidth = AndesEastX - AndesWestX + 1
		AndesHeight = AndesNorthY - AndesSouthY + 1

		AndesWater = 35+sea

		self.generatePlotsInRegion(AndesWater,
				   AndesWidth, AndesHeight,
				   AndesWestX, AndesSouthY,
				   GatherGrain, GatherGrain,
				   self.iVertFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - British Isles.
		NiTextOut("Generating Europe (Python Earth2) ...")
		# Set dimensions of British Isles
		EmeraldWestX = int(self.iW * EmeraldWestLon)
		EmeraldEastX = int(self.iW * EmeraldEastLon)
		EmeraldNorthY = int(self.iH * EmeraldNorthLat)
		EmeraldSouthY = int(self.iH * EmeraldSouthLat)
		EmeraldWidth = EmeraldEastX - EmeraldWestX + 1
		EmeraldHeight = EmeraldNorthY - EmeraldSouthY + 1

		EmeraldWater = 80+sea

		self.generatePlotsInRegion(EmeraldWater,
				   EmeraldWidth, EmeraldHeight,
				   EmeraldWestX, EmeraldSouthY,
				   BalanceGrain, ScatterGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(EmeraldWater,
				   EmeraldWidth, EmeraldHeight,
				   EmeraldWestX, EmeraldSouthY,
				   BalanceGrain, ScatterGrain,
				   self.iVertFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - Northern Europe.
		NiTextOut("Generating Europe (Python Earth2) ...")
		# Set dimensions of Northern Europe
		NEuropeWestX = int(self.iW * NEuropeWestLon)
		NEuropeEastX = int(self.iW * NEuropeEastLon)
		NEuropeNorthY = int(self.iH * NEuropeNorthLat)
		NEuropeSouthY = int(self.iH * NEuropeSouthLat)
		NEuropeWidth = NEuropeEastX - NEuropeWestX + 1
		NEuropeHeight = NEuropeNorthY - NEuropeSouthY + 1

		NEuropeWater = 50+sea

		self.generatePlotsInRegion(NEuropeWater,
				   NEuropeWidth, NEuropeHeight,
				   NEuropeWestX, NEuropeSouthY,
				   BalanceGrain, GatherGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - Central Europe.
		NiTextOut("Generating Europe (Python Earth2) ...")
		# Set dimensions of Central Europe
		CEuropeWestX = int(self.iW * CEuropeWestLon)
		CEuropeEastX = int(self.iW * CEuropeEastLon)
		CEuropeNorthY = int(self.iH * CEuropeNorthLat)
		CEuropeSouthY = int(self.iH * CEuropeSouthLat)
		CEuropeWidth = CEuropeEastX - CEuropeWestX + 1
		CEuropeHeight = CEuropeNorthY - CEuropeSouthY + 1

		CEuropeWater = 35+sea

		self.generatePlotsInRegion(CEuropeWater,
				   CEuropeWidth, CEuropeHeight,
				   CEuropeWestX, CEuropeSouthY,
				   GatherGrain, GatherGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(CEuropeWater,
				   CEuropeWidth, CEuropeHeight,
				   CEuropeWestX, CEuropeSouthY,
				   GatherGrain, GatherGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - Iberia.
		NiTextOut("Generating Europe (Python Earth2) ...")
		# Set dimensions of Iberia
		IberiaWestX = int(self.iW * IberiaWestLon)
		IberiaEastX = int(self.iW * IberiaEastLon)
		IberiaNorthY = int(self.iH * IberiaNorthLat)
		IberiaSouthY = int(self.iH * IberiaSouthLat)
		IberiaWidth = IberiaEastX - IberiaWestX + 1
		IberiaHeight = IberiaNorthY - IberiaSouthY + 1

		IberiaWater = 20+sea

		self.generatePlotsInRegion(IberiaWater,
				   IberiaWidth, IberiaHeight,
				   IberiaWestX, IberiaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - Mediterranean.
		NiTextOut("Generating Europe (Python Earth2) ...")
		# Set dimensions of Mediterranean
		MediWestX = int(self.iW * MediWestLon)
		MediEastX = int(self.iW * MediEastLon)
		MediNorthY = int(self.iH * MediNorthLat)
		MediSouthY = int(self.iH * MediSouthLat)
		MediWidth = MediEastX - MediWestX + 1
		MediHeight = MediNorthY - MediSouthY + 1

		MediWater = 80+sea

		self.generatePlotsInRegion(MediWater,
				   MediWidth, MediHeight,
				   MediWestX, MediSouthY,
				   ScatterGrain, ScatterGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - North Africa.
		NiTextOut("Generating Africa (Python Earth2) ...")
		# Set dimensions of North Africa
		AfricaWestX = int(self.iW * AfricaWestLon)
		AfricaEastX = int(self.iW * AfricaEastLon)
		AfricaNorthY = int(self.iH * AfricaNorthLat)
		AfricaSouthY = int(self.iH * AfricaSouthLat)
		AfricaWidth = AfricaEastX - AfricaWestX + 1
		AfricaHeight = AfricaNorthY - AfricaSouthY + 1

		AfricaWater = 50+sea

		self.generatePlotsInRegion(AfricaWater,
				   AfricaWidth, AfricaHeight,
				   AfricaWestX, AfricaSouthY,
				   GatherGrain, ScatterGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(AfricaWater,
				   AfricaWidth, AfricaHeight,
				   AfricaWestX, AfricaSouthY,
				   GatherGrain, ScatterGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(AfricaWater,
				   AfricaWidth, AfricaHeight,
				   AfricaWestX, AfricaSouthY,
				   GatherGrain, ScatterGrain,
				   self.iVertFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - Numidia.
		NiTextOut("Generating Africa (Python Earth2) ...")
		# Set dimensions of Numidia
		NumidiaWestX = int(self.iW * NumidiaWestLon)
		NumidiaEastX = int(self.iW * NumidiaEastLon)
		NumidiaNorthY = int(self.iH * NumidiaNorthLat)
		NumidiaSouthY = int(self.iH * NumidiaSouthLat)
		NumidiaWidth = NumidiaEastX - NumidiaWestX + 1
		NumidiaHeight = NumidiaNorthY - NumidiaSouthY + 1

		NumidiaWater = 20+sea

		self.generatePlotsInRegion(NumidiaWater,
				   NumidiaWidth, NumidiaHeight,
				   NumidiaWestX, NumidiaSouthY,
				   GatherGrain, GatherGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - Central Africa.
		NiTextOut("Generating Africa (Python Earth2) ...")
		# Set dimensions of Central Africa
		CAfricaWestX = int(self.iW * CAfricaWestLon)
		CAfricaEastX = int(self.iW * CAfricaEastLon)
		CAfricaNorthY = int(self.iH * CAfricaNorthLat)
		CAfricaSouthY = int(self.iH * CAfricaSouthLat)
		CAfricaWidth = CAfricaEastX - CAfricaWestX + 1
		CAfricaHeight = CAfricaNorthY - CAfricaSouthY + 1

		CAfricaWater = 35+sea

		self.generatePlotsInRegion(CAfricaWater,
				   CAfricaWidth, CAfricaHeight,
				   CAfricaWestX, CAfricaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(CAfricaWater,
				   CAfricaWidth, CAfricaHeight,
				   CAfricaWestX, CAfricaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - South Africa.
		NiTextOut("Generating Africa (Python Earth2) ...")
		# Set dimensions of South Africa
		SAfricaWestX = int(self.iW * SAfricaWestLon)
		SAfricaEastX = int(self.iW * SAfricaEastLon)
		SAfricaNorthY = int(self.iH * SAfricaNorthLat)
		SAfricaSouthY = int(self.iH * SAfricaSouthLat)
		SAfricaWidth = SAfricaEastX - SAfricaWestX + 1
		SAfricaHeight = SAfricaNorthY - SAfricaSouthY + 1

		SAfricaWater = 45+sea

		self.generatePlotsInRegion(SAfricaWater,
				   SAfricaWidth, SAfricaHeight,
				   SAfricaWestX, SAfricaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(SAfricaWater,
				   SAfricaWidth, SAfricaHeight,
				   SAfricaWestX, SAfricaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - Siberia.
		NiTextOut("Generating Asia (Python Earth2) ...")
		# Set dimensions of Siberia
		SiberiaWestX = int(self.iW * SiberiaWestLon)
		SiberiaEastX = int(self.iW * SiberiaEastLon)
		SiberiaNorthY = int(self.iH * SiberiaNorthLat)
		SiberiaSouthY = int(self.iH * SiberiaSouthLat)
		SiberiaWidth = SiberiaEastX - SiberiaWestX + 1
		SiberiaHeight = SiberiaNorthY - SiberiaSouthY + 1

		SiberiaWater = 25+sea

		self.generatePlotsInRegion(SiberiaWater,
				   SiberiaWidth, SiberiaHeight,
				   SiberiaWestX, SiberiaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - Steppe.
		NiTextOut("Generating Asia (Python Earth2) ...")
		# Set dimensions of Steppe
		SteppeWestX = int(self.iW * SteppeWestLon)
		SteppeEastX = int(self.iW * SteppeEastLon)
		SteppeNorthY = int(self.iH * SteppeNorthLat)
		SteppeSouthY = int(self.iH * SteppeSouthLat)
		SteppeWidth = SteppeEastX - SteppeWestX + 1
		SteppeHeight = SteppeNorthY - SteppeSouthY + 1

		SteppeWater = 6+sea

		self.generatePlotsInRegion(SteppeWater,
				   SteppeWidth, SteppeHeight,
				   SteppeWestX, SteppeSouthY,
				   GatherGrain, GatherGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - Near East.
		NiTextOut("Generating Asia (Python Earth2) ...")
		# Set dimensions of Near East
		NearEastWestX = int(self.iW * NearEastWestLon)
		NearEastEastX = int(self.iW * NearEastEastLon)
		NearEastNorthY = int(self.iH * NearEastNorthLat)
		NearEastSouthY = int(self.iH * NearEastSouthLat)
		NearEastWidth = NearEastEastX - NearEastWestX + 1
		NearEastHeight = NearEastNorthY - NearEastSouthY + 1

		NearEastWater = 50+sea

		self.generatePlotsInRegion(NearEastWater,
				   NearEastWidth, NearEastHeight,
				   NearEastWestX, NearEastSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(NearEastWater,
				   NearEastWidth, NearEastHeight,
				   NearEastWestX, NearEastSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - Arabia.
		NiTextOut("Generating Asia (Python Earth2) ...")
		# Set dimensions of Arabia
		ArabiaWestX = int(self.iW * ArabiaWestLon)
		ArabiaEastX = int(self.iW * ArabiaEastLon)
		ArabiaNorthY = int(self.iH * ArabiaNorthLat)
		ArabiaSouthY = int(self.iH * ArabiaSouthLat)
		ArabiaWidth = ArabiaEastX - ArabiaWestX + 1
		ArabiaHeight = ArabiaNorthY - ArabiaSouthY + 1

		ArabiaWater = 50+sea

		self.generatePlotsInRegion(ArabiaWater,
				   ArabiaWidth, ArabiaHeight,
				   ArabiaWestX, ArabiaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iVertFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - India.
		NiTextOut("Generating Asia (Python Earth2) ...")
		# Set dimensions of India
		IndiaWestX = int(self.iW * IndiaWestLon)
		IndiaEastX = int(self.iW * IndiaEastLon)
		IndiaNorthY = int(self.iH * IndiaNorthLat)
		IndiaSouthY = int(self.iH * IndiaSouthLat)
		IndiaWidth = IndiaEastX - IndiaWestX + 1
		IndiaHeight = IndiaNorthY - IndiaSouthY + 1

		IndiaWater = 33+sea

		self.generatePlotsInRegion(IndiaWater,
				   IndiaWidth, IndiaHeight,
				   IndiaWestX, IndiaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(IndiaWater,
				   IndiaWidth, IndiaHeight,
				   IndiaWestX, IndiaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - China.
		NiTextOut("Generating Asia (Python Earth2) ...")
		# Set dimensions of China
		ChinaWestX = int(self.iW * ChinaWestLon)
		ChinaEastX = int(self.iW * ChinaEastLon)
		ChinaNorthY = int(self.iH * ChinaNorthLat)
		ChinaSouthY = int(self.iH * ChinaSouthLat)
		ChinaWidth = ChinaEastX - ChinaWestX + 1
		ChinaHeight = ChinaNorthY - ChinaSouthY + 1

		ChinaWater = 65+sea

		self.generatePlotsInRegion(ChinaWater,
				   ChinaWidth, ChinaHeight,
				   ChinaWestX, ChinaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(ChinaWater,
				   ChinaWidth, ChinaHeight,
				   ChinaWestX, ChinaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iHorzFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(ChinaWater,
				   ChinaWidth, ChinaHeight,
				   ChinaWestX, ChinaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iHorzFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - IndoChina.
		NiTextOut("Generating Asia (Python Earth2) ...")
		# Set dimensions of IndoChina
		IndoChinaWestX = int(self.iW * IndoChinaWestLon)
		IndoChinaEastX = int(self.iW * IndoChinaEastLon)
		IndoChinaNorthY = int(self.iH * IndoChinaNorthLat)
		IndoChinaSouthY = int(self.iH * IndoChinaSouthLat)
		IndoChinaWidth = IndoChinaEastX - IndoChinaWestX + 1
		IndoChinaHeight = IndoChinaNorthY - IndoChinaSouthY + 1

		IndoChinaWater = 82+sea

		self.generatePlotsInRegion(IndoChinaWater,
				   IndoChinaWidth, IndoChinaHeight,
				   IndoChinaWestX, IndoChinaSouthY,
				   ScatterGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(IndoChinaWater,
				   IndoChinaWidth, IndoChinaHeight,
				   IndoChinaWestX, IndoChinaSouthY,
				   ScatterGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - Japan.
		NiTextOut("Generating Asia (Python Earth2) ...")
		# Set dimensions of Japan
		JapanWestX = int(self.iW * JapanWestLon)
		JapanEastX = int(self.iW * JapanEastLon)
		JapanNorthY = int(self.iH * JapanNorthLat)
		JapanSouthY = int(self.iH * JapanSouthLat)
		JapanWidth = JapanEastX - JapanWestX + 1
		JapanHeight = JapanNorthY - JapanSouthY + 1

		JapanWater = 92+sea

		self.generatePlotsInRegion(JapanWater,
				   JapanWidth, JapanHeight,
				   JapanWestX, JapanSouthY,
				   BalanceGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(JapanWater,
				   JapanWidth, JapanHeight,
				   JapanWestX, JapanSouthY,
				   BalanceGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the Eastern Hemisphere - Australia.
		NiTextOut("Generating Australia (Python Earth2) ...")
		# Set dimensions of Australia
		AustraliaWestX = int(self.iW * AustraliaWestLon)
		AustraliaEastX = int(self.iW * AustraliaEastLon)
		AustraliaNorthY = int(self.iH * AustraliaNorthLat)
		AustraliaSouthY = int(self.iH * AustraliaSouthLat)
		AustraliaWidth = AustraliaEastX - AustraliaWestX + 1
		AustraliaHeight = AustraliaNorthY - AustraliaSouthY + 1

		AustraliaWater = 45+sea

		self.generatePlotsInRegion(AustraliaWater,
				   AustraliaWidth, AustraliaHeight,
				   AustraliaWestX, AustraliaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		self.generatePlotsInRegion(AustraliaWater,
				   AustraliaWidth, AustraliaHeight,
				   AustraliaWestX, AustraliaSouthY,
				   GatherGrain, BalanceGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# Simulate the South Pacific - South Pacific.
		NiTextOut("Generating Pacific (Python Earth2) ...")
		# Set dimensions of South Pacific
		SouthPacificWestX = int(self.iW * SouthPacificWestLon)
		SouthPacificEastX = int(self.iW * SouthPacificEastLon)
		SouthPacificNorthY = int(self.iH * SouthPacificNorthLat)
		SouthPacificSouthY = int(self.iH * SouthPacificSouthLat)
		SouthPacificWidth = SouthPacificEastX - SouthPacificWestX + 1
		SouthPacificHeight = SouthPacificNorthY - SouthPacificSouthY + 1

		SouthPacificWater = 94+sea

		self.generatePlotsInRegion(SouthPacificWater,
				   SouthPacificWidth, SouthPacificHeight,
				   SouthPacificWestX, SouthPacificSouthY,
				   ScatterGrain, ScatterGrain,
				   self.iRoundFlags, self.iTerrainFlags,
				   5, 5,
				   True, 5,
				   -1, False,
				   False
				   )

		# All regions have been processed. Plot Type generation completed.
		return self.wholeworldPlotTypes